// [x86data.js]
// X86/X64 instruction-set data.
//
// [License]
// Public Domain.

// DESCRIPTION
// -----------
//
// These tables can be used for any purpose. The idea is to have an instruction
// database that can be used to generate tables needed by assemblers and disasm
// engines, and also to have data that can be used to generate instruction tables
// for developers.

// INSTRUCTIONS
// ------------
//
// Each instruction definition consists of 5 strings:
//
//   [0] - Instruction name.
//   [1] - Instruction operands.
//   [2] - Instruction encoding.
//   [3] - Instruction opcodes.
//   [4] - Instruction tags - EFLAGs, CPU requirements, and some other metadata.
//
// The definition tries to match Intel and AMD instruction set manuals, but there
// are small differences to make the definition more informative and compact.

// REGISTERS
// ---------
//
//   "reg"   - Register, if a specific register (like "cl") is used, it means
//             that the instruction uses this register implicitly and is not
//             used to encode the instruction. Assemblers and disassemblers
//             may decide how to show this operand.
//
//   "<reg>" - Designed for AsmJit, implicit register that is omitted in
//             X86Assembler, but has to be specified in X86Compiler (there
//             is no other way, when using X86Compiler all registers, implicit
//             or explicit, has to be passed.

// READ/WRITE INFORMATION
// ----------------------
//
// Instruction tables also contain overrides that can be used to specify which
// operands are used as a destination, and which operands are source-only. Each
// instruction in general assumes that the first operand is always read/write
// and all following operands are read-only. However, this assumption is not
// valid for all instructions, thus instructions that don't match this assumption
// use these decorators:
//
//   "R:operand" - The operand is read-only.
//   "W:operand" - The operand is write-only.
//   "X:operand" - The operand is read/write.
//
// If a "R|W|X[A:B]" decorator is used, it describes which bytes are read, written,
// or read and written. This is important in case that you want to generate tables
// that will be used to schedule/reorder the instruction stream. Also note that this
// DB always uses BYTE indexes, not BITS, as used in Intel instruction set manuals.

// WHAT IS MISSING
// ---------------
//
// Here is a list of missing instructions to keep track of it:
//
// [ ] call (far)
// [ ] jmp (far)
// [ ] ret (far)
// [ ] iret/iretd
// [ ] lds/les/lfs/lgs/lss
// [ ] rsm
// [ ] xacquire (prefix)
// [ ] xrelease (prefix)
// [ ] xabort
// [ ] xbegin
// [ ] xend
// [ ] xlat/xlatb
// [ ] xtest
// [ ] VMX/SVM extensions
// [ ] TSX extensions

(function($export, $as) {
"use strict";

$export[$as] = {
  architectures: [
    "ANY",
    "X86",
    "X64"
  ],

  features: [
    "3DNOW",
    "3DNOW2",
    "ADX",
    "AES",
    "AVX",
    "AVX2",
    "AVX512BW",
    "AVX512CD",
    "AVX512DQ",
    "AVX512ER",
    "AVX512F",
    "AVX512IFMA",
    "AVX512PF",
    "AVX512VBMI",
    "AVX512VL",
    "BMI",
    "BMI2",
    "CLFLUSH",
    "CLFLUSH_OPT",
    "CLWB",
    "CMOV",
    "CMPXCHG8B",
    "CMPXCHG16B",
    "F16C",
    "FMA",
    "FMA4",
    "FSGSBASE",
    "FXSR",
    "GEODE",
    "I486",
    "LAHFSAHF",
    "LZCNT",
    "MPX",
    "MMX",
    "MONITOR",
    "MOVBE",
    "MSR",
    "PCLMULQDQ",
    "PCOMMIT",
    "POPCNT",
    "PREFETCHW",
    "PREFETCHWT1",
    "RDRAND",
    "RDSEED",
    "RDTSC",
    "RDTSCP",
    "SHA",
    "SMAP",
    "SSE",
    "SSE2",
    "SSE3",
    "SSE4_1",
    "SSE4_2",
    "SSE4A",
    "SSSE3",
    "TBM",
    "VMX",
    "XOP",
    "XSAVE",
    "XSAVE_OPT"
  ],

  instructions: [
    // X86/X64.
    ["aaa"              , "<ax>"                                     , "NONE" , "37"                               , "X86 OF=U SF=U ZF=U AF=W PF=U CF=W"],
    ["aas"              , "<ax>"                                     , "NONE" , "3F"                               , "X86 OF=U SF=U ZF=U AF=W PF=U CF=W"],
    ["aad"              , "<ax>, ib"                                 , "I"    , "D5 ib"                            , "X86 OF=U SF=W ZF=W AF=U PF=W CF=U"],
    ["aam"              , "<ax>, ib"                                 , "I"    , "D4 ib"                            , "X86 OF=U SF=W ZF=W AF=U PF=W CF=U"],

    ["adc"              , "al, ib"                                   , "I"    , "14 ib"                            , "ANY LOCK OF=W SF=W ZF=W AF=W PF=W CF=X"],
    ["adc"              , "ax, iw"                                   , "I"    , "66 15 iw"                         , "ANY LOCK OF=W SF=W ZF=W AF=W PF=W CF=X"],
    ["adc"              , "eax, id"                                  , "I"    , "15 id"                            , "ANY LOCK OF=W SF=W ZF=W AF=W PF=W CF=X"],
    ["adc"              , "rax, id"                                  , "I"    , "REX.W 15 id"                      , "X64 LOCK OF=W SF=W ZF=W AF=W PF=W CF=X"],

    ["adc"              , "r8/m8, ib"                                , "MI"   , "80 /2 ib"                         , "ANY LOCK OF=W SF=W ZF=W AF=W PF=W CF=X"],
    ["adc"              , "r16/m16, iw"                              , "MI"   , "66 81 /2 iw"                      , "ANY LOCK OF=W SF=W ZF=W AF=W PF=W CF=X"],
    ["adc"              , "r32/m32, id"                              , "MI"   , "81 /2 id"                         , "ANY LOCK OF=W SF=W ZF=W AF=W PF=W CF=X"],
    ["adc"              , "r64/m64, id"                              , "MI"   , "REX.W 81 /2 id"                   , "X64 LOCK OF=W SF=W ZF=W AF=W PF=W CF=X"],

    ["adc"              , "r16/m16, ib"                              , "MI"   , "66 83 /2 ib"                      , "ANY LOCK OF=W SF=W ZF=W AF=W PF=W CF=X"],
    ["adc"              , "r32/m32, ib"                              , "MI"   , "83 /2 ib"                         , "ANY LOCK OF=W SF=W ZF=W AF=W PF=W CF=X"],
    ["adc"              , "r64/m64, ib"                              , "MI"   , "REX.W 83 /2 ib"                   , "X64 LOCK OF=W SF=W ZF=W AF=W PF=W CF=X"],

    ["adc"              , "r8/m8, r8"                                , "MR"   , "10 /r"                            , "ANY LOCK OF=W SF=W ZF=W AF=W PF=W CF=X"],
    ["adc"              , "r16/m16, r16"                             , "MR"   , "66 11 /r"                         , "ANY LOCK OF=W SF=W ZF=W AF=W PF=W CF=X"],
    ["adc"              , "r32/m32, r32"                             , "MR"   , "11 /r"                            , "ANY LOCK OF=W SF=W ZF=W AF=W PF=W CF=X"],
    ["adc"              , "r64/m64, r64"                             , "MR"   , "REX.W 11 /r"                      , "X64 LOCK OF=W SF=W ZF=W AF=W PF=W CF=X"],

    ["adc"              , "r8, r8/m8"                                , "RM"   , "12 /r"                            , "ANY LOCK OF=W SF=W ZF=W AF=W PF=W CF=X"],
    ["adc"              , "r16, r16/m16"                             , "RM"   , "66 13 /r"                         , "ANY LOCK OF=W SF=W ZF=W AF=W PF=W CF=X"],
    ["adc"              , "r32, r32/m32"                             , "RM"   , "13 /r"                            , "ANY LOCK OF=W SF=W ZF=W AF=W PF=W CF=X"],
    ["adc"              , "r64, r64/m64"                             , "RM"   , "REX.W 13 /r"                      , "X64 LOCK OF=W SF=W ZF=W AF=W PF=W CF=X"],

    ["add"              , "al, ib"                                   , "I"    , "04 ib"                            , "ANY LOCK OF=W SF=W ZF=W AF=W PF=W CF=W"],
    ["add"              , "ax, iw"                                   , "I"    , "66 05 iw"                         , "ANY LOCK OF=W SF=W ZF=W AF=W PF=W CF=W"],
    ["add"              , "eax, id"                                  , "I"    , "05 id"                            , "ANY LOCK OF=W SF=W ZF=W AF=W PF=W CF=W"],
    ["add"              , "rax, id"                                  , "I"    , "REX.W 05 id"                      , "X64 LOCK OF=W SF=W ZF=W AF=W PF=W CF=W"],

    ["add"              , "r8/m8, ib"                                , "MI"   , "80 /0 ib"                         , "ANY LOCK OF=W SF=W ZF=W AF=W PF=W CF=W"],
    ["add"              , "r16/m16, iw"                              , "MI"   , "66 81 /0 iw"                      , "ANY LOCK OF=W SF=W ZF=W AF=W PF=W CF=W"],
    ["add"              , "r32/m32, id"                              , "MI"   , "81 /0 id"                         , "ANY LOCK OF=W SF=W ZF=W AF=W PF=W CF=W"],
    ["add"              , "r64/m64, id"                              , "MI"   , "REX.W 81 /0 id"                   , "X64 LOCK OF=W SF=W ZF=W AF=W PF=W CF=W"],

    ["add"              , "r16/m16, ib"                              , "MI"   , "66 83 /0 ib"                      , "ANY LOCK OF=W SF=W ZF=W AF=W PF=W CF=W"],
    ["add"              , "r32/m32, ib"                              , "MI"   , "83 /0 ib"                         , "ANY LOCK OF=W SF=W ZF=W AF=W PF=W CF=W"],
    ["add"              , "r64/m64, ib"                              , "MI"   , "REX.W 83 /0 ib"                   , "X64 LOCK OF=W SF=W ZF=W AF=W PF=W CF=W"],

    ["add"              , "r8/m8, r8"                                , "MR"   , "00 /r"                            , "ANY LOCK OF=W SF=W ZF=W AF=W PF=W CF=W"],
    ["add"              , "r16/m16, r16"                             , "MR"   , "66 01 /r"                         , "ANY LOCK OF=W SF=W ZF=W AF=W PF=W CF=W"],
    ["add"              , "r32/m32, r32"                             , "MR"   , "01 /r"                            , "ANY LOCK OF=W SF=W ZF=W AF=W PF=W CF=W"],
    ["add"              , "r64/m64, r64"                             , "MR"   , "REX.W 01 /r"                      , "X64 LOCK OF=W SF=W ZF=W AF=W PF=W CF=W"],

    ["add"              , "r8, r8/m8"                                , "RM"   , "02 /r"                            , "ANY LOCK OF=W SF=W ZF=W AF=W PF=W CF=W"],
    ["add"              , "r16, r16/m16"                             , "RM"   , "66 03 /r"                         , "ANY LOCK OF=W SF=W ZF=W AF=W PF=W CF=W"],
    ["add"              , "r32, r32/m32"                             , "RM"   , "03 /r"                            , "ANY LOCK OF=W SF=W ZF=W AF=W PF=W CF=W"],
    ["add"              , "r64, r64/m64"                             , "RM"   , "REX.W 03 /r"                      , "X64 LOCK OF=W SF=W ZF=W AF=W PF=W CF=W"],

    ["and"              , "al, ib"                                   , "I"    , "24 ib"                            , "ANY LOCK OF=0 SF=W ZF=W AF=U PF=W CF=0"],
    ["and"              , "ax, iw"                                   , "I"    , "66 25 iw"                         , "ANY LOCK OF=0 SF=W ZF=W AF=U PF=W CF=0"],
    ["and"              , "eax, id"                                  , "I"    , "25 id"                            , "ANY LOCK OF=0 SF=W ZF=W AF=U PF=W CF=0"],
    ["and"              , "rax, id"                                  , "I"    , "REX.W 25 id"                      , "X64 LOCK OF=0 SF=W ZF=W AF=U PF=W CF=0"],

    ["and"              , "r8/m8, ib"                                , "MI"   , "80 /4 ib"                         , "ANY LOCK OF=0 SF=W ZF=W AF=U PF=W CF=0"],
    ["and"              , "r16/m16, iw"                              , "MI"   , "66 81 /4 iw"                      , "ANY LOCK OF=0 SF=W ZF=W AF=U PF=W CF=0"],
    ["and"              , "r32/m32, id"                              , "MI"   , "81 /4 id"                         , "ANY LOCK OF=0 SF=W ZF=W AF=U PF=W CF=0"],
    ["and"              , "r64/m64, id"                              , "MI"   , "REX.W 81 /4 id"                   , "X64 LOCK OF=0 SF=W ZF=W AF=U PF=W CF=0"],

    ["and"              , "r16/m16, ib"                              , "MI"   , "66 83 /4 ib"                      , "ANY LOCK OF=0 SF=W ZF=W AF=U PF=W CF=0"],
    ["and"              , "r32/m32, ib"                              , "MI"   , "83 /4 ib"                         , "ANY LOCK OF=0 SF=W ZF=W AF=U PF=W CF=0"],
    ["and"              , "r64/m64, ib"                              , "MI"   , "REX.W 83 /4 ib"                   , "X64 LOCK OF=0 SF=W ZF=W AF=U PF=W CF=0"],

    ["and"              , "r8/m8, r8"                                , "MR"   , "20 /r"                            , "ANY LOCK OF=0 SF=W ZF=W AF=U PF=W CF=0"],
    ["and"              , "r16/m16, r16"                             , "MR"   , "66 21 /r"                         , "ANY LOCK OF=0 SF=W ZF=W AF=U PF=W CF=0"],
    ["and"              , "r32/m32, r32"                             , "MR"   , "21 /r"                            , "ANY LOCK OF=0 SF=W ZF=W AF=U PF=W CF=0"],
    ["and"              , "r64/m64, r64"                             , "MR"   , "REX.W 21 /r"                      , "X64 LOCK OF=0 SF=W ZF=W AF=U PF=W CF=0"],

    ["and"              , "r8, r8/m8"                                , "RM"   , "22 /r"                            , "ANY LOCK OF=0 SF=W ZF=W AF=U PF=W CF=0"],
    ["and"              , "r16, r16/m16"                             , "RM"   , "66 23 /r"                         , "ANY LOCK OF=0 SF=W ZF=W AF=U PF=W CF=0"],
    ["and"              , "r32, r32/m32"                             , "RM"   , "23 /r"                            , "ANY LOCK OF=0 SF=W ZF=W AF=U PF=W CF=0"],
    ["and"              , "r64, r64/m64"                             , "RM"   , "REX.W 23 /r"                      , "X64 LOCK OF=0 SF=W ZF=W AF=U PF=W CF=0"],

    ["bsf"              , "r16, r16/m16"                             , "RM"   , "66 0F BC /r"                      , "ANY OF=U SF=U ZF=W AF=U PF=U CF=U"],
    ["bsf"              , "r32, r32/m32"                             , "RM"   , "0F BC /r"                         , "ANY OF=U SF=U ZF=W AF=U PF=U CF=U"],
    ["bsf"              , "r64, r64/m64"                             , "RM"   , "REX.W 0F BC /r"                   , "X64 OF=U SF=U ZF=W AF=U PF=U CF=U"],

    ["bsr"              , "r16, r16/m16"                             , "RM"   , "66 0F BD /r"                      , "ANY OF=U SF=U ZF=W AF=U PF=U CF=U"],
    ["bsr"              , "r32, r32/m32"                             , "RM"   , "0F BD /r"                         , "ANY OF=U SF=U ZF=W AF=U PF=U CF=U"],
    ["bsr"              , "r64, r64/m64"                             , "RM"   , "REX.W 0F BD /r"                   , "X64 OF=U SF=U ZF=W AF=U PF=U CF=U"],

    ["bswap"            , "r32"                                      , "O"    , "0F C8+r"                          , "ANY"],
    ["bswap"            , "r64"                                      , "O"    , "REX.W 0F C8+r"                    , "X64"],

    ["bt"               , "R:r16/m16, r16"                           , "MR"   , "66 0F A3 /r"                      , "ANY OF=U SF=U AF=U PF=U CF=W"],
    ["bt"               , "R:r32/m32, r32"                           , "MR"   , "0F A3 /r"                         , "ANY OF=U SF=U AF=U PF=U CF=W"],
    ["bt"               , "R:r64/m64, r64"                           , "MR"   , "REX.W 0F A3 /r"                   , "X64 OF=U SF=U AF=U PF=U CF=W"],

    ["bt"               , "R:r16/m16, ib"                            , "MI"   , "66 0F BA /4 ib"                   , "ANY OF=U SF=U AF=U PF=U CF=W"],
    ["bt"               , "R:r32/m32, ib"                            , "MI"   , "0F BA /4 ib"                      , "ANY OF=U SF=U AF=U PF=U CF=W"],
    ["bt"               , "R:r64/m64, ib"                            , "MI"   , "REX.W 0F BA /4 ib"                , "X64 OF=U SF=U AF=U PF=U CF=W"],

    ["btc"              , "r16/m16, r16"                             , "MR"   , "66 0F BB /r"                      , "ANY LOCK OF=U SF=U AF=U PF=U CF=W"],
    ["btc"              , "r32/m32, r32"                             , "MR"   , "0F BB /r"                         , "ANY LOCK OF=U SF=U AF=U PF=U CF=W"],
    ["btc"              , "r64/m64, r64"                             , "MR"   , "REX.W 0F BB /r"                   , "X64 LOCK OF=U SF=U AF=U PF=U CF=W"],

    ["btc"              , "r16/m16, ib"                              , "MI"   , "66 0F BA /7 ib"                   , "ANY LOCK OF=U SF=U AF=U PF=U CF=W"],
    ["btc"              , "r32/m32, ib"                              , "MI"   , "0F BA /7 ib"                      , "ANY LOCK OF=U SF=U AF=U PF=U CF=W"],
    ["btc"              , "r64/m64, ib"                              , "MI"   , "REX.W 0F BA /7 ib"                , "X64 LOCK OF=U SF=U AF=U PF=U CF=W"],

    ["btr"              , "r16/m16, r16"                             , "MR"   , "66 0F B3 /r"                      , "ANY LOCK OF=U SF=U AF=U PF=U CF=W"],
    ["btr"              , "r32/m32, r32"                             , "MR"   , "0F B3 /r"                         , "ANY LOCK OF=U SF=U AF=U PF=U CF=W"],
    ["btr"              , "r64/m64, r64"                             , "MR"   , "REX.W 0F B3 /r"                   , "X64 LOCK OF=U SF=U AF=U PF=U CF=W"],

    ["btr"              , "r16/m16, ib"                              , "MI"   , "66 0F BA /6 ib"                   , "ANY LOCK OF=U SF=U AF=U PF=U CF=W"],
    ["btr"              , "r32/m32, ib"                              , "MI"   , "0F BA /6 ib"                      , "ANY LOCK OF=U SF=U AF=U PF=U CF=W"],
    ["btr"              , "r64/m64, ib"                              , "MI"   , "REX.W 0F BA /6 ib"                , "X64 LOCK OF=U SF=U AF=U PF=U CF=W"],

    ["bts"              , "r16/m16, r16"                             , "MR"   , "66 0F AB /r"                      , "ANY LOCK OF=U SF=U AF=U PF=U CF=W"],
    ["bts"              , "r32/m32, r32"                             , "MR"   , "0F AB /r"                         , "ANY LOCK OF=U SF=U AF=U PF=U CF=W"],
    ["bts"              , "r64/m64, r64"                             , "MR"   , "REX.W 0F AB /r"                   , "X64 LOCK OF=U SF=U AF=U PF=U CF=W"],

    ["bts"              , "r16/m16, ib"                              , "MI"   , "66 0F BA /5 ib"                   , "ANY LOCK OF=U SF=U AF=U PF=U CF=W"],
    ["bts"              , "r32/m32, ib"                              , "MI"   , "0F BA /5 ib"                      , "ANY LOCK OF=U SF=U AF=U PF=U CF=W"],
    ["bts"              , "r64/m64, ib"                              , "MI"   , "REX.W 0F BA /5 ib"                , "X64 LOCK OF=U SF=U AF=U PF=U CF=W"],

    ["call"             , "rel32"                                    , "M"    , "E8 cd"                            , "ANY VOLATILE OF=U SF=U ZF=U AF=U PF=U CF=U"],
    ["call"             , "R:r32/m32"                                , "M"    , "FF /2"                            , "X86 VOLATILE OF=U SF=U ZF=U AF=U PF=U CF=U"],
    ["call"             , "R:r64/m64"                                , "M"    , "FF /2"                            , "X64 VOLATILE OF=U SF=U ZF=U AF=U PF=U CF=U"],

    ["cbw"              , "<ax>"                                     , "NONE" , "66 98"                            , "ANY"],
    ["cwde"             , "<eax>"                                    , "NONE" , "98"                               , "ANY"],
    ["cdqe"             , "<rax>"                                    , "NONE" , "REX.W 98"                         , "X64"],

    ["cwd"              , "W:<dx>, <ax>"                             , "NONE" , "66 99"                            , "ANY"],
    ["cdq"              , "W:<edx>, <eax>"                           , "NONE" , "99"                               , "ANY"],
    ["cqo"              , "W:<rdx>, <rax>"                           , "NONE" , "REX.W 99"                         , "X64"],

    ["clac"             , ""                                         , "NONE" , "0F 01 CA"                         , "ANY VOLATILE SMAP AC=0"],
    ["clc"              , ""                                         , "NONE" , "F8"                               , "ANY CF=0"],
    ["cld"              , ""                                         , "NONE" , "FC"                               , "ANY DF=0"],
    ["cmc"              , ""                                         , "NONE" , "F5"                               , "ANY CF=X"],

    ["cmovo"            , "r16, r16/m16"                             , "RM"   , "66 0F 40 /r"                      , "ANY CMOV OF=R"],
    ["cmovo"            , "r32, r32/m32"                             , "RM"   , "0F 40 /r"                         , "ANY CMOV OF=R"],
    ["cmovo"            , "r64, r64/m64"                             , "RM"   , "REX.W 0F 40 /r"                   , "X64 CMOV OF=R"],
    ["cmovno"           , "r16, r16/m16"                             , "RM"   , "66 0F 41 /r"                      , "ANY CMOV OF=R"],
    ["cmovno"           , "r32, r32/m32"                             , "RM"   , "0F 41 /r"                         , "ANY CMOV OF=R"],
    ["cmovno"           , "r64, r64/m64"                             , "RM"   , "REX.W 0F 41 /r"                   , "X64 CMOV OF=R"],
    ["cmovc"            , "r16, r16/m16"                             , "RM"   , "66 0F 42 /r"                      , "ANY CMOV CF=R"],
    ["cmovc"            , "r32, r32/m32"                             , "RM"   , "0F 42 /r"                         , "ANY CMOV CF=R"],
    ["cmovc"            , "r64, r64/m64"                             , "RM"   , "REX.W 0F 42 /r"                   , "X64 CMOV CF=R"],
    ["cmovnc"           , "r16, r16/m16"                             , "RM"   , "66 0F 43 /r"                      , "ANY CMOV CF=R"],
    ["cmovnc"           , "r32, r32/m32"                             , "RM"   , "0F 43 /r"                         , "ANY CMOV CF=R"],
    ["cmovnc"           , "r64, r64/m64"                             , "RM"   , "REX.W 0F 43 /r"                   , "X64 CMOV CF=R"],
    ["cmovb/cmovnae"    , "r16, r16/m16"                             , "RM"   , "66 0F 42 /r"                      , "ANY CMOV CF=R"],
    ["cmovb/cmovnae"    , "r32, r32/m32"                             , "RM"   , "0F 42 /r"                         , "ANY CMOV CF=R"],
    ["cmovb/cmovnae"    , "r64, r64/m64"                             , "RM"   , "REX.W 0F 42 /r"                   , "X64 CMOV CF=R"],
    ["cmovae/cmovnb"    , "r16, r16/m16"                             , "RM"   , "66 0F 43 /r"                      , "ANY CMOV CF=R"],
    ["cmovae/cmovnb"    , "r32, r32/m32"                             , "RM"   , "0F 43 /r"                         , "ANY CMOV CF=R"],
    ["cmovae/cmovnb"    , "r64, r64/m64"                             , "RM"   , "REX.W 0F 43 /r"                   , "X64 CMOV CF=R"],
    ["cmove/cmovz"      , "r16, r16/m16"                             , "RM"   , "66 0F 44 /r"                      , "ANY CMOV ZF=R"],
    ["cmove/cmovz"      , "r32, r32/m32"                             , "RM"   , "0F 44 /r"                         , "ANY CMOV ZF=R"],
    ["cmove/cmovz"      , "r64, r64/m64"                             , "RM"   , "REX.W 0F 44 /r"                   , "X64 CMOV ZF=R"],
    ["cmovne/cmovnz"    , "r16, r16/m16"                             , "RM"   , "66 0F 45 /r"                      , "ANY CMOV ZF=R"],
    ["cmovne/cmovnz"    , "r32, r32/m32"                             , "RM"   , "0F 45 /r"                         , "ANY CMOV ZF=R"],
    ["cmovne/cmovnz"    , "r64, r64/m64"                             , "RM"   , "REX.W 0F 45 /r"                   , "X64 CMOV ZF=R"],
    ["cmovbe/cmovna"    , "r16, r16/m16"                             , "RM"   , "66 0F 46 /r"                      , "ANY CMOV CF=R ZF=R"],
    ["cmovbe/cmovna"    , "r32, r32/m32"                             , "RM"   , "0F 46 /r"                         , "ANY CMOV CF=R ZF=R"],
    ["cmovbe/cmovna"    , "r64, r64/m64"                             , "RM"   , "REX.W 0F 46 /r"                   , "X64 CMOV CF=R ZF=R"],
    ["cmova/cmovnbe"    , "r16, r16/m16"                             , "RM"   , "66 0F 47 /r"                      , "ANY CMOV CF=R ZF=R"],
    ["cmova/cmovnbe"    , "r32, r32/m32"                             , "RM"   , "0F 47 /r"                         , "ANY CMOV CF=R ZF=R"],
    ["cmova/cmovnbe"    , "r64, r64/m64"                             , "RM"   , "REX.W 0F 47 /r"                   , "X64 CMOV CF=R ZF=R"],
    ["cmovs"            , "r16, r16/m16"                             , "RM"   , "66 0F 48 /r"                      , "ANY CMOV SF=R"],
    ["cmovs"            , "r32, r32/m32"                             , "RM"   , "0F 48 /r"                         , "ANY CMOV SF=R"],
    ["cmovs"            , "r64, r64/m64"                             , "RM"   , "REX.W 0F 48 /r"                   , "X64 CMOV SF=R"],
    ["cmovns"           , "r16, r16/m16"                             , "RM"   , "66 0F 49 /r"                      , "ANY CMOV SF=R"],
    ["cmovns"           , "r32, r32/m32"                             , "RM"   , "0F 49 /r"                         , "ANY CMOV SF=R"],
    ["cmovns"           , "r64, r64/m64"                             , "RM"   , "REX.W 0F 49 /r"                   , "X64 CMOV SF=R"],
    ["cmovp/cmovpe"     , "r16, r16/m16"                             , "RM"   , "66 0F 4A /r"                      , "ANY CMOV PF=R"],
    ["cmovp/cmovpe"     , "r32, r32/m32"                             , "RM"   , "0F 4A /r"                         , "ANY CMOV PF=R"],
    ["cmovp/cmovpe"     , "r64, r64/m64"                             , "RM"   , "REX.W 0F 4A /r"                   , "X64 CMOV PF=R"],
    ["cmovnp/cmovpo"    , "r16, r16/m16"                             , "RM"   , "66 0F 4B /r"                      , "ANY CMOV PF=R"],
    ["cmovnp/cmovpo"    , "r32, r32/m32"                             , "RM"   , "0F 4B /r"                         , "ANY CMOV PF=R"],
    ["cmovnp/cmovpo"    , "r64, r64/m64"                             , "RM"   , "REX.W 0F 4B /r"                   , "X64 CMOV PF=R"],
    ["cmovl/cmovnge"    , "r16, r16/m16"                             , "RM"   , "66 0F 4C /r"                      , "ANY CMOV SF=R OF=R"],
    ["cmovl/cmovnge"    , "r32, r32/m32"                             , "RM"   , "0F 4C /r"                         , "ANY CMOV SF=R OF=R"],
    ["cmovl/cmovnge"    , "r64, r64/m64"                             , "RM"   , "REX.W 0F 4C /r"                   , "X64 CMOV SF=R OF=R"],
    ["cmovge/cmovnl"    , "r16, r16/m16"                             , "RM"   , "66 0F 4D /r"                      , "ANY CMOV SF=R OF=R"],
    ["cmovge/cmovnl"    , "r32, r32/m32"                             , "RM"   , "0F 4D /r"                         , "ANY CMOV SF=R OF=R"],
    ["cmovge/cmovnl"    , "r64, r64/m64"                             , "RM"   , "REX.W 0F 4D /r"                   , "X64 CMOV SF=R OF=R"],
    ["cmovle/cmovng"    , "r16, r16/m16"                             , "RM"   , "66 0F 4E /r"                      , "ANY CMOV ZF=R SF=R OF=R"],
    ["cmovle/cmovng"    , "r32, r32/m32"                             , "RM"   , "0F 4E /r"                         , "ANY CMOV ZF=R SF=R OF=R"],
    ["cmovle/cmovng"    , "r64, r64/m64"                             , "RM"   , "REX.W 0F 4E /r"                   , "X64 CMOV ZF=R SF=R OF=R"],
    ["cmovg/cmovnle"    , "r16, r16/m16"                             , "RM"   , "66 0F 4F /r"                      , "ANY CMOV ZF=R SF=R OF=R"],
    ["cmovg/cmovnle"    , "r32, r32/m32"                             , "RM"   , "0F 4F /r"                         , "ANY CMOV ZF=R SF=R OF=R"],
    ["cmovg/cmovnle"    , "r64, r64/m64"                             , "RM"   , "REX.W 0F 4F /r"                   , "X64 CMOV ZF=R SF=R OF=R"],

    ["cmp"              , "R:al, ib"                                 , "I"    , "3C ib"                            , "ANY OF=W SF=W ZF=W AF=W PF=W CF=W"],
    ["cmp"              , "R:ax, iw"                                 , "I"    , "66 3D iw"                         , "ANY OF=W SF=W ZF=W AF=W PF=W CF=W"],
    ["cmp"              , "R:eax, id"                                , "I"    , "3D id"                            , "ANY OF=W SF=W ZF=W AF=W PF=W CF=W"],
    ["cmp"              , "R:rax, id"                                , "I"    , "REX.W 3D id"                      , "X64 OF=W SF=W ZF=W AF=W PF=W CF=W"],

    ["cmp"              , "R:r8/m8, ib"                              , "MI"   , "80 /7 ib"                         , "ANY OF=W SF=W ZF=W AF=W PF=W CF=W"],
    ["cmp"              , "R:r16/m16, iw"                            , "MI"   , "66 81 /7 iw"                      , "ANY OF=W SF=W ZF=W AF=W PF=W CF=W"],
    ["cmp"              , "R:r32/m32, id"                            , "MI"   , "81 /7 id"                         , "ANY OF=W SF=W ZF=W AF=W PF=W CF=W"],
    ["cmp"              , "R:r64/m64, id"                            , "MI"   , "REX.W 81 /7 id"                   , "X64 OF=W SF=W ZF=W AF=W PF=W CF=W"],

    ["cmp"              , "R:r16/m16, ib"                            , "MI"   , "66 83 /7 ib"                      , "ANY OF=W SF=W ZF=W AF=W PF=W CF=W"],
    ["cmp"              , "R:r32/m32, ib"                            , "MI"   , "83 /7 ib"                         , "ANY OF=W SF=W ZF=W AF=W PF=W CF=W"],
    ["cmp"              , "R:r64/m64, ib"                            , "MI"   , "REX.W 83 /7 ib"                   , "X64 OF=W SF=W ZF=W AF=W PF=W CF=W"],

    ["cmp"              , "R:r8/m8, r8"                              , "MR"   , "38 /r"                            , "ANY OF=W SF=W ZF=W AF=W PF=W CF=W"],
    ["cmp"              , "R:r16/m16, r16"                           , "MR"   , "66 39 /r"                         , "ANY OF=W SF=W ZF=W AF=W PF=W CF=W"],
    ["cmp"              , "R:r32/m32, r32"                           , "MR"   , "39 /r"                            , "ANY OF=W SF=W ZF=W AF=W PF=W CF=W"],
    ["cmp"              , "R:r64/m64, r64"                           , "MR"   , "REX.W 39 /r"                      , "X64 OF=W SF=W ZF=W AF=W PF=W CF=W"],

    ["cmp"              , "R:r8, r8/m8"                              , "RM"   , "3A /r"                            , "ANY OF=W SF=W ZF=W AF=W PF=W CF=W"],
    ["cmp"              , "R:r16, r16/m16"                           , "RM"   , "66 3B /r"                         , "ANY OF=W SF=W ZF=W AF=W PF=W CF=W"],
    ["cmp"              , "R:r32, r32/m32"                           , "RM"   , "3B /r"                            , "ANY OF=W SF=W ZF=W AF=W PF=W CF=W"],
    ["cmp"              , "R:r64, r64/m64"                           , "RM"   , "REX.W 3B /r"                      , "X64 OF=W SF=W ZF=W AF=W PF=W CF=W"],

    ["cmpsb"            , "X:<ds:zsi>, X:<es:zdi>"                   , "NONE" , "A6"                               , "ANY REP OF=W SF=W ZF=W AF=W PF=W CF=W DF=R"],
    ["cmpsw"            , "X:<ds:zsi>, X:<es:zdi>"                   , "NONE" , "66 A7"                            , "ANY REP OF=W SF=W ZF=W AF=W PF=W CF=W DF=R"],
    ["cmpsd"            , "X:<ds:zsi>, X:<es:zdi>"                   , "NONE" , "A7"                               , "ANY REP OF=W SF=W ZF=W AF=W PF=W CF=W DF=R"],
    ["cmpsq"            , "X:<ds:zsi>, X:<es:zdi>"                   , "NONE" , "REX.W A7"                         , "X64 REP OF=W SF=W ZF=W AF=W PF=W CF=W DF=R"],

    ["cmpxchg"          , "r8/m8, r8, <al>"                          , "MR"   , "0F B0 /r"                         , "ANY I486 LOCK OF=W SF=W ZF=W AF=W PF=W CF=W"],
    ["cmpxchg"          , "r16/m16, r16, <ax>"                       , "MR"   , "66 0F B1 /r"                      , "ANY I486 LOCK OF=W SF=W ZF=W AF=W PF=W CF=W"],
    ["cmpxchg"          , "r32/m32, r32, <eax>"                      , "MR"   , "0F B1 /r"                         , "ANY I486 LOCK OF=W SF=W ZF=W AF=W PF=W CF=W"],
    ["cmpxchg"          , "r64/m64, r64, <rax>"                      , "MR"   , "REX.W 0F B1 /r"                   , "X64 I486 LOCK OF=W SF=W ZF=W AF=W PF=W CF=W"],

    ["cmpxchg8b"        , "X:m64, X:<edx>, X:<eax>, <ecx>, <ebx>"    , "M"    , "0F C7 /1"                         , "ANY VOLATILE CMPXCHG8B LOCK ZF=W"],
    ["cmpxchg16b"       , "X:m128, X:<rdx>, X:<rax>, <rcx>, <rbx>"   , "M"    , "REX.W 0F C7 /1"                   , "X64 VOLATILE CMPXCHG16B LOCK ZF=W"],

    ["cpuid"            , "X:<eax>, W:<ebx>, X:<ecx>, W:<edx>"       , "NONE" , "0F A2"                            , "ANY VOLATILE I486"],

    ["daa"              , "<ax>"                                     , "NONE" , "27"                               , "X86 OF=U SF=W ZF=W AF=W PF=W CF=W"],
    ["das"              , "<ax>"                                     , "NONE" , "2F"                               , "X86 OF=U SF=W ZF=W AF=W PF=W CF=W"],

    ["dec"              , "r16"                                      , "O"    , "66 48+r"                          , "X86 LOCK OF=W SF=W ZF=W AF=W PF=W"],
    ["dec"              , "r32"                                      , "O"    , "48+r"                             , "X86 LOCK OF=W SF=W ZF=W AF=W PF=W"],
    ["dec"              , "r8/m8"                                    , "M"    , "FE /1"                            , "ANY LOCK OF=W SF=W ZF=W AF=W PF=W"],
    ["dec"              , "r16/m16"                                  , "M"    , "66 FF /1"                         , "ANY LOCK OF=W SF=W ZF=W AF=W PF=W"],
    ["dec"              , "r32/m32"                                  , "M"    , "FF /1"                            , "ANY LOCK OF=W SF=W ZF=W AF=W PF=W"],
    ["dec"              , "r64/m64"                                  , "M"    , "REX.W FF /1"                      , "X64 LOCK OF=W SF=W ZF=W AF=W PF=W"],

    ["div"              , "X:<ax>, r8/m8"                            , "M"    , "F6 /6"                            , "ANY OF=U SF=U ZF=U AF=U PF=U CF=U"],
    ["div"              , "X:<ax>, X:<dx>, r16/m16"                  , "M"    , "66 F7 /6"                         , "ANY OF=U SF=U ZF=U AF=U PF=U CF=U"],
    ["div"              , "X:<eax>, X:<edx>, r32/m32"                , "M"    , "F7 /6"                            , "ANY OF=U SF=U ZF=U AF=U PF=U CF=U"],
    ["div"              , "X:<rax>, X:<rdx>, r64/m64"                , "M"    , "REX.W F7 /6"                      , "X64 OF=U SF=U ZF=U AF=U PF=U CF=U"],

    ["enter"            , "iw, ib"                                   , "II"   , "C8 iw ib"                         , "ANY VOLATILE"],

    ["idiv"             , "X:<ax>, r8/m8"                            , "M"    , "F6 /7"                            , "ANY OF=U SF=U ZF=U AF=U PF=U CF=U"],
    ["idiv"             , "X:<dx>, X:<ax>, r16/m16"                  , "M"    , "66 F7 /7"                         , "ANY OF=U SF=U ZF=U AF=U PF=U CF=U"],
    ["idiv"             , "X:<edx>, X:<eax>, r32/m32"                , "M"    , "F7 /7"                            , "ANY OF=U SF=U ZF=U AF=U PF=U CF=U"],
    ["idiv"             , "X:<rdx>, X:<rax>, r64/m64"                , "M"    , "REX.W F7 /7"                      , "X64 OF=U SF=U ZF=U AF=U PF=U CF=U"],

    ["imul"             , "X:<ax>, r8/m8"                            , "M"    , "F6 /5"                            , "ANY OF=W SF=W ZF=U AF=U PF=U CF=W"],
    ["imul"             , "W:<dx>, X:<ax>, r16/m16"                  , "M"    , "66 F7 /5"                         , "ANY OF=W SF=W ZF=U AF=U PF=U CF=W"],
    ["imul"             , "W:<edx>, X:<eax>, r32/m32"                , "M"    , "F7 /5"                            , "ANY OF=W SF=W ZF=U AF=U PF=U CF=W"],
    ["imul"             , "W:<rdx>, X:<rax>, r64/m64"                , "M"    , "REX.W F7 /5"                      , "X64 OF=W SF=W ZF=U AF=U PF=U CF=W"],
    ["imul"             , "r16, r16/m16"                             , "RM"   , "66 0F AF /r"                      , "ANY OF=W SF=W ZF=U AF=U PF=U CF=W"],
    ["imul"             , "r32, r32/m32"                             , "RM"   , "0F AF /r"                         , "ANY OF=W SF=W ZF=U AF=U PF=U CF=W"],
    ["imul"             , "r64, r64/m64"                             , "RM"   , "REX.W 0F AF /r"                   , "X64 OF=W SF=W ZF=U AF=U PF=U CF=W"],
    ["imul"             , "W:r16, r16/m16, ib"                       , "RMI"  , "66 6B /r ib"                      , "ANY OF=W SF=W ZF=U AF=U PF=U CF=W"],
    ["imul"             , "W:r32, r32/m32, ib"                       , "RMI"  , "6B /r ib"                         , "ANY OF=W SF=W ZF=U AF=U PF=U CF=W"],
    ["imul"             , "W:r64, r64/m64, ib"                       , "RMI"  , "REX.W 6B /r ib"                   , "X64 OF=W SF=W ZF=U AF=U PF=U CF=W"],
    ["imul"             , "W:r16, r16/m16, iw"                       , "RMI"  , "66 69 /r iw"                      , "ANY OF=W SF=W ZF=U AF=U PF=U CF=W"],
    ["imul"             , "W:r32, r32/m32, id"                       , "RMI"  , "69 /r id"                         , "ANY OF=W SF=W ZF=U AF=U PF=U CF=W"],
    ["imul"             , "W:r64, r64/m64, id"                       , "RMI"  , "REX.W 69 /r id"                   , "X64 OF=W SF=W ZF=U AF=U PF=U CF=W"],

    ["inc"              , "r16"                                      , "O"    , "66 40+r"                          , "X86 LOCK OF=W SF=W ZF=W AF=W PF=W"],
    ["inc"              , "r32"                                      , "O"    , "40+r"                             , "X86 LOCK OF=W SF=W ZF=W AF=W PF=W"],
    ["inc"              , "r8/m8"                                    , "M"    , "FE /0"                            , "ANY LOCK OF=W SF=W ZF=W AF=W PF=W"],
    ["inc"              , "r16/m16"                                  , "M"    , "66 FF /0"                         , "ANY LOCK OF=W SF=W ZF=W AF=W PF=W"],
    ["inc"              , "r32/m32"                                  , "M"    , "FF /0"                            , "ANY LOCK OF=W SF=W ZF=W AF=W PF=W"],
    ["inc"              , "r64/m64"                                  , "M"    , "REX.W FF /0"                      , "X64 LOCK OF=W SF=W ZF=W AF=W PF=W"],

    ["in"               , "W:al, ib"                                 , "I"    , "E4 ib"                            , "ANY VOLATILE"],
    ["in"               , "W:ax, ib"                                 , "I"    , "66 E5 ib"                         , "ANY VOLATILE"],
    ["in"               , "W:eax, ib"                                , "I"    , "E5 ib"                            , "ANY VOLATILE"],
    ["in"               , "W:al, dx"                                 , "NONE" , "EC"                               , "ANY VOLATILE"],
    ["in"               , "W:ax, dx"                                 , "NONE" , "66 ED"                            , "ANY VOLATILE"],
    ["in"               , "W:eax, dx"                                , "NONE" , "ED"                               , "ANY VOLATILE"],
    ["insb"             , "W:es:zdi, dx"                             , "NONE" , "6C"                               , "ANY VOLATILE REP"],
    ["insw"             , "W:es:zdi, dx"                             , "NONE" , "66 6D"                            , "ANY VOLATILE REP"],
    ["insd"             , "W:es:zdi, dx"                             , "NONE" , "6D"                               , "ANY VOLATILE REP"],

    ["int"              , "ib"                                       , "I"    , "CD ib"                            , "ANY VOLATILE"],
    ["int3"             , ""                                         , "NONE" , "CC"                               , "ANY VOLATILE"],
    ["into"             , ""                                         , "NONE" , "CE"                               , "ANY VOLATILE"],

    ["jo"               , "rel8"                                     , "D"    , "70 cb"                            , "ANY VOLATILE OF=R"],
    ["jno"              , "rel8"                                     , "D"    , "71 cb"                            , "ANY VOLATILE OF=R"],
    ["jb/jnae/jc"       , "rel8"                                     , "D"    , "72 cb"                            , "ANY VOLATILE CF=R"],
    ["jae/jnb/jnc"      , "rel8"                                     , "D"    , "73 cb"                            , "ANY VOLATILE CF=R"],
    ["je/jz"            , "rel8"                                     , "D"    , "74 cb"                            , "ANY VOLATILE ZF=R"],
    ["jne/jnz"          , "rel8"                                     , "D"    , "75 cb"                            , "ANY VOLATILE ZF=R"],
    ["jbe/jna"          , "rel8"                                     , "D"    , "76 cb"                            , "ANY VOLATILE CF=R ZF=R"],
    ["ja/jnbe"          , "rel8"                                     , "D"    , "77 cb"                            , "ANY VOLATILE CF=R ZF=R"],
    ["js"               , "rel8"                                     , "D"    , "78 cb"                            , "ANY VOLATILE SF=R"],
    ["jns"              , "rel8"                                     , "D"    , "79 cb"                            , "ANY VOLATILE SF=R"],
    ["jp/jpe"           , "rel8"                                     , "D"    , "7A cb"                            , "ANY VOLATILE PF=R"],
    ["jnp/jpo"          , "rel8"                                     , "D"    , "7B cb"                            , "ANY VOLATILE PF=R"],
    ["jl/jnge"          , "rel8"                                     , "D"    , "7C cb"                            , "ANY VOLATILE SF=R OF=R"],
    ["jge/jnl"          , "rel8"                                     , "D"    , "7D cb"                            , "ANY VOLATILE SF=R OF=R"],
    ["jle/jng"          , "rel8"                                     , "D"    , "7E cb"                            , "ANY VOLATILE ZF=R SF=R OF=R"],
    ["jg/jnle"          , "rel8"                                     , "D"    , "7F cb"                            , "ANY VOLATILE ZF=R SF=R OF=R"],

    ["jo"               , "rel32"                                    , "D"    , "0F 80 cd"                         , "ANY VOLATILE OF=R"],
    ["jno"              , "rel32"                                    , "D"    , "0F 81 cd"                         , "ANY VOLATILE OF=R"],
    ["jb/jnae"          , "rel32"                                    , "D"    , "0F 82 cd"                         , "ANY VOLATILE CF=R"],
    ["jae/jnb"          , "rel32"                                    , "D"    , "0F 83 cd"                         , "ANY VOLATILE CF=R"],
    ["je/jz"            , "rel32"                                    , "D"    , "0F 84 cd"                         , "ANY VOLATILE ZF=R"],
    ["jne/jnz"          , "rel32"                                    , "D"    , "0F 85 cd"                         , "ANY VOLATILE ZF=R"],
    ["jbe/jna"          , "rel32"                                    , "D"    , "0F 86 cd"                         , "ANY VOLATILE CF=R ZF=R"],
    ["ja/jnbe"          , "rel32"                                    , "D"    , "0F 87 cd"                         , "ANY VOLATILE CF=R ZF=R"],
    ["js"               , "rel32"                                    , "D"    , "0F 88 cd"                         , "ANY VOLATILE SF=R"],
    ["jns"              , "rel32"                                    , "D"    , "0F 89 cd"                         , "ANY VOLATILE SF=R"],
    ["jp/jpe"           , "rel32"                                    , "D"    , "0F 8A cd"                         , "ANY VOLATILE PF=R"],
    ["jnp/jpo"          , "rel32"                                    , "D"    , "0F 8B cd"                         , "ANY VOLATILE PF=R"],
    ["jl/jnge"          , "rel32"                                    , "D"    , "0F 8C cd"                         , "ANY VOLATILE SF=R OF=R"],
    ["jge/jnl"          , "rel32"                                    , "D"    , "0F 8D cd"                         , "ANY VOLATILE SF=R OF=R"],
    ["jle/jng"          , "rel32"                                    , "D"    , "0F 8E cd"                         , "ANY VOLATILE ZF=R SF=R OF=R"],
    ["jg/jnle"          , "rel32"                                    , "D"    , "0F 8F cd"                         , "ANY VOLATILE ZF=R SF=R OF=R"],

    ["jecxz"            , "R:cx, rel8"                               , "D"    , "67 E3 cb"                         , "X86 VOLATILE"],
    ["jecxz"            , "R:ecx, rel8"                              , "D"    , "E3 cb"                            , "X86 VOLATILE"],
    ["jecxz"            , "R:ecx, rel8"                              , "D"    , "67 E3 cb"                         , "X64 VOLATILE"],
    ["jecxz"            , "R:rcx, rel8"                              , "D"    , "E3 cb"                            , "X64 VOLATILE"],

    ["jmp"              , "rel8"                                     , "D"    , "EB cb"                            , "ANY VOLATILE"],
    ["jmp"              , "rel32"                                    , "D"    , "E9 cd"                            , "ANY VOLATILE"],
    ["jmp"              , "R:r32/m32"                                , "D"    , "FF /4"                            , "X86 VOLATILE"],
    ["jmp"              , "R:r64/m64"                                , "D"    , "FF /4"                            , "X64 VOLATILE"],

    ["lahf"             , "W:<ah>"                                   , "NONE" , "9F"                               , "ANY VOLATILE LAHFSAHF SF=R ZF=R AF=R PF=R CF=R"],

    ["lea"              , "W:r16, mem"                               , "RM"   , "67 8D /r"                         , "ANY"],
    ["lea"              , "W:r32, mem"                               , "RM"   , "8D /r"                            , "ANY"],
    ["lea"              , "W:r64, mem"                               , "RM"   , "REX.W 8D /r"                      , "X64"],

    ["leave"            , ""                                         , "NONE" , "C9"                               , "ANY VOLATILE"],

    ["lodsb"            , "W:<al>, X:<ds:zsi>"                       , "NONE" , "AC"                               , "ANY REP DF=R"],
    ["lodsw"            , "W:<ax>, X:<ds:zsi>"                       , "NONE" , "66 AD"                            , "ANY REP DF=R"],
    ["lodsd"            , "W:<eax>, X:<ds:zsi>"                      , "NONE" , "AD"                               , "ANY REP DF=R"],
    ["lodsq"            , "W:<rax>, X:<ds:zsi>"                      , "NONE" , "REX.W AD"                         , "X64 REP DF=R"],

    ["loop"             , "X:cx, rel8"                               , "D"    , "67 E2 cb"                         , "X86 VOLATILE"],
    ["loop"             , "X:ecx, rel8"                              , "D"    , "E2 cb"                            , "X86 VOLATILE"],
    ["loop"             , "X:ecx, rel8"                              , "D"    , "67 E2 cb"                         , "X64 VOLATILE"],
    ["loop"             , "X:rcx, rel8"                              , "D"    , "E2 cb"                            , "X64 VOLATILE"],

    ["loope"            , "X:cx, rel8"                               , "D"    , "67 E1 cb"                         , "X86 VOLATILE ZF=R"],
    ["loope"            , "X:ecx, rel8"                              , "D"    , "E1 cb"                            , "X86 VOLATILE ZF=R"],
    ["loope"            , "X:ecx, rel8"                              , "D"    , "67 E1 cb"                         , "X64 VOLATILE ZF=R"],
    ["loope"            , "X:rcx, rel8"                              , "D"    , "E1 cb"                            , "X64 VOLATILE ZF=R"],

    ["loopne"           , "X:cx, rel8"                               , "D"    , "67 E0 cb"                         , "X86 VOLATILE ZF=R"],
    ["loopne"           , "X:ecx, rel8"                              , "D"    , "E0 cb"                            , "X86 VOLATILE ZF=R"],
    ["loopne"           , "X:ecx, rel8"                              , "D"    , "67 E0 cb"                         , "X64 VOLATILE ZF=R"],
    ["loopne"           , "X:rcx, rel8"                              , "D"    , "E0 cb"                            , "X64 VOLATILE ZF=R"],

    ["mov"              , "W:r8/m8, r8"                              , "MR"   , "88 /r"                            , "ANY"],
    ["mov"              , "W:r16/m16, r16"                           , "MR"   , "66 89 /r"                         , "ANY"],
    ["mov"              , "W:r32/m32, r32"                           , "MR"   , "89 /r"                            , "ANY"],
    ["mov"              , "W:r64/m64, r64"                           , "MR"   , "REX.W 89 /r"                      , "X64"],

    ["mov"              , "W:r8, r8/m8"                              , "RM"   , "8A /r"                            , "ANY"],
    ["mov"              , "W:r16, r16/m16"                           , "RM"   , "66 8B /r"                         , "ANY"],
    ["mov"              , "W:r32, r32/m32"                           , "RM"   , "8B /r"                            , "ANY"],
    ["mov"              , "W:r64, r64/m64"                           , "RM"   , "REX.W 8B /r"                      , "X64"],

    ["mov"              , "W:r16/m16, sreg"                          , "MR"   , "8C /r"                            , "ANY"],
    ["mov"              , "W:r64/m64, sreg"                          , "MR"   , "REX.W 8C /r"                      , "X64"],
    ["mov"              , "W:sreg, r16/m16"                          , "RM"   , "8E /r"                            , "ANY"],
    ["mov"              , "W:sreg, r64/m64"                          , "RM"   , "REX.W 8E /r"                      , "X64"],

    ["mov"              , "W:al, moff8"                              , "NONE" , "A0"                               , "ANY"],
    ["mov"              , "W:ax, moff16"                             , "NONE" , "66 A1"                            , "ANY"],
    ["mov"              , "W:eax, moff32"                            , "NONE" , "A1"                               , "ANY"],
    ["mov"              , "W:rax, moff64"                            , "NONE" , "REX.W A1"                         , "X64"],

    ["mov"              , "W:moff8, al"                              , "NONE" , "A2"                               , "ANY"],
    ["mov"              , "W:moff16, ax"                             , "NONE" , "66 A3"                            , "ANY"],
    ["mov"              , "W:moff32, eax"                            , "NONE" , "A3"                               , "ANY"],
    ["mov"              , "W:moff64, rax"                            , "NONE" , "REX.W A3"                         , "X64"],

    ["mov"              , "W:r8, ib"                                 , "I"    , "B0+r ib"                          , "ANY"],
    ["mov"              , "W:r16, iw"                                , "I"    , "66 B8+r iw"                       , "ANY"],
    ["mov"              , "W:r32, id"                                , "I"    , "B8+r id"                          , "ANY"],
    ["mov"              , "W:r64, iq"                                , "I"    , "REX.W B8+r iq"                    , "X64"],

    ["mov"              , "W:r8/m8, ib"                              , "MI"   , "C6 /0 ib"                         , "ANY"],
    ["mov"              , "W:r16/m16, iw"                            , "MI"   , "66 C7 /0 iw"                      , "ANY"],
    ["mov"              , "W:r32/m32, id"                            , "MI"   , "C7 /0 id"                         , "ANY"],
    ["mov"              , "W:r64/m64, id"                            , "MI"   , "REX.W C7 /0 id"                   , "X64"],

    ["mov"              , "W:r32, creg"                              , "MR"   , "0F 20 /r"                         , "X86 OF=U SF=U ZF=U AF=U PF=U CF=U"],
    ["mov"              , "W:r64, creg"                              , "MR"   , "0F 20 /r"                         , "X64 OF=U SF=U ZF=U AF=U PF=U CF=U"],
    ["mov"              , "W:creg, r32"                              , "RM"   , "0F 22 /r"                         , "X86 OF=U SF=U ZF=U AF=U PF=U CF=U"],
    ["mov"              , "W:creg, r64"                              , "RM"   , "0F 22 /r"                         , "X64 OF=U SF=U ZF=U AF=U PF=U CF=U"],

    ["mov"              , "W:r32, dreg"                              , "MR"   , "0F 21 /r"                         , "X86 OF=U SF=U ZF=U AF=U PF=U CF=U"],
    ["mov"              , "W:r64, dreg"                              , "MR"   , "0F 21 /r"                         , "X64 OF=U SF=U ZF=U AF=U PF=U CF=U"],
    ["mov"              , "W:dreg, r32"                              , "RM"   , "0F 23 /r"                         , "X86 OF=U SF=U ZF=U AF=U PF=U CF=U"],
    ["mov"              , "W:dreg, r64"                              , "RM"   , "0F 23 /r"                         , "X64 OF=U SF=U ZF=U AF=U PF=U CF=U"],

    ["movsb"            , "X:<es:zdi>, X:<ds:zsi>"                   , "NONE" , "A4"                               , "ANY REP DF=R"],
    ["movsw"            , "X:<es:zdi>, X:<ds:zsi>"                   , "NONE" , "66 A5"                            , "ANY REP DF=R"],
    ["movsd"            , "X:<es:zdi>, X:<ds:zsi>"                   , "NONE" , "A5"                               , "ANY REP DF=R"],
    ["movsq"            , "X:<es:zdi>, X:<ds:zsi>"                   , "NONE" , "REX.W A5"                         , "X64 REP DF=R"],

    ["movsx"            , "W:r16, r8/m8"                             , "RM"   , "66 0F BE /r"                      , "ANY"],
    ["movsx"            , "W:r32, r8/m8"                             , "RM"   , "0F BE /r"                         , "ANY"],
    ["movsx"            , "W:r64, r8/m8"                             , "RM"   , "REX.W 0F BE /r"                   , "X64"],
    ["movsx"            , "W:r32, r16/m16"                           , "RM"   , "0F BF /r"                         , "ANY"],
    ["movsx"            , "W:r64, r16/m16"                           , "RM"   , "REX.W 0F BF /r"                   , "X64"],
    ["movsxd"           , "W:r64, r32/m32"                           , "RM"   , "REX.W 63 /r"                      , "X64"],

    ["movzx"            , "W:r16, r8/m8"                             , "RM"   , "66 0F B6 /r"                      , "ANY"],
    ["movzx"            , "W:r32, r8/m8"                             , "RM"   , "0F B6 /r"                         , "ANY"],
    ["movzx"            , "W:r64, r8/m8"                             , "RM"   , "REX.W 0F B6 /r"                   , "X64"],
    ["movzx"            , "W:r32, r16/m16"                           , "RM"   , "0F B7 /r"                         , "ANY"],
    ["movzx"            , "W:r64, r16/m16"                           , "RM"   , "REX.W 0F B7 /r"                   , "X64"],

    ["mul"              , "X:<ax>, r8/m8"                            , "M"    , "F6 /4"                            , "ANY OF=W SF=U ZF=U AF=U PF=U CF=W"],
    ["mul"              , "W:<dx>, X:<ax>, r16/m16"                  , "M"    , "66 F7 /4"                         , "ANY OF=W SF=U ZF=U AF=U PF=U CF=W"],
    ["mul"              , "W:<edx>, X:<eax>, r32/m32"                , "M"    , "F7 /4"                            , "ANY OF=W SF=U ZF=U AF=U PF=U CF=W"],
    ["mul"              , "W:<rdx>, X:<rax>, r64/m64"                , "M"    , "REX.W F7 /4"                      , "X64 OF=W SF=U ZF=U AF=U PF=U CF=W"],

    ["neg"              , "r8/m8"                                    , "M"    , "F6 /3"                            , "ANY OF=W SF=W ZF=W AF=W PF=W CF=W"],
    ["neg"              , "r16/m16"                                  , "M"    , "66 F7 /3"                         , "ANY OF=W SF=W ZF=W AF=W PF=W CF=W"],
    ["neg"              , "r32/m32"                                  , "M"    , "F7 /3"                            , "ANY OF=W SF=W ZF=W AF=W PF=W CF=W"],
    ["neg"              , "r64/m64"                                  , "M"    , "REX.W F7 /3"                      , "X64 OF=W SF=W ZF=W AF=W PF=W CF=W"],

    ["nop"              , ""                                         , "NONE" , "90"                               , "ANY"],
    ["nop"              , "R:r16/m16"                                , "M"    , "66 0F 1F /0"                      , "ANY"],
    ["nop"              , "R:r32/m32"                                , "M"    , "0F 1F /0"                         , "ANY"],

    ["not"              , "r8/m8"                                    , "M"    , "F6 /2"                            , "ANY"],
    ["not"              , "r16/m16"                                  , "M"    , "66 F7 /2"                         , "ANY"],
    ["not"              , "r32/m32"                                  , "M"    , "F7 /2"                            , "ANY"],
    ["not"              , "r64/m64"                                  , "M"    , "REX.W F7 /2"                      , "X64"],

    ["or"               , "al, ib"                                   , "I"    , "0C ib"                            , "ANY LOCK OF=0 SF=W ZF=W AF=U PF=W CF=0"],
    ["or"               , "ax, iw"                                   , "I"    , "66 0D iw"                         , "ANY LOCK OF=0 SF=W ZF=W AF=U PF=W CF=0"],
    ["or"               , "eax, id"                                  , "I"    , "0D id"                            , "ANY LOCK OF=0 SF=W ZF=W AF=U PF=W CF=0"],
    ["or"               , "rax, id"                                  , "I"    , "REX.W 0D id"                      , "X64 LOCK OF=0 SF=W ZF=W AF=U PF=W CF=0"],

    ["or"               , "r8/m8, ib"                                , "MI"   , "80 /1 ib"                         , "ANY LOCK OF=0 SF=W ZF=W AF=U PF=W CF=0"],
    ["or"               , "r16/m16, iw"                              , "MI"   , "66 81 /1 iw"                      , "ANY LOCK OF=0 SF=W ZF=W AF=U PF=W CF=0"],
    ["or"               , "r32/m32, id"                              , "MI"   , "81 /1 id"                         , "ANY LOCK OF=0 SF=W ZF=W AF=U PF=W CF=0"],
    ["or"               , "r64/m64, id"                              , "MI"   , "REX.W 81 /1 id"                   , "X64 LOCK OF=0 SF=W ZF=W AF=U PF=W CF=0"],

    ["or"               , "r16/m16, ib"                              , "MI"   , "66 83 /1 ib"                      , "ANY LOCK OF=0 SF=W ZF=W AF=U PF=W CF=0"],
    ["or"               , "r32/m32, ib"                              , "MI"   , "83 /1 ib"                         , "ANY LOCK OF=0 SF=W ZF=W AF=U PF=W CF=0"],
    ["or"               , "r64/m64, ib"                              , "MI"   , "REX.W 83 /1 ib"                   , "X64 LOCK OF=0 SF=W ZF=W AF=U PF=W CF=0"],

    ["or"               , "r8/m8, r8"                                , "MR"   , "08 /r"                            , "ANY LOCK OF=0 SF=W ZF=W AF=U PF=W CF=0"],
    ["or"               , "r16/m16, r16"                             , "MR"   , "66 09 /r"                         , "ANY LOCK OF=0 SF=W ZF=W AF=U PF=W CF=0"],
    ["or"               , "r32/m32, r32"                             , "MR"   , "09 /r"                            , "ANY LOCK OF=0 SF=W ZF=W AF=U PF=W CF=0"],
    ["or"               , "r64/m64, r64"                             , "MR"   , "REX.W 09 /r"                      , "X64 LOCK OF=0 SF=W ZF=W AF=U PF=W CF=0"],

    ["or"               , "r8, r8/m8"                                , "RM"   , "0A /r"                            , "ANY LOCK OF=0 SF=W ZF=W AF=U PF=W CF=0"],
    ["or"               , "r16, r16/m16"                             , "RM"   , "66 0B /r"                         , "ANY LOCK OF=0 SF=W ZF=W AF=U PF=W CF=0"],
    ["or"               , "r32, r32/m32"                             , "RM"   , "0B /r"                            , "ANY LOCK OF=0 SF=W ZF=W AF=U PF=W CF=0"],
    ["or"               , "r64, r64/m64"                             , "RM"   , "REX.W 0B /r"                      , "X64 LOCK OF=0 SF=W ZF=W AF=U PF=W CF=0"],

    ["out"              , "ib, al"                                   , "I"    , "E6 ib"                            , "ANY VOLATILE"],
    ["out"              , "ib, ax"                                   , "I"    , "66 E7 ib"                         , "ANY VOLATILE"],
    ["out"              , "ib, eax"                                  , "I"    , "E7 ib"                            , "ANY VOLATILE"],
    ["out"              , "R:dx, R:al"                               , "NONE" , "EE"                               , "ANY VOLATILE"],
    ["out"              , "R:dx, R:ax"                               , "NONE" , "66 EF"                            , "ANY VOLATILE"],
    ["out"              , "R:dx, R:eax"                              , "NONE" , "EF"                               , "ANY VOLATILE"],
    ["outsb"            , "R:dx, R:ds:zsi"                           , "NONE" , "6E"                               , "ANY VOLATILE REP"],
    ["outsw"            , "R:dx, R:ds:zsi"                           , "NONE" , "66 6F"                            , "ANY VOLATILE REP"],
    ["outsd"            , "R:dx, R:ds:zsi"                           , "NONE" , "6F"                               , "ANY VOLATILE REP"],

    ["pause"            , ""                                         , "NONE" , "F3 90"                            , "ANY VOLATILE"],

    ["pop"              , "W:r16/m16"                                , "M"    , "66 8F /0"                         , "ANY VOLATILE"],
    ["pop"              , "W:r32/m32"                                , "M"    , "8F /0"                            , "X86 VOLATILE"],
    ["pop"              , "W:r64/m64"                                , "M"    , "8F /0"                            , "X64 VOLATILE"],

    ["pop"              , "W:r16"                                    , "O"    , "66 58+r"                          , "ANY VOLATILE"],
    ["pop"              , "W:r32"                                    , "O"    , "58+r"                             , "X86 VOLATILE"],
    ["pop"              , "W:r64"                                    , "O"    , "58+r"                             , "X64 VOLATILE"],

    ["pop"              , "W:ds"                                     , "NONE" , "1F"                               , "X86 VOLATILE"],
    ["pop"              , "W:es"                                     , "NONE" , "07"                               , "X86 VOLATILE"],
    ["pop"              , "W:ss"                                     , "NONE" , "17"                               , "X86 VOLATILE"],
    ["pop"              , "W:fs"                                     , "NONE" , "0F A1"                            , "ANY VOLATILE"],
    ["pop"              , "W:gs"                                     , "NONE" , "0F A9"                            , "ANY VOLATILE"],

    ["popa"             , ""                                         , "NONE" , "61"                               , "X86 VOLATILE"],
    ["popf"             , ""                                         , "NONE" , "9D"                               , "ANY VOLATILE OF=W SF=W ZF=W AF=W PF=W CF=W"],

    ["push"             , "R:r16/m16"                                , "M"    , "66 FF /6"                         , "ANY VOLATILE"],
    ["push"             , "R:r32/m32"                                , "M"    , "FF /6"                            , "X86 VOLATILE"],
    ["push"             , "R:r64/m64"                                , "M"    , "FF /6"                            , "X64 VOLATILE"],

    ["push"             , "R:r16"                                    , "O"    , "66 50+r"                          , "ANY VOLATILE"],
    ["push"             , "R:r32"                                    , "O"    , "50+r"                             , "X86 VOLATILE"],
    ["push"             , "R:r64"                                    , "O"    , "50+r"                             , "X64 VOLATILE"],

    ["push"             , "ib"                                       , "I"    , "6A ib"                            , "ANY VOLATILE"],
    ["push"             , "iw"                                       , "I"    , "66 68 iw"                         , "ANY VOLATILE"],
    ["push"             , "id"                                       , "I"    , "68 id"                            , "ANY VOLATILE"],

    ["push"             , "R:cs"                                     , "NONE" , "0E"                               , "X86 VOLATILE"],
    ["push"             , "R:ss"                                     , "NONE" , "16"                               , "X86 VOLATILE"],
    ["push"             , "R:ds"                                     , "NONE" , "1E"                               , "X86 VOLATILE"],
    ["push"             , "R:es"                                     , "NONE" , "06"                               , "X86 VOLATILE"],
    ["push"             , "R:fs"                                     , "NONE" , "0F A0"                            , "ANY VOLATILE"],
    ["push"             , "R:gs"                                     , "NONE" , "0F A8"                            , "ANY VOLATILE"],

    ["pusha"            , ""                                         , "NONE" , "60"                               , "X86 VOLATILE"],
    ["pushf"            , ""                                         , "NONE" , "9C"                               , "ANY VOLATILE"],

    ["rcl"              , "r8/m8, 1"                                 , "M"    , "D0 /2"                            , "ANY CF=W OF=W"],
    ["rcl"              , "r8/m8, cl"                                , "M"    , "D2 /2"                            , "ANY CF=W OF=W"],
    ["rcl"              , "r8/m8, ib"                                , "MI"   , "C0 /2 ib"                         , "ANY CF=W OF=W"],

    ["rcl"              , "r16/m16, 1"                               , "M"    , "66 D1 /2"                         , "ANY CF=W OF=W"],
    ["rcl"              , "r16/m16, cl"                              , "M"    , "66 D3 /2"                         , "ANY CF=W OF=W"],
    ["rcl"              , "r16/m16, ib"                              , "MI"   , "66 C1 /2 ib"                      , "ANY CF=W OF=W"],

    ["rcl"              , "r32/m32, 1"                               , "M"    , "D1 /2"                            , "ANY CF=W OF=W"],
    ["rcl"              , "r32/m32, cl"                              , "M"    , "D3 /2"                            , "ANY CF=W OF=W"],
    ["rcl"              , "r32/m32, ib"                              , "MI"   , "C1 /2 ib"                         , "ANY CF=W OF=W"],

    ["rcl"              , "r64/m64, 1"                               , "M"    , "REX.W D1 /2"                      , "X64 CF=W OF=W"],
    ["rcl"              , "r64/m64, cl"                              , "M"    , "REX.W D3 /2"                      , "X64 CF=W OF=W"],
    ["rcl"              , "r64/m64, ib"                              , "MI"   , "REX.W C1 /2 ib"                   , "X64 CF=W OF=W"],

    ["rcr"              , "r8/m8, 1"                                 , "M"    , "D0 /3"                            , "ANY CF=W OF=W"],
    ["rcr"              , "r8/m8, cl"                                , "M"    , "D2 /3"                            , "ANY CF=W OF=W"],
    ["rcr"              , "r8/m8, ib"                                , "MI"   , "C0 /3 ib"                         , "ANY CF=W OF=W"],

    ["rcr"              , "r16/m16, 1"                               , "M"    , "66 D1 /3"                         , "ANY CF=W OF=W"],
    ["rcr"              , "r16/m16, cl"                              , "M"    , "66 D3 /3"                         , "ANY CF=W OF=W"],
    ["rcr"              , "r16/m16, ib"                              , "MI"   , "66 C1 /3 ib"                      , "ANY CF=W OF=W"],

    ["rcr"              , "r32/m32, 1"                               , "M"    , "D1 /3"                            , "ANY CF=W OF=W"],
    ["rcr"              , "r32/m32, cl"                              , "M"    , "D3 /3"                            , "ANY CF=W OF=W"],
    ["rcr"              , "r32/m32, ib"                              , "MI"   , "C1 /3 ib"                         , "ANY CF=W OF=W"],

    ["rcr"              , "r64/m64, 1"                               , "M"    , "REX.W D1 /3"                      , "X64 CF=W OF=W"],
    ["rcr"              , "r64/m64, cl"                              , "M"    , "REX.W D3 /3"                      , "X64 CF=W OF=W"],
    ["rcr"              , "r64/m64, ib"                              , "MI"   , "REX.W C1 /3 ib"                   , "X64 CF=W OF=W"],

    ["rol"              , "r8/m8, 1"                                 , "M"    , "D0 /0"                            , "ANY CF=W OF=W"],
    ["rol"              , "r8/m8, cl"                                , "M"    , "D2 /0"                            , "ANY CF=W OF=W"],
    ["rol"              , "r8/m8, ib"                                , "MI"   , "C0 /0 ib"                         , "ANY CF=W OF=W"],

    ["rol"              , "r16/m16, 1"                               , "M"    , "66 D1 /0"                         , "ANY CF=W OF=W"],
    ["rol"              , "r16/m16, cl"                              , "M"    , "66 D3 /0"                         , "ANY CF=W OF=W"],
    ["rol"              , "r16/m16, ib"                              , "MI"   , "66 C1 /0 ib"                      , "ANY CF=W OF=W"],

    ["rol"              , "r32/m32, 1"                               , "M"    , "D1 /0"                            , "ANY CF=W OF=W"],
    ["rol"              , "r32/m32, cl"                              , "M"    , "D3 /0"                            , "ANY CF=W OF=W"],
    ["rol"              , "r32/m32, ib"                              , "MI"   , "C1 /0 ib"                         , "ANY CF=W OF=W"],

    ["rol"              , "r64/m64, 1"                               , "M"    , "REX.W D1 /0"                      , "X64 CF=W OF=W"],
    ["rol"              , "r64/m64, cl"                              , "M"    , "REX.W D3 /0"                      , "X64 CF=W OF=W"],
    ["rol"              , "r64/m64, ib"                              , "MI"   , "REX.W C1 /0 ib"                   , "X64 CF=W OF=W"],

    ["ror"              , "r8/m8, 1"                                 , "M"    , "D0 /1"                            , "ANY CF=W OF=W"],
    ["ror"              , "r8/m8, cl"                                , "M"    , "D2 /1"                            , "ANY CF=W OF=W"],
    ["ror"              , "r8/m8, ib"                                , "MI"   , "C0 /1 ib"                         , "ANY CF=W OF=W"],

    ["ror"              , "r16/m16, 1"                               , "M"    , "66 D1 /1"                         , "ANY CF=W OF=W"],
    ["ror"              , "r16/m16, cl"                              , "M"    , "66 D3 /1"                         , "ANY CF=W OF=W"],
    ["ror"              , "r16/m16, ib"                              , "MI"   , "66 C1 /1 ib"                      , "ANY CF=W OF=W"],

    ["ror"              , "r32/m32, 1"                               , "M"    , "D1 /1"                            , "ANY CF=W OF=W"],
    ["ror"              , "r32/m32, cl"                              , "M"    , "D3 /1"                            , "ANY CF=W OF=W"],
    ["ror"              , "r32/m32, ib"                              , "MI"   , "C1 /1 ib"                         , "ANY CF=W OF=W"],

    ["ror"              , "r64/m64, 1"                               , "M"    , "REX.W D1 /1"                      , "X64 CF=W OF=W"],
    ["ror"              , "r64/m64, cl"                              , "M"    , "REX.W D3 /1"                      , "X64 CF=W OF=W"],
    ["ror"              , "r64/m64, ib"                              , "MI"   , "REX.W C1 /1 ib"                   , "X64 CF=W OF=W"],

    ["ret"              , ""                                         , "NONE" , "C3"                               , "ANY"],
    ["ret"              , "iw"                                       , "NONE" , "C2 iw"                            , "ANY"],

    ["sahf"             , "R:<ah>"                                   , "NONE" , "9E"                               , "ANY VOLATILE LAHFSAHF SF=W ZF=W AF=W PF=W CF=W"],

    ["sal/shl"          , "r8/m8, 1"                                 , "M"    , "D0 /4"                            , "ANY OF=W SF=W ZF=W AF=W PF=W CF=W"],
    ["sal/shl"          , "r8/m8, cl"                                , "M"    , "D2 /4"                            , "ANY OF=W SF=W ZF=W AF=W PF=W CF=W"],
    ["sal/shl"          , "r8/m8, ib"                                , "MI"   , "C0 /4 ib"                         , "ANY OF=W SF=W ZF=W AF=W PF=W CF=W"],

    ["sal/shl"          , "r16/m16, 1"                               , "M"    , "66 D1 /4"                         , "ANY OF=W SF=W ZF=W AF=W PF=W CF=W"],
    ["sal/shl"          , "r16/m16, cl"                              , "M"    , "66 D3 /4"                         , "ANY OF=W SF=W ZF=W AF=W PF=W CF=W"],
    ["sal/shl"          , "r16/m16, ib"                              , "MI"   , "66 C1 /4 ib"                      , "ANY OF=W SF=W ZF=W AF=W PF=W CF=W"],

    ["sal/shl"          , "r32/m32, 1"                               , "M"    , "D1 /4"                            , "ANY OF=W SF=W ZF=W AF=W PF=W CF=W"],
    ["sal/shl"          , "r32/m32, cl"                              , "M"    , "D3 /4"                            , "ANY OF=W SF=W ZF=W AF=W PF=W CF=W"],
    ["sal/shl"          , "r32/m32, ib"                              , "MI"   , "C1 /4 ib"                         , "ANY OF=W SF=W ZF=W AF=W PF=W CF=W"],

    ["sal/shl"          , "r64/m64, 1"                               , "M"    , "REX.W D1 /4"                      , "X64 OF=W SF=W ZF=W AF=W PF=W CF=W"],
    ["sal/shl"          , "r64/m64, cl"                              , "M"    , "REX.W D3 /4"                      , "X64 OF=W SF=W ZF=W AF=W PF=W CF=W"],
    ["sal/shl"          , "r64/m64, ib"                              , "MI"   , "REX.W C1 /4 ib"                   , "X64 OF=W SF=W ZF=W AF=W PF=W CF=W"],

    ["sar"              , "r8/m8, 1"                                 , "M"    , "D0 /7"                            , "ANY OF=W SF=W ZF=W AF=W PF=W CF=W"],
    ["sar"              , "r8/m8, cl"                                , "M"    , "D2 /7"                            , "ANY OF=W SF=W ZF=W AF=W PF=W CF=W"],
    ["sar"              , "r8/m8, ib"                                , "MI"   , "C0 /7 ib"                         , "ANY OF=W SF=W ZF=W AF=W PF=W CF=W"],

    ["sar"              , "r16/m16, 1"                               , "M"    , "66 D1 /7"                         , "ANY OF=W SF=W ZF=W AF=W PF=W CF=W"],
    ["sar"              , "r16/m16, cl"                              , "M"    , "66 D3 /7"                         , "ANY OF=W SF=W ZF=W AF=W PF=W CF=W"],
    ["sar"              , "r16/m16, ib"                              , "MI"   , "66 C1 /7 ib"                      , "ANY OF=W SF=W ZF=W AF=W PF=W CF=W"],

    ["sar"              , "r32/m32, 1"                               , "M"    , "D1 /7"                            , "ANY OF=W SF=W ZF=W AF=W PF=W CF=W"],
    ["sar"              , "r32/m32, cl"                              , "M"    , "D3 /7"                            , "ANY OF=W SF=W ZF=W AF=W PF=W CF=W"],
    ["sar"              , "r32/m32, ib"                              , "MI"   , "C1 /7 ib"                         , "ANY OF=W SF=W ZF=W AF=W PF=W CF=W"],

    ["sar"              , "r64/m64, 1"                               , "M"    , "REX.W D1 /7"                      , "X64 OF=W SF=W ZF=W AF=W PF=W CF=W"],
    ["sar"              , "r64/m64, cl"                              , "M"    , "REX.W D3 /7"                      , "X64 OF=W SF=W ZF=W AF=W PF=W CF=W"],
    ["sar"              , "r64/m64, ib"                              , "MI"   , "REX.W C1 /7 ib"                   , "X64 OF=W SF=W ZF=W AF=W PF=W CF=W"],

    ["shr"              , "r8/m8, 1"                                 , "M"    , "D0 /5"                            , "ANY OF=W SF=W ZF=W AF=W PF=W CF=W"],
    ["shr"              , "r8/m8, cl"                                , "M"    , "D2 /5"                            , "ANY OF=W SF=W ZF=W AF=W PF=W CF=W"],
    ["shr"              , "r8/m8, ib"                                , "MI"   , "C0 /5 ib"                         , "ANY OF=W SF=W ZF=W AF=W PF=W CF=W"],

    ["shr"              , "r16/m16, 1"                               , "M"    , "66 D1 /5"                         , "ANY OF=W SF=W ZF=W AF=W PF=W CF=W"],
    ["shr"              , "r16/m16, cl"                              , "M"    , "66 D3 /5"                         , "ANY OF=W SF=W ZF=W AF=W PF=W CF=W"],
    ["shr"              , "r16/m16, ib"                              , "MI"   , "66 C1 /5 ib"                      , "ANY OF=W SF=W ZF=W AF=W PF=W CF=W"],

    ["shr"              , "r32/m32, 1"                               , "M"    , "D1 /5"                            , "ANY OF=W SF=W ZF=W AF=W PF=W CF=W"],
    ["shr"              , "r32/m32, cl"                              , "M"    , "D3 /5"                            , "ANY OF=W SF=W ZF=W AF=W PF=W CF=W"],
    ["shr"              , "r32/m32, ib"                              , "MI"   , "C1 /5 ib"                         , "ANY OF=W SF=W ZF=W AF=W PF=W CF=W"],

    ["shr"              , "r64/m64, 1"                               , "M"    , "REX.W D1 /5"                      , "X64 OF=W SF=W ZF=W AF=W PF=W CF=W"],
    ["shr"              , "r64/m64, cl"                              , "M"    , "REX.W D3 /5"                      , "X64 OF=W SF=W ZF=W AF=W PF=W CF=W"],
    ["shr"              , "r64/m64, ib"                              , "MI"   , "REX.W C1 /5 ib"                   , "X64 OF=W SF=W ZF=W AF=W PF=W CF=W"],

    ["sbb"              , "al, ib"                                   , "I"    , "1C ib"                            , "ANY LOCK OF=W SF=W ZF=W AF=W PF=W CF=X"],
    ["sbb"              , "ax, iw"                                   , "I"    , "66 1D iw"                         , "ANY LOCK OF=W SF=W ZF=W AF=W PF=W CF=X"],
    ["sbb"              , "eax, id"                                  , "I"    , "1D id"                            , "ANY LOCK OF=W SF=W ZF=W AF=W PF=W CF=X"],
    ["sbb"              , "rax, id"                                  , "I"    , "REX.W 1D id"                      , "X64 LOCK OF=W SF=W ZF=W AF=W PF=W CF=X"],

    ["sbb"              , "r8/m8, ib"                                , "MI"   , "80 /3 ib"                         , "ANY LOCK OF=W SF=W ZF=W AF=W PF=W CF=X"],
    ["sbb"              , "r16/m16, iw"                              , "MI"   , "66 81 /3 iw"                      , "ANY LOCK OF=W SF=W ZF=W AF=W PF=W CF=X"],
    ["sbb"              , "r32/m32, id"                              , "MI"   , "81 /3 id"                         , "ANY LOCK OF=W SF=W ZF=W AF=W PF=W CF=X"],
    ["sbb"              , "r64/m64, id"                              , "MI"   , "REX.W 81 /3 id"                   , "X64 LOCK OF=W SF=W ZF=W AF=W PF=W CF=X"],

    ["sbb"              , "r16/m16, ib"                              , "MI"   , "66 83 /3 ib"                      , "ANY LOCK OF=W SF=W ZF=W AF=W PF=W CF=X"],
    ["sbb"              , "r32/m32, ib"                              , "MI"   , "83 /3 ib"                         , "ANY LOCK OF=W SF=W ZF=W AF=W PF=W CF=X"],
    ["sbb"              , "r64/m64, ib"                              , "MI"   , "REX.W 83 /3 ib"                   , "X64 LOCK OF=W SF=W ZF=W AF=W PF=W CF=X"],

    ["sbb"              , "r8/m8, r8"                                , "MR"   , "18 /r"                            , "ANY LOCK OF=W SF=W ZF=W AF=W PF=W CF=X"],
    ["sbb"              , "r16/m16, r16"                             , "MR"   , "66 19 /r"                         , "ANY LOCK OF=W SF=W ZF=W AF=W PF=W CF=X"],
    ["sbb"              , "r32/m32, r32"                             , "MR"   , "19 /r"                            , "ANY LOCK OF=W SF=W ZF=W AF=W PF=W CF=X"],
    ["sbb"              , "r64/m64, r64"                             , "MR"   , "REX.W 19 /r"                      , "X64 LOCK OF=W SF=W ZF=W AF=W PF=W CF=X"],

    ["sbb"              , "r8, r8/m8"                                , "RM"   , "1A /r"                            , "ANY LOCK OF=W SF=W ZF=W AF=W PF=W CF=X"],
    ["sbb"              , "r16, r16/m16"                             , "RM"   , "66 1B /r"                         , "ANY LOCK OF=W SF=W ZF=W AF=W PF=W CF=X"],
    ["sbb"              , "r32, r32/m32"                             , "RM"   , "1B /r"                            , "ANY LOCK OF=W SF=W ZF=W AF=W PF=W CF=X"],
    ["sbb"              , "r64, r64/m64"                             , "RM"   , "REX.W 1B /r"                      , "X64 LOCK OF=W SF=W ZF=W AF=W PF=W CF=X"],

    ["scasb"            , "R:<al>, X:<es:zdi>"                       , "NONE" , "AE"                               , "ANY REP OF=W SF=W ZF=W AF=W PF=W CF=W DF=R"],
    ["scasw"            , "R:<ax>, X:<es:zdi>"                       , "NONE" , "66 AF"                            , "ANY REP OF=W SF=W ZF=W AF=W PF=W CF=W DF=R"],
    ["scasd"            , "R:<eax>, X:<es:zdi>"                      , "NONE" , "AF"                               , "ANY REP OF=W SF=W ZF=W AF=W PF=W CF=W DF=R"],
    ["scasq"            , "R:<rax>, X:<es:zdi>"                      , "NONE" , "REX.W AF"                         , "X64 REP OF=W SF=W ZF=W AF=W PF=W CF=W DF=R"],

    ["seto"             , "W:r8/m8"                                  , "M"    , "0F 90 /r"                         , "ANY OF=R"],
    ["setno"            , "W:r8/m8"                                  , "M"    , "0F 91 /r"                         , "ANY OF=R"],
    ["setb/setnae/setc" , "W:r8/m8"                                  , "M"    , "0F 92 /r"                         , "ANY CF=R"],
    ["setae/setnb/setnc", "W:r8/m8"                                  , "M"    , "0F 93 /r"                         , "ANY CF=R"],
    ["sete/setz"        , "W:r8/m8"                                  , "M"    , "0F 94 /r"                         , "ANY ZF=R"],
    ["setne/setnz"      , "W:r8/m8"                                  , "M"    , "0F 95 /r"                         , "ANY ZF=R"],
    ["setbe/setna"      , "W:r8/m8"                                  , "M"    , "0F 96 /r"                         , "ANY CF=R ZF=R"],
    ["seta/setnbe"      , "W:r8/m8"                                  , "M"    , "0F 97 /r"                         , "ANY CF=R ZF=R"],
    ["sets"             , "W:r8/m8"                                  , "M"    , "0F 98 /r"                         , "ANY SF=R"],
    ["setns"            , "W:r8/m8"                                  , "M"    , "0F 99 /r"                         , "ANY SF=R"],
    ["setp/setpe"       , "W:r8/m8"                                  , "M"    , "0F 9A /r"                         , "ANY PF=R"],
    ["setnp/setpo"      , "W:r8/m8"                                  , "M"    , "0F 9B /r"                         , "ANY PF=R"],
    ["setl/setnge"      , "W:r8/m8"                                  , "M"    , "0F 9C /r"                         , "ANY SF=R OF=R"],
    ["setge/setnl"      , "W:r8/m8"                                  , "M"    , "0F 9D /r"                         , "ANY SF=R OF=R"],
    ["setle/setng"      , "W:r8/m8"                                  , "M"    , "0F 9E /r"                         , "ANY ZF=R SF=R OF=R"],
    ["setg/setnle"      , "W:r8/m8"                                  , "M"    , "0F 9F /r"                         , "ANY ZF=R SF=R OF=R"],

    ["shld"             , "r16/m16, r16, ib"                         , "MRI"  , "66 0F A4 /r ib"                   , "ANY OF=W SF=W ZF=W AF=U PF=W CF=W"],
    ["shld"             , "r16/m16, r16, cl"                         , "MR"   , "66 0F A5 /r"                      , "ANY OF=W SF=W ZF=W AF=U PF=W CF=W"],
    ["shld"             , "r32/m32, r32, ib"                         , "MRI"  , "0F A4 /r ib"                      , "ANY OF=W SF=W ZF=W AF=U PF=W CF=W"],
    ["shld"             , "r32/m32, r32, cl"                         , "MR"   , "0F A5 /r"                         , "ANY OF=W SF=W ZF=W AF=U PF=W CF=W"],
    ["shld"             , "r64/m64, r64, ib"                         , "MRI"  , "REX.W 0F A4 /r ib"                , "X64 OF=W SF=W ZF=W AF=U PF=W CF=W"],
    ["shld"             , "r64/m64, r64, cl"                         , "MR"   , "REX.W 0F A5 /r"                   , "X64 OF=W SF=W ZF=W AF=U PF=W CF=W"],

    ["shrd"             , "r16/m16, r16, ib"                         , "MRI"  , "66 0F AC /r ib"                   , "ANY OF=W SF=W ZF=W AF=U PF=W CF=W"],
    ["shrd"             , "r16/m16, r16, cl"                         , "MR"   , "66 0F AD /r"                      , "ANY OF=W SF=W ZF=W AF=U PF=W CF=W"],
    ["shrd"             , "r32/m32, r32, ib"                         , "MRI"  , "0F AC /r ib"                      , "ANY OF=W SF=W ZF=W AF=U PF=W CF=W"],
    ["shrd"             , "r32/m32, r32, cl"                         , "MR"   , "0F AD /r"                         , "ANY OF=W SF=W ZF=W AF=U PF=W CF=W"],
    ["shrd"             , "r64/m64, r64, ib"                         , "MRI"  , "REX.W 0F AC /r ib"                , "X64 OF=W SF=W ZF=W AF=U PF=W CF=W"],
    ["shrd"             , "r64/m64, r64, cl"                         , "MR"   , "REX.W 0F AD /r"                   , "X64 OF=W SF=W ZF=W AF=U PF=W CF=W"],

    ["stac"             , ""                                         , "NONE" , "0F 01 CB"                         , "ANY VOLATILE SMAP AC=1"],
    ["stc"              , ""                                         , "NONE" , "F9"                               , "ANY CF=1"],
    ["std"              , ""                                         , "NONE" , "FD"                               , "ANY DF=1"],
    ["sti"              , ""                                         , "NONE" , "FB"                               , "ANY VOLATILE IF=1"],

    ["stosb"            , "X:<es:zdi>, R:<al>"                       , "NONE" , "AA"                               , "ANY REP DF=R"],
    ["stosw"            , "X:<es:zdi>, R:<ax>"                       , "NONE" , "66 AB"                            , "ANY REP DF=R"],
    ["stosd"            , "X:<es:zdi>, R:<eax>"                      , "NONE" , "AB"                               , "ANY REP DF=R"],
    ["stosq"            , "X:<es:zdi>, R:<rax>"                      , "NONE" , "REX.W AB"                         , "X64 REP DF=R"],

    ["sub"              , "al, ib"                                   , "I"    , "2C ib"                            , "ANY LOCK OF=W SF=W ZF=W AF=W PF=W CF=W"],
    ["sub"              , "ax, iw"                                   , "I"    , "66 2D iw"                         , "ANY LOCK OF=W SF=W ZF=W AF=W PF=W CF=W"],
    ["sub"              , "eax, id"                                  , "I"    , "2D id"                            , "ANY LOCK OF=W SF=W ZF=W AF=W PF=W CF=W"],
    ["sub"              , "rax, id"                                  , "I"    , "REX.W 2D id"                      , "X64 LOCK OF=W SF=W ZF=W AF=W PF=W CF=W"],

    ["sub"              , "r8/m8, ib"                                , "MI"   , "80 /5 ib"                         , "ANY LOCK OF=W SF=W ZF=W AF=W PF=W CF=W"],
    ["sub"              , "r16/m16, iw"                              , "MI"   , "66 81 /5 iw"                      , "ANY LOCK OF=W SF=W ZF=W AF=W PF=W CF=W"],
    ["sub"              , "r32/m32, id"                              , "MI"   , "81 /5 id"                         , "ANY LOCK OF=W SF=W ZF=W AF=W PF=W CF=W"],
    ["sub"              , "r64/m64, id"                              , "MI"   , "REX.W 81 /5 id"                   , "X64 LOCK OF=W SF=W ZF=W AF=W PF=W CF=W"],

    ["sub"              , "r16/m16, ib"                              , "MI"   , "66 83 /5 ib"                      , "ANY LOCK OF=W SF=W ZF=W AF=W PF=W CF=W"],
    ["sub"              , "r32/m32, ib"                              , "MI"   , "83 /5 ib"                         , "ANY LOCK OF=W SF=W ZF=W AF=W PF=W CF=W"],
    ["sub"              , "r64/m64, ib"                              , "MI"   , "REX.W 83 /5 ib"                   , "X64 LOCK OF=W SF=W ZF=W AF=W PF=W CF=W"],

    ["sub"              , "r8/m8, r8"                                , "MR"   , "28 /r"                            , "ANY LOCK OF=W SF=W ZF=W AF=W PF=W CF=W"],
    ["sub"              , "r16/m16, r16"                             , "MR"   , "66 29 /r"                         , "ANY LOCK OF=W SF=W ZF=W AF=W PF=W CF=W"],
    ["sub"              , "r32/m32, r32"                             , "MR"   , "29 /r"                            , "ANY LOCK OF=W SF=W ZF=W AF=W PF=W CF=W"],
    ["sub"              , "r64/m64, r64"                             , "MR"   , "REX.W 29 /r"                      , "X64 LOCK OF=W SF=W ZF=W AF=W PF=W CF=W"],

    ["sub"              , "r8, r8/m8"                                , "RM"   , "2A /r"                            , "ANY LOCK OF=W SF=W ZF=W AF=W PF=W CF=W"],
    ["sub"              , "r16, r16/m16"                             , "RM"   , "66 2B /r"                         , "ANY LOCK OF=W SF=W ZF=W AF=W PF=W CF=W"],
    ["sub"              , "r32, r32/m32"                             , "RM"   , "2B /r"                            , "ANY LOCK OF=W SF=W ZF=W AF=W PF=W CF=W"],
    ["sub"              , "r64, r64/m64"                             , "RM"   , "REX.W 2B /r"                      , "X64 LOCK OF=W SF=W ZF=W AF=W PF=W CF=W"],

    ["swapgs"           , ""                                         , "NONE" , "0F 01 F8"                         , "X64 VOLATILE"],

    ["test"             , "R:al, ib"                                 , "I"    , "A8 ib"                            , "ANY LOCK OF=0 SF=W ZF=W AF=U PF=W CF=0"],
    ["test"             , "R:ax, iw"                                 , "I"    , "66 A9 iw"                         , "ANY LOCK OF=0 SF=W ZF=W AF=U PF=W CF=0"],
    ["test"             , "R:eax, id"                                , "I"    , "A9 id"                            , "ANY LOCK OF=0 SF=W ZF=W AF=U PF=W CF=0"],
    ["test"             , "R:rax, id"                                , "I"    , "REX.W A9 id"                      , "X64 LOCK OF=0 SF=W ZF=W AF=U PF=W CF=0"],

    ["test"             , "R:r8/m8, ib"                              , "MI"   , "F6 /0 ib"                         , "ANY LOCK OF=0 SF=W ZF=W AF=U PF=W CF=0"],
    ["test"             , "R:r16/m16, iw"                            , "MI"   , "66 F7 /0 iw"                      , "ANY LOCK OF=0 SF=W ZF=W AF=U PF=W CF=0"],
    ["test"             , "R:r32/m32, id"                            , "MI"   , "F7 /0 id"                         , "ANY LOCK OF=0 SF=W ZF=W AF=U PF=W CF=0"],
    ["test"             , "R:r64/m64, id"                            , "MI"   , "REX.W F7 /0 id"                   , "X64 LOCK OF=0 SF=W ZF=W AF=U PF=W CF=0"],

    ["test"             , "R:r8/m8, r8"                              , "MR"   , "84 /r"                            , "ANY LOCK OF=0 SF=W ZF=W AF=U PF=W CF=0"],
    ["test"             , "R:r16/m16, r16"                           , "MR"   , "66 85 /r"                         , "ANY LOCK OF=0 SF=W ZF=W AF=U PF=W CF=0"],
    ["test"             , "R:r32/m32, r32"                           , "MR"   , "85 /r"                            , "ANY LOCK OF=0 SF=W ZF=W AF=U PF=W CF=0"],
    ["test"             , "R:r64/m64, r64"                           , "MR"   , "REX.W 85 /r"                      , "X64 LOCK OF=0 SF=W ZF=W AF=U PF=W CF=0"],

    ["ud2"              , ""                                         , "NONE" , "0F 0B"                            , "ANY"],

    ["xadd"             , "X:r8/m8, X:r8"                            , "MR"   , "0F C0 /r"                         , "ANY I486 LOCK OF=W SF=W ZF=W AF=W PF=W CF=W"],
    ["xadd"             , "X:r16/m16, X:r16"                         , "MR"   , "66 0F C1 /r"                      , "ANY I486 LOCK OF=W SF=W ZF=W AF=W PF=W CF=W"],
    ["xadd"             , "X:r32/m32, X:r32"                         , "MR"   , "0F C1 /r"                         , "ANY I486 LOCK OF=W SF=W ZF=W AF=W PF=W CF=W"],
    ["xadd"             , "X:r64/m64, X:r64"                         , "MR"   , "REX.W 0F C1 /r"                   , "X64 I486 LOCK OF=W SF=W ZF=W AF=W PF=W CF=W"],

    ["xchg"             , "X:ax, X:r16"                              , "O"    , "66 90+r"                          , "ANY LOCK"],
    ["xchg"             , "X:eax, X:r32"                             , "O"    , "90+r"                             , "ANY LOCK"],
    ["xchg"             , "X:rax, X:r64"                             , "O"    , "REX.W 90+r"                       , "X64 LOCK"],

    ["xchg"             , "X:r16, X:ax"                              , "O"    , "66 90+r"                          , "ANY LOCK"],
    ["xchg"             , "X:r32, X:eax"                             , "O"    , "90+r"                             , "ANY LOCK"],
    ["xchg"             , "X:r64, X:rax"                             , "O"    , "REX.W 90+r"                       , "X64 LOCK"],

    ["xchg"             , "X:r8/m8, X:r8"                            , "MR"   , "86 /r"                            , "ANY LOCK"],
    ["xchg"             , "X:r16/m16, X:r16"                         , "MR"   , "66 87 /r"                         , "ANY LOCK"],
    ["xchg"             , "X:r32/m32, X:r32"                         , "MR"   , "87 /r"                            , "ANY LOCK"],
    ["xchg"             , "X:r64/m64, X:r64"                         , "MR"   , "REX.W 87 /r"                      , "X64 LOCK"],

    ["xchg"             , "X:r8, X:r8/m8"                            , "RM"   , "86 /r"                            , "ANY LOCK"],
    ["xchg"             , "X:r16, X:r16/m16"                         , "RM"   , "66 87 /r"                         , "ANY LOCK"],
    ["xchg"             , "X:r32, X:r32/m32"                         , "RM"   , "87 /r"                            , "ANY LOCK"],
    ["xchg"             , "X:r64, X:r64/m64"                         , "RM"   , "REX.W 87 /r"                      , "X64 LOCK"],

    ["xor"              , "al, ib"                                   , "I"    , "34 ib"                            , "ANY LOCK OF=0 SF=W ZF=W AF=U PF=W CF=0"],
    ["xor"              , "ax, iw"                                   , "I"    , "66 35 iw"                         , "ANY LOCK OF=0 SF=W ZF=W AF=U PF=W CF=0"],
    ["xor"              , "eax, id"                                  , "I"    , "35 id"                            , "ANY LOCK OF=0 SF=W ZF=W AF=U PF=W CF=0"],
    ["xor"              , "rax, id"                                  , "I"    , "REX.W 35 id"                      , "X64 LOCK OF=0 SF=W ZF=W AF=U PF=W CF=0"],

    ["xor"              , "r8/m8, ib"                                , "MI"   , "80 /6 ib"                         , "ANY LOCK OF=0 SF=W ZF=W AF=U PF=W CF=0"],
    ["xor"              , "r16/m16, iw"                              , "MI"   , "66 81 /6 iw"                      , "ANY LOCK OF=0 SF=W ZF=W AF=U PF=W CF=0"],
    ["xor"              , "r32/m32, id"                              , "MI"   , "81 /6 id"                         , "ANY LOCK OF=0 SF=W ZF=W AF=U PF=W CF=0"],
    ["xor"              , "r64/m64, id"                              , "MI"   , "REX.W 81 /6 id"                   , "X64 LOCK OF=0 SF=W ZF=W AF=U PF=W CF=0"],

    ["xor"              , "r16/m16, ib"                              , "MI"   , "66 83 /6 ib"                      , "ANY LOCK OF=0 SF=W ZF=W AF=U PF=W CF=0"],
    ["xor"              , "r32/m32, ib"                              , "MI"   , "83 /6 ib"                         , "ANY LOCK OF=0 SF=W ZF=W AF=U PF=W CF=0"],
    ["xor"              , "r64/m64, ib"                              , "MI"   , "REX.W 83 /6 ib"                   , "X64 LOCK OF=0 SF=W ZF=W AF=U PF=W CF=0"],

    ["xor"              , "r8/m8, r8"                                , "MR"   , "30 /r"                            , "ANY LOCK OF=0 SF=W ZF=W AF=U PF=W CF=0"],
    ["xor"              , "r16/m16, r16"                             , "MR"   , "66 31 /r"                         , "ANY LOCK OF=0 SF=W ZF=W AF=U PF=W CF=0"],
    ["xor"              , "r32/m32, r32"                             , "MR"   , "31 /r"                            , "ANY LOCK OF=0 SF=W ZF=W AF=U PF=W CF=0"],
    ["xor"              , "r64/m64, r64"                             , "MR"   , "REX.W 31 /r"                      , "X64 LOCK OF=0 SF=W ZF=W AF=U PF=W CF=0"],

    ["xor"              , "r8, r8/m8"                                , "RM"   , "32 /r"                            , "ANY LOCK OF=0 SF=W ZF=W AF=U PF=W CF=0"],
    ["xor"              , "r16, r16/m16"                             , "RM"   , "66 33 /r"                         , "ANY LOCK OF=0 SF=W ZF=W AF=U PF=W CF=0"],
    ["xor"              , "r32, r32/m32"                             , "RM"   , "33 /r"                            , "ANY LOCK OF=0 SF=W ZF=W AF=U PF=W CF=0"],
    ["xor"              , "r64, r64/m64"                             , "RM"   , "REX.W 33 /r"                      , "X64 LOCK OF=0 SF=W ZF=W AF=U PF=W CF=0"],

    // X86/X64 (Special / Privileged).
/*
    ["arpl"             , "X:r16/m16, R:r16"                         , "MR"   , "63 /r"                            , "X86 VOLATILE ZF=W"],
    ["bound"            , "R:r16, R:m32"                             , "RM"   , "66 62 /r"                         , "X86 VOLATILE"],
    ["bound"            , "R:r32, R:m64"                             , "RM"   , "62 /r"                            , "X86 VOLATILE"],
    ["cli"              , ""                                         , "NP"   , "FA"                               , "ANY VOLATILE IF=W"],
    ["clts"             , ""                                         , "NP"   , "0F 06"                            , "ANY VOLATILE PRIVILEGE=L0"],
    ["hlt"              , ""                                         , "NONE" , "F4"                               , "ANY VOLATILE PRIVILEGE=L0"],
    ["invd"             , ""                                         , "NONE" , "0F 08"                            , "ANY VOLATILE PRIVILEGE=L0 I486"],
    ["invlpg"           , "R:mem"                                    , "M"    , "0F 01 /7"                         , "ANY VOLATILE PRIVILEGE=L0 I486"],
    ["invpcid"          , "R:rxx, R:m128"                            , "RM"   , "66 0F 38 82 /r"                   , "ANY VOLATILE PRIVILEGE=L0 I486"],
    ["lar"              , "W:r16, R:r16/m16"                         , "RM"   , "66 0F 02 /r"                      , "ANY VOLATILE ZF=W"],
    ["lar"              , "W:r32, R:r32/m16"                         , "RM"   , "0F 02 /r"                         , "ANY VOLATILE ZF=W"],
    ["lgdt"             , "R:mem"                                    , "M"    , "0F 01 /2"                         , "ANY VOLATILE PRIVILEGE=L0"],
    ["lidt"             , "R:mem"                                    , "M"    , "0F 01 /3"                         , "ANY VOLATILE PRIVILEGE=L0"],
    ["lldt"             , "R:rxx/m16"                                , "M"    , "0F 00 /2"                         , "ANY VOLATILE PRIVILEGE=L0"],
    ["lmsw"             , "R:rxx/m16"                                , "M"    , "0F 01 /6"                         , "ANY VOLATILE PRIVILEGE=L0"],
    ["lsl"              , "W:rxx, R:r16/m16"                         , "RM"   , "0F 03 /r"                         , "ANY VOLATILE ZF=W"],
    ["ltr"              , "R:rxx/m16"                                , "M"    , "0F 00 /3"                         , "ANY VOLATILE PRIVILEGE=L0"],
    ["monitor"          , "R:<ds:zax>, R:<ecx>, R:<edx>"             , "NONE" , "0F 01 C8"                         , "ANY VOLATILE PRIVILEGE=L0 MONITOR"],
    ["mwait"            , "R:<eax>, R:<ecx>"                         , "NONE" , "0F 01 C9"                         , "ANY VOLATILE PRIVILEGE=L0 MONITOR"],
    ["sgdt"             , "W:mem"                                    , "M"    , "0F 01 /0"                         , "ANY VOLATILE"],
    ["sidt"             , "W:mem"                                    , "M"    , "0F 01 /1"                         , "ANY VOLATILE"],
    ["sldt"             , "W:r32/m16"                                , "M"    , "0F 00 /0"                         , "ANY VOLATILE"],
    ["sldt"             , "W:r64/m16"                                , "M"    , "REX.W 0F 00 /0"                   , "X64 VOLATILE"],
    ["smsw"             , "W:r16/m16"                                , "M"    , "66 0F 01 /4"                      , "ANY VOLATILE"],
    ["smsw"             , "W:r32/m16"                                , "M"    , "0F 01 /4"                         , "ANY VOLATILE"],
    ["smsw"             , "W:r64/m16"                                , "M"    , "REX.W 0F 01 /4"                   , "X64 VOLATILE"],
    ["syscall"          , ""                                         , "NONE" , "0F 05"                            , "X64 VOLATILE"],
    ["sysenter"         , ""                                         , "NONE" , "0F 34"                            , "ANY VOLATILE"],
    ["sysexit"          , ""                                         , "NONE" , "0F 35"                            , "ANY VOLATILE"],
    ["sysexit64"        , ""                                         , "NONE" , "REX.W 0F 35"                      , "ANY VOLATILE"],
    ["sysret"           , ""                                         , "NONE" , "0F 07"                            , "X64 VOLATILE"],
    ["sysret64"         , ""                                         , "NONE" , "REX.W 0F 07"                      , "X64 VOLATILE"],
    ["rdmsr"            , "W:<edx>,W:<eax>,R:<ecx>"                  , "NONE" , "0F 32"                            , "ANY VOLATILE PRIVILEGE=L0"],
    ["rdpmc"            , "W:<edx>,W:<eax>,R:<ecx>"                  , "NONE" , "0F 33"                            , "ANY VOLATILE PRIVILEGE=L0"],
    ["str"              , "W:rxx/m16"                                , "M"    , "0F 00 /1"                         , "ANY VOLATILE"],
    ["verr"             , "rxx/m16"                                  , "M"    , "0F 00 /4"                         , "ANY VOLATILE ZF=W"],
    ["verw"             , "rxx/m16"                                  , "M"    , "0F 00 /5"                         , "ANY VOLATILE ZF=W"],
    ["wbinvd"           , ""                                         , "NONE" , "0F 09"                            , "ANY VOLATILE PRIVILEGE=L0"],
    ["wrmsr"            , "R:<edx>,R:<eax>,R:<ecx>"                  , "NONE" , "0F 30"                            , "ANY VOLATILE PRIVILEGE=L0"],
*/
    // FPU.
    ["f2xm1"            , ""                                         , "NONE" , "D9 F0"                            , "ANY FPU        C0=U C1=W C2=U C3=U"],
    ["fabs"             , ""                                         , "NONE" , "D9 E1"                            , "ANY FPU        C0=U C1=0 C2=U C3=U"],
    ["fadd"             , "m32fp"                                    , "M"    , "D8 /0"                            , "ANY FPU        C0=U C1=W C2=U C3=U"],
    ["fadd"             , "m64fp"                                    , "M"    , "DC /0"                            , "ANY FPU        C0=U C1=W C2=U C3=U"],
    ["fadd"             , "st(0), st(i)"                             , "O"    , "D8 C0+i"                          , "ANY FPU        C0=U C1=W C2=U C3=U"],
    ["fadd"             , "st(i), st(0)"                             , "O"    , "DC C0+i"                          , "ANY FPU        C0=U C1=W C2=U C3=U"],
    ["faddp"            , ""                                         , "NONE" , "DE C1"                            , "ANY FPU_POP    C0=U C1=W C2=U C3=U"],
    ["faddp"            , "st(i)"                                    , "O"    , "DE C0+i"                          , "ANY FPU_POP    C0=U C1=W C2=U C3=U"],
    ["fbld"             , "m80dec"                                   , "M"    , "DF /4"                            , "ANY FPU_PUSH   C0=U C1=W C2=U C3=U"],
    ["fbstp"            , "m80bcd"                                   , "M"    , "DF /6"                            , "ANY FPU_POP    C0=U C1=W C2=U C3=U"],
    ["fchs"             , ""                                         , "NONE" , "D9 E0"                            , "ANY FPU        C0=U C1=0 C2=U C3=U"],
    ["fclex"            , ""                                         , "NONE" , "9B DB E2"                         , "ANY FPU        C0=U C1=U C2=U C3=U"],
    ["fcmovb"           , "st(i)"                                    , "O"    , "DA C0+i"                          , "ANY FPU CMOV   C0=U C1=W C2=U C3=U"],
    ["fcmovbe"          , "st(i)"                                    , "O"    , "DA D0+i"                          , "ANY FPU CMOV   C0=U C1=W C2=U C3=U"],
    ["fcmove"           , "st(i)"                                    , "O"    , "DA C8+i"                          , "ANY FPU CMOV   C0=U C1=W C2=U C3=U"],
    ["fcmovnb"          , "st(i)"                                    , "O"    , "DB C0+i"                          , "ANY FPU CMOV   C0=U C1=W C2=U C3=U"],
    ["fcmovnbe"         , "st(i)"                                    , "O"    , "DB D0+i"                          , "ANY FPU CMOV   C0=U C1=W C2=U C3=U"],
    ["fcmovne"          , "st(i)"                                    , "O"    , "DB C8+i"                          , "ANY FPU CMOV   C0=U C1=W C2=U C3=U"],
    ["fcmovnu"          , "st(i)"                                    , "O"    , "DB D8+i"                          , "ANY FPU CMOV   C0=U C1=W C2=U C3=U"],
    ["fcmovu"           , "st(i)"                                    , "O"    , "DA D8+i"                          , "ANY FPU CMOV   C0=U C1=W C2=U C3=U"],
    ["fcom"             , ""                                         , "NONE" , "D8 D1"                            , "ANY FPU        C0=W C1=0 C2=W C3=W"],
    ["fcom"             , "m32fp"                                    , "M"    , "D8 /2"                            , "ANY FPU        C0=W C1=0 C2=W C3=W"],
    ["fcom"             , "m64fp"                                    , "M"    , "DC /2"                            , "ANY FPU        C0=W C1=0 C2=W C3=W"],
    ["fcom"             , "st(i)"                                    , "O"    , "D8 D0+i"                          , "ANY FPU        C0=W C1=0 C2=W C3=W"],
    ["fcomi"            , "st(i)"                                    , "O"    , "DB F0+i"                          , "ANY FPU        C1=0 ZF=W PF=W CF=W"],
    ["fcomip"           , "st(i)"                                    , "O"    , "DF F0+i"                          , "ANY FPU_POP    C1=0 ZF=W PF=W CF=W"],
    ["fcomp"            , ""                                         , "NONE" , "D8 D9"                            , "ANY FPU_POP    C0=W C1=W C2=W C3=W"],
    ["fcomp"            , "m32fp"                                    , "M"    , "D8 /3"                            , "ANY FPU_POP    C0=W C1=W C2=W C3=W"],
    ["fcomp"            , "m64fp"                                    , "M"    , "DC /3"                            , "ANY FPU_POP    C0=W C1=W C2=W C3=W"],
    ["fcomp"            , "st(i)"                                    , "O"    , "D8 D8+i"                          , "ANY FPU_POP    C0=W C1=W C2=W C3=W"],
    ["fcompp"           , ""                                         , "NONE" , "DE D9"                            , "ANY FPU_POP=2  C0=W C1=W C2=W C3=W"],
    ["fcos"             , ""                                         , "NONE" , "D9 FF"                            , "ANY FPU        C0=U C1=W C2=W C3=U"],
    ["fdecstp"          , ""                                         , "NONE" , "D9 F6"                            , "ANY FPU_TOP=-1 C0=U C1=W C2=U C3=U"],
    ["fdiv"             , "m32fp"                                    , "M"    , "D8 /6"                            , "ANY FPU        C0=U C1=W C2=U C3=U"],
    ["fdiv"             , "m64fp"                                    , "M"    , "DC /6"                            , "ANY FPU        C0=U C1=W C2=U C3=U"],
    ["fdiv"             , "st(0), st(i)"                             , "O"    , "D8 F0+i"                          , "ANY FPU        C0=U C1=W C2=U C3=U"],
    ["fdiv"             , "st(i), st(0)"                             , "O"    , "DC F8+i"                          , "ANY FPU        C0=U C1=W C2=U C3=U"],
    ["fdivp"            , ""                                         , "NONE" , "DE F9"                            , "ANY FPU_POP    C0=U C1=W C2=U C3=U"],
    ["fdivp"            , "st(i)"                                    , "O"    , "DE F8+i"                          , "ANY FPU_POP    C0=U C1=W C2=U C3=U"],
    ["fdivr"            , "m32fp"                                    , "M"    , "D8 /7"                            , "ANY FPU        C0=U C1=W C2=U C3=U"],
    ["fdivr"            , "m64fp"                                    , "M"    , "DC /7"                            , "ANY FPU        C0=U C1=W C2=U C3=U"],
    ["fdivr"            , "st(0), st(i)"                             , "O"    , "D8 F8+i"                          , "ANY FPU        C0=U C1=W C2=U C3=U"],
    ["fdivr"            , "st(i), st(0)"                             , "O"    , "DC F0+i"                          , "ANY FPU        C0=U C1=W C2=U C3=U"],
    ["fdivrp"           , ""                                         , "NONE" , "DE F1"                            , "ANY FPU_POP    C0=U C1=W C2=U C3=U"],
    ["fdivrp"           , "st(i)"                                    , "O"    , "DE F0+i"                          , "ANY FPU_POP    C0=U C1=W C2=U C3=U"],
    ["ffree"            , "st(i)"                                    , "O"    , "DD C0+i"                          , "ANY FPU        C0=U C1=U C2=U C3=U"],
    ["fiadd"            , "m16int"                                   , "M"    , "DE /0"                            , "ANY FPU        C0=U C1=W C2=U C3=U"],
    ["fiadd"            , "m32int"                                   , "M"    , "DA /0"                            , "ANY FPU        C0=U C1=W C2=U C3=U"],
    ["ficom"            , "m16int"                                   , "M"    , "DE /2"                            , "ANY FPU        C0=W C1=0 C2=W C3=W"],
    ["ficom"            , "m32int"                                   , "M"    , "DA /2"                            , "ANY FPU        C0=W C1=0 C2=W C3=W"],
    ["ficomp"           , "m16int"                                   , "M"    , "DE /3"                            , "ANY FPU        C0=W C1=0 C2=W C3=W"],
    ["ficomp"           , "m32int"                                   , "M"    , "DA /3"                            , "ANY FPU        C0=W C1=0 C2=W C3=W"],
    ["fidiv"            , "m16int"                                   , "M"    , "DE /6"                            , "ANY FPU        C0=U C1=W C2=U C3=U"],
    ["fidiv"            , "m32int"                                   , "M"    , "DA /6"                            , "ANY FPU        C0=U C1=W C2=U C3=U"],
    ["fidivr"           , "m16int"                                   , "M"    , "DE /7"                            , "ANY FPU        C0=U C1=W C2=U C3=U"],
    ["fidivr"           , "m32int"                                   , "M"    , "DA /7"                            , "ANY FPU        C0=U C1=W C2=U C3=U"],
    ["fild"             , "m16int"                                   , "M"    , "DF /0"                            , "ANY FPU_PUSH   C0=U C1=W C2=U C3=U"],
    ["fild"             , "m32int"                                   , "M"    , "DB /0"                            , "ANY FPU_PUSH   C0=U C1=W C2=U C3=U"],
    ["fild"             , "m64int"                                   , "M"    , "DF /5"                            , "ANY FPU_PUSH   C0=U C1=W C2=U C3=U"],
    ["fimul"            , "m16int"                                   , "M"    , "DE /1"                            , "ANY FPU        C0=U C1=W C2=U C3=U"],
    ["fimul"            , "m32int"                                   , "M"    , "DA /1"                            , "ANY FPU        C0=U C1=W C2=U C3=U"],
    ["fincstp"          , ""                                         , "NONE" , "D9 F7"                            , "ANY FPU_TOP=+1 C0=U C1=0 C2=U C3=U"],
    ["finit"            , ""                                         , "NONE" , "9B DB E3"                         , "ANY FPU        C0=0 C1=0 C2=0 C3=0"],
    ["fist"             , "m16int"                                   , "M"    , "DF /2"                            , "ANY FPU        C0=U C1=W C2=U C3=U"],
    ["fist"             , "m32int"                                   , "M"    , "DB /2"                            , "ANY FPU        C0=U C1=W C2=U C3=U"],
    ["fistp"            , "m16int"                                   , "M"    , "DF /3"                            , "ANY FPU_POP    C0=U C1=W C2=U C3=U"],
    ["fistp"            , "m32int"                                   , "M"    , "DB /3"                            , "ANY FPU_POP    C0=U C1=W C2=U C3=U"],
    ["fistp"            , "m64int"                                   , "M"    , "DF /7"                            , "ANY FPU_POP    C0=U C1=W C2=U C3=U"],
    ["fisttp"           , "m16int"                                   , "M"    , "DF /1"                            , "ANY FPU_POP    C0=U C1=0 C2=U C3=U SSE3"],
    ["fisttp"           , "m32int"                                   , "M"    , "DB /1"                            , "ANY FPU_POP    C0=U C1=0 C2=U C3=U SSE3"],
    ["fisttp"           , "m64int"                                   , "M"    , "DD /1"                            , "ANY FPU_POP    C0=U C1=0 C2=U C3=U SSE3"],
    ["fisub"            , "m16int"                                   , "M"    , "DE /4"                            , "ANY FPU        C0=U C1=W C2=U C3=U"],
    ["fisub"            , "m32int"                                   , "M"    , "DA /4"                            , "ANY FPU        C0=U C1=W C2=U C3=U"],
    ["fisubr"           , "m16int"                                   , "M"    , "DE /5"                            , "ANY FPU        C0=U C1=W C2=U C3=U"],
    ["fisubr"           , "m32int"                                   , "M"    , "DA /5"                            , "ANY FPU        C0=U C1=W C2=U C3=U"],
    ["fld"              , "m32fp"                                    , "M"    , "D9 /0"                            , "ANY FPU_PUSH   C0=U C1=W C2=U C3=U"],
    ["fld"              , "m64fp"                                    , "M"    , "DD /0"                            , "ANY FPU_PUSH   C0=U C1=W C2=U C3=U"],
    ["fld"              , "m80fp"                                    , "M"    , "DB /5"                            , "ANY FPU_PUSH   C0=U C1=W C2=U C3=U"],
    ["fld"              , "st(i)"                                    , "O"    , "D9 C0+i"                          , "ANY FPU_PUSH   C0=U C1=W C2=U C3=U"],
    ["fld1"             , ""                                         , "NONE" , "D9 E8"                            , "ANY FPU_PUSH   C0=U C1=W C2=U C3=U"],
    ["fldcw"            , "m16"                                      , "M"    , "D9 /5"                            , "ANY FPU        C0=U C1=U C2=U C3=U"],
    ["fldenv"           , "mem"                                      , "M"    , "D9 /4"                            , "ANY FPU        C0=W C1=W C2=W C3=W"],
    ["fldl2e"           , ""                                         , "NONE" , "D9 EA"                            , "ANY FPU_PUSH   C0=U C1=W C2=U C3=U"],
    ["fldl2t"           , ""                                         , "NONE" , "D9 E9"                            , "ANY FPU_PUSH   C0=U C1=W C2=U C3=U"],
    ["fldlg2"           , ""                                         , "NONE" , "D9 EC"                            , "ANY FPU_PUSH   C0=U C1=W C2=U C3=U"],
    ["fldln2"           , ""                                         , "NONE" , "D9 ED"                            , "ANY FPU_PUSH   C0=U C1=W C2=U C3=U"],
    ["fldpi"            , ""                                         , "NONE" , "D9 EB"                            , "ANY FPU_PUSH   C0=U C1=W C2=U C3=U"],
    ["fldz"             , ""                                         , "NONE" , "D9 EE"                            , "ANY FPU_PUSH   C0=U C1=W C2=U C3=U"],
    ["fmul"             , "m32fp"                                    , "M"    , "D8 /1"                            , "ANY FPU        C0=U C1=W C2=U C3=U"],
    ["fmul"             , "m64fp"                                    , "M"    , "DC /1"                            , "ANY FPU        C0=U C1=W C2=U C3=U"],
    ["fmul"             , "st(0), st(i)"                             , "O"    , "D8 C8+i"                          , "ANY FPU        C0=U C1=W C2=U C3=U"],
    ["fmul"             , "st(i), st(0)"                             , "O"    , "DC C8+i"                          , "ANY FPU        C0=U C1=W C2=U C3=U"],
    ["fmulp"            , ""                                         , "NONE" , "DE C9"                            , "ANY FPU_POP    C0=U C1=W C2=U C3=U"],
    ["fmulp"            , "st(i)"                                    , "O"    , "DE C8+i"                          , "ANY FPU_POP    C0=U C1=W C2=U C3=U"],
    ["fnclex"           , ""                                         , "NONE" , "DB E2"                            , "ANY FPU        C0=U C1=U C2=U C3=U"],
    ["fninit"           , ""                                         , "NONE" , "DB E3"                            , "ANY FPU        C0=0 C1=0 C2=0 C3=0"],
    ["fnop"             , ""                                         , "NONE" , "D9 D0"                            , "ANY FPU        C0=U C1=U C2=U C3=U"],
    ["fnsave"           , "mem"                                      , "M"    , "DD /6"                            , "ANY FPU        C0=W C1=W C2=W C3=W"],
    ["fnstcw"           , "m16"                                      , "M"    , "D9 /7"                            , "ANY FPU        C0=U C1=U C2=U C3=U"],
    ["fnstenv"          , "mem"                                      , "M"    , "D9 /6"                            , "ANY FPU        C0=U C1=U C2=U C3=U"],
    ["fnstsw"           , "ax"                                       , "NONE" , "DF E0"                            , "ANY FPU        C0=U C1=U C2=U C3=U"],
    ["fnstsw"           , "m16"                                      , "M"    , "DD /7"                            , "ANY FPU        C0=U C1=U C2=U C3=U"],
    ["fpatan"           , ""                                         , "NONE" , "D9 F3"                            , "ANY FPU_POP    C0=U C1=W C2=U C3=U"],
    ["fprem"            , ""                                         , "NONE" , "D9 F8"                            , "ANY FPU        C0=U C1=W C2=U C3=U"],
    ["fprem1"           , ""                                         , "NONE" , "D9 F5"                            , "ANY FPU        C0=W C1=W C2=W C3=W"],
    ["fptan"            , ""                                         , "NONE" , "D9 F2"                            , "ANY FPU_POP    C0=U C1=W C2=W C3=U"],
    ["frndint"          , ""                                         , "NONE" , "D9 FC"                            , "ANY FPU        C0=U C1=W C2=U C3=U"],
    ["frstor"           , "mem"                                      , "M"    , "DD /4"                            , "ANY FPU        C0=W C1=W C2=W C3=W"],
    ["fsave"            , "mem"                                      , "M"    , "9B DD /6"                         , "ANY FPU        C0=W C1=W C2=W C3=W"],
    ["fscale"           , ""                                         , "NONE" , "D9 FD"                            , "ANY FPU        C0=U C1=W C2=U C3=U"],
    ["fsin"             , ""                                         , "NONE" , "D9 FE"                            , "ANY FPU        C0=U C1=W C2=W C3=U"],
    ["fsincos"          , ""                                         , "NONE" , "D9 FB"                            , "ANY FPU_PUSH   C0=U C1=W C2=W C3=U"],
    ["fsqrt"            , ""                                         , "NONE" , "D9 FE"                            , "ANY FPU        C0=U C1=W C2=U C3=U"],
    ["fst"              , "m32fp"                                    , "M"    , "D9 /2"                            , "ANY FPU        C0=U C1=W C2=U C3=U"],
    ["fst"              , "m64fp"                                    , "M"    , "DD /2"                            , "ANY FPU        C0=U C1=W C2=U C3=U"],
    ["fst"              , "st(i)"                                    , "O"    , "DD D0+i"                          , "ANY FPU        C0=U C1=W C2=U C3=U"],
    ["fstcw"            , "m16"                                      , "M"    , "9B D9 /7"                         , "ANY FPU        C0=U C1=U C2=U C3=U"],
    ["fstenv"           , "mem"                                      , "M"    , "9B D9 /6"                         , "ANY FPU        C0=U C1=U C2=U C3=U"],
    ["fstp"             , "m32fp"                                    , "M"    , "D9 /3"                            , "ANY FPU_POP    C0=U C1=W C2=U C3=U"],
    ["fstp"             , "m64fp"                                    , "M"    , "DD /3"                            , "ANY FPU_POP    C0=U C1=W C2=U C3=U"],
    ["fstp"             , "m80fp"                                    , "M"    , "DB /7"                            , "ANY FPU_POP    C0=U C1=W C2=U C3=U"],
    ["fstp"             , "st(i)"                                    , "O"    , "DD D8+i"                          , "ANY FPU_POP    C0=U C1=W C2=U C3=U"],
    ["fstsw"            , "ax"                                       , "NONE" , "9B DF E0"                         , "ANY FPU        C0=U C1=U C2=U C3=U"],
    ["fstsw"            , "m16"                                      , "M"    , "9B DD /7"                         , "ANY FPU        C0=U C1=U C2=U C3=U"],
    ["fsub"             , "m32fp"                                    , "M"    , "D8 /4"                            , "ANY FPU        C0=U C1=W C2=U C3=U"],
    ["fsub"             , "m64fp"                                    , "M"    , "DC /4"                            , "ANY FPU        C0=U C1=W C2=U C3=U"],
    ["fsub"             , "st(0), st(i)"                             , "O"    , "D8 E0+i"                          , "ANY FPU        C0=U C1=W C2=U C3=U"],
    ["fsub"             , "st(i), st(0)"                             , "O"    , "DC E8+i"                          , "ANY FPU        C0=U C1=W C2=U C3=U"],
    ["fsubp"            , ""                                         , "NONE" , "DE E9"                            , "ANY FPU_POP    C0=U C1=W C2=U C3=U"],
    ["fsubp"            , "st(i)"                                    , "O"    , "DE E8+i"                          , "ANY FPU_POP    C0=U C1=W C2=U C3=U"],
    ["fsubr"            , "m32fp"                                    , "M"    , "D8 /5"                            , "ANY FPU        C0=U C1=W C2=U C3=U"],
    ["fsubr"            , "m64fp"                                    , "M"    , "DC /5"                            , "ANY FPU        C0=U C1=W C2=U C3=U"],
    ["fsubr"            , "st(0), st(i)"                             , "O"    , "D8 E8+i"                          , "ANY FPU        C0=U C1=W C2=U C3=U"],
    ["fsubr"            , "st(i), st(0)"                             , "O"    , "DC E0+i"                          , "ANY FPU        C0=U C1=W C2=U C3=U"],
    ["fsubrp"           , ""                                         , "NONE" , "DE E1"                            , "ANY FPU_POP    C0=U C1=W C2=U C3=U"],
    ["fsubrp"           , "st(i)"                                    , "O"    , "DE E0+i"                          , "ANY FPU_POP    C0=U C1=W C2=U C3=U"],
    ["ftst"             , ""                                         , "NONE" , "D9 E4"                            , "ANY FPU        C0=W C1=0 C2=W C3=W"],
    ["fucom"            , ""                                         , "NONE" , "DD E1"                            , "ANY FPU        C0=W C1=0 C2=W C3=W"],
    ["fucom"            , "st(i)"                                    , "O"    , "DD E0+i"                          , "ANY FPU        C0=W C1=0 C2=W C3=W"],
    ["fucomi"           , "st(i)"                                    , "O"    , "DB E8+i"                          , "ANY FPU        C1=0 ZF=W PF=W CF=W"],
    ["fucomip"          , "st(i)"                                    , "O"    , "DF E8+i"                          , "ANY FPU_POP    C1=0 ZF=W PF=W CF=W"],
    ["fucomp"           , ""                                         , "NONE" , "DD E9"                            , "ANY FPU_POP    C0=W C1=W C2=W C3=W"],
    ["fucomp"           , "st(i)"                                    , "O"    , "DD E8+i"                          , "ANY FPU_POP    C0=W C1=W C2=W C3=W"],
    ["fucompp"          , ""                                         , "NONE" , "DA E9"                            , "ANY FPU_POP=2  C0=W C1=W C2=W C3=W"],
    ["fwait/wait"       , ""                                         , "NONE" , "DB"                               , "ANY FPU        C0=U C1=U C2=U C3=U"],
    ["fxam"             , ""                                         , "NONE" , "D9 E5"                            , "ANY FPU        C0=W C1=W C2=W C3=W"],
    ["fxch"             , ""                                         , "NONE" , "D9 C9"                            , "ANY FPU        C0=U C1=0 C2=U C3=U"],
    ["fxch"             , "st(i)"                                    , "O"    , "D9 C8+i"                          , "ANY FPU        C0=U C1=0 C2=U C3=U"],
    ["fxtract"          , ""                                         , "NONE" , "D9 F4"                            , "ANY FPU_PUSH   C0=U C1=W C2=U C3=U"],
    ["fyl2x"            , ""                                         , "NONE" , "D9 F1"                            , "ANY FPU_POP    C0=U C1=W C2=U C3=U"],
    ["fyl2xp1"          , ""                                         , "NONE" , "D9 F9"                            , "ANY FPU_POP    C0=U C1=W C2=U C3=U"],

    // FXSR.
    ["fxrstor"          , "mem"                                      , "NONE" , "0F AE /1"                         , "ANY VOLATILE FXSR C0=W C1=W C2=W C3=W"],
    ["fxrstor64"        , "mem"                                      , "NONE" , "REX.W 0F AE /1"                   , "X64 VOLATILE FXSR C0=W C1=W C2=W C3=W"],
    ["fxsave"           , "mem"                                      , "NONE" , "0F AE /0"                         , "ANY VOLATILE FXSR"],
    ["fxsave64"         , "mem"                                      , "NONE" , "REX.W 0F AE /0"                   , "X64 VOLATILE FXSR"],

    // ADX / BMI / *CNT.
    ["adcx"             , "r32, r32/m32"                             , "RM"   , "66 0F 38 F6 /r"                   , "ANY ADX CF=X"],
    ["adcx"             , "r64, r64/m64"                             , "RM"   , "REX.W 66 0F 38 F6 /r"             , "X64 ADX CF=X"],
    ["adox"             , "r32, r32/m32"                             , "RM"   , "F3 0F 38 F6 /r"                   , "ANY ADX OF=X"],
    ["adox"             , "r64, r64/m64"                             , "RM"   , "REX.W F3 0F 38 F6 /r"             , "X64 ADX OF=X"],
    ["andn"             , "W:r32, r32, r32/m32"                      , "RVM"  , "VEX.NDS.LZ.0F38.W0 F2 /r"         , "ANY BMI OF=0 SF=W ZF=W AF=U PF=U CF=0"],
    ["andn"             , "W:r64, r64, r64/m64"                      , "RVM"  , "VEX.NDS.LZ.0F38.W1 F2 /r"         , "X64 BMI OF=0 SF=W ZF=W AF=U PF=U CF=0"],
    ["bextr"            , "W:r32, r32/m32, r32"                      , "RMV"  , "VEX.NDS.LZ.0F38.W0 F7 /r"         , "ANY BMI OF=0 SF=U ZF=W AF=U PF=U CF=0"],
    ["bextr"            , "W:r64, r64/m64, r64"                      , "RMV"  , "VEX.NDS.LZ.0F38.W0 F7 /r"         , "X64 BMI OF=0 SF=U ZF=W AF=U PF=U CF=0"],
    ["blsi"             , "W:r32, r32/m32"                           , "VM"   , "VEX.NDD.LZ.0F38.W0 F3 /3"         , "ANY BMI OF=0 SF=W ZF=W AF=U PF=U CF=W"],
    ["blsi"             , "W:r64, r64/m64"                           , "VM"   , "VEX.NDD.LZ.0F38.W1 F3 /3"         , "X64 BMI OF=0 SF=W ZF=W AF=U PF=U CF=W"],
    ["blsmsk"           , "W:r32, r32/m32"                           , "VM"   , "VEX.NDD.LZ.0F38.W0 F3 /2"         , "ANY BMI OF=0 SF=W ZF=0 AF=U PF=U CF=W"],
    ["blsmsk"           , "W:r64, r64/m64"                           , "VM"   , "VEX.NDD.LZ.0F38.W1 F3 /2"         , "X64 BMI OF=0 SF=W ZF=0 AF=U PF=U CF=W"],
    ["blsr"             , "W:r32, r32/m32"                           , "VM"   , "VEX.NDD.LZ.0F38.W0 F3 /1"         , "ANY BMI OF=0 SF=W ZF=W AF=U PF=U CF=W"],
    ["blsr"             , "W:r64, r64/m64"                           , "VM"   , "VEX.NDD.LZ.0F38.W1 F3 /1"         , "X64 BMI OF=0 SF=W ZF=W AF=U PF=U CF=W"],
    ["bzhi"             , "W:r32, r32/m32, r32"                      , "RMV"  , "VEX.NDS.LZ.0F38.W0 F5 /r"         , "ANY BMI2 OF=0 SF=W ZF=W AF=U PF=U CF=W"],
    ["bzhi"             , "W:r64, r64/m64, r64"                      , "RMV"  , "VEX.NDS.LZ.0F38.W1 F5 /r"         , "X64 BMI2 OF=0 SF=W ZF=W AF=U PF=U CF=W"],
    ["lzcnt"            , "W:r16, r16/m16"                           , "RM"   , "66 F3 0F BD /r"                   , "ANY LZCNT OF=U SF=U ZF=W AF=U PF=U CF=W"],
    ["lzcnt"            , "W:r32, r32/m32"                           , "RM"   , "F3 0F BD /r"                      , "ANY LZCNT OF=U SF=U ZF=W AF=U PF=U CF=W"],
    ["lzcnt"            , "W:r64, r64/m64"                           , "RM"   , "REX.W F3 0F BD /r"                , "X64 LZCNT OF=U SF=U ZF=W AF=U PF=U CF=W"],
    ["mulx"             , "W:r32, W:r32, r32/m32, <edx>"             , "RVM"  , "VEX.NDD.LZ.F2.0F38.W0 F6 /r"      , "ANY BMI2"],
    ["mulx"             , "W:r64, W:r64, r64/m64, <rdx>"             , "RVM"  , "VEX.NDD.LZ.F2.0F38.W1 F6 /r"      , "X64 BMI2"],
    ["pdep"             , "W:r32, r32, r32/m32"                      , "RVM"  , "VEX.NDS.LZ.F2.0F38.W0 F5 /r"      , "ANY BMI2"],
    ["pdep"             , "W:r64, r64, r64/m64"                      , "RVM"  , "VEX.NDS.LZ.F2.0F38.W1 F5 /r"      , "X64 BMI2"],
    ["pext"             , "W:r32, r32, r32/m32"                      , "RVM"  , "VEX.NDS.LZ.F3.0F38.W0 F5 /r"      , "ANY BMI2"],
    ["pext"             , "W:r64, r64, r64/m64"                      , "RVM"  , "VEX.NDS.LZ.F3.0F38.W1 F5 /r"      , "X64 BMI2"],
    ["popcnt"           , "W:r16, r16/m16"                           , "RM"   , "66 F3 0F B8 /r"                   , "ANY POPCNT OF=0 SF=0 ZF=W AF=0 PF=0 CF=0"],
    ["popcnt"           , "W:r32, r32/m32"                           , "RM"   , "F3 0F B8 /r"                      , "ANY POPCNT OF=0 SF=0 ZF=W AF=0 PF=0 CF=0"],
    ["popcnt"           , "W:r64, r64/m64"                           , "RM"   , "REX.W F3 0F B8 /r"                , "X64 POPCNT OF=0 SF=0 ZF=W AF=0 PF=0 CF=0"],
    ["rorx"             , "W:r32, r32/m32, ib"                       , "RMI"  , "VEX.LZ.F2.0F3A.W0 F0 /r ib"       , "ANY BMI2"],
    ["rorx"             , "W:r64, r64/m64, ib"                       , "RMI"  , "VEX.LZ.F2.0F3A.W1 F0 /r ib"       , "X64 BMI2"],
    ["sarx"             , "W:r32, r32/m32, r32"                      , "RMV"  , "VEX.NDS.LZ.F3.0F38.W0 F7 /r"      , "ANY BMI2"],
    ["sarx"             , "W:r64, r64/m64, r64"                      , "RMV"  , "VEX.NDS.LZ.F3.0F38.W1 F7 /r"      , "X64 BMI2"],
    ["shlx"             , "W:r32, r32/m32, r32"                      , "RMV"  , "VEX.NDS.LZ.66.0F38.W0 F7 /r"      , "ANY BMI2"],
    ["shlx"             , "W:r64, r64/m64, r64"                      , "RMV"  , "VEX.NDS.LZ.66.0F38.W1 F7 /r"      , "X64 BMI2"],
    ["shrx"             , "W:r32, r32/m32, r32"                      , "RMV"  , "VEX.NDS.LZ.F2.0F38.W0 F7 /r"      , "ANY BMI2"],
    ["shrx"             , "W:r64, r64/m64, r64"                      , "RMV"  , "VEX.NDS.LZ.F2.0F38.W1 F7 /r"      , "X64 BMI2"],
    ["tzcnt"            , "W:r16, r16/m16"                           , "RM"   , "66 F3 0F BC /r"                   , "ANY BMI OF=U SF=U ZF=W AF=U PF=U CF=W"],
    ["tzcnt"            , "W:r32, r32/m32"                           , "RM"   , "F3 0F BC /r"                      , "ANY BMI OF=U SF=U ZF=W AF=U PF=U CF=W"],
    ["tzcnt"            , "W:r64, r64/m64"                           , "RM"   , "REX.W F3 0F BC /r"                , "X64 BMI OF=U SF=U ZF=W AF=U PF=U CF=W"],

    // TBM.
    ["blci"             , "W:r32, r32/m32"                           , "VM"   , "XOP.NDD.LZ.M9.W0 02 /6"           , "ANY TBM"],
    ["blci"             , "W:r64, r64/m64"                           , "VM"   , "XOP.NDD.LZ.M9.W1 02 /6"           , "X64 TBM"],
    ["blcic"            , "W:r32, r32/m32"                           , "VM"   , "XOP.NDD.LZ.M9.W0 01 /5"           , "ANY TBM"],
    ["blcic"            , "W:r64, r64/m64"                           , "VM"   , "XOP.NDD.LZ.M9.W1 01 /5"           , "X64 TBM"],
    ["blsic"            , "W:r32, r32/m32"                           , "VM"   , "XOP.NDD.LZ.M9.W0 01 /6"           , "ANY TBM"],
    ["blsic"            , "W:r64, r64/m64"                           , "VM"   , "XOP.NDD.LZ.M9.W1 01 /6"           , "X64 TBM"],
    ["blcfill"          , "W:r32, r32/m32"                           , "VM"   , "XOP.NDD.LZ.M9.W0 01 /1"           , "ANY TBM"],
    ["blcfill"          , "W:r64, r64/m64"                           , "VM"   , "XOP.NDD.LZ.M9.W1 01 /1"           , "X64 TBM"],
    ["blsfill"          , "W:r32, r32/m32"                           , "VM"   , "XOP.NDD.LZ.M9.W0 01 /2"           , "ANY TBM"],
    ["blsfill"          , "W:r64, r64/m64"                           , "VM"   , "XOP.NDD.LZ.M9.W1 01 /2"           , "X64 TBM"],
    ["blcmsk"           , "W:r32, r32/m32"                           , "VM"   , "XOP.NDD.LZ.M9.W0 02 /1"           , "ANY TBM"],
    ["blcmsk"           , "W:r64, r64/m64"                           , "VM"   , "XOP.NDD.LZ.M9.W1 02 /1"           , "X64 TBM"],
    ["blcs"             , "W:r32, r32/m32"                           , "VM"   , "XOP.NDD.LZ.M9.W0 01 /3"           , "ANY TBM"],
    ["blcs"             , "W:r64, r64/m64"                           , "VM"   , "XOP.NDD.LZ.M9.W1 01 /3"           , "X64 TBM"],
    ["tzmsk"            , "W:r32, r32/m32"                           , "VM"   , "XOP.NDD.LZ.M9.W0 01 /4"           , "ANY TBM"],
    ["tzmsk"            , "W:r64, r64/m64"                           , "VM"   , "XOP.NDD.LZ.M9.W1 01 /4"           , "X64 TBM"],
    ["t1mskc"           , "W:r32, r32/m32"                           , "VM"   , "XOP.NDD.LZ.M9.W0 01 /7"           , "ANY TBM"],
    ["t1mskc"           , "W:r64, r64/m64"                           , "VM"   , "XOP.NDD.LZ.M9.W1 01 /7"           , "X64 TBM"],

    // CRC32.
    ["crc32"            , "r32, r8/m8"                               , "RM"   , "F2 0F 38 F0 /r"                   , "ANY SSE4_2"],
    ["crc32"            , "r32, r16/m16"                             , "RM"   , "66 F2 0F 38 F1 /r"                , "ANY SSE4_2"],
    ["crc32"            , "r32, r32/m32"                             , "RM"   , "F2 0F 38 F1 /r"                   , "ANY SSE4_2"],
    ["crc32"            , "r64, r8/m8"                               , "RM"   , "REX.W F2 0F 38 F0 /r"             , "X64 SSE4_2"],
    ["crc32"            , "r64, r64/m64"                             , "RM"   , "REX.W F2 0F 38 F1 /r"             , "X64 SSE4_2"],

    // MOVBE.
    ["movbe"            , "W[7:0]:r16, m16"                          , "RM"   , "66 0F 38 F0 /r"                   , "ANY MOVBE"],
    ["movbe"            , "W[7:0]:r32, m32"                          , "RM"   , "0F 38 F0 /r"                      , "ANY MOVBE"],
    ["movbe"            , "W[7:0]:r64, m64"                          , "RM"   , "REX.W 0F 38 F0 /r"                , "X64 MOVBE"],
    ["movbe"            , "W:m16, r16"                               , "MR"   , "66 0F 38 F1 /r"                   , "ANY MOVBE"],
    ["movbe"            , "W:m32, r32"                               , "MR"   , "0F 38 F1 /r"                      , "ANY MOVBE"],
    ["movbe"            , "W:m64, r64"                               , "MR"   , "REX.W 0F 38 F1 /r"                , "X64 MOVBE"],

    // CLFLUSH / CLFLUSH_OPT.
    ["clflush"          , "R:mem"                                    , "M"    , "0F AE /7"                         , "ANY VOLATILE CLFLUSH"],
    ["clflushopt"       , "R:mem"                                    , "M"    , "66 0F AE /7"                      , "ANY VOLATILE CLFLUSH_OPT"],

    // PREFETCHW / PREFETCHWT1.
    ["prefetchw"        , "R:mem"                                    , "M"    , "0F 0D /1"                         , "ANY PREFETCHW OF=U SF=U ZF=U AF=U PF=U CF=U"],
    ["prefetchwt1"      , "R:mem"                                    , "M"    , "0F 0D /2"                         , "ANY PREFETCHWT1 OF=U SF=U ZF=U AF=U PF=U CF=U"],

    // CLWB / PCOMMIT.
    ["clwb"             , "R:mem"                                    , "M"    , "66 0F AE /6"                      , "ANY CLWB"],
    ["pcommit"          , ""                                         , "NONE" , "66 0F AE F8"                      , "ANY PCOMMIT"],

    // RDTSC/RDTSCP.
    ["rdtsc"            , "W:<edx>, W:<eax>"                         , "NONE" , "0F 31"                            , "ANY VOLATILE RDTSC"],
    ["rdtscp"           , "W:<edx>, W:<eax>, W:<ecx>"                , "NONE" , "0F 01 F9"                         , "ANY VOLATILE RDTSCP"],

    // FSGSBASE.
    ["rdfsbase"         , "W:r32"                                    , "M"    , "F3 0F AE /0"                      , "X64 VOLATILE FSGSBASE"],
    ["rdfsbase"         , "W:r64"                                    , "M"    , "REX.W F3 0F AE /0"                , "X64 VOLATILE FSGSBASE"],
    ["rdgsbase"         , "W:r32"                                    , "M"    , "F3 0F AE /1"                      , "X64 VOLATILE FSGSBASE"],
    ["rdgsbase"         , "W:r64"                                    , "M"    , "REX.W F3 0F AE /1"                , "X64 VOLATILE FSGSBASE"],
    ["wrfsbase"         , "R:r32"                                    , "M"    , "F3 0F AE /2"                      , "X64 VOLATILE FSGSBASE"],
    ["wrfsbase"         , "R:r64"                                    , "M"    , "REX.W F3 0F AE /2"                , "X64 VOLATILE FSGSBASE"],
    ["wrgsbase"         , "R:r32"                                    , "M"    , "F3 0F AE /3"                      , "X64 VOLATILE FSGSBASE"],
    ["wrgsbase"         , "R:r64"                                    , "M"    , "REX.W F3 0F AE /3"                , "X64 VOLATILE FSGSBASE"],

    // RDRAND / RDSEED.
    ["rdrand"           , "W:r16"                                    , "M"    , "66 0F C7 /6"                      , "ANY RDRAND OF=0 SF=0 ZF=0 AF=0 PF=0 CF=W"],
    ["rdrand"           , "W:r32"                                    , "M"    , "0F C7 /6"                         , "ANY RDRAND OF=0 SF=0 ZF=0 AF=0 PF=0 CF=W"],
    ["rdrand"           , "W:r64"                                    , "M"    , "REX.W 0F C7 /6"                   , "X64 RDRAND OF=0 SF=0 ZF=0 AF=0 PF=0 CF=W"],
    ["rdseed"           , "W:r16"                                    , "M"    , "66 0F C7 /7"                      , "ANY RDSEED OF=0 SF=0 ZF=0 AF=0 PF=0 CF=W"],
    ["rdseed"           , "W:r32"                                    , "M"    , "0F C7 /7"                         , "ANY RDSEED OF=0 SF=0 ZF=0 AF=0 PF=0 CF=W"],
    ["rdseed"           , "W:r64"                                    , "M"    , "REX.W 0F C7 /7"                   , "X64 RDSEED OF=0 SF=0 ZF=0 AF=0 PF=0 CF=W"],

    // XSAVE.
    ["xgetbv"           , "R:<ecx>, W:<edx>, W:<eax>"                , "NONE" , "0F 01 D0"                         , "ANY VOLATILE XSAVE XCR=R"],
    ["xsetbv"           , "R:<ecx>, R:<edx>, R:<eax>"                , "NONE" , "0F 01 D1"                         , "ANY VOLATILE XSAVE XCR=W PRIVILEGE=L0"],

    ["xrstor"           , "mem, <edx>, <eax>"                        , "M"    , "0F AE /5"                         , "ANY VOLATILE XSAVE XCR=R"],
    ["xrstor64"         , "mem, <edx>, <eax>"                        , "M"    , "REX.W 0F AE /5"                   , "X64 VOLATILE XSAVE XCR=R"],
    ["xrstors"          , "mem, <edx>, <eax>"                        , "M"    , "0F C7 /3"                         , "ANY VOLATILE XSAVE XCR=R"],
    ["xrstors64"        , "mem, <edx>, <eax>"                        , "M"    , "REX.W 0F C7 /3"                   , "X64 VOLATILE XSAVE XCR=R"],
    ["xsave"            , "mem, <edx>, <eax>"                        , "M"    , "0F AE /4"                         , "ANY VOLATILE XSAVE XCR=R"],
    ["xsave64"          , "mem, <edx>, <eax>"                        , "M"    , "REX.W 0F AE /4"                   , "X64 VOLATILE XSAVE XCR=R"],
    ["xsavec"           , "mem, <edx>, <eax>"                        , "M"    , "0F C7 /4"                         , "ANY VOLATILE XSAVE XCR=R"],
    ["xsavec64"         , "mem, <edx>, <eax>"                        , "M"    , "REX.W 0F C7 /4"                   , "X64 VOLATILE XSAVE XCR=R"],
    ["xsaveopt"         , "mem, <edx>, <eax>"                        , "M"    , "0F AE /6"                         , "ANY VOLATILE XSAVE_OPT XCR=R"],
    ["xsaveopt64"       , "mem, <edx>, <eax>"                        , "M"    , "REX.W 0F AE /6"                   , "X64 VOLATILE XSAVE_OPT XCR=R"],
    ["xsaves"           , "mem, <edx>, <eax>"                        , "M"    , "0F C7 /5"                         , "ANY VOLATILE XSAVE XCR=R"],
    ["xsaves64"         , "mem, <edx>, <eax>"                        , "M"    , "REX.W 0F C7 /5"                   , "X64 VOLATILE XSAVE XCR=R"],

    // MPX.
    ["bndcl"            , "R:bnd, r32/m32"                           , "RM"   , "F3 0F 1A /r"                      , "X86 VOLATILE MPX"],
    ["bndcl"            , "R:bnd, r64/m64"                           , "RM"   , "F3 0F 1A /r"                      , "X64 VOLATILE MPX"],
    ["bndcn"            , "R:bnd, r32/m32"                           , "RM"   , "F2 0F 1B /r"                      , "X86 VOLATILE MPX"],
    ["bndcn"            , "R:bnd, r64/m64"                           , "RM"   , "F2 0F 1B /r"                      , "X64 VOLATILE MPX"],
    ["bndcu"            , "R:bnd, r32/m32"                           , "RM"   , "F2 0F 1A /r"                      , "X86 VOLATILE MPX"],
    ["bndcu"            , "R:bnd, r64/m64"                           , "RM"   , "F2 0F 1A /r"                      , "X64 VOLATILE MPX"],
    ["bndldx"           , "W:bnd, mem"                               , "RM"   , "0F 1A /r"                         , "ANY VOLATILE MPX"],
    ["bndmk"            , "W:bnd, mem"                               , "RM"   , "F3 0F 1B /r"                      , "ANY VOLATILE MPX"],
    ["bndmov"           , "W:bnd, bnd/mem"                           , "RM"   , "66 0F 1A /r"                      , "ANY VOLATILE MPX"],
    ["bndmov"           , "W:bnd/mem, bnd"                           , "MR"   , "66 0F 1B /r"                      , "ANY VOLATILE MPX"],
    ["bndstx"           , "W:mem, bnd"                               , "MR"   , "0F 1B /r"                         , "ANY VOLATILE MPX"],

    // MMX & SSE.
    ["addpd"            , "X:xmm, xmm/m128"                          , "RM"   , "66 0F 58 /r"                      , "ANY SSE2"],
    ["addps"            , "X:xmm, xmm/m128"                          , "RM"   , "0F 58 /r"                         , "ANY SSE"],
    ["addsd"            , "X:xmm, xmm/m64"                           , "RM"   , "F2 0F 58 /r"                      , "ANY SSE2"],
    ["addss"            , "X:xmm, xmm/m32"                           , "RM"   , "F3 0F 58 /r"                      , "ANY SSE"],
    ["addsubpd"         , "X:xmm, xmm/m128"                          , "RM"   , "66 0F D0 /r"                      , "ANY SSE3"],
    ["addsubps"         , "X:xmm, xmm/m128"                          , "RM"   , "F2 0F D0 /r"                      , "ANY SSE3"],
    ["andnpd"           , "X:xmm, xmm/m128"                          , "RM"   , "66 0F 55 /r"                      , "ANY SSE2"],
    ["andnps"           , "X:xmm, xmm/m128"                          , "RM"   , "0F 55 /r"                         , "ANY SSE"],
    ["andpd"            , "X:xmm, xmm/m128"                          , "RM"   , "66 0F 54 /r"                      , "ANY SSE2"],
    ["andps"            , "X:xmm, xmm/m128"                          , "RM"   , "0F 54 /r"                         , "ANY SSE"],
    ["blendpd"          , "X:xmm, xmm/m128, ib"                      , "RMI"  , "66 0F 3A 0D /r ib"                , "ANY SSE4_1"],
    ["blendps"          , "X:xmm, xmm/m128, ib"                      , "RMI"  , "66 0F 3A 0C /r ib"                , "ANY SSE4_1"],
    ["blendvpd"         , "X:xmm, xmm/m128, <xmm0>"                  , "RM"   , "66 0F 38 15 /r"                   , "ANY SSE4_1"],
    ["blendvps"         , "X:xmm, xmm/m128, <xmm0>"                  , "RM"   , "66 0F 38 14 /r"                   , "ANY SSE4_1"],
    ["cmppd"            , "X:xmm, xmm/m128, ib"                      , "RM"   , "66 0F C2 /r ib"                   , "ANY SSE2"],
    ["cmpps"            , "X:xmm, xmm/m128, ib"                      , "RM"   , "0F C2 /r ib"                      , "ANY SSE"],
    ["cmpsd"            , "X:xmm, xmm/m64, ib"                       , "RM"   , "F2 0F C2 /r ib"                   , "ANY SSE2"],
    ["cmpss"            , "X:xmm, xmm/m32, ib"                       , "RM"   , "F3 0F C2 /r ib"                   , "ANY SSE"],
    ["comisd"           , "R:xmm, xmm/m64"                           , "RM"   , "66 0F 2F /r"                      , "ANY SSE2 OF=0 SF=0 ZF=W AF=0 PF=W CF=W"],
    ["comiss"           , "R:xmm, xmm/m32"                           , "RM"   , "0F 2F /r"                         , "ANY SSE OF=0 SF=0 ZF=W AF=0 PF=W CF=W"],
    ["cvtdq2pd"         , "W:xmm, xmm/m64"                           , "RM"   , "F3 0F E6 /r"                      , "ANY SSE2"],
    ["cvtdq2ps"         , "W:xmm, xmm/m128"                          , "RM"   , "0F 5B /r"                         , "ANY SSE2"],
    ["cvtpd2dq"         , "W:xmm, xmm/m128"                          , "RM"   , "F2 0F E6 /r"                      , "ANY SSE2"],
    ["cvtpd2pi"         , "W[7:0]:mm, xmm/m128"                      , "RM"   , "66 0F 2D /r"                      , "ANY SSE2"],
    ["cvtpd2ps"         , "W:xmm, xmm/m128"                          , "RM"   , "66 0F 5A /r"                      , "ANY SSE2"],
    ["cvtpi2pd"         , "W:xmm, mm/m64"                            , "RM"   , "66 0F 2A /r"                      , "ANY SSE2"],
    ["cvtpi2ps"         , "W[7:0]:xmm, mm/m64"                       , "RM"   , "0F 2A /r"                         , "ANY SSE"],
    ["cvtps2dq"         , "W:xmm, xmm/m128"                          , "RM"   , "66 0F 5B /r"                      , "ANY SSE2"],
    ["cvtps2pd"         , "W:xmm, xmm/m64"                           , "RM"   , "0F 5A /r"                         , "ANY SSE2"],
    ["cvtps2pi"         , "W[7:0]:mm, xmm/m64"                       , "RM"   , "0F 2D /r"                         , "ANY SSE"],
    ["cvtsd2si"         , "W[7:0]:r32, xmm/m64"                      , "RM"   , "F2 0F 2D /r"                      , "ANY SSE2"],
    ["cvtsd2si"         , "W[7:0]:r64, xmm/m64"                      , "RM"   , "REX.W F2 0F 2D /r"                , "X64 SSE2"],
    ["cvtsd2ss"         , "W[3:0]:xmm, xmm/m64"                      , "RM"   , "F2 0F 5A /r"                      , "ANY SSE2"],
    ["cvtsi2sd"         , "W[7:0]:xmm, r32/m32"                      , "RM"   , "F2 0F 2A /r"                      , "ANY SSE2"],
    ["cvtsi2sd"         , "W[7:0]:xmm, r64/m64"                      , "RM"   , "REX.W F2 0F 2A /r"                , "X64 SSE2"],
    ["cvtsi2ss"         , "W[3:0]:xmm, r32/m32"                      , "RM"   , "F3 0F 2A /r"                      , "ANY SSE"],
    ["cvtsi2ss"         , "W[3:0]:xmm, r64/m64"                      , "RM"   , "REX.W F3 0F 2A /r"                , "X64 SSE"],
    ["cvtss2sd"         , "W[7:0]:xmm, xmm/m32"                      , "RM"   , "F3 0F 5A /r"                      , "ANY SSE2"],
    ["cvtss2si"         , "W[7:0]:r32, xmm/m32"                      , "RM"   , "F3 0F 2D /r"                      , "ANY SSE"],
    ["cvtss2si"         , "W[7:0]:r64, xmm/m32"                      , "RM"   , "REX.W F3 0F 2D /r"                , "X64 SSE"],
    ["cvttpd2dq"        , "W:xmm, xmm/m128"                          , "RM"   , "66 0F E6 /r"                      , "ANY SSE2"],
    ["cvttpd2pi"        , "W[7:0]:mm, xmm/m128"                      , "RM"   , "66 0F 2C /r"                      , "ANY SSE2"],
    ["cvttps2dq"        , "W:xmm, xmm/m128"                          , "RM"   , "F3 0F 5B /r"                      , "ANY SSE2"],
    ["cvttps2pi"        , "W[7:0]:mm, xmm/m64"                       , "RM"   , "0F 2C /r"                         , "ANY SSE"],
    ["cvttsd2si"        , "W[7:0]:r32, xmm/m64"                      , "RM"   , "F2 0F 2C /r"                      , "ANY SSE2"],
    ["cvttsd2si"        , "W[7:0]:r64, xmm/m64"                      , "RM"   , "REX.W F2 0F 2C /r"                , "X64 SSE2"],
    ["cvttss2si"        , "W[7:0]:r32, xmm/m32"                      , "RM"   , "F3 0F 2C /r"                      , "ANY SSE"],
    ["cvttss2si"        , "W[7:0]:r64, xmm/m32"                      , "RM"   , "REX.W F3 0F 2C /r"                , "X64 SSE"],
    ["divpd"            , "X:xmm, xmm/m128"                          , "RM"   , "66 0F 5E /r"                      , "ANY SSE2"],
    ["divps"            , "X:xmm, xmm/m128"                          , "RM"   , "0F 5E /r"                         , "ANY SSE"],
    ["divsd"            , "X:xmm, xmm/m64"                           , "RM"   , "F2 0F 5E /r"                      , "ANY SSE2"],
    ["divss"            , "X:xmm, xmm/m32"                           , "RM"   , "F3 0F 5E /r"                      , "ANY SSE"],
    ["dppd"             , "X:xmm, xmm/m128, ib"                      , "RMI"  , "66 0F 3A 41 /r ib"                , "ANY SSE4_1"],
    ["dpps"             , "X:xmm, xmm/m128, ib"                      , "RMI"  , "66 0F 3A 40 /r ib"                , "ANY SSE4_1"],
    ["emms"             , ""                                         , "NONE" , "0F 77"                            , "ANY VOLATILE MMX"],
    ["extractps"        , "W:r32/m32, xmm, ib"                       , "MRI"  , "66 0F 3A 17 /r ib"                , "ANY SSE4_1"],
    ["extrq"            , "xmm, ib, ib"                              , "RII"  , "66 0F 78 /0 ib ib"                , "ANY SSE4A"],
    ["extrq"            , "xmm, xmm"                                 , "RM"   , "66 0F 79 /r"                      , "ANY SSE4A"],
    ["haddpd"           , "xmm, xmm/m128"                            , "RM"   , "66 0F 7C /r"                      , "ANY SSE3"],
    ["haddps"           , "xmm, xmm/m128"                            , "RM"   , "F2 0F 7C /r"                      , "ANY SSE3"],
    ["hsubpd"           , "xmm, xmm/m128"                            , "RM"   , "66 0F 7D /r"                      , "ANY SSE3"],
    ["hsubps"           , "xmm, xmm/m128"                            , "RM"   , "F2 0F 7D /r"                      , "ANY SSE3"],
    ["insertps"         , "xmm, xmm/m32, ib"                         , "RMI"  , "66 0F 3A 21 /r ib"                , "ANY SSE4_1"],
    ["insertq"          , "xmm, xmm"                                 , "RM"   , "F2 0F 79 /r"                      , "ANY SSE4A"],
    ["insertq"          , "xmm, xmm, ib, ib"                         , "RMII" , "F2 0F 78 /r ib ib"                , "ANY SSE4A"],
    ["lddqu"            , "W:xmm, m128"                              , "RM"   , "F2 0F F0 /r"                      , "ANY SSE3"],
    ["ldmxcsr"          , "R:m32"                                    , "M"    , "0F AE /2"                         , "ANY SSE"],
    ["lfence"           , ""                                         , "NONE" , "0F AE /5"                         , "ANY VOLATILE SSE2"],
    ["maskmovdqu"       , "X:xmm, xmm, <ds:zdi>"                     , "RM"   , "66 0F F7 /r"                      , "ANY SSE2"],
    ["maskmovq"         , "X:mm, mm, <ds:zdi>"                       , "RM"   , "0F F7 /r"                         , "ANY SSE"],
    ["maxpd"            , "X:xmm, xmm/m128"                          , "RM"   , "66 0F 5F /r"                      , "ANY SSE2"],
    ["maxps"            , "X:xmm, xmm/m128"                          , "RM"   , "0F 5F /r"                         , "ANY SSE"],
    ["maxsd"            , "X:xmm, xmm/m64"                           , "RM"   , "F2 0F 5F /r"                      , "ANY SSE2"],
    ["maxss"            , "X:xmm, xmm/m32"                           , "RM"   , "F3 0F 5F /r"                      , "ANY SSE"],
    ["mfence"           , ""                                         , "NONE" , "0F AE /6"                         , "ANY VOLATILE SSE2"],
    ["minpd"            , "X:xmm, xmm/m128"                          , "RM"   , "66 0F 5D /r"                      , "ANY SSE2"],
    ["minps"            , "X:xmm, xmm/m128"                          , "RM"   , "0F 5D /r"                         , "ANY SSE"],
    ["minsd"            , "X:xmm, xmm/m64"                           , "RM"   , "F2 0F 5D /r"                      , "ANY SSE2"],
    ["minss"            , "X:xmm, xmm/m32"                           , "RM"   , "F3 0F 5D /r"                      , "ANY SSE"],
    ["movapd"           , "W:xmm, xmm/m128"                          , "RM"   , "66 0F 28 /r"                      , "ANY SSE2"],
    ["movapd"           , "W:xmm/m128, xmm"                          , "MR"   , "66 0F 29 /r"                      , "ANY SSE2"],
    ["movaps"           , "W:xmm, xmm/m128"                          , "RM"   , "0F 28 /r"                         , "ANY SSE"],
    ["movaps"           , "W:xmm/m128, xmm"                          , "MR"   , "0F 29 /r"                         , "ANY SSE"],
    ["movd"             , "W[7:0]:mm, r32/m32"                       , "RM"   , "0F 6E /r"                         , "ANY MMX"],
    ["movd"             , "W[7:0]:r32/m32, mm"                       , "MR"   , "0F 7E /r"                         , "ANY MMX"],
    ["movd"             , "W[7:0]:r32/m32, xmm"                      , "MR"   , "66 0F 7E /r"                      , "ANY SSE2"],
    ["movd"             , "W:xmm, r32/m32"                           , "RM"   , "66 0F 6E /r"                      , "ANY SSE2"],
    ["movddup"          , "W:xmm, xmm/m64"                           , "RM"   , "F2 0F 12 /r"                      , "ANY SSE3"],
    ["movdq2q"          , "W[7:0]:mm, xmm"                           , "RM"   , "F2 0F D6 /r"                      , "ANY SSE2"],
    ["movdqa"           , "W:xmm, xmm/m128"                          , "RM"   , "66 0F 6F /r"                      , "ANY SSE2"],
    ["movdqa"           , "W:xmm/m128, xmm"                          , "MR"   , "66 0F 7F /r"                      , "ANY SSE2"],
    ["movdqu"           , "W:xmm, xmm/m128"                          , "RM"   , "F3 0F 6F /r"                      , "ANY SSE2"],
    ["movdqu"           , "W:xmm/m128, xmm"                          , "MR"   , "F3 0F 7F /r"                      , "ANY SSE2"],
    ["movhlps"          , "W[7:0]:xmm, xmm"                          , "RM"   , "0F 12 /r"                         , "ANY SSE"],
    ["movhpd"           , "W:m64, xmm"                               , "MR"   , "66 0F 17 /r"                      , "ANY SSE2"],
    ["movhpd"           , "W[15:8]:xmm, m64"                         , "RM"   , "66 0F 16 /r"                      , "ANY SSE2"],
    ["movhps"           , "W:m64, xmm"                               , "MR"   , "0F 17 /r"                         , "ANY SSE"],
    ["movhps"           , "W[15:8]:xmm, m64"                         , "RM"   , "0F 16 /r"                         , "ANY SSE"],
    ["movlhps"          , "W[15:8]:xmm, xmm"                         , "RM"   , "0F 16 /r"                         , "ANY SSE"],
    ["movlpd"           , "W:m64, xmm"                               , "MR"   , "66 0F 13 /r"                      , "ANY SSE2"],
    ["movlpd"           , "W[7:0]:xmm, m64"                          , "RM"   , "66 0F 12 /r"                      , "ANY SSE2"],
    ["movlps"           , "W:m64, xmm"                               , "MR"   , "0F 13 /r"                         , "ANY SSE"],
    ["movlps"           , "W[7:0]:xmm, m64"                          , "RM"   , "0F 12 /r"                         , "ANY SSE"],
    ["movmskpd"         , "W:r32, xmm"                               , "RM"   , "66 0F 50 /r"                      , "ANY SSE2"],
    ["movmskpd"         , "W:r64, xmm"                               , "RM"   , "66 0F 50 /r"                      , "X64 SSE2"],
    ["movmskps"         , "W:r32, xmm"                               , "RM"   , "0F 50 /r"                         , "ANY SSE"],
    ["movmskps"         , "W:r64, xmm"                               , "RM"   , "0F 50 /r"                         , "X64 SSE"],
    ["movntdq"          , "W:m128, xmm"                              , "MR"   , "66 0F E7 /r"                      , "ANY SSE2"],
    ["movntdqa"         , "W:xmm, m128"                              , "RM"   , "66 0F 38 2A /r"                   , "ANY SSE4_1"],
    ["movnti"           , "W:m32, r32"                               , "MR"   , "0F C3 /r"                         , "ANY SSE2"],
    ["movnti"           , "W:m64, r64"                               , "MR"   , "REX.W 0F C3 /r"                   , "X64 SSE2"],
    ["movntpd"          , "W:m128, xmm"                              , "MR"   , "66 0F 2B /r"                      , "ANY SSE2"],
    ["movntps"          , "W:m128, xmm"                              , "MR"   , "0F 2B /r"                         , "ANY SSE"],
    ["movntq"           , "W:m64, mm"                                , "MR"   , "0F E7 /r"                         , "ANY MMX"],
    ["movntsd"          , "W:m64, xmm"                               , "RM"   , "F2 0F 2B /r"                      , "ANY SSE4A"],
    ["movntss"          , "W:m32, xmm"                               , "RM"   , "F3 0F 2B /r"                      , "ANY SSE4A"],
    ["movq"             , "W:mm, mm/m64"                             , "RM"   , "0F 6F /r"                         , "ANY MMX"],
    ["movq"             , "W:mm, r64/m64"                            , "RM"   , "REX.W 0F 6E /r"                   , "X64 MMX"],
    ["movq"             , "W:mm/m64, mm"                             , "MR"   , "0F 7F /r"                         , "ANY MMX"],
    ["movq"             , "W:r64/m64, mm"                            , "MR"   , "REX.W 0F 7E /r"                   , "X64 MMX"],
    ["movq"             , "W:r64/m64, xmm"                           , "MR"   , "REX.W 66 0F 7E /r"                , "X64 SSE2"],
    ["movq"             , "W:xmm, r64/m64"                           , "RM"   , "REX.W 66 0F 6E /r"                , "X64 SSE2"],
    ["movq"             , "W:xmm, xmm/m64"                           , "RM"   , "F3 0F 7E /r"                      , "ANY SSE2"],
    ["movq"             , "W:xmm/m64, xmm"                           , "MR"   , "66 0F D6 /r"                      , "ANY SSE2"],
    ["movq2dq"          , "W:xmm, mm"                                , "RM"   , "F3 0F D6 /r"                      , "ANY SSE2"],
    ["movsd"            , "W[7:0]:xmm, xmm"                          , "RM"   , "F2 0F 10 /r"                      , "ANY SSE2"],
    ["movsd"            , "W:xmm, m64"                               , "RM"   , "F2 0F 10 /r"                      , "ANY SSE2"],
    ["movsd"            , "W:m64, xmm"                               , "MR"   , "F2 0F 11 /r"                      , "ANY SSE2"],
    ["movshdup"         , "W:xmm, xmm/m128"                          , "RM"   , "F3 0F 16 /r"                      , "ANY SSE3"],
    ["movsldup"         , "W:xmm, xmm/m128"                          , "RM"   , "F3 0F 12 /r"                      , "ANY SSE3"],
    ["movss"            , "W[3:0]:xmm, xmm"                          , "RM"   , "F3 0F 10 /r"                      , "ANY SSE"],
    ["movss"            , "W:xmm, m32"                               , "RM"   , "F3 0F 10 /r"                      , "ANY SSE"],
    ["movss"            , "W[3:0]:m32, xmm"                          , "MR"   , "F3 0F 11 /r"                      , "ANY SSE"],
    ["movups"           , "W:xmm, xmm/m128"                          , "RM"   , "0F 10 /r"                         , "ANY SSE"],
    ["movups"           , "W:xmm/m128, xmm"                          , "MR"   , "0F 11 /r"                         , "ANY SSE"],
    ["movupd"           , "W:xmm, xmm/m128"                          , "RM"   , "66 0F 10 /r"                      , "ANY SSE2"],
    ["movupd"           , "W:xmm/m128, xmm"                          , "MR"   , "66 0F 11 /r"                      , "ANY SSE2"],
    ["mpsadbw"          , "X:xmm, xmm/m128, ib"                      , "RMI"  , "66 0F 3A 42 /r ib"                , "ANY SSE4_1"],
    ["mulpd"            , "X:xmm, xmm/m128"                          , "RM"   , "66 0F 59 /r"                      , "ANY SSE2"],
    ["mulps"            , "X:xmm, xmm/m128"                          , "RM"   , "0F 59 /r"                         , "ANY SSE"],
    ["mulsd"            , "X:xmm, xmm/m64"                           , "RM"   , "F2 0F 59 /r"                      , "ANY SSE2"],
    ["mulss"            , "X:xmm, xmm/m32"                           , "RM"   , "F3 0F 59 /r"                      , "ANY SSE"],
    ["orpd"             , "X:xmm, xmm/m128"                          , "RM"   , "66 0F 56 /r"                      , "ANY SSE2"],
    ["orps"             , "X:xmm, xmm/m128"                          , "RM"   , "0F 56 /r"                         , "ANY SSE"],
    ["pabsb"            , "X:mm, mm/m64"                             , "RM"   , "0F 38 1C /r"                      , "ANY SSSE3"],
    ["pabsb"            , "X:xmm, xmm/m128"                          , "RM"   , "66 0F 38 1C /r"                   , "ANY SSSE3"],
    ["pabsd"            , "X:mm, mm/m64"                             , "RM"   , "0F 38 1E /r"                      , "ANY SSSE3"],
    ["pabsd"            , "X:xmm, xmm/m128"                          , "RM"   , "66 0F 38 1E /r"                   , "ANY SSSE3"],
    ["pabsw"            , "X:mm, mm/m64"                             , "RM"   , "0F 38 1D /r"                      , "ANY SSSE3"],
    ["pabsw"            , "X:xmm, xmm/m128"                          , "RM"   , "66 0F 38 1D /r"                   , "ANY SSSE3"],
    ["packssdw"         , "X:mm, mm/m64"                             , "RM"   , "0F 6B /r"                         , "ANY MMX"],
    ["packssdw"         , "X:xmm, xmm/m128"                          , "RM"   , "66 0F 6B /r"                      , "ANY SSE2"],
    ["packsswb"         , "X:mm, mm/m64"                             , "RM"   , "0F 63 /r"                         , "ANY MMX"],
    ["packsswb"         , "X:xmm, xmm/m128"                          , "RM"   , "66 0F 63 /r"                      , "ANY SSE2"],
    ["packusdw"         , "X:xmm, xmm/m128"                          , "RM"   , "66 0F 38 2B /r"                   , "ANY SSE4_1"],
    ["packuswb"         , "X:mm, mm/m64"                             , "RM"   , "0F 67 /r"                         , "ANY MMX"],
    ["packuswb"         , "X:xmm, xmm/m128"                          , "RM"   , "66 0F 67 /r"                      , "ANY SSE2"],
    ["paddb"            , "X:mm, mm/m64"                             , "RM"   , "0F FC /r"                         , "ANY MMX"],
    ["paddb"            , "X:xmm, xmm/m128"                          , "RM"   , "66 0F FC /r"                      , "ANY SSE2"],
    ["paddd"            , "X:mm, mm/m64"                             , "RM"   , "0F FE /r"                         , "ANY MMX"],
    ["paddd"            , "X:xmm, xmm/m128"                          , "RM"   , "66 0F FE /r"                      , "ANY SSE2"],
    ["paddq"            , "X:mm, mm/m64"                             , "RM"   , "0F D4 /r"                         , "ANY SSE2"],
    ["paddq"            , "X:xmm, xmm/m128"                          , "RM"   , "66 0F D4 /r"                      , "ANY SSE2"],
    ["paddsb"           , "X:mm, mm/m64"                             , "RM"   , "0F EC /r"                         , "ANY MMX"],
    ["paddsb"           , "X:xmm, xmm/m128"                          , "RM"   , "66 0F EC /r"                      , "ANY SSE2"],
    ["paddsw"           , "X:mm, mm/m64"                             , "RM"   , "0F ED /r"                         , "ANY MMX"],
    ["paddsw"           , "X:xmm, xmm/m128"                          , "RM"   , "66 0F ED /r"                      , "ANY SSE2"],
    ["paddusb"          , "X:mm, mm/m64"                             , "RM"   , "0F DC /r"                         , "ANY MMX"],
    ["paddusb"          , "X:xmm, xmm/m128"                          , "RM"   , "66 0F DC /r"                      , "ANY SSE2"],
    ["paddusw"          , "X:mm, mm/m64"                             , "RM"   , "0F DD /r"                         , "ANY MMX"],
    ["paddusw"          , "X:xmm, xmm/m128"                          , "RM"   , "66 0F DD /r"                      , "ANY SSE2"],
    ["paddw"            , "X:mm, mm/m64"                             , "RM"   , "0F FD /r"                         , "ANY MMX"],
    ["paddw"            , "X:xmm, xmm/m128"                          , "RM"   , "66 0F FD /r"                      , "ANY SSE2"],
    ["palignr"          , "X:mm, mm/m64, ib"                         , "RMI"  , "0F 3A 0F /r ib"                   , "ANY SSE3"],
    ["palignr"          , "X:xmm, xmm/m128, ib"                      , "RMI"  , "66 0F 3A 0F /r ib"                , "ANY SSE3"],
    ["pand"             , "X:mm, mm/m64"                             , "RM"   , "0F DB /r"                         , "ANY MMX"],
    ["pand"             , "X:xmm, xmm/m128"                          , "RM"   , "66 0F DB /r"                      , "ANY SSE2"],
    ["pandn"            , "X:mm, mm/m64"                             , "RM"   , "0F DF /r"                         , "ANY MMX"],
    ["pandn"            , "X:xmm, xmm/m128"                          , "RM"   , "66 0F DF /r"                      , "ANY SSE2"],
    ["pavgb"            , "X:mm, mm/m64"                             , "RM"   , "0F E0 /r"                         , "ANY SSE"],
    ["pavgb"            , "X:xmm, xmm/m128"                          , "RM"   , "66 0F E0 /r"                      , "ANY SSE2"],
    ["pavgw"            , "X:mm, mm/m64"                             , "RM"   , "0F E3 /r"                         , "ANY SSE"],
    ["pavgw"            , "X:xmm, xmm/m128"                          , "RM"   , "66 0F E3 /r"                      , "ANY SSE2"],
    ["pblendvb"         , "X:xmm, xmm/m128, <xmm0>"                  , "RM"   , "66 0F E0 /r"                      , "ANY SSE4_1"],
    ["pblendw"          , "X:xmm, xmm/m128, ib"                      , "RMI"  , "66 0F 3A 0E /r ib"                , "ANY SSE4_1"],
    ["pclmulqdq"        , "X:xmm, xmm/m128, ib"                      , "RMI"  , "66 0F 3A 44 /r ib"                , "ANY PCLMULQDQ"],
    ["pcmpeqb"          , "X:mm, mm/m64"                             , "RM"   , "0F 74 /r"                         , "ANY MMX"],
    ["pcmpeqb"          , "X:xmm, xmm/m128"                          , "RM"   , "66 0F 74 /r"                      , "ANY SSE2"],
    ["pcmpeqd"          , "X:mm, mm/m64"                             , "RM"   , "0F 76 /r"                         , "ANY MMX"],
    ["pcmpeqd"          , "X:xmm, xmm/m128"                          , "RM"   , "66 0F 76 /r"                      , "ANY SSE2"],
    ["pcmpeqq"          , "X:xmm, xmm/m128"                          , "RM"   , "66 0F 38 29 /r"                   , "ANY SSE4_1"],
    ["pcmpeqw"          , "X:mm, mm/m64"                             , "RM"   , "0F 75 /r"                         , "ANY MMX"],
    ["pcmpeqw"          , "X:xmm, xmm/m128"                          , "RM"   , "66 0F 75 /r"                      , "ANY SSE2"],
    ["pcmpestri"        , "W:<ecx>, xmm, <eax>, xmm/m128, <edx>, ib" , "RMI"  , "66 0F 3A 61 /r ib"                , "ANY SSE4_2 OF=W SF=W ZF=W AF=0 PF=0 CF=W"],
    ["pcmpestrm"        , "W:<xmm0>, xmm, <eax>, xmm/m128, <edx>, ib", "RMI"  , "66 0F 3A 60 /r ib"                , "ANY SSE4_2 OF=W SF=W ZF=W AF=0 PF=0 CF=W"],
    ["pcmpgtb"          , "X:mm, mm/m64"                             , "RM"   , "0F 64 /r"                         , "ANY MMX"],
    ["pcmpgtb"          , "X:xmm, xmm/m128"                          , "RM"   , "66 0F 64 /r"                      , "ANY SSE2"],
    ["pcmpgtd"          , "X:mm, mm/m64"                             , "RM"   , "0F 66 /r"                         , "ANY MMX"],
    ["pcmpgtd"          , "X:xmm, xmm/m128"                          , "RM"   , "66 0F 66 /r"                      , "ANY SSE2"],
    ["pcmpgtq"          , "X:xmm, xmm/m128"                          , "RM"   , "66 0F 38 37 /r"                   , "ANY SSE4_1"],
    ["pcmpgtw"          , "X:mm, mm/m64"                             , "RM"   , "0F 65 /r"                         , "ANY MMX"],
    ["pcmpgtw"          , "X:xmm, xmm/m128"                          , "RM"   , "66 0F 65 /r"                      , "ANY SSE2"],
    ["pcmpistri"        , "W:<ecx>, xmm, xmm/m128, ib"               , "RMI"  , "66 0F 3A 63 /r ib"                , "ANY SSE4_2 OF=W SF=W ZF=W AF=0 PF=0 CF=W"],
    ["pcmpistrm"        , "W:<xmm0>, xmm, xmm/m128, ib"              , "RMI"  , "66 0F 3A 62 /r ib"                , "ANY SSE4_2 OF=W SF=W ZF=W AF=0 PF=0 CF=W"],
    ["pextrb"           , "W:r32/m8, xmm, ib"                        , "MRI"  , "66 0F 3A 14 /r"                   , "ANY SSE4_1"],
    ["pextrd"           , "W:r32/m32, xmm, ib"                       , "MRI"  , "66 0F 3A 16 /r"                   , "ANY SSE4_1"],
    ["pextrq"           , "W:r64/m64, xmm, ib"                       , "MRI"  , "REX.W 66 0F 3A 16 /r ib"          , "X64 SSE4_1"],
    ["pextrw"           , "W:r32, mm, ib"                            , "RMI"  , "0F C5 /r ib"                      , "ANY SSE"],
    ["pextrw"           , "W:r32, xmm, ib"                           , "RMI"  , "66 0F C5 /r ib"                   , "ANY SSE2"],
    ["pextrw"           , "W:r32/m16, xmm, ib"                       , "MRI"  , "66 0F 3A 15 /r ib"                , "ANY SSE4_1"],
    ["phaddd"           , "X:mm, mm/m64"                             , "RM"   , "0F 38 02 /r"                      , "ANY SSSE3"],
    ["phaddd"           , "X:xmm, xmm/m128"                          , "RM"   , "66 0F 38 02 /r"                   , "ANY SSSE3"],
    ["phaddsw"          , "X:mm, mm/m64"                             , "RM"   , "0F 38 03 /r"                      , "ANY SSSE3"],
    ["phaddsw"          , "X:xmm, xmm/m128"                          , "RM"   , "66 0F 38 03 /r"                   , "ANY SSSE3"],
    ["phaddw"           , "X:mm, mm/m64"                             , "RM"   , "0F 38 01 /r"                      , "ANY SSSE3"],
    ["phaddw"           , "X:xmm, xmm/m128"                          , "RM"   , "66 0F 38 01 /r"                   , "ANY SSSE3"],
    ["phminposuw"       , "X:xmm, xmm/m128"                          , "RM"   , "66 0F 38 41 /r"                   , "ANY SSE4_1"],
    ["phsubd"           , "X:mm, mm/m64"                             , "RM"   , "0F 38 06 /r"                      , "ANY SSSE3"],
    ["phsubd"           , "X:xmm, xmm/m128"                          , "RM"   , "66 0F 38 06 /r"                   , "ANY SSSE3"],
    ["phsubsw"          , "X:mm, mm/m64"                             , "RM"   , "0F 38 07 /r"                      , "ANY SSSE3"],
    ["phsubsw"          , "X:xmm, xmm/m128"                          , "RM"   , "66 0F 38 07 /r"                   , "ANY SSSE3"],
    ["phsubw"           , "X:mm, mm/m64"                             , "RM"   , "0F 38 05 /r"                      , "ANY SSSE3"],
    ["phsubw"           , "X:xmm, xmm/m128"                          , "RM"   , "66 0F 38 05 /r"                   , "ANY SSSE3"],
    ["pinsrb"           , "X:xmm, r32/m8, ib"                        , "RMI"  , "66 0F 3A 20 /r ib"                , "ANY SSE4_1"],
    ["pinsrd"           , "X:xmm, r32/m32, ib"                       , "RMI"  , "66 0F 3A 22 /r ib"                , "ANY SSE4_1"],
    ["pinsrq"           , "X:xmm, r64/m64, ib"                       , "RMI"  , "REX.W 66 0F 3A 22 /r ib"          , "X64 SSE4_1"],
    ["pinsrw"           , "X:mm, r32/m16, ib"                        , "RMI"  , "0F C4 /r ib"                      , "ANY SSE"],
    ["pinsrw"           , "X:xmm, r32/m16, ib"                       , "RMI"  , "66 0F C4 /r ib"                   , "ANY SSE2"],
    ["pmaddubsw"        , "X:mm, mm/m64"                             , "RM"   , "0F 38 04 /r"                      , "ANY SSSE3"],
    ["pmaddubsw"        , "X:xmm, xmm/m128"                          , "RM"   , "66 0F 38 04 /r"                   , "ANY SSSE3"],
    ["pmaddwd"          , "X:mm, mm/m64"                             , "RM"   , "0F F5 /r"                         , "ANY MMX"],
    ["pmaddwd"          , "X:xmm, xmm/m128"                          , "RM"   , "66 0F F5 /r"                      , "ANY SSE2"],
    ["pmaxsb"           , "X:xmm, xmm/m128"                          , "RM"   , "66 0F 38 3C /r"                   , "ANY SSE4_1"],
    ["pmaxsd"           , "X:xmm, xmm/m128"                          , "RM"   , "66 0F 38 3D /r"                   , "ANY SSE4_1"],
    ["pmaxsw"           , "X:mm, mm/m64"                             , "RM"   , "0F EE /r"                         , "ANY SSE"],
    ["pmaxsw"           , "X:xmm, xmm/m128"                          , "RM"   , "66 0F EE /r"                      , "ANY SSE2"],
    ["pmaxub"           , "X:mm, mm/m64"                             , "RM"   , "0F DE /r"                         , "ANY SSE"],
    ["pmaxub"           , "X:xmm, xmm/m128"                          , "RM"   , "66 0F DE /r"                      , "ANY SSE2"],
    ["pmaxud"           , "X:xmm, xmm/m128"                          , "RM"   , "66 0F 38 3F /r"                   , "ANY SSE4_1"],
    ["pmaxuw"           , "X:xmm, xmm/m128"                          , "RM"   , "66 0F 38 3E /r"                   , "ANY SSE4_1"],
    ["pminsb"           , "X:xmm, xmm/m128"                          , "RM"   , "66 0F 38 38 /r"                   , "ANY SSE4_1"],
    ["pminsd"           , "X:xmm, xmm/m128"                          , "RM"   , "66 0F 38 39 /r"                   , "ANY SSE4_1"],
    ["pminsw"           , "X:mm, mm/m64"                             , "RM"   , "0F EA /r"                         , "ANY SSE"],
    ["pminsw"           , "X:xmm, xmm/m128"                          , "RM"   , "66 0F EA /r"                      , "ANY SSE2"],
    ["pminub"           , "X:mm, mm/m64"                             , "RM"   , "0F DA /r"                         , "ANY SSE"],
    ["pminub"           , "X:xmm, xmm/m128"                          , "RM"   , "66 0F DA /r"                      , "ANY SSE2"],
    ["pminud"           , "X:xmm, xmm/m128"                          , "RM"   , "66 0F 38 3B /r"                   , "ANY SSE4_1"],
    ["pminuw"           , "X:xmm, xmm/m128"                          , "RM"   , "66 0F 38 3A /r"                   , "ANY SSE4_1"],
    ["pmovmskb"         , "W:r32, mm"                                , "RM"   , "0F D7 /r"                         , "ANY SSE"],
    ["pmovmskb"         , "W:r64, mm"                                , "RM"   , "0F D7 /r"                         , "X64 SSE"],
    ["pmovmskb"         , "W:r32, xmm"                               , "RM"   , "66 0F D7 /r"                      , "ANY SSE2"],
    ["pmovmskb"         , "W:r64, xmm"                               , "RM"   , "66 0F D7 /r"                      , "X64 SSE2"],
    ["pmovsxbd"         , "W:xmm, xmm/m32"                           , "RM"   , "66 0F 38 21 /r"                   , "ANY SSE4_1"],
    ["pmovsxbq"         , "W:xmm, xmm/m16"                           , "RM"   , "66 0F 38 22 /r"                   , "ANY SSE4_1"],
    ["pmovsxbw"         , "W:xmm, xmm/m64"                           , "RM"   , "66 0F 38 20 /r"                   , "ANY SSE4_1"],
    ["pmovsxdq"         , "W:xmm, xmm/m64"                           , "RM"   , "66 0F 38 25 /r"                   , "ANY SSE4_1"],
    ["pmovsxwd"         , "W:xmm, xmm/m64"                           , "RM"   , "66 0F 38 23 /r"                   , "ANY SSE4_1"],
    ["pmovsxwq"         , "W:xmm, xmm/m32"                           , "RM"   , "66 0F 38 24 /r"                   , "ANY SSE4_1"],
    ["pmovzxbd"         , "W:xmm, xmm/m32"                           , "RM"   , "66 0F 38 31 /r"                   , "ANY SSE4_1"],
    ["pmovzxbq"         , "W:xmm, xmm/m16"                           , "RM"   , "66 0F 38 32 /r"                   , "ANY SSE4_1"],
    ["pmovzxbw"         , "W:xmm, xmm/m64"                           , "RM"   , "66 0F 38 30 /r"                   , "ANY SSE4_1"],
    ["pmovzxdq"         , "W:xmm, xmm/m64"                           , "RM"   , "66 0F 38 35 /r"                   , "ANY SSE4_1"],
    ["pmovzxwd"         , "W:xmm, xmm/m64"                           , "RM"   , "66 0F 38 33 /r"                   , "ANY SSE4_1"],
    ["pmovzxwq"         , "W:xmm, xmm/m32"                           , "RM"   , "66 0F 38 34 /r"                   , "ANY SSE4_1"],
    ["pmuldq"           , "X:xmm, xmm/m128"                          , "RM"   , "66 0F 38 28 /r"                   , "ANY SSE4_1"],
    ["pmulhrsw"         , "X:mm, mm/m64"                             , "RM"   , "0F 38 0B /r"                      , "ANY SSSE3"],
    ["pmulhrsw"         , "X:xmm, xmm/m128"                          , "RM"   , "66 0F 38 0B /r"                   , "ANY SSSE3"],
    ["pmulhuw"          , "X:mm, mm/m64"                             , "RM"   , "0F E4 /r"                         , "ANY SSE"],
    ["pmulhuw"          , "X:xmm, xmm/m128"                          , "RM"   , "66 0F E4 /r"                      , "ANY SSE2"],
    ["pmulhw"           , "X:mm, mm/m64"                             , "RM"   , "0F E5 /r"                         , "ANY MMX"],
    ["pmulhw"           , "X:xmm, xmm/m128"                          , "RM"   , "66 0F E5 /r"                      , "ANY SSE2"],
    ["pmulld"           , "X:xmm, xmm/m128"                          , "RM"   , "66 0F 38 40 /r"                   , "ANY SSE4_1"],
    ["pmullw"           , "X:mm, mm/m64"                             , "RM"   , "0F D5 /r"                         , "ANY MMX"],
    ["pmullw"           , "X:xmm, xmm/m128"                          , "RM"   , "66 0F D5 /r"                      , "ANY SSE2"],
    ["pmuludq"          , "X:mm, mm/m64"                             , "RM"   , "0F F4 /r"                         , "ANY SSE2"],
    ["pmuludq"          , "X:xmm, xmm/m128"                          , "RM"   , "66 0F F4 /r"                      , "ANY SSE2"],
    ["por"              , "X:mm, mm/m64"                             , "RM"   , "0F EB /r"                         , "ANY MMX"],
    ["por"              , "X:xmm, xmm/m128"                          , "RM"   , "66 0F EB /r"                      , "ANY SSE2"],
    ["prefetchnta"      , "R:mem"                                    , "M"    , "0F 18 /0"                         , "ANY SSE"],
    ["prefetcht0"       , "R:mem"                                    , "M"    , "0F 18 /1"                         , "ANY SSE"],
    ["prefetcht1"       , "R:mem"                                    , "M"    , "0F 18 /2"                         , "ANY SSE"],
    ["prefetcht2"       , "R:mem"                                    , "M"    , "0F 18 /3"                         , "ANY SSE"],
    ["psadbw"           , "X:mm, mm/m64"                             , "RM"   , "0F F6 /r"                         , "ANY SSE"],
    ["psadbw"           , "X:xmm, xmm/m128"                          , "RM"   , "66 0F F6 /r"                      , "ANY SSE2"],
    ["pshufb"           , "W:mm, mm/m64"                             , "RM"   , "0F 38 00 /r"                      , "ANY SSSE3"],
    ["pshufb"           , "W:xmm, xmm/m128"                          , "RM"   , "66 0F 38 00 /r"                   , "ANY SSSE3"],
    ["pshufd"           , "W:xmm, xmm/m128, ib"                      , "RMI"  , "66 0F 70 /r ib"                   , "ANY SSE2"],
    ["pshufhw"          , "W:xmm, xmm/m128, ib"                      , "RMI"  , "F3 0F 70 /r ib"                   , "ANY SSE2"],
    ["pshuflw"          , "W:xmm, xmm/m128, ib"                      , "RMI"  , "F2 0F 70 /r ib"                   , "ANY SSE2"],
    ["pshufw"           , "W:mm, mm/m64, ib"                         , "RMI"  , "0F 70 /r ib"                      , "ANY MMX"],
    ["psignb"           , "X:mm, mm/m64"                             , "RM"   , "0F 38 08 /r"                      , "ANY SSSE3"],
    ["psignb"           , "X:xmm, xmm/m128"                          , "RM"   , "66 0F 38 08 /r"                   , "ANY SSSE3"],
    ["psignd"           , "X:mm, mm/m64"                             , "RM"   , "0F 38 0A /r"                      , "ANY SSSE3"],
    ["psignd"           , "X:xmm, xmm/m128"                          , "RM"   , "66 0F 38 0A /r"                   , "ANY SSSE3"],
    ["psignw"           , "X:mm, mm/m64"                             , "RM"   , "0F 38 09 /r"                      , "ANY SSSE3"],
    ["psignw"           , "X:xmm, xmm/m128"                          , "RM"   , "66 0F 38 09 /r"                   , "ANY SSSE3"],
    ["pslld"            , "X:mm, ib"                                 , "MI"   , "0F 72 /6 ib"                      , "ANY MMX"],
    ["pslld"            , "X:mm, mm/m64"                             , "RM"   , "0F F2 /r"                         , "ANY MMX"],
    ["pslld"            , "X:xmm, ib"                                , "MI"   , "66 0F 72 /6 ib"                   , "ANY SSE2"],
    ["pslld"            , "X:xmm, xmm/m128"                          , "RM"   , "66 0F F2 /r"                      , "ANY SSE2"],
    ["pslldq"           , "X:xmm, ib"                                , "MI"   , "66 0F 73 /7 ib"                   , "ANY SSE2"],
    ["psllq"            , "X:mm, ib"                                 , "MI"   , "0F 73 /6 ib"                      , "ANY MMX"],
    ["psllq"            , "X:mm, mm/m64"                             , "RM"   , "0F F3 /r"                         , "ANY MMX"],
    ["psllq"            , "X:xmm, ib"                                , "MI"   , "66 0F 73 /6 ib"                   , "ANY SSE2"],
    ["psllq"            , "X:xmm, xmm/m128"                          , "RM"   , "66 0F F3 /r"                      , "ANY SSE2"],
    ["psllw"            , "X:mm, ib"                                 , "MI"   , "0F 71 /6 ib"                      , "ANY MMX"],
    ["psllw"            , "X:mm, mm/m64"                             , "RM"   , "0F F1 /r"                         , "ANY MMX"],
    ["psllw"            , "X:xmm, ib"                                , "MI"   , "66 0F 71 /6 ib"                   , "ANY SSE2"],
    ["psllw"            , "X:xmm, xmm/m128"                          , "RM"   , "66 0F F1 /r"                      , "ANY SSE2"],
    ["psrad"            , "X:mm, ib"                                 , "MI"   , "0F 72 /4 ib"                      , "ANY MMX"],
    ["psrad"            , "X:mm, mm/m64"                             , "RM"   , "0F E2 /r"                         , "ANY MMX"],
    ["psrad"            , "X:xmm, ib"                                , "MI"   , "66 0F 72 /4 ib"                   , "ANY SSE2"],
    ["psrad"            , "X:xmm, xmm/m128"                          , "RM"   , "66 0F E2 /r"                      , "ANY SSE2"],
    ["psraw"            , "X:mm, ib"                                 , "MI"   , "0F 71 /4 ib"                      , "ANY MMX"],
    ["psraw"            , "X:mm, mm/m64"                             , "RM"   , "0F E1 /r"                         , "ANY MMX"],
    ["psraw"            , "X:xmm, ib"                                , "MI"   , "66 0F 71 /4 ib"                   , "ANY SSE2"],
    ["psraw"            , "X:xmm, xmm/m128"                          , "RM"   , "66 0F E1 /r"                      , "ANY SSE2"],
    ["psrld"            , "X:mm, ib"                                 , "MI"   , "0F 72 /2 ib"                      , "ANY MMX"],
    ["psrld"            , "X:mm, mm/m64"                             , "RM"   , "0F D2 /r"                         , "ANY MMX"],
    ["psrld"            , "X:xmm, ib"                                , "MI"   , "66 0F 72 /2 ib"                   , "ANY SSE2"],
    ["psrld"            , "X:xmm, xmm/m128"                          , "RM"   , "66 0F D2 /r"                      , "ANY SSE2"],
    ["psrldq"           , "X:xmm, ib"                                , "MI"   , "66 0F 73 /3 ib"                   , "ANY SSE2"],
    ["psrlq"            , "X:mm, ib"                                 , "MI"   , "0F 73 /2 ib"                      , "ANY MMX"],
    ["psrlq"            , "X:mm, mm/m64"                             , "RM"   , "0F D3 /r"                         , "ANY MMX"],
    ["psrlq"            , "X:xmm, ib"                                , "MI"   , "66 0F 73 /2 ib"                   , "ANY SSE2"],
    ["psrlq"            , "X:xmm, xmm/m128"                          , "RM"   , "66 0F D3 /r"                      , "ANY SSE2"],
    ["psrlw"            , "X:mm, ib"                                 , "MI"   , "0F 71 /2 ib"                      , "ANY MMX"],
    ["psrlw"            , "X:mm, mm/m64"                             , "RM"   , "0F D1 /r"                         , "ANY MMX"],
    ["psrlw"            , "X:xmm, ib"                                , "MI"   , "66 0F 71 /2 ib"                   , "ANY SSE2"],
    ["psrlw"            , "X:xmm, xmm/m128"                          , "RM"   , "66 0F D1 /r"                      , "ANY SSE2"],
    ["psubb"            , "X:mm, mm/m64"                             , "RM"   , "0F F8 /r"                         , "ANY MMX"],
    ["psubb"            , "X:xmm, xmm/m128"                          , "RM"   , "66 0F F8 /r"                      , "ANY SSE2"],
    ["psubd"            , "X:mm, mm/m64"                             , "RM"   , "0F FA /r"                         , "ANY MMX"],
    ["psubd"            , "X:xmm, xmm/m128"                          , "RM"   , "66 0F FA /r"                      , "ANY SSE2"],
    ["psubq"            , "X:mm, mm/m64"                             , "RM"   , "0F FB /r"                         , "ANY SSE2"],
    ["psubq"            , "X:xmm, xmm/m128"                          , "RM"   , "66 0F FB /r"                      , "ANY SSE2"],
    ["psubsb"           , "X:mm, mm/m64"                             , "RM"   , "0F E8 /r"                         , "ANY MMX"],
    ["psubsb"           , "X:xmm, xmm/m128"                          , "RM"   , "66 0F E8 /r"                      , "ANY SSE2"],
    ["psubsw"           , "X:mm, mm/m64"                             , "RM"   , "0F E9 /r"                         , "ANY MMX"],
    ["psubsw"           , "X:xmm, xmm/m128"                          , "RM"   , "66 0F E9 /r"                      , "ANY SSE2"],
    ["psubusb"          , "X:mm, mm/m64"                             , "RM"   , "0F D8 /r"                         , "ANY MMX"],
    ["psubusb"          , "X:xmm, xmm/m128"                          , "RM"   , "66 0F D8 /r"                      , "ANY SSE2"],
    ["psubusw"          , "X:mm, mm/m64"                             , "RM"   , "0F D9 /r"                         , "ANY MMX"],
    ["psubusw"          , "X:xmm, xmm/m128"                          , "RM"   , "66 0F D9 /r"                      , "ANY SSE2"],
    ["psubw"            , "X:mm, mm/m64"                             , "RM"   , "0F F9 /r"                         , "ANY MMX"],
    ["psubw"            , "X:xmm, xmm/m128"                          , "RM"   , "66 0F F9 /r"                      , "ANY SSE2"],
    ["ptest"            , "R:xmm, xmm/m128"                          , "RM"   , "66 0F 38 17 /r"                   , "ANY SSE4_1 OF=0 SF=0 ZF=W AF=0 PF=0 CF=W"],
    ["punpckhbw"        , "X:mm, mm/m64"                             , "RM"   , "0F 68 /r"                         , "ANY MMX"],
    ["punpckhbw"        , "X:xmm, xmm/m128"                          , "RM"   , "66 0F 68 /r"                      , "ANY SSE2"],
    ["punpckhdq"        , "X:mm, mm/m64"                             , "RM"   , "0F 6A /r"                         , "ANY MMX"],
    ["punpckhdq"        , "X:xmm, xmm/m128"                          , "RM"   , "66 0F 6A /r"                      , "ANY SSE2"],
    ["punpckhqdq"       , "X:xmm, xmm/m128"                          , "RM"   , "66 0F 6D /r"                      , "ANY SSE2"],
    ["punpckhwd"        , "X:mm, mm/m64"                             , "RM"   , "0F 69 /r"                         , "ANY MMX"],
    ["punpckhwd"        , "X:xmm, xmm/m128"                          , "RM"   , "66 0F 69 /r"                      , "ANY SSE2"],
    ["punpcklbw"        , "X:mm, mm/m64"                             , "RM"   , "0F 60 /r"                         , "ANY MMX"],
    ["punpcklbw"        , "X:xmm, xmm/m128"                          , "RM"   , "66 0F 60 /r"                      , "ANY SSE2"],
    ["punpckldq"        , "X:mm, mm/m64"                             , "RM"   , "0F 62 /r"                         , "ANY MMX"],
    ["punpckldq"        , "X:xmm, xmm/m128"                          , "RM"   , "66 0F 62 /r"                      , "ANY SSE2"],
    ["punpcklqdq"       , "X:xmm, xmm/m128"                          , "RM"   , "66 0F 6C /r"                      , "ANY SSE2"],
    ["punpcklwd"        , "X:mm, mm/m64"                             , "RM"   , "0F 61 /r"                         , "ANY MMX"],
    ["punpcklwd"        , "X:xmm, xmm/m128"                          , "RM"   , "66 0F 61 /r"                      , "ANY SSE2"],
    ["pxor"             , "X:mm, mm/m64"                             , "RM"   , "0F EF /r"                         , "ANY MMX"],
    ["pxor"             , "X:xmm, xmm/m128"                          , "RM"   , "66 0F EF /r"                      , "ANY SSE2"],
    ["rcpps"            , "W:xmm, xmm/m128"                          , "RM"   , "0F 53 /r"                         , "ANY SSE"],
    ["rcpss"            , "W[3:0]:xmm, xmm/m32"                      , "RM"   , "F3 0F 53 /r"                      , "ANY SSE"],
    ["roundpd"          , "W:xmm, xmm/m128, ib"                      , "RMI"  , "66 0F 3A 09 /r ib"                , "ANY SSE4_1"],
    ["roundps"          , "W:xmm, xmm/m128, ib"                      , "RMI"  , "66 0F 3A 08 /r ib"                , "ANY SSE4_1"],
    ["roundsd"          , "W[7:0]:xmm, xmm/m64, ib"                  , "RMI"  , "66 0F 3A 0B /r ib"                , "ANY SSE4_1"],
    ["roundss"          , "W[3:0]:xmm, xmm/m32, ib"                  , "RMI"  , "66 0F 3A 0A /r ib"                , "ANY SSE4_1"],
    ["rsqrtps"          , "W:xmm, xmm/m128"                          , "RM"   , "0F 52 /r"                         , "ANY SSE"],
    ["rsqrtss"          , "W[3:0]:xmm, xmm/m32"                      , "RM"   , "F3 0F 52 /r"                      , "ANY SSE"],
    ["sfence"           , ""                                         , "NONE" , "0F AE /7"                         , "ANY VOLATILE SSE"],
    ["shufpd"           , "X:xmm, xmm/m128, ib"                      , "RMI"  , "66 0F C6 /r ib"                   , "ANY SSE2"],
    ["shufps"           , "X:xmm, xmm/m128, ib"                      , "RMI"  , "0F C6 /r ib"                      , "ANY SSE"],
    ["sqrtpd"           , "W:xmm, xmm/m128"                          , "RM"   , "66 0F 51 /r"                      , "ANY SSE2"],
    ["sqrtps"           , "W:xmm, xmm/m128"                          , "RM"   , "0F 51 /r"                         , "ANY SSE"],
    ["sqrtsd"           , "W[7:0]:xmm, xmm/m64"                      , "RM"   , "F2 0F 51 /r"                      , "ANY SSE2"],
    ["sqrtss"           , "W[3:0]:xmm, xmm/m32"                      , "RM"   , "F3 0F 51 /r"                      , "ANY SSE"],
    ["stmxcsr"          , "W:m32"                                    , "M"    , "0F AE /3"                         , "ANY SSE"],
    ["subpd"            , "X:xmm, xmm/m128"                          , "RM"   , "66 0F 5C /r"                      , "ANY SSE2"],
    ["subps"            , "X:xmm, xmm/m128"                          , "RM"   , "0F 5C /r"                         , "ANY SSE"],
    ["subsd"            , "X:xmm, xmm/m64"                           , "RM"   , "F2 0F 5C /r"                      , "ANY SSE2"],
    ["subss"            , "X:xmm, xmm/m32"                           , "RM"   , "F3 0F 5C /r"                      , "ANY SSE"],
    ["ucomisd"          , "R:xmm, xmm/m64"                           , "RM"   , "66 0F 2E /r"                      , "ANY SSE2 OF=0 SF=0 ZF=W AF=0 PF=W CF=W"],
    ["ucomiss"          , "R:xmm, xmm/m32"                           , "RM"   , "0F 2E /r"                         , "ANY SSE OF=0 SF=0 ZF=W AF=0 PF=W CF=W"],
    ["unpckhpd"         , "X:xmm, xmm/m128"                          , "RM"   , "66 0F 15 /r"                      , "ANY SSE2"],
    ["unpckhps"         , "X:xmm, xmm/m128"                          , "RM"   , "0F 15 /r"                         , "ANY SSE"],
    ["unpcklpd"         , "X:xmm, xmm/m128"                          , "RM"   , "66 0F 14 /r"                      , "ANY SSE2"],
    ["unpcklps"         , "X:xmm, xmm/m128"                          , "RM"   , "0F 14 /r"                         , "ANY SSE"],
    ["xorpd"            , "X:xmm, xmm/m128"                          , "RM"   , "66 0F 57 /r"                      , "ANY SSE2"],
    ["xorps"            , "X:xmm, xmm/m128"                          , "RM"   , "0F 57 /r"                         , "ANY SSE"],

    // 3DNOW.
    ["femms"            , ""                                         , "NONE" , "0F 0E"                            , "ANY VOLATILE 3DNOW"],
    ["pavgusb"          , "X:mm, mm/m64"                             , "RM"   , "0F 0F /r BF"                      , "ANY 3DNOW"],
    ["pf2id"            , "W:mm, mm/m64"                             , "RM"   , "0F 0F /r 1D"                      , "ANY 3DNOW"],
    ["pf2iw"            , "W:mm, mm/m64"                             , "RM"   , "0F 0F /r 1C"                      , "ANY 3DNOW2"],
    ["pfacc"            , "X:mm, mm/m64"                             , "RM"   , "0F 0F /r AE"                      , "ANY 3DNOW"],
    ["pfadd"            , "X:mm, mm/m64"                             , "RM"   , "0F 0F /r 9E"                      , "ANY 3DNOW"],
    ["pfcmpeq"          , "X:mm, mm/m64"                             , "RM"   , "0F 0F /r B0"                      , "ANY 3DNOW"],
    ["pfcmpge"          , "X:mm, mm/m64"                             , "RM"   , "0F 0F /r 90"                      , "ANY 3DNOW"],
    ["pfcmpgt"          , "X:mm, mm/m64"                             , "RM"   , "0F 0F /r A0"                      , "ANY 3DNOW"],
    ["pfmax"            , "X:mm, mm/m64"                             , "RM"   , "0F 0F /r A4"                      , "ANY 3DNOW"],
    ["pfmin"            , "X:mm, mm/m64"                             , "RM"   , "0F 0F /r 94"                      , "ANY 3DNOW"],
    ["pfmul"            , "X:mm, mm/m64"                             , "RM"   , "0F 0F /r B4"                      , "ANY 3DNOW"],
    ["pfnacc"           , "X:mm, mm/m64"                             , "RM"   , "0F 0F /r 8A"                      , "ANY 3DNOW2"],
    ["pfpnacc"          , "X:mm, mm/m64"                             , "RM"   , "0F 0F /r 8E"                      , "ANY 3DNOW2"],
    ["pfrcp"            , "W:mm, mm/m64"                             , "RM"   , "0F 0F /r 96"                      , "ANY 3DNOW"],
    ["pfrcpit1"         , "X:mm, mm/m64"                             , "RM"   , "0F 0F /r A6"                      , "ANY 3DNOW"],
    ["pfrcpit2"         , "X:mm, mm/m64"                             , "RM"   , "0F 0F /r B6"                      , "ANY 3DNOW"],
    ["pfrcpv"           , "X:mm, mm/m64"                             , "RM"   , "0F 0F /r 86"                      , "ANY GEODE"],
    ["pfrsqit1"         , "W:mm, mm/m64"                             , "RM"   , "0F 0F /r A7"                      , "ANY 3DNOW"],
    ["pfrsqrt"          , "W:mm, mm/m64"                             , "RM"   , "0F 0F /r 97"                      , "ANY 3DNOW"],
    ["pfrsqrtv"         , "X:mm, mm/m64"                             , "RM"   , "0F 0F /r 87"                      , "ANY GEODE"],
    ["pfsub"            , "X:mm, mm/m64"                             , "RM"   , "0F 0F /r 9A"                      , "ANY 3DNOW"],
    ["pfsubr"           , "X:mm, mm/m64"                             , "RM"   , "0F 0F /r AA"                      , "ANY 3DNOW"],
    ["pi2fd"            , "W:mm, mm/m64"                             , "RM"   , "0F 0F /r 0D"                      , "ANY 3DNOW"],
    ["pi2fw"            , "W:mm, mm/m64"                             , "RM"   , "0F 0F /r 0C"                      , "ANY 3DNOW2"],
    ["pmulhrw"          , "X:mm, mm/m64"                             , "RM"   , "0F 0F /r B7"                      , "ANY 3DNOW"],
    ["prefetch3dnow"    , "R:mem"                                    , "M"    , "0F 0D /0"                         , "ANY 3DNOW"],
    ["pswapd"           , "W:mm, mm/m64"                             , "RM"   , "0F 0F /r BB"                      , "ANY 3DNOW2"],

    // AES.
    ["aesdec"           , "X:xmm, xmm/m128"                          , "RM"   , "66 0F 38 DE /r"                   , "ANY AES"],
    ["aesdeclast"       , "X:xmm, xmm/m128"                          , "RM"   , "66 0F 38 DF /r"                   , "ANY AES"],
    ["aesenc"           , "X:xmm, xmm/m128"                          , "RM"   , "66 0F 38 DC /r"                   , "ANY AES"],
    ["aesenclast"       , "X:xmm, xmm/m128"                          , "RM"   , "66 0F 38 DD /r"                   , "ANY AES"],
    ["aesimc"           , "W:xmm, xmm/m128"                          , "RM"   , "66 0F 38 DB /r"                   , "ANY AES"],
    ["aeskeygenassist"  , "W:xmm, xmm/m128, ib"                      , "RM"   , "66 0F 3A DF /r ib"                , "ANY AES"],

    // SHA.
    ["sha1msg1"         , "xmm, xmm/m128"                            , "RM"   , "0F 38 C9 /r"                      , "ANY SHA"],
    ["sha1msg2"         , "xmm, xmm/m128"                            , "RM"   , "0F 38 CA /r"                      , "ANY SHA"],
    ["sha1nexte"        , "xmm, xmm/m128"                            , "RM"   , "0F 38 C8 /r"                      , "ANY SHA"],
    ["sha1rnds4"        , "xmm, xmm/m128, ib"                        , "RMI"  , "0F 3A CC /r ib"                   , "ANY SHA"],
    ["sha256msg1"       , "xmm, xmm/m128"                            , "RM"   , "0F 38 CC /r"                      , "ANY SHA"],
    ["sha256msg2"       , "xmm, xmm/m128"                            , "RM"   , "0F 38 CD /r"                      , "ANY SHA"],
    ["sha256rnds2"      , "xmm, xmm/m128, <xmm0>"                    , "RM"   , "0F 38 CB /r"                      , "ANY SHA"],

    // AVX.
    ["vaddpd"           , "W:xmm, xmm, xmm/m128"                     , "RVM"  , "VEX.NDS.128.66.0F.WIG 58 /r"      , "ANY AVX"],
    ["vaddpd"           , "W:ymm, ymm, ymm/m256"                     , "RVM"  , "VEX.NDS.256.66.0F.WIG 58 /r"      , "ANY AVX"],
    ["vaddps"           , "W:xmm, xmm, xmm/m128"                     , "RVM"  , "VEX.NDS.128.0F.WIG 58 /r"         , "ANY AVX"],
    ["vaddps"           , "W:ymm, ymm, ymm/m256"                     , "RVM"  , "VEX.NDS.256.0F.WIG 58 /r"         , "ANY AVX"],
    ["vaddsd"           , "W:xmm, xmm, xmm/m64"                      , "RVM"  , "VEX.NDS.LIG.F2.0F.WIG 58 /r"      , "ANY AVX"],
    ["vaddss"           , "W:xmm, xmm, xmm/m32"                      , "RVM"  , "VEX.NDS.LIG.F3.0F.WIG 58 /r"      , "ANY AVX"],
    ["vaddsubpd"        , "W:xmm, xmm, xmm/m128"                     , "RVM"  , "VEX.NDS.128.66.0F.WIG D0 /r"      , "ANY AVX"],
    ["vaddsubpd"        , "W:ymm, ymm, ymm/m256"                     , "RVM"  , "VEX.NDS.256.66.0F.WIG D0 /r"      , "ANY AVX"],
    ["vaddsubps"        , "W:xmm, xmm, xmm/m128"                     , "RVM"  , "VEX.NDS.128.F2.0F.WIG D0 /r"      , "ANY AVX"],
    ["vaddsubps"        , "W:ymm, ymm, ymm/m256"                     , "RVM"  , "VEX.NDS.256.F2.0F.WIG D0 /r"      , "ANY AVX"],
    ["vaesdec"          , "W:xmm, xmm, xmm/m128"                     , "RVM"  , "VEX.NDS.128.66.0F38.WIG DE /r"    , "ANY AVX AES"],
    ["vaesdeclast"      , "W:xmm, xmm, xmm/m128"                     , "RVM"  , "VEX.NDS.128.66.0F38.WIG DF /r"    , "ANY AVX AES"],
    ["vaesenc"          , "W:xmm, xmm, xmm/m128"                     , "RVM"  , "VEX.NDS.128.66.0F38.WIG DC /r"    , "ANY AVX AES"],
    ["vaesenclast"      , "W:xmm, xmm, xmm/m128"                     , "RVM"  , "VEX.NDS.128.66.0F38.WIG DD /r"    , "ANY AVX AES"],
    ["vaesimc"          , "W:xmm, xmm/m128"                          , "RVM"  , "VEX.128.66.0F38.WIG DB /r"        , "ANY AVX AES"],
    ["vaeskeygenassist" , "W:xmm, xmm/m128, ib"                      , "RVM"  , "VEX.128.66.0F3A.WIG DF /r ib"     , "ANY AVX AES"],
    ["vandnpd"          , "W:xmm, xmm, xmm/m128"                     , "RVM"  , "VEX.NDS.128.66.0F.WIG 55 /r"      , "ANY AVX"],
    ["vandnpd"          , "W:ymm, ymm, ymm/m256"                     , "RVM"  , "VEX.NDS.256.66.0F.WIG 55 /r"      , "ANY AVX"],
    ["vandnps"          , "W:xmm, xmm, xmm/m128"                     , "RVM"  , "VEX.NDS.128.0F.WIG 55 /r"         , "ANY AVX"],
    ["vandnps"          , "W:ymm, ymm, ymm/m256"                     , "RVM"  , "VEX.NDS.256.0F.WIG 55 /r"         , "ANY AVX"],
    ["vandpd"           , "W:xmm, xmm, xmm/m128"                     , "RVM"  , "VEX.NDS.128.66.0F.WIG 54 /r"      , "ANY AVX"],
    ["vandpd"           , "W:ymm, ymm, ymm/m256"                     , "RVM"  , "VEX.NDS.256.66.0F.WIG 54 /r"      , "ANY AVX"],
    ["vandps"           , "W:xmm, xmm, xmm/m128"                     , "RVM"  , "VEX.NDS.128.0F.WIG 54 /r"         , "ANY AVX"],
    ["vandps"           , "W:ymm, ymm, ymm/m256"                     , "RVM"  , "VEX.NDS.256.0F.WIG 54 /r"         , "ANY AVX"],
    ["vblendpd"         , "W:xmm, xmm, xmm/m128, ib"                 , "RVMI" , "VEX.NDS.128.66.0F3A.WIG 0D /r ib" , "ANY AVX"],
    ["vblendpd"         , "W:ymm, ymm, ymm/m256, ib"                 , "RVMI" , "VEX.NDS.256.66.0F3A.WIG 0D /r ib" , "ANY AVX"],
    ["vblendps"         , "W:xmm, xmm, xmm/m128, ib"                 , "RVMI" , "VEX.NDS.128.66.0F3A.WIG 0C /r ib" , "ANY AVX"],
    ["vblendps"         , "W:ymm, ymm, ymm/m256, ib"                 , "RVMI" , "VEX.NDS.256.66.0F3A.WIG 0C /r ib" , "ANY AVX"],
    ["vblendvpd"        , "W:xmm, xmm, xmm/m128, xmm"                , "RVMS" , "VEX.NDS.128.66.0F3A.W0 4B /r /is4", "ANY AVX"],
    ["vblendvpd"        , "W:ymm, ymm, ymm/m256, ymm"                , "RVMS" , "VEX.NDS.256.66.0F3A.W0 4B /r /is4", "ANY AVX"],
    ["vblendvps"        , "W:xmm, xmm, xmm/m128, xmm"                , "RVMS" , "VEX.NDS.128.66.0F3A.W0 4A /r /is4", "ANY AVX"],
    ["vblendvps"        , "W:ymm, ymm, ymm/m256, ymm"                , "RVMS" , "VEX.NDS.256.66.0F3A.W0 4A /r /is4", "ANY AVX"],
    ["vbroadcastf128"   , "W:ymm, m128"                              , "RM"   , "VEX.256.66.0F38.W0 1A /r"         , "ANY AVX"],
    ["vbroadcasti128"   , "W:ymm, m128"                              , "RM"   , "VEX.256.66.0F38.W0 5A /r"         , "ANY AVX2"],
    ["vbroadcastsd"     , "W:ymm, m64"                               , "RM"   , "VEX.256.66.0F38.W0 19 /r"         , "ANY AVX"],
    ["vbroadcastsd"     , "W:ymm, xmm"                               , "RM"   , "VEX.256.66.0F38.W0 19 /r"         , "ANY AVX2"],
    ["vbroadcastss"     , "W:xmm, m32"                               , "RM"   , "VEX.128.66.0F38.W0 18 /r"         , "ANY AVX"],
    ["vbroadcastss"     , "W:xmm, xmm"                               , "RM"   , "VEX.128.66.0F38.W0 18 /r"         , "ANY AVX2"],
    ["vbroadcastss"     , "W:ymm, m32"                               , "RM"   , "VEX.256.66.0F38.W0 18 /r"         , "ANY AVX"],
    ["vbroadcastss"     , "W:ymm, xmm"                               , "RM"   , "VEX.256.66.0F38.W0 18 /r"         , "ANY AVX2"],
    ["vcmppd"           , "W:xmm, xmm, xmm/m128, ib"                 , "RVM"  , "VEX.NDS.128.66.0F.WIG C2 /r ib"   , "ANY AVX"],
    ["vcmppd"           , "W:ymm, ymm, ymm/m256, ib"                 , "RVM"  , "VEX.NDS.256.66.0F.WIG C2 /r ib"   , "ANY AVX"],
    ["vcmpps"           , "W:xmm, xmm, xmm/m128, ib"                 , "RVM"  , "VEX.NDS.128.0F.WIG C2 /r ib"      , "ANY AVX"],
    ["vcmpps"           , "W:ymm, ymm, ymm/m256, ib"                 , "RVM"  , "VEX.NDS.256.0F.WIG C2 /r ib"      , "ANY AVX"],
    ["vcmpsd"           , "W:xmm, xmm, xmm/m64, ib"                  , "RVM"  , "VEX.NDS.LIG.F2.0F.WIG C2 /r ib"   , "ANY AVX"],
    ["vcmpss"           , "W:xmm, xmm, xmm/m32, ib"                  , "RVM"  , "VEX.NDS.LIG.F3.0F.WIG C2 /r ib"   , "ANY AVX"],
    ["vcomisd"          , "R:xmm, xmm/m64"                           , "RM"   , "VEX.LIG.66.0F.WIG 2F /r"          , "ANY AVX OF=0 SF=0 ZF=W AF=0 PF=W CF=W"],
    ["vcomiss"          , "R:xmm, xmm/m32"                           , "RM"   , "VEX.LIG.0F.WIG 2F /r"             , "ANY AVX OF=0 SF=0 ZF=W AF=0 PF=W CF=W"],
    ["vcvtdq2pd"        , "W:xmm, xmm/m64"                           , "RM"   , "VEX.128.F3.0F.WIG E6 /r"          , "ANY AVX"],
    ["vcvtdq2pd"        , "W:ymm, xmm/m128"                          , "RM"   , "VEX.256.F3.0F.WIG E6 /r"          , "ANY AVX"],
    ["vcvtdq2ps"        , "W:xmm, xmm/m128"                          , "RM"   , "VEX.128.0F.WIG 5B /r"             , "ANY AVX"],
    ["vcvtdq2ps"        , "W:ymm, ymm/m256"                          , "RM"   , "VEX.256.0F.WIG 5B /r"             , "ANY AVX"],
    ["vcvtpd2dq"        , "W:xmm, xmm/m128"                          , "RM"   , "VEX.128.F2.0F.WIG E6 /r"          , "ANY AVX"],
    ["vcvtpd2dq"        , "W:xmm, ymm/m256"                          , "RM"   , "VEX.256.F2.0F.WIG E6 /r"          , "ANY AVX"],
    ["vcvtpd2ps"        , "W:xmm, xmm/m128"                          , "RM"   , "VEX.128.66.0F.WIG 5A /r"          , "ANY AVX"],
    ["vcvtpd2ps"        , "W:xmm, ymm/m256"                          , "RM"   , "VEX.256.66.0F.WIG 5A /r"          , "ANY AVX"],
    ["vcvtps2dq"        , "W:xmm, xmm/m128"                          , "RM"   , "VEX.128.66.0F.WIG 5B /r"          , "ANY AVX"],
    ["vcvtps2dq"        , "W:ymm, ymm/m256"                          , "RM"   , "VEX.256.66.0F.WIG 5B /r"          , "ANY AVX"],
    ["vcvtps2pd"        , "W:xmm, xmm/m64"                           , "RM"   , "VEX.128.0F.WIG 5A /r"             , "ANY AVX"],
    ["vcvtps2pd"        , "W:ymm, xmm/m128"                          , "RM"   , "VEX.256.0F.WIG 5A /r"             , "ANY AVX"],
    ["vcvtsd2si"        , "W:r32, xmm/m64"                           , "RM"   , "VEX.LIG.F2.0F.W0 2D /r"           , "ANY AVX"],
    ["vcvtsd2si"        , "W:r64, xmm/m64"                           , "RM"   , "VEX.LIG.F2.0F.W1 2D /r"           , "X64 AVX"],
    ["vcvtsd2ss"        , "W:xmm, xmm, xmm/m64"                      , "RVM"  , "VEX.NDS.LIG.F2.0F.WIG 5A /r"      , "ANY AVX"],
    ["vcvtsi2sd"        , "W:xmm, xmm, r32/m32"                      , "RVM"  , "VEX.NDS.LIG.F2.0F.W0 2A /r"       , "ANY AVX"],
    ["vcvtsi2sd"        , "W:xmm, xmm, r64/m64"                      , "RVM"  , "VEX.NDS.LIG.F2.0F.W1 2A /r"       , "X64 AVX"],
    ["vcvtsi2ss"        , "W:xmm, xmm, r32/m32"                      , "RVM"  , "VEX.NDS.LIG.F3.0F.W0 2A /r"       , "ANY AVX"],
    ["vcvtsi2ss"        , "W:xmm, xmm, r64/m64"                      , "RVM"  , "VEX.NDS.LIG.F3.0F.W1 2A /r"       , "X64 AVX"],
    ["vcvtss2sd"        , "W:xmm, xmm, xmm/m32"                      , "RVM"  , "VEX.NDS.LIG.F3.0F.WIG 5A /r"      , "ANY AVX"],
    ["vcvtss2si"        , "W:r32, xmm/m32"                           , "RM"   , "VEX.LIG.F3.0F.W0 2D /r"           , "ANY AVX"],
    ["vcvtss2si"        , "W:r64, xmm/m32"                           , "RM"   , "VEX.LIG.F3.0F.W1 2D /r"           , "X64 AVX"],
    ["vcvttpd2dq"       , "W:xmm, xmm/m128"                          , "RM"   , "VEX.128.66.0F.WIG E6 /r"          , "ANY AVX"],
    ["vcvttpd2dq"       , "W:xmm, ymm/m256"                          , "RM"   , "VEX.256.66.0F.WIG E6 /r"          , "ANY AVX"],
    ["vcvttps2dq"       , "W:xmm, xmm/m128"                          , "RM"   , "VEX.128.F3.0F.WIG 5B /r"          , "ANY AVX"],
    ["vcvttps2dq"       , "W:ymm, ymm/m256"                          , "RM"   , "VEX.256.F3.0F.WIG 5B /r"          , "ANY AVX"],
    ["vcvttsd2si"       , "W:r32, xmm/m64"                           , "RM"   , "VEX.LIG.F2.0F.W0 2C /r"           , "ANY AVX"],
    ["vcvttsd2si"       , "W:r64, xmm/m64"                           , "RM"   , "VEX.LIG.F2.0F.W1 2C /r"           , "X64 AVX"],
    ["vcvttss2si"       , "W:r32, xmm/m32"                           , "RM"   , "VEX.LIG.F3.0F.W0 2C /r"           , "ANY AVX"],
    ["vcvttss2si"       , "W:r64, xmm/m32"                           , "RM"   , "VEX.LIG.F3.0F.W1 2C /r"           , "X64 AVX"],
    ["vdivpd"           , "W:xmm, xmm, xmm/m128"                     , "RVM"  , "VEX.NDS.128.66.0F.WIG 5E /r"      , "ANY AVX"],
    ["vdivpd"           , "W:ymm, ymm, ymm/m256"                     , "RVM"  , "VEX.NDS.256.66.0F.WIG 5E /r"      , "ANY AVX"],
    ["vdivps"           , "W:xmm, xmm, xmm/m128"                     , "RVM"  , "VEX.NDS.128.0F.WIG 5E /r"         , "ANY AVX"],
    ["vdivps"           , "W:ymm, ymm, ymm/m256"                     , "RVM"  , "VEX.NDS.256.0F.WIG 5E /r"         , "ANY AVX"],
    ["vdivsd"           , "W:xmm, xmm, xmm/m64"                      , "RVM"  , "VEX.NDS.LIG.F2.0F.WIG 5E /r"      , "ANY AVX"],
    ["vdivss"           , "W:xmm, xmm, xmm/m32"                      , "RVM"  , "VEX.NDS.LIG.F3.0F.WIG 5E /r"      , "ANY AVX"],
    ["vdppd"            , "W:xmm, xmm, xmm/m128, ib"                 , "RVMI" , "VEX.NDS.128.66.0F3A.WIG 41 /r ib" , "ANY AVX"],
    ["vdppd"            , "W:ymm, ymm, ymm/m256, ib"                 , "RVMI" , "VEX.NDS.256.66.0F3A.WIG 41 /r ib" , "ANY AVX"],
    ["vdpps"            , "W:xmm, xmm, xmm/m128, ib"                 , "RVMI" , "VEX.NDS.128.66.0F3A.WIG 40 /r ib" , "ANY AVX"],
    ["vdpps"            , "W:ymm, ymm, ymm/m256, ib"                 , "RVMI" , "VEX.NDS.256.66.0F3A.WIG 40 /r ib" , "ANY AVX"],
    ["vextractf128"     , "W:xmm/m128, ymm, ib"                      , "MRI"  , "VEX.256.66.0F3A.W0 19 /r ib"      , "ANY AVX"],
    ["vextracti128"     , "W:xmm/m128, ymm, ib"                      , "MRI"  , "VEX.256.66.0F3A.W0 39 /r ib"      , "ANY AVX2"],
    ["vextractps"       , "W:r32/m32, xmm, ib"                       , "MRI"  , "VEX.128.66.0F3A.WIG 17 /r ib"     , "ANY AVX"],
    ["vgatherdpd"       , "W:xmm, vm32x, xmm"                        , "RMV"  , "VEX.DDS.128.66.0F38.W1 92 /r"     , "ANY AVX2"],
    ["vgatherdpd"       , "W:ymm, vm32x, ymm"                        , "RMV"  , "VEX.DDS.256.66.0F38.W1 92 /r"     , "ANY AVX2"],
    ["vgatherdps"       , "W:xmm, vm32x, xmm"                        , "RMV"  , "VEX.DDS.128.66.0F38.W0 92 /r"     , "ANY AVX2"],
    ["vgatherdps"       , "W:ymm, vm32y, ymm"                        , "RMV"  , "VEX.DDS.256.66.0F38.W0 92 /r"     , "ANY AVX2"],
    ["vgatherqpd"       , "W:xmm, vm64x, xmm"                        , "RMV"  , "VEX.DDS.128.66.0F38.W1 93 /r"     , "ANY AVX2"],
    ["vgatherqpd"       , "W:ymm, vm64y, ymm"                        , "RMV"  , "VEX.DDS.256.66.0F38.W1 93 /r"     , "ANY AVX2"],
    ["vgatherqps"       , "W:xmm, vm64x, xmm"                        , "RMV"  , "VEX.DDS.128.66.0F38.W0 93 /r"     , "ANY AVX2"],
    ["vgatherqps"       , "W:xmm, vm64y, xmm"                        , "RMV"  , "VEX.DDS.256.66.0F38.W0 93 /r"     , "ANY AVX2"],
    ["vhaddpd"          , "W:xmm, xmm, xmm/m128"                     , "RVM"  , "VEX.NDS.128.66.0F.WIG 7C /r"      , "ANY AVX"],
    ["vhaddpd"          , "W:ymm, ymm, ymm/m256"                     , "RVM"  , "VEX.NDS.256.66.0F.WIG 7C /r"      , "ANY AVX"],
    ["vhaddps"          , "W:xmm, xmm, xmm/m128"                     , "RVM"  , "VEX.NDS.128.F2.0F.WIG 7C /r"      , "ANY AVX"],
    ["vhaddps"          , "W:ymm, ymm, ymm/m256"                     , "RVM"  , "VEX.NDS.256.F2.0F.WIG 7C /r"      , "ANY AVX"],
    ["vhsubpd"          , "W:xmm, xmm, xmm/m128"                     , "RVM"  , "VEX.NDS.128.66.0F.WIG 7D /r"      , "ANY AVX"],
    ["vhsubpd"          , "W:ymm, ymm, ymm/m256"                     , "RVM"  , "VEX.NDS.256.66.0F.WIG 7D /r"      , "ANY AVX"],
    ["vhsubps"          , "W:xmm, xmm, xmm/m128"                     , "RVM"  , "VEX.NDS.128.F2.0F.WIG 7D /r"      , "ANY AVX"],
    ["vhsubps"          , "W:ymm, ymm, ymm/m256"                     , "RVM"  , "VEX.NDS.256.F2.0F.WIG 7D /r"      , "ANY AVX"],
    ["vinsertf128"      , "W:ymm, ymm, xmm/m128, ib"                 , "RVMI" , "VEX.NDS.256.66.0F3A.W0 18 /r ib"  , "ANY AVX"],
    ["vinserti128"      , "W:ymm, ymm, xmm/m128, ib"                 , "RVMI" , "VEX.NDS.256.66.0F3A.W0 38 /r ib"  , "ANY AVX2"],
    ["vinsertps"        , "W:xmm, xmm, xmm/m32, ib"                  , "RVMI" , "VEX.NDS.128.66.0F3A.WIG 21 /r ib" , "ANY AVX"],
    ["vlddqu"           , "W:xmm, m128"                              , "RM"   , "VEX.128.F2.0F.WIG F0 /r"          , "ANY AVX"],
    ["vlddqu"           , "W:ymm, m256"                              , "RM"   , "VEX.256.F2.0F.WIG F0 /r"          , "ANY AVX"],
    ["vldmxcsr"         , "R:m32"                                    , "M"    , "VEX.LZ.0F.WIG AE /2"              , "ANY AVX"],
    ["vmaskmovdqu"      , "R:xmm, xmm, <ds:zdi>"                     , "RM"   , "VEX.128.66.0F.WIG F7 /r"          , "ANY AVX"],
    ["vmaskmovpd"       , "W:m128, xmm, xmm"                         , "MVR"  , "VEX.NDS.128.66.0F38.W0 2F /r"     , "ANY AVX"],
    ["vmaskmovpd"       , "W:m256, ymm, ymm"                         , "MVR"  , "VEX.NDS.256.66.0F38.W0 2F /r"     , "ANY AVX"],
    ["vmaskmovpd"       , "W:xmm, xmm, m128"                         , "RVM"  , "VEX.NDS.128.66.0F38.W0 2D /r"     , "ANY AVX"],
    ["vmaskmovpd"       , "W:ymm, ymm, m256"                         , "RVM"  , "VEX.NDS.256.66.0F38.W0 2D /r"     , "ANY AVX"],
    ["vmaskmovps"       , "W:m128, xmm, xmm"                         , "MVR"  , "VEX.NDS.128.66.0F38.W0 2E /r"     , "ANY AVX"],
    ["vmaskmovps"       , "W:m256, ymm, ymm"                         , "MVR"  , "VEX.NDS.256.66.0F38.W0 2E /r"     , "ANY AVX"],
    ["vmaskmovps"       , "W:xmm, xmm, m128"                         , "RVM"  , "VEX.NDS.128.66.0F38.W0 2C /r"     , "ANY AVX"],
    ["vmaskmovps"       , "W:ymm, ymm, m256"                         , "RVM"  , "VEX.NDS.256.66.0F38.W0 2C /r"     , "ANY AVX"],
    ["vmaxpd"           , "W:xmm, xmm, xmm/m128"                     , "RVM"  , "VEX.NDS.128.66.0F.WIG 5F /r"      , "ANY AVX"],
    ["vmaxpd"           , "W:ymm, ymm, ymm/m256"                     , "RVM"  , "VEX.NDS.256.66.0F.WIG 5F /r"      , "ANY AVX"],
    ["vmaxps"           , "W:xmm, xmm, xmm/m128"                     , "RVM"  , "VEX.NDS.128.0F.WIG 5F /r"         , "ANY AVX"],
    ["vmaxps"           , "W:ymm, ymm, ymm/m256"                     , "RVM"  , "VEX.NDS.256.0F.WIG 5F /r"         , "ANY AVX"],
    ["vmaxsd"           , "W:xmm, xmm, xmm/m64"                      , "RVM"  , "VEX.NDS.LIG.F2.0F.WIG 5F /r"      , "ANY AVX"],
    ["vmaxss"           , "W:xmm, xmm, xmm/m32"                      , "RVM"  , "VEX.NDS.LIG.F3.0F.WIG 5F /r"      , "ANY AVX"],
    ["vminpd"           , "W:xmm, xmm, xmm/m128"                     , "RVM"  , "VEX.NDS.128.66.0F.WIG 5D /r"      , "ANY AVX"],
    ["vminpd"           , "W:ymm, ymm, ymm/m256"                     , "RVM"  , "VEX.NDS.256.66.0F.WIG 5D /r"      , "ANY AVX"],
    ["vminps"           , "W:xmm, xmm, xmm/m128"                     , "RVM"  , "VEX.NDS.128.0F.WIG 5D /r"         , "ANY AVX"],
    ["vminps"           , "W:ymm, ymm, ymm/m256"                     , "RVM"  , "VEX.NDS.256.0F.WIG 5D /r"         , "ANY AVX"],
    ["vminsd"           , "W:xmm, xmm, xmm/m64"                      , "RVM"  , "VEX.NDS.LIG.F2.0F.WIG 5D /r"      , "ANY AVX"],
    ["vminss"           , "W:xmm, xmm, xmm/m32"                      , "RVM"  , "VEX.NDS.LIG.F3.0F.WIG 5D /r"      , "ANY AVX"],
    ["vmovapd"          , "W:xmm, xmm/m128"                          , "RM"   , "VEX.128.66.0F.WIG 28 /r"          , "ANY AVX"],
    ["vmovapd"          , "W:xmm/m128, xmm"                          , "MR"   , "VEX.128.66.0F.WIG 29 /r"          , "ANY AVX"],
    ["vmovapd"          , "W:ymm, ymm/m256"                          , "RM"   , "VEX.256.66.0F.WIG 28 /r"          , "ANY AVX"],
    ["vmovapd"          , "W:ymm/m256, ymm"                          , "MR"   , "VEX.256.66.0F.WIG 29 /r"          , "ANY AVX"],
    ["vmovaps"          , "W:xmm, xmm/m128"                          , "RM"   , "VEX.128.0F.WIG 28 /r"             , "ANY AVX"],
    ["vmovaps"          , "W:xmm/m128, xmm"                          , "MR"   , "VEX.128.0F.WIG 29 /r"             , "ANY AVX"],
    ["vmovaps"          , "W:ymm, ymm/m256"                          , "RM"   , "VEX.256.0F.WIG 28 /r"             , "ANY AVX"],
    ["vmovaps"          , "W:ymm/m256, ymm"                          , "MR"   , "VEX.256.0F.WIG 29 /r"             , "ANY AVX"],
    ["vmovd"            , "W:r32/m32, xmm"                           , "MR"   , "VEX.128.66.0F.W0 7E /r"           , "ANY AVX"],
    ["vmovd"            , "W:xmm, r32/m32"                           , "RM"   , "VEX.128.66.0F.W0 6E /r"           , "ANY AVX"],
    ["vmovddup"         , "W:xmm, xmm/m64"                           , "RM"   , "VEX.128.F2.0F.WIG 12 /r"          , "ANY AVX"],
    ["vmovddup"         , "W:ymm, ymm/m256"                          , "RM"   , "VEX.256.F2.0F.WIG 12 /r"          , "ANY AVX"],
    ["vmovdqa"          , "W:xmm, xmm/m128"                          , "RM"   , "VEX.128.66.0F.WIG 6F /r"          , "ANY AVX"],
    ["vmovdqa"          , "W:xmm/m128, xmm"                          , "MR"   , "VEX.128.66.0F.WIG 7F /r"          , "ANY AVX"],
    ["vmovdqa"          , "W:ymm, ymm/m256"                          , "RM"   , "VEX.256.66.0F.WIG 6F /r"          , "ANY AVX"],
    ["vmovdqa"          , "W:ymm/m256, ymm"                          , "MR"   , "VEX.256.66.0F.WIG 7F /r"          , "ANY AVX"],
    ["vmovdqu"          , "W:xmm, xmm/m128"                          , "RM"   , "VEX.128.F3.0F.WIG 6F /r"          , "ANY AVX"],
    ["vmovdqu"          , "W:xmm/m128, xmm"                          , "MR"   , "VEX.128.F3.0F.WIG 7F /r"          , "ANY AVX"],
    ["vmovdqu"          , "W:ymm, ymm/m256"                          , "RM"   , "VEX.256.F3.0F.WIG 6F /r"          , "ANY AVX"],
    ["vmovdqu"          , "W:ymm/m256, ymm"                          , "MR"   , "VEX.256.F3.0F.WIG 7F /r"          , "ANY AVX"],
    ["vmovhlps"         , "W:xmm, xmm, xmm"                          , "RVM"  , "VEX.NDS.128.0F.WIG 12 /r"         , "ANY AVX"],
    ["vmovhpd"          , "W:m64, xmm"                               , "MR"   , "VEX.128.66.0F.WIG 17 /r"          , "ANY AVX"],
    ["vmovhpd"          , "W:xmm, xmm, m64"                          , "RVM"  , "VEX.NDS.128.66.0F.WIG 16 /r"      , "ANY AVX"],
    ["vmovhps"          , "W:m64, xmm"                               , "MR"   , "VEX.128.0F.WIG 17 /r"             , "ANY AVX"],
    ["vmovhps"          , "W:xmm, xmm, m64"                          , "RVM"  , "VEX.NDS.128.0F.WIG 16 /r"         , "ANY AVX"],
    ["vmovlhps"         , "W:xmm, xmm, xmm"                          , "RVM"  , "VEX.NDS.128.0F.WIG 16 /r"         , "ANY AVX"],
    ["vmovlpd"          , "W:m64, xmm"                               , "MR"   , "VEX.128.66.0F.WIG 13 /r"          , "ANY AVX"],
    ["vmovlpd"          , "W:xmm, xmm, m64"                          , "RVM"  , "VEX.NDS.128.66.0F.WIG 12 /r"      , "ANY AVX"],
    ["vmovlps"          , "W:m64, xmm"                               , "MR"   , "VEX.128.0F.WIG 13 /r"             , "ANY AVX"],
    ["vmovlps"          , "W:xmm, xmm, m64"                          , "RVM"  , "VEX.NDS.128.0F.WIG 12 /r"         , "ANY AVX"],
    ["vmovmskpd"        , "W:r32, xmm"                               , "RM"   , "VEX.128.66.0F.WIG 50 /r"          , "ANY AVX"],
    ["vmovmskpd"        , "W:r64, xmm"                               , "RM"   , "VEX.128.66.0F.WIG 50 /r"          , "X64 AVX"],
    ["vmovmskpd"        , "W:r32, ymm"                               , "RM"   , "VEX.256.66.0F.WIG 50 /r"          , "ANY AVX"],
    ["vmovmskpd"        , "W:r64, ymm"                               , "RM"   , "VEX.256.66.0F.WIG 50 /r"          , "X64 AVX"],
    ["vmovmskps"        , "W:r32, xmm"                               , "RM"   , "VEX.128.0F.WIG 50 /r"             , "ANY AVX"],
    ["vmovmskps"        , "W:r64, xmm"                               , "RM"   , "VEX.128.0F.WIG 50 /r"             , "X64 AVX"],
    ["vmovmskps"        , "W:r32, ymm"                               , "RM"   , "VEX.256.0F.WIG 50 /r"             , "ANY AVX"],
    ["vmovmskps"        , "W:r64, ymm"                               , "RM"   , "VEX.256.0F.WIG 50 /r"             , "X64 AVX"],
    ["vmovntdq"         , "W:m128, xmm"                              , "MR"   , "VEX.128.66.0F.WIG E7 /r"          , "ANY AVX"],
    ["vmovntdq"         , "W:m256, ymm"                              , "MR"   , "VEX.256.66.0F.WIG E7 /r"          , "ANY AVX"],
    ["vmovntdqa"        , "W:xmm, m128"                              , "RM"   , "VEX.128.66.0F38.WIG 2A /r"        , "ANY AVX"],
    ["vmovntdqa"        , "W:ymm, m256"                              , "RM"   , "VEX.256.66.0F38.WIG 2A /r"        , "ANY AVX2"],
    ["vmovntpd"         , "W:m128, xmm"                              , "MR"   , "VEX.128.66.0F.WIG 2B /r"          , "ANY AVX"],
    ["vmovntpd"         , "W:m256, ymm"                              , "MR"   , "VEX.256.66.0F.WIG 2B /r"          , "ANY AVX"],
    ["vmovntps"         , "W:m128, xmm"                              , "MR"   , "VEX.128.0F.WIG 2B /r"             , "ANY AVX"],
    ["vmovntps"         , "W:m256, ymm"                              , "MR"   , "VEX.256.0F.WIG 2B /r"             , "ANY AVX"],
    ["vmovq"            , "W:r64/m64, xmm"                           , "MR"   , "VEX.128.66.0F.W1 7E /r"           , "X64 AVX"],
    ["vmovq"            , "W:xmm, m64"                               , "RM"   , "VEX.128.F3.0F.WIG 7E /r"          , "ANY AVX"],
    ["vmovq"            , "W:xmm, r64/m64"                           , "RM"   , "VEX.128.66.0F.W1 6E /r"           , "X64 AVX"],
    ["vmovq"            , "W:xmm, xmm"                               , "RM"   , "VEX.128.F3.0F.WIG 7E /r"          , "ANY AVX"],
    ["vmovq"            , "W:xmm/m64, xmm"                           , "MR"   , "VEX.128.66.0F.WIG D6 /r"          , "ANY AVX"],
    ["vmovsd"           , "W:m64, xmm"                               , "MR"   , "VEX.LIG.F2.0F.WIG 11 /r"          , "ANY AVX"],
    ["vmovsd"           , "W:xmm, m64"                               , "RM"   , "VEX.LIG.F2.0F.WIG 10 /r"          , "ANY AVX"],
    ["vmovsd"           , "W:xmm, xmm, xmm"                          , "MVR"  , "VEX.NDS.LIG.F2.0F.WIG 11 /r"      , "ANY AVX"],
    ["vmovsd"           , "W:xmm, xmm, xmm"                          , "RVM"  , "VEX.NDS.LIG.F2.0F.WIG 10 /r"      , "ANY AVX"],
    ["vmovshdup"        , "W:xmm, xmm/m128"                          , "RM"   , "VEX.128.F3.0F.WIG 16 /r"          , "ANY AVX"],
    ["vmovshdup"        , "W:ymm, ymm/m256"                          , "RM"   , "VEX.256.F3.0F.WIG 16 /r"          , "ANY AVX"],
    ["vmovsldup"        , "W:xmm, xmm/m128"                          , "RM"   , "VEX.128.F3.0F.WIG 12 /r"          , "ANY AVX"],
    ["vmovsldup"        , "W:ymm, ymm/m256"                          , "RM"   , "VEX.256.F3.0F.WIG 12 /r"          , "ANY AVX"],
    ["vmovss"           , "W:m32, xmm"                               , "MR"   , "VEX.LIG.F3.0F.WIG 11 /r"          , "ANY AVX"],
    ["vmovss"           , "W:xmm, m32"                               , "RM"   , "VEX.LIG.F3.0F.WIG 10 /r"          , "ANY AVX"],
    ["vmovss"           , "W:xmm, xmm, xmm"                          , "MVR"  , "VEX.NDS.LIG.F3.0F.WIG 11 /r"      , "ANY AVX"],
    ["vmovss"           , "W:xmm, xmm, xmm"                          , "RVM"  , "VEX.NDS.LIG.F3.0F.WIG 10 /r"      , "ANY AVX"],
    ["vmovupd"          , "W:xmm, xmm/m128"                          , "RM"   , "VEX.128.0F.WIG 10 /r"             , "ANY AVX"],
    ["vmovupd"          , "W:xmm/m128, xmm"                          , "MR"   , "VEX.128.0F.WIG 11 /r"             , "ANY AVX"],
    ["vmovupd"          , "W:ymm, ymm/m256"                          , "RM"   , "VEX.256.0F.WIG 10 /r"             , "ANY AVX"],
    ["vmovupd"          , "W:ymm/m256, ymm"                          , "MR"   , "VEX.256.0F.WIG 11 /r"             , "ANY AVX"],
    ["vmovups"          , "W:xmm, xmm/m128"                          , "RM"   , "VEX.128.66.0F.WIG 10 /r"          , "ANY AVX"],
    ["vmovups"          , "W:xmm/m128, xmm"                          , "MR"   , "VEX.128.66.0F.WIG 11 /r"          , "ANY AVX"],
    ["vmovups"          , "W:ymm, ymm/m256"                          , "RM"   , "VEX.256.66.0F.WIG 10 /r"          , "ANY AVX"],
    ["vmovups"          , "W:ymm/m256, ymm"                          , "MR"   , "VEX.256.66.0F.WIG 11 /r"          , "ANY AVX"],
    ["vmpsadbw"         , "W:xmm, xmm, xmm/m128, ib"                 , "RVMI" , "VEX.NDS.128.66.0F3A.WIG 42 /r ib" , "ANY AVX"],
    ["vmpsadbw"         , "W:ymm, ymm, ymm/m256, ib"                 , "RVMI" , "VEX.NDS.256.66.0F3A.WIG 42 /r ib" , "ANY AVX2"],
    ["vmulpd"           , "W:xmm, xmm, xmm/m128"                     , "RVM"  , "VEX.NDS.128.66.0F.WIG 59 /r"      , "ANY AVX"],
    ["vmulpd"           , "W:ymm, ymm, ymm/m256"                     , "RVM"  , "VEX.NDS.256.66.0F.WIG 59 /r"      , "ANY AVX2"],
    ["vmulps"           , "W:xmm, xmm, xmm/m128"                     , "RVM"  , "VEX.NDS.128.0F.WIG 59 /r"         , "ANY AVX"],
    ["vmulps"           , "W:ymm, ymm, ymm/m256"                     , "RVM"  , "VEX.NDS.256.0F.WIG 59 /r"         , "ANY AVX2"],
    ["vmulsd"           , "W:xmm, xmm, xmm/m64"                      , "RVM"  , "VEX.NDS.LIG.F2.0F.WIG 59 /r"      , "ANY AVX"],
    ["vmulss"           , "W:xmm, xmm, xmm/m32"                      , "RVM"  , "VEX.NDS.LIG.F3.0F.WIG 59 /r"      , "ANY AVX"],
    ["vorpd"            , "W:xmm, xmm, xmm/m128"                     , "RVM"  , "VEX.NDS.128.66.0F.WIG 56 /r"      , "ANY AVX"],
    ["vorpd"            , "W:ymm, ymm, ymm/m256"                     , "RVM"  , "VEX.NDS.256.66.0F.WIG 56 /r"      , "ANY AVX"],
    ["vorps"            , "W:xmm, xmm, xmm/m128"                     , "RVM"  , "VEX.NDS.128.0F.WIG 56 /r"         , "ANY AVX"],
    ["vorps"            , "W:ymm, ymm, ymm/m256"                     , "RVM"  , "VEX.NDS.256.0F.WIG 56 /r"         , "ANY AVX"],
    ["vpabsb"           , "W:xmm, xmm/m128"                          , "RM"   , "VEX.128.66.0F38.WIG 1C /r"        , "ANY AVX"],
    ["vpabsb"           , "W:ymm, ymm/m256"                          , "RM"   , "VEX.256.66.0F38.WIG 1C /r"        , "ANY AVX2"],
    ["vpabsd"           , "W:xmm, xmm/m128"                          , "RM"   , "VEX.128.66.0F38.WIG 1E /r"        , "ANY AVX"],
    ["vpabsd"           , "W:ymm, ymm/m256"                          , "RM"   , "VEX.256.66.0F38.WIG 1E /r"        , "ANY AVX2"],
    ["vpabsw"           , "W:xmm, xmm/m128"                          , "RM"   , "VEX.128.66.0F38.WIG 1D /r"        , "ANY AVX"],
    ["vpabsw"           , "W:ymm, ymm/m256"                          , "RM"   , "VEX.256.66.0F38.WIG 1D /r"        , "ANY AVX2"],
    ["vpackssdw"        , "W:xmm, xmm, xmm/m128"                     , "RVM"  , "VEX.NDS.128.66.0F.WIG 6B /r"      , "ANY AVX"],
    ["vpackssdw"        , "W:ymm, ymm, ymm/m256"                     , "RVM"  , "VEX.NDS.256.66.0F.WIG 6B /r"      , "ANY AVX2"],
    ["vpacksswb"        , "W:xmm, xmm, xmm/m128"                     , "RVM"  , "VEX.NDS.128.66.0F.WIG 63 /r"      , "ANY AVX"],
    ["vpacksswb"        , "W:ymm, ymm, ymm/m256"                     , "RVM"  , "VEX.NDS.256.66.0F.WIG 63 /r"      , "ANY AVX2"],
    ["vpackusdw"        , "W:xmm, xmm, xmm/m128"                     , "RVM"  , "VEX.NDS.128.66.0F38.WIG 2B /r"    , "ANY AVX"],
    ["vpackusdw"        , "W:ymm, ymm, ymm/m256"                     , "RVM"  , "VEX.NDS.256.66.0F38.WIG 2B /r"    , "ANY AVX2"],
    ["vpackuswb"        , "W:xmm, xmm, xmm/m128"                     , "RVM"  , "VEX.NDS.128.66.0F.WIG 67 /r"      , "ANY AVX"],
    ["vpackuswb"        , "W:ymm, ymm, ymm/m256"                     , "RVM"  , "VEX.NDS.256.66.0F.WIG 67 /r"      , "ANY AVX2"],
    ["vpaddb"           , "W:xmm, xmm, xmm/m128"                     , "RVM"  , "VEX.NDS.128.66.0F.WIG FC /r"      , "ANY AVX"],
    ["vpaddb"           , "W:ymm, ymm, ymm/m256"                     , "RVM"  , "VEX.NDS.256.66.0F.WIG FC /r"      , "ANY AVX2"],
    ["vpaddd"           , "W:xmm, xmm, xmm/m128"                     , "RVM"  , "VEX.NDS.128.66.0F.WIG FE /r"      , "ANY AVX"],
    ["vpaddd"           , "W:ymm, ymm, ymm/m256"                     , "RVM"  , "VEX.NDS.256.66.0F.WIG FE /r"      , "ANY AVX2"],
    ["vpaddq"           , "W:xmm, xmm, xmm/m128"                     , "RVM"  , "VEX.NDS.128.66.0F.WIG D4 /r"      , "ANY AVX"],
    ["vpaddq"           , "W:ymm, ymm, ymm/m256"                     , "RVM"  , "VEX.NDS.256.66.0F.WIG D4 /r"      , "ANY AVX2"],
    ["vpaddsb"          , "W:xmm, xmm, xmm/m128"                     , "RVM"  , "VEX.NDS.128.66.0F.WIG EC /r"      , "ANY AVX"],
    ["vpaddsb"          , "W:ymm, ymm, ymm/m256"                     , "RVM"  , "VEX.NDS.256.66.0F.WIG EC /r"      , "ANY AVX2"],
    ["vpaddsw"          , "W:xmm, xmm, xmm/m128"                     , "RVM"  , "VEX.NDS.128.66.0F.WIG ED /r"      , "ANY AVX"],
    ["vpaddsw"          , "W:ymm, ymm, ymm/m256"                     , "RVM"  , "VEX.NDS.256.66.0F.WIG ED /r"      , "ANY AVX2"],
    ["vpaddusb"         , "W:xmm, xmm, xmm/m128"                     , "RVM"  , "VEX.NDS.128.66.0F.WIG DC /r"      , "ANY AVX"],
    ["vpaddusb"         , "W:ymm, ymm, ymm/m256"                     , "RVM"  , "VEX.NDS.256.66.0F.WIG DC /r"      , "ANY AVX2"],
    ["vpaddusw"         , "W:xmm, xmm, xmm/m128"                     , "RVM"  , "VEX.NDS.128.66.0F.WIG DD /r"      , "ANY AVX"],
    ["vpaddusw"         , "W:ymm, ymm, ymm/m256"                     , "RVM"  , "VEX.NDS.256.66.0F.WIG DD /r"      , "ANY AVX2"],
    ["vpaddw"           , "W:xmm, xmm, xmm/m128"                     , "RVM"  , "VEX.NDS.128.66.0F.WIG FD /r"      , "ANY AVX"],
    ["vpaddw"           , "W:ymm, ymm, ymm/m256"                     , "RVM"  , "VEX.NDS.256.66.0F.WIG FD /r"      , "ANY AVX2"],
    ["vpalignr"         , "W:xmm, xmm, xmm/m128, ib"                 , "RVMI" , "VEX.NDS.128.66.0F3A.WIG 0F /r ib" , "ANY AVX"],
    ["vpalignr"         , "W:ymm, ymm, ymm/m256, ib"                 , "RVMI" , "VEX.NDS.256.66.0F3A.WIG 0F /r ib" , "ANY AVX2"],
    ["vpand"            , "W:xmm, xmm, xmm/m128"                     , "RVM"  , "VEX.NDS.128.66.0F.WIG DB /r"      , "ANY AVX"],
    ["vpand"            , "W:ymm, ymm, ymm/m256"                     , "RVM"  , "VEX.NDS.256.66.0F.WIG DB /r"      , "ANY AVX2"],
    ["vpandn"           , "W:xmm, xmm, xmm/m128"                     , "RVM"  , "VEX.NDS.128.66.0F.WIG DF /r"      , "ANY AVX"],
    ["vpandn"           , "W:ymm, ymm, ymm/m256"                     , "RVM"  , "VEX.NDS.256.66.0F.WIG DF /r"      , "ANY AVX2"],
    ["vpavgb"           , "W:xmm, xmm, xmm/m128"                     , "RVM"  , "VEX.NDS.128.66.0F.WIG E0 /r"      , "ANY AVX"],
    ["vpavgb"           , "W:ymm, ymm, ymm/m256"                     , "RVM"  , "VEX.NDS.256.66.0F.WIG E0 /r"      , "ANY AVX"],
    ["vpavgw"           , "W:xmm, xmm, xmm/m128"                     , "RVM"  , "VEX.NDS.128.66.0F.WIG E3 /r"      , "ANY AVX2"],
    ["vpavgw"           , "W:ymm, ymm, ymm/m256"                     , "RVM"  , "VEX.NDS.256.66.0F.WIG E3 /r"      , "ANY AVX2"],
    ["vpblendd"         , "W:xmm, xmm, xmm/m128, ib"                 , "RVMI" , "VEX.NDS.128.66.0F3A.W0 02 /r ib"  , "ANY AVX2"],
    ["vpblendd"         , "W:ymm, ymm, ymm/m256, ib"                 , "RVMI" , "VEX.NDS.256.66.0F3A.W0 02 /r ib"  , "ANY AVX2"],
    ["vpblendvb"        , "W:xmm, xmm, xmm/m128, xmm"                , "RVMS" , "VEX.NDS.128.66.0F3A.W0 4C /r /is4", "ANY AVX"],
    ["vpblendvb"        , "W:ymm, ymm, ymm/m256, ymm"                , "RVMS" , "VEX.NDS.256.66.0F3A.W0 4C /r /is4", "ANY AVX2"],
    ["vpblendw"         , "W:xmm, xmm, xmm/m128, xmm, ib"            , "RVMI" , "VEX.NDS.128.66.0F3A.WIG 0E /r ib" , "ANY AVX"],
    ["vpblendw"         , "W:ymm, ymm, ymm/m256, ymm, ib"            , "RVMI" , "VEX.NDS.256.66.0F3A.WIG 0E /r ib" , "ANY AVX2"],
    ["vpbroadcastb"     , "W:xmm, xmm/m8"                            , "RM"   , "VEX.128.66.0F38.W0 78 /r"         , "ANY AVX2"],
    ["vpbroadcastb"     , "W:ymm, xmm/m8"                            , "RM"   , "VEX.256.66.0F38.W0 78 /r"         , "ANY AVX2"],
    ["vpbroadcastd"     , "W:xmm, xmm/m32"                           , "RM"   , "VEX.128.66.0F38.W0 58 /r"         , "ANY AVX2"],
    ["vpbroadcastd"     , "W:ymm, xmm/m32"                           , "RM"   , "VEX.256.66.0F38.W0 58 /r"         , "ANY AVX2"],
    ["vpbroadcastq"     , "W:xmm, xmm/m64"                           , "RM"   , "VEX.128.66.0F38.W0 59 /r"         , "ANY AVX2"],
    ["vpbroadcastq"     , "W:ymm, xmm/m64"                           , "RM"   , "VEX.256.66.0F38.W0 59 /r"         , "ANY AVX2"],
    ["vpbroadcastw"     , "W:xmm, xmm/m16"                           , "RM"   , "VEX.128.66.0F38.W0 79 /r"         , "ANY AVX2"],
    ["vpbroadcastw"     , "W:ymm, xmm/m16"                           , "RM"   , "VEX.256.66.0F38.W0 79 /r"         , "ANY AVX2"],
    ["vpclmulqdq"       , "W:xmm, xmm, xmm/m128, ib"                 , "RVMI" , "VEX.NDS.128.66.0F3A.WIG 44 /r ib" , "ANY AVX PCLMULQDQ"],
    ["vpcmpeqb"         , "W:xmm, xmm, xmm/m128"                     , "RVM"  , "VEX.NDS.128.66.0F.WIG 74 /r"      , "ANY AVX"],
    ["vpcmpeqb"         , "W:ymm, ymm, ymm/m256"                     , "RVM"  , "VEX.NDS.256.66.0F.WIG 74 /r"      , "ANY AVX2"],
    ["vpcmpeqd"         , "W:xmm, xmm, xmm/m128"                     , "RVM"  , "VEX.NDS.128.66.0F.WIG 76 /r"      , "ANY AVX"],
    ["vpcmpeqd"         , "W:ymm, ymm, ymm/m256"                     , "RVM"  , "VEX.NDS.256.66.0F.WIG 76 /r"      , "ANY AVX2"],
    ["vpcmpeqq"         , "W:xmm, xmm, xmm/m128"                     , "RVM"  , "VEX.NDS.128.66.0F38.WIG 29 /r"    , "ANY AVX"],
    ["vpcmpeqq"         , "W:ymm, ymm, ymm/m256"                     , "RVM"  , "VEX.NDS.256.66.0F38.WIG 29 /r"    , "ANY AVX2"],
    ["vpcmpeqw"         , "W:xmm, xmm, xmm/m128"                     , "RVM"  , "VEX.NDS.128.66.0F.WIG 75 /r"      , "ANY AVX"],
    ["vpcmpeqw"         , "W:ymm, ymm, ymm/m256"                     , "RVM"  , "VEX.NDS.256.66.0F.WIG 75 /r"      , "ANY AVX2"],
    ["vpcmpestri"       , "W:<ecx>, xmm, <eax>, xmm/m128, <edx>, ib" , "RMI"  , "VEX.128.66.0F3A.WIG 61 /r ib"     , "ANY AVX OF=W SF=W ZF=W AF=0 PF=0 CF=W"],
    ["vpcmpestrm"       , "W:<xmm0>, xmm, <eax>, xmm/m128, <edx>, ib", "RMI"  , "VEX.128.66.0F3A.WIG 60 /r ib"     , "ANY AVX OF=W SF=W ZF=W AF=0 PF=0 CF=W"],
    ["vpcmpgtb"         , "W:xmm, xmm, xmm/m128"                     , "RVM"  , "VEX.NDS.128.66.0F.WIG 64 /r"      , "ANY AVX"],
    ["vpcmpgtb"         , "W:ymm, ymm, ymm/m256"                     , "RVM"  , "VEX.NDS.256.66.0F.WIG 64 /r"      , "ANY AVX2"],
    ["vpcmpgtd"         , "W:xmm, xmm, xmm/m128"                     , "RVM"  , "VEX.NDS.128.66.0F.WIG 66 /r"      , "ANY AVX"],
    ["vpcmpgtd"         , "W:ymm, ymm, ymm/m256"                     , "RVM"  , "VEX.NDS.256.66.0F.WIG 66 /r"      , "ANY AVX2"],
    ["vpcmpgtq"         , "W:xmm, xmm, xmm/m128"                     , "RVM"  , "VEX.NDS.128.66.0F38.WIG 37 /r"    , "ANY AVX"],
    ["vpcmpgtq"         , "W:ymm, ymm, ymm/m256"                     , "RVM"  , "VEX.NDS.256.66.0F38.WIG 37 /r"    , "ANY AVX2"],
    ["vpcmpgtw"         , "W:xmm, xmm, xmm/m128"                     , "RVM"  , "VEX.NDS.128.66.0F.WIG 65 /r"      , "ANY AVX"],
    ["vpcmpgtw"         , "W:ymm, ymm, ymm/m256"                     , "RVM"  , "VEX.NDS.256.66.0F.WIG 65 /r"      , "ANY AVX2"],
    ["vpcmpistri"       , "W:<ecx>, xmm, xmm/m128, ib"               , "RMI"  , "VEX.128.66.0F3A.WIG 63 /r ib"     , "ANY AVX OF=W SF=W ZF=W AF=0 PF=0 CF=W"],
    ["vpcmpistrm"       , "W:<xmm0>, xmm, xmm/m128, ib"              , "RMI"  , "VEX.128.66.0F3A.WIG 62 /r ib"     , "ANY AVX OF=W SF=W ZF=W AF=0 PF=0 CF=W"],
    ["vperm2f128"       , "W:ymm, ymm, ymm/m256, ib"                 , "RVMI" , "VEX.NDS.256.66.0F3A.W0 06 /r ib"  , "ANY AVX"],
    ["vperm2i128"       , "W:ymm, ymm, ymm/m256, ib"                 , "RVMI" , "VEX.NDS.256.66.0F3A.W0 46 /r ib"  , "ANY AVX2"],
    ["vpermd"           , "W:ymm, ymm, ymm/m256"                     , "RVM"  , "VEX.NDS.256.66.0F38.W0 36 /r"     , "ANY AVX2"],
    ["vpermilpd"        , "W:xmm, xmm, xmm/m128"                     , "RVM"  , "VEX.NDS.128.66.0F38.W0 0D /r"     , "ANY AVX"],
    ["vpermilpd"        , "W:xmm, xmm/m128, ib"                      , "RMI"  , "VEX.128.66.0F3A.W0 05 /r ib"      , "ANY AVX"],
    ["vpermilpd"        , "W:ymm, ymm, ymm/m256"                     , "RVM"  , "VEX.NDS.256.66.0F38.W0 0D /r"     , "ANY AVX"],
    ["vpermilpd"        , "W:ymm, ymm/m256, ib"                      , "RMI"  , "VEX.256.66.0F3A.W0 05 /r ib"      , "ANY AVX"],
    ["vpermilps"        , "W:xmm, xmm, xmm/m128"                     , "RVM"  , "VEX.NDS.128.66.0F38.W0 0C /r"     , "ANY AVX"],
    ["vpermilps"        , "W:xmm, xmm/m128, ib"                      , "RMI"  , "VEX.128.66.0F3A.W0 04 /r ib"      , "ANY AVX"],
    ["vpermilps"        , "W:ymm, ymm, ymm/m256"                     , "RVM"  , "VEX.NDS.256.66.0F38.W0 0C /r"     , "ANY AVX"],
    ["vpermilps"        , "W:ymm, ymm/m256, ib"                      , "RMI"  , "VEX.256.66.0F3A.W0 04 /r ib"      , "ANY AVX"],
    ["vpermpd"          , "W:ymm, ymm/m256, ib"                      , "RMI"  , "VEX.256.66.0F3A.W1 01 /r ib"      , "ANY AVX2"],
    ["vpermps"          , "W:ymm, ymm, ymm/m256"                     , "RVM"  , "VEX.NDS.256.66.0F38.W0 16 /r"     , "ANY AVX2"],
    ["vpermq"           , "W:ymm, ymm/m256, ib"                      , "RMI"  , "VEX.256.66.0F3A.W1 00 /r ib"      , "ANY AVX2"],
    ["vpextrb"          , "W:r32/m8, xmm, ib"                        , "MRI"  , "VEX.128.66.0F3A.W0 14 /r ib"      , "ANY AVX"],
    ["vpextrd"          , "W:r32/m32, xmm, ib"                       , "MRI"  , "VEX.128.66.0F3A.W0 16 /r ib"      , "ANY AVX"],
    ["vpextrq"          , "W:r64/m64, xmm, ib"                       , "MRI"  , "VEX.128.66.0F3A.W1 16 /r ib"      , "X64 AVX"],
    ["vpextrw"          , "W:r32, xmm, ib"                           , "RMI"  , "VEX.128.66.0F.W0 C5 /r ib"        , "ANY AVX"],
    ["vpextrw"          , "W:r32/m16, xmm, ib"                       , "MRI"  , "VEX.128.66.0F3A.W0 15 /r ib"      , "ANY AVX"],
    ["vpgatherdd"       , "W:xmm, vm32x, xmm"                        , "RMV"  , "VEX.DDS.128.66.0F38.W0 90 /r"     , "ANY AVX2"],
    ["vpgatherdd"       , "W:ymm, vm32y, ymm"                        , "RMV"  , "VEX.DDS.256.66.0F38.W0 90 /r"     , "ANY AVX2"],
    ["vpgatherdq"       , "W:xmm, vm32x, xmm"                        , "RMV"  , "VEX.DDS.128.66.0F38.W1 90 /r"     , "ANY AVX2"],
    ["vpgatherdq"       , "W:ymm, vm32x, ymm"                        , "RMV"  , "VEX.DDS.256.66.0F38.W1 90 /r"     , "ANY AVX2"],
    ["vpgatherqd"       , "W:xmm, vm64x, xmm"                        , "RMV"  , "VEX.DDS.128.66.0F38.W0 91 /r"     , "ANY AVX2"],
    ["vpgatherqd"       , "W:xmm, vm64y, xmm"                        , "RMV"  , "VEX.DDS.256.66.0F38.W0 91 /r"     , "ANY AVX2"],
    ["vpgatherqq"       , "W:xmm, vm64x, xmm"                        , "RMV"  , "VEX.DDS.128.66.0F38.W1 91 /r"     , "ANY AVX2"],
    ["vpgatherqq"       , "W:ymm, vm64y, ymm"                        , "RMV"  , "VEX.DDS.256.66.0F38.W1 91 /r"     , "ANY AVX2"],
    ["vphaddd"          , "W:xmm, xmm, xmm/m128"                     , "RVM"  , "VEX.NDS.128.66.0F38.WIG 02 /r"    , "ANY AVX"],
    ["vphaddd"          , "W:ymm, ymm, ymm/m256"                     , "RVM"  , "VEX.NDS.256.66.0F38.WIG 02 /r"    , "ANY AVX2"],
    ["vphaddsw"         , "W:xmm, xmm, xmm/m128"                     , "RVM"  , "VEX.NDS.128.66.0F38.WIG 03 /r"    , "ANY AVX"],
    ["vphaddsw"         , "W:ymm, ymm, ymm/m256"                     , "RVM"  , "VEX.NDS.256.66.0F38.WIG 03 /r"    , "ANY AVX2"],
    ["vphaddw"          , "W:xmm, xmm, xmm/m128"                     , "RVM"  , "VEX.NDS.128.66.0F38.WIG 01 /r"    , "ANY AVX"],
    ["vphaddw"          , "W:ymm, ymm, ymm/m256"                     , "RVM"  , "VEX.NDS.256.66.0F38.WIG 01 /r"    , "ANY AVX2"],
    ["vphminposuw"      , "W:xmm, xmm/m128"                          , "RM"   , "VEX.128.66.0F38.WIG 41 /r"        , "ANY AVX"],
    ["vphsubd"          , "W:xmm, xmm, xmm/m128"                     , "RVM"  , "VEX.NDS.128.66.0F38.WIG 06 /r"    , "ANY AVX"],
    ["vphsubd"          , "W:ymm, ymm, ymm/m256"                     , "RVM"  , "VEX.NDS.256.66.0F38.WIG 06 /r"    , "ANY AVX2"],
    ["vphsubsw"         , "W:xmm, xmm, xmm/m128"                     , "RVM"  , "VEX.NDS.128.66.0F38.WIG 07 /r"    , "ANY AVX"],
    ["vphsubsw"         , "W:ymm, ymm, ymm/m256"                     , "RVM"  , "VEX.NDS.256.66.0F38.WIG 07 /r"    , "ANY AVX2"],
    ["vphsubw"          , "W:xmm, xmm, xmm/m128"                     , "RVM"  , "VEX.NDS.128.66.0F38.WIG 05 /r"    , "ANY AVX"],
    ["vphsubw"          , "W:ymm, ymm, ymm/m256"                     , "RVM"  , "VEX.NDS.256.66.0F38.WIG 05 /r"    , "ANY AVX2"],
    ["vpinsrb"          , "W:xmm, r32/m8, ib"                        , "RMI"  , "VEX.NDS.128.66.0F3A.W0 20 /r ib"  , "ANY AVX"],
    ["vpinsrd"          , "W:xmm, r32/m32, ib"                       , "RMI"  , "VEX.NDS.128.66.0F3A.W0 22 /r ib"  , "ANY AVX"],
    ["vpinsrq"          , "W:xmm, r64/m64, ib"                       , "RMI"  , "VEX.NDS.128.66.0F3A.W1 22 /r ib"  , "X64 AVX"],
    ["vpinsrw"          , "W:xmm, xmm, r32/m16, ib"                  , "RVMI" , "VEX.NDS.128.66.0F.W0 C4 /r ib"    , "ANY AVX"],
    ["vpmaddubsw"       , "W:xmm, xmm, xmm/m128"                     , "RVM"  , "VEX.NDS.128.66.0F38.WIG 04 /r"    , "ANY AVX"],
    ["vpmaddubsw"       , "W:ymm, ymm, ymm/m256"                     , "RVM"  , "VEX.NDS.256.66.0F38.WIG 04 /r"    , "ANY AVX2"],
    ["vpmaddwd"         , "W:xmm, xmm, xmm/m128"                     , "RVM"  , "VEX.NDS.128.66.0F.WIG F5 /r"      , "ANY AVX"],
    ["vpmaddwd"         , "W:ymm, ymm, ymm/m256"                     , "RVM"  , "VEX.NDS.256.66.0F.WIG F5 /r"      , "ANY AVX2"],
    ["vpmaskmovd"       , "W:m128, xmm, xmm"                         , "MVR"  , "VEX.NDS.128.66.0F38.W0 8E /r"     , "ANY AVX2"],
    ["vpmaskmovd"       , "W:m256, ymm, ymm"                         , "MVR"  , "VEX.NDS.256.66.0F38.W0 8E /r"     , "ANY AVX2"],
    ["vpmaskmovd"       , "W:xmm, xmm, m128"                         , "RVM"  , "VEX.NDS.128.66.0F38.W0 8C /r"     , "ANY AVX2"],
    ["vpmaskmovd"       , "W:ymm, ymm, m256"                         , "RVM"  , "VEX.NDS.256.66.0F38.W0 8C /r"     , "ANY AVX2"],
    ["vpmaskmovq"       , "W:m128, xmm, xmm"                         , "MVR"  , "VEX.NDS.128.66.0F38.W1 8E /r"     , "ANY AVX2"],
    ["vpmaskmovq"       , "W:m256, ymm, ymm"                         , "MVR"  , "VEX.NDS.256.66.0F38.W1 8E /r"     , "ANY AVX2"],
    ["vpmaskmovq"       , "W:xmm, xmm, m128"                         , "RVM"  , "VEX.NDS.128.66.0F38.W1 8C /r"     , "ANY AVX2"],
    ["vpmaskmovq"       , "W:ymm, ymm, m256"                         , "RVM"  , "VEX.NDS.256.66.0F38.W1 8C /r"     , "ANY AVX2"],
    ["vpmaxsb"          , "W:xmm, xmm, xmm/m128"                     , "RVM"  , "VEX.NDS.128.66.0F38.WIG 3C /r"    , "ANY AVX"],
    ["vpmaxsb"          , "W:ymm, ymm, ymm/m256"                     , "RVM"  , "VEX.NDS.256.66.0F38.WIG 3C /r"    , "ANY AVX2"],
    ["vpmaxsd"          , "W:xmm, xmm, xmm/m128"                     , "RVM"  , "VEX.NDS.128.66.0F38.WIG 3D /r"    , "ANY AVX"],
    ["vpmaxsd"          , "W:ymm, ymm, ymm/m256"                     , "RVM"  , "VEX.NDS.256.66.0F38.WIG 3D /r"    , "ANY AVX2"],
    ["vpmaxsw"          , "W:xmm, xmm, xmm/m128"                     , "RVM"  , "VEX.NDS.128.66.0F.WIG EE /r"      , "ANY AVX"],
    ["vpmaxsw"          , "W:ymm, ymm, ymm/m256"                     , "RVM"  , "VEX.NDS.256.66.0F.WIG EE /r"      , "ANY AVX2"],
    ["vpmaxub"          , "W:xmm, xmm, xmm/m128"                     , "RVM"  , "VEX.NDS.128.66.0F.WIG DE /r"      , "ANY AVX"],
    ["vpmaxub"          , "W:ymm, ymm, ymm/m256"                     , "RVM"  , "VEX.NDS.256.66.0F.WIG DE /r"      , "ANY AVX2"],
    ["vpmaxud"          , "W:xmm, xmm, xmm/m128"                     , "RVM"  , "VEX.NDS.128.66.0F38.WIG 3F /r"    , "ANY AVX"],
    ["vpmaxud"          , "W:ymm, ymm, ymm/m256"                     , "RVM"  , "VEX.NDS.256.66.0F38.WIG 3F /r"    , "ANY AVX2"],
    ["vpmaxuw"          , "W:xmm, xmm, xmm/m128"                     , "RVM"  , "VEX.NDS.128.66.0F38.WIG 3E /r"    , "ANY AVX"],
    ["vpmaxuw"          , "W:ymm, ymm, ymm/m256"                     , "RVM"  , "VEX.NDS.256.66.0F38.WIG 3E /r"    , "ANY AVX2"],
    ["vpminsb"          , "W:xmm, xmm, xmm/m128"                     , "RVM"  , "VEX.NDS.128.66.0F38.WIG 38 /r"    , "ANY AVX"],
    ["vpminsb"          , "W:ymm, ymm, ymm/m256"                     , "RVM"  , "VEX.NDS.256.66.0F38.WIG 38 /r"    , "ANY AVX2"],
    ["vpminsd"          , "W:xmm, xmm, xmm/m128"                     , "RVM"  , "VEX.NDS.128.66.0F38.WIG 39 /r"    , "ANY AVX"],
    ["vpminsd"          , "W:ymm, ymm, ymm/m256"                     , "RVM"  , "VEX.NDS.256.66.0F38.WIG 39 /r"    , "ANY AVX2"],
    ["vpminsw"          , "W:xmm, xmm, xmm/m128"                     , "RVM"  , "VEX.NDS.128.66.0F.WIG EA /r"      , "ANY AVX"],
    ["vpminsw"          , "W:ymm, ymm, ymm/m256"                     , "RVM"  , "VEX.NDS.256.66.0F.WIG EA /r"      , "ANY AVX2"],
    ["vpminub"          , "W:xmm, xmm, xmm/m128"                     , "RVM"  , "VEX.NDS.128.66.0F.WIG DA /r"      , "ANY AVX"],
    ["vpminub"          , "W:ymm, ymm, ymm/m256"                     , "RVM"  , "VEX.NDS.256.66.0F.WIG DA /r"      , "ANY AVX2"],
    ["vpminud"          , "W:xmm, xmm, xmm/m128"                     , "RVM"  , "VEX.NDS.128.66.0F38.WIG 3B /r"    , "ANY AVX"],
    ["vpminud"          , "W:ymm, ymm, ymm/m256"                     , "RVM"  , "VEX.NDS.256.66.0F38.WIG 3B /r"    , "ANY AVX2"],
    ["vpminuw"          , "W:xmm, xmm, xmm/m128"                     , "RVM"  , "VEX.NDS.128.66.0F38.WIG 3A /r"    , "ANY AVX"],
    ["vpminuw"          , "W:ymm, ymm, ymm/m256"                     , "RVM"  , "VEX.NDS.256.66.0F38.WIG 3A /r"    , "ANY AVX2"],
    ["vpmovmskb"        , "W:r32, xmm"                               , "RVM"  , "VEX.128.66.0F.WIG D7 /r"          , "ANY AVX"],
    ["vpmovmskb"        , "W:r64, xmm"                               , "RVM"  , "VEX.128.66.0F.WIG D7 /r"          , "X64 AVX"],
    ["vpmovmskb"        , "W:r32, ymm"                               , "RVM"  , "VEX.256.66.0F.WIG D7 /r"          , "ANY AVX2"],
    ["vpmovmskb"        , "W:r64, ymm"                               , "RVM"  , "VEX.256.66.0F.WIG D7 /r"          , "X64 AVX2"],
    ["vpmovsxbd"        , "W:xmm, xmm/m32"                           , "RM"   , "VEX.128.66.0F38.WIG 21 /r"        , "ANY AVX"],
    ["vpmovsxbd"        , "W:ymm, xmm/m64"                           , "RM"   , "VEX.256.66.0F38.WIG 21 /r"        , "ANY AVX2"],
    ["vpmovsxbq"        , "W:xmm, xmm/m16"                           , "RM"   , "VEX.128.66.0F38.WIG 22 /r"        , "ANY AVX"],
    ["vpmovsxbq"        , "W:ymm, xmm/m32"                           , "RM"   , "VEX.256.66.0F38.WIG 22 /r"        , "ANY AVX2"],
    ["vpmovsxbw"        , "W:xmm, xmm/m64"                           , "RM"   , "VEX.128.66.0F38.WIG 20 /r"        , "ANY AVX"],
    ["vpmovsxbw"        , "W:ymm, xmm/m128"                          , "RM"   , "VEX.256.66.0F38.WIG 20 /r"        , "ANY AVX2"],
    ["vpmovsxdq"        , "W:xmm, xmm/m64"                           , "RM"   , "VEX.128.66.0F38.WIG 25 /r"        , "ANY AVX"],
    ["vpmovsxdq"        , "W:ymm, xmm/m128"                          , "RM"   , "VEX.256.66.0F38.WIG 25 /r"        , "ANY AVX2"],
    ["vpmovsxwd"        , "W:xmm, xmm/m64"                           , "RM"   , "VEX.128.66.0F38.WIG 23 /r"        , "ANY AVX"],
    ["vpmovsxwd"        , "W:ymm, xmm/m128"                          , "RM"   , "VEX.256.66.0F38.WIG 23 /r"        , "ANY AVX2"],
    ["vpmovsxwq"        , "W:xmm, xmm/m32"                           , "RM"   , "VEX.128.66.0F38.WIG 24 /r"        , "ANY AVX"],
    ["vpmovsxwq"        , "W:ymm, xmm/m64"                           , "RM"   , "VEX.256.66.0F38.WIG 24 /r"        , "ANY AVX2"],
    ["vpmovzxbd"        , "W:xmm, xmm/m32"                           , "RM"   , "VEX.128.66.0F38.WIG 31 /r"        , "ANY AVX"],
    ["vpmovzxbd"        , "W:ymm, xmm/m64"                           , "RM"   , "VEX.256.66.0F38.WIG 31 /r"        , "ANY AVX2"],
    ["vpmovzxbq"        , "W:xmm, xmm/m16"                           , "RM"   , "VEX.128.66.0F38.WIG 32 /r"        , "ANY AVX"],
    ["vpmovzxbq"        , "W:ymm, xmm/m32"                           , "RM"   , "VEX.256.66.0F38.WIG 32 /r"        , "ANY AVX2"],
    ["vpmovzxbw"        , "W:xmm, xmm/m64"                           , "RM"   , "VEX.128.66.0F38.WIG 30 /r"        , "ANY AVX"],
    ["vpmovzxbw"        , "W:ymm, xmm/m128"                          , "RM"   , "VEX.256.66.0F38.WIG 30 /r"        , "ANY AVX2"],
    ["vpmovzxdq"        , "W:xmm, xmm/m64"                           , "RM"   , "VEX.128.66.0F38.WIG 35 /r"        , "ANY AVX"],
    ["vpmovzxdq"        , "W:ymm, xmm/m128"                          , "RM"   , "VEX.256.66.0F38.WIG 35 /r"        , "ANY AVX2"],
    ["vpmovzxwd"        , "W:xmm, xmm/m64"                           , "RM"   , "VEX.128.66.0F38.WIG 33 /r"        , "ANY AVX"],
    ["vpmovzxwd"        , "W:ymm, xmm/m128"                          , "RM"   , "VEX.256.66.0F38.WIG 33 /r"        , "ANY AVX2"],
    ["vpmovzxwq"        , "W:xmm, xmm/m32"                           , "RM"   , "VEX.128.66.0F38.WIG 34 /r"        , "ANY AVX"],
    ["vpmovzxwq"        , "W:ymm, xmm/m64"                           , "RM"   , "VEX.256.66.0F38.WIG 34 /r"        , "ANY AVX2"],
    ["vpmuldq"          , "W:xmm, xmm, xmm/m128"                     , "RVM"  , "VEX.NDS.128.66.0F38.WIG 28 /r"    , "ANY AVX"],
    ["vpmuldq"          , "W:ymm, ymm, ymm/m256"                     , "RVM"  , "VEX.NDS.256.66.0F38.WIG 28 /r"    , "ANY AVX2"],
    ["vpmulhrsw"        , "W:xmm, xmm, xmm/m128"                     , "RVM"  , "VEX.NDS.128.66.0F38.WIG 0B /r"    , "ANY AVX"],
    ["vpmulhrsw"        , "W:ymm, ymm, ymm/m256"                     , "RVM"  , "VEX.NDS.256.66.0F38.WIG 0B /r"    , "ANY AVX2"],
    ["vpmulhuw"         , "W:xmm, xmm, xmm/m128"                     , "RVM"  , "VEX.NDS.128.66.0F.WIG E4 /r"      , "ANY AVX"],
    ["vpmulhuw"         , "W:ymm, ymm, ymm/m256"                     , "RVM"  , "VEX.NDS.256.66.0F.WIG E4 /r"      , "ANY AVX2"],
    ["vpmulhw"          , "W:xmm, xmm, xmm/m128"                     , "RVM"  , "VEX.NDS.128.66.0F.WIG E5 /r"      , "ANY AVX"],
    ["vpmulhw"          , "W:ymm, ymm, ymm/m256"                     , "RVM"  , "VEX.NDS.256.66.0F.WIG E5 /r"      , "ANY AVX2"],
    ["vpmulld"          , "W:xmm, xmm, xmm/m128"                     , "RVM"  , "VEX.NDS.128.66.0F38.WIG 40 /r"    , "ANY AVX"],
    ["vpmulld"          , "W:ymm, ymm, ymm/m256"                     , "RVM"  , "VEX.NDS.256.66.0F38.WIG 40 /r"    , "ANY AVX2"],
    ["vpmullw"          , "W:xmm, xmm, xmm/m128"                     , "RVM"  , "VEX.NDS.128.66.0F.WIG D5 /r"      , "ANY AVX"],
    ["vpmullw"          , "W:ymm, ymm, ymm/m256"                     , "RVM"  , "VEX.NDS.256.66.0F.WIG D5 /r"      , "ANY AVX2"],
    ["vpmuludq"         , "W:xmm, xmm, xmm/m128"                     , "RVM"  , "VEX.NDS.128.66.0F.WIG F4 /r"      , "ANY AVX"],
    ["vpmuludq"         , "W:ymm, ymm, ymm/m256"                     , "RVM"  , "VEX.NDS.256.66.0F.WIG F4 /r"      , "ANY AVX2"],
    ["vpor"             , "W:xmm, xmm, xmm/m128"                     , "RVM"  , "VEX.NDS.128.66.0F.WIG EB /r"      , "ANY AVX"],
    ["vpor"             , "W:ymm, ymm, ymm/m256"                     , "RVM"  , "VEX.NDS.256.66.0F.WIG EB /r"      , "ANY AVX2"],
    ["vpsadbw"          , "W:xmm, xmm, xmm/m128"                     , "RVM"  , "VEX.NDS.128.66.0F.WIG F6 /r"      , "ANY AVX"],
    ["vpsadbw"          , "W:ymm, ymm, ymm/m256"                     , "RVM"  , "VEX.NDS.256.66.0F.WIG F6 /r"      , "ANY AVX2"],
    ["vpshufb"          , "W:xmm, xmm, xmm/m128"                     , "RVM"  , "VEX.NDS.128.66.0F38.WIG 00 /r"    , "ANY AVX"],
    ["vpshufb"          , "W:ymm, ymm, ymm/m256"                     , "RVM"  , "VEX.NDS.256.66.0F38.WIG 00 /r"    , "ANY AVX2"],
    ["vpshufd"          , "W:xmm, xmm, xmm/m128, ib"                 , "RVMI" , "VEX.128.66.0F.WIG 70 /r ib"       , "ANY AVX"],
    ["vpshufd"          , "W:ymm, ymm, ymm/m256, ib"                 , "RVMI" , "VEX.256.66.0F.WIG 70 /r ib"       , "ANY AVX2"],
    ["vpshufhw"         , "W:xmm, xmm, xmm/m128, ib"                 , "RVMI" , "VEX.128.F3.0F.WIG 70 /r ib"       , "ANY AVX"],
    ["vpshufhw"         , "W:ymm, ymm, ymm/m256, ib"                 , "RVMI" , "VEX.256.F3.0F.WIG 70 /r ib"       , "ANY AVX2"],
    ["vpshuflw"         , "W:xmm, xmm, xmm/m128, ib"                 , "RVMI" , "VEX.128.F2.0F.WIG 70 /r ib"       , "ANY AVX"],
    ["vpshuflw"         , "W:ymm, ymm, ymm/m256, ib"                 , "RVMI" , "VEX.256.F2.0F.WIG 70 /r ib"       , "ANY AVX2"],
    ["vpsignb"          , "W:xmm, xmm, xmm/m128"                     , "RVM"  , "VEX.NDS.128.66.0F38.WIG 08 /r"    , "ANY AVX"],
    ["vpsignb"          , "W:ymm, ymm, ymm/m256"                     , "RVM"  , "VEX.NDS.256.66.0F38.WIG 08 /r"    , "ANY AVX2"],
    ["vpsignd"          , "W:xmm, xmm, xmm/m128"                     , "RVM"  , "VEX.NDS.128.66.0F38.WIG 0A /r"    , "ANY AVX"],
    ["vpsignd"          , "W:ymm, ymm, ymm/m256"                     , "RVM"  , "VEX.NDS.256.66.0F38.WIG 0A /r"    , "ANY AVX2"],
    ["vpsignw"          , "W:xmm, xmm, xmm/m128"                     , "RVM"  , "VEX.NDS.128.66.0F38.WIG 09 /r"    , "ANY AVX"],
    ["vpsignw"          , "W:ymm, ymm, ymm/m256"                     , "RVM"  , "VEX.NDS.256.66.0F38.WIG 09 /r"    , "ANY AVX2"],
    ["vpslld"           , "W:xmm, xmm, ib"                           , "VMI"  , "VEX.NDD.128.66.0F.WIG 72 /6 ib"   , "ANY AVX"],
    ["vpslld"           , "W:xmm, xmm, xmm/m128"                     , "RVM"  , "VEX.NDS.128.66.0F.WIG F2 /r"      , "ANY AVX"],
    ["vpslld"           , "W:ymm, ymm, ib"                           , "VMI"  , "VEX.NDD.256.66.0F.WIG 72 /6 ib"   , "ANY AVX2"],
    ["vpslld"           , "W:ymm, ymm, xmm/m128"                     , "RVM"  , "VEX.NDS.256.66.0F.WIG F2 /r"      , "ANY AVX2"],
    ["vpslldq"          , "W:xmm, xmm, ib"                           , "VMI"  , "VEX.NDD.128.66.0F.WIG 73 /7 ib"   , "ANY AVX"],
    ["vpslldq"          , "W:ymm, ymm, ib"                           , "VMI"  , "VEX.NDD.256.66.0F.WIG 73 /7 ib"   , "ANY AVX2"],
    ["vpsllq"           , "W:xmm, xmm, ib"                           , "VMI"  , "VEX.NDD.128.66.0F.WIG 73 /6 ib"   , "ANY AVX"],
    ["vpsllq"           , "W:xmm, xmm, xmm/m128"                     , "RVM"  , "VEX.NDS.128.66.0F.WIG F3 /r"      , "ANY AVX"],
    ["vpsllq"           , "W:ymm, ymm, ib"                           , "VMI"  , "VEX.NDD.256.66.0F.WIG 73 /6 ib"   , "ANY AVX2"],
    ["vpsllq"           , "W:ymm, ymm, xmm/m128"                     , "RVM"  , "VEX.NDS.256.66.0F.WIG F3 /r"      , "ANY AVX2"],
    ["vpsllvd"          , "W:xmm, xmm, xmm/m128"                     , "RVM"  , "VEX.NDS.128.66.0F38.W0 47 /r"     , "ANY AVX2"],
    ["vpsllvd"          , "W:ymm, ymm, ymm/m256"                     , "RVM"  , "VEX.NDS.256.66.0F38.W0 47 /r"     , "ANY AVX2"],
    ["vpsllvq"          , "W:xmm, xmm, xmm/m128"                     , "RVM"  , "VEX.NDS.128.66.0F38.W1 47 /r"     , "ANY AVX2"],
    ["vpsllvq"          , "W:ymm, ymm, ymm/m256"                     , "RVM"  , "VEX.NDS.256.66.0F38.W1 47 /r"     , "ANY AVX2"],
    ["vpsllw"           , "W:xmm, xmm, ib"                           , "VMI"  , "VEX.NDD.128.66.0F.WIG 71 /6 ib"   , "ANY AVX"],
    ["vpsllw"           , "W:xmm, xmm, xmm/m128"                     , "RVM"  , "VEX.NDS.128.66.0F.WIG F1 /r"      , "ANY AVX"],
    ["vpsllw"           , "W:ymm, ymm, ib"                           , "VMI"  , "VEX.NDD.256.66.0F.WIG 71 /6 ib"   , "ANY AVX2"],
    ["vpsllw"           , "W:ymm, ymm, xmm/m128"                     , "RVM"  , "VEX.NDS.256.66.0F.WIG F1 /r"      , "ANY AVX2"],
    ["vpsrad"           , "W:xmm, xmm, ib"                           , "VMI"  , "VEX.NDD.128.66.0F.WIG 72 /4 ib"   , "ANY AVX"],
    ["vpsrad"           , "W:xmm, xmm, xmm/m128"                     , "RVM"  , "VEX.NDS.128.66.0F.WIG E2 /r"      , "ANY AVX"],
    ["vpsrad"           , "W:ymm, ymm, ib"                           , "VMI"  , "VEX.NDD.256.66.0F.WIG 72 /4 ib"   , "ANY AVX2"],
    ["vpsrad"           , "W:ymm, ymm, xmm/m128"                     , "RVM"  , "VEX.NDS.256.66.0F.WIG E2 /r"      , "ANY AVX2"],
    ["vpsravd"          , "W:xmm, xmm, xmm/m128"                     , "RVM"  , "VEX.NDS.128.66.0F38.W0 46 /r"     , "ANY AVX2"],
    ["vpsravd"          , "W:ymm, ymm, ymm/m256"                     , "RVM"  , "VEX.NDS.256.66.0F38.W0 46 /r"     , "ANY AVX2"],
    ["vpsraw"           , "W:xmm, xmm, ib"                           , "VMI"  , "VEX.NDD.128.66.0F.WIG 71 /4 ib"   , "ANY AVX"],
    ["vpsraw"           , "W:xmm, xmm, xmm/m128"                     , "RVM"  , "VEX.NDS.128.66.0F.WIG E1 /r"      , "ANY AVX"],
    ["vpsraw"           , "W:ymm, ymm, ib"                           , "VMI"  , "VEX.NDD.256.66.0F.WIG 71 /4 ib"   , "ANY AVX2"],
    ["vpsraw"           , "W:ymm, ymm, xmm/m128"                     , "RVM"  , "VEX.NDS.256.66.0F.WIG E1 /r"      , "ANY AVX2"],
    ["vpsrld"           , "W:xmm, xmm, ib"                           , "VMI"  , "VEX.NDD.128.66.0F.WIG 72 /2 ib"   , "ANY AVX"],
    ["vpsrld"           , "W:xmm, xmm, xmm/m128"                     , "RVM"  , "VEX.NDS.128.66.0F.WIG D2 /r"      , "ANY AVX"],
    ["vpsrld"           , "W:ymm, ymm, ib"                           , "VMI"  , "VEX.NDD.256.66.0F.WIG 72 /2 ib"   , "ANY AVX2"],
    ["vpsrld"           , "W:ymm, ymm, xmm/m128"                     , "RVM"  , "VEX.NDS.256.66.0F.WIG D2 /r"      , "ANY AVX2"],
    ["vpsrldq"          , "W:xmm, xmm, ib"                           , "VMI"  , "VEX.NDD.128.66.0F.WIG 73 /3 ib"   , "ANY AVX"],
    ["vpsrldq"          , "W:ymm, ymm, ib"                           , "VMI"  , "VEX.NDD.256.66.0F.WIG 73 /3 ib"   , "ANY AVX2"],
    ["vpsrlq"           , "W:xmm, xmm, ib"                           , "VMI"  , "VEX.NDD.128.66.0F.WIG 73 /2 ib"   , "ANY AVX"],
    ["vpsrlq"           , "W:xmm, xmm, xmm/m128"                     , "RVM"  , "VEX.NDS.128.66.0F.WIG D3 /r"      , "ANY AVX"],
    ["vpsrlq"           , "W:ymm, ymm, ib"                           , "VMI"  , "VEX.NDD.256.66.0F.WIG 73 /2 ib"   , "ANY AVX2"],
    ["vpsrlq"           , "W:ymm, ymm, xmm/m128"                     , "RVM"  , "VEX.NDS.256.66.0F.WIG D3 /r"      , "ANY AVX2"],
    ["vpsrlvd"          , "W:xmm, xmm, xmm/m128"                     , "RVM"  , "VEX.NDS.128.66.0F38.W0 45 /r"     , "ANY AVX2"],
    ["vpsrlvd"          , "W:ymm, ymm, ymm/m256"                     , "RVM"  , "VEX.NDS.256.66.0F38.W0 45 /r"     , "ANY AVX2"],
    ["vpsrlvq"          , "W:xmm, xmm, xmm/m128"                     , "RVM"  , "VEX.NDS.128.66.0F38.W1 45 /r"     , "ANY AVX2"],
    ["vpsrlvq"          , "W:ymm, ymm, ymm/m256"                     , "RVM"  , "VEX.NDS.256.66.0F38.W1 45 /r"     , "ANY AVX2"],
    ["vpsrlw"           , "W:xmm, xmm, ib"                           , "VMI"  , "VEX.NDD.128.66.0F.WIG 71 /2 ib"   , "ANY AVX"],
    ["vpsrlw"           , "W:xmm, xmm, xmm/m128"                     , "RVM"  , "VEX.NDS.128.66.0F.WIG D1 /r"      , "ANY AVX"],
    ["vpsrlw"           , "W:ymm, ymm, ib"                           , "VMI"  , "VEX.NDD.256.66.0F.WIG 71 /2 ib"   , "ANY AVX2"],
    ["vpsrlw"           , "W:ymm, ymm, xmm/m128"                     , "RVM"  , "VEX.NDS.256.66.0F.WIG D1 /r"      , "ANY AVX2"],
    ["vpsubb"           , "W:xmm, xmm, xmm/m128"                     , "RVM"  , "VEX.NDS.128.66.0F.WIG F8 /r"      , "ANY AVX"],
    ["vpsubb"           , "W:ymm, ymm, ymm/m256"                     , "RVM"  , "VEX.NDS.256.66.0F.WIG F8 /r"      , "ANY AVX2"],
    ["vpsubd"           , "W:xmm, xmm, xmm/m128"                     , "RVM"  , "VEX.NDS.128.66.0F.WIG FA /r"      , "ANY AVX"],
    ["vpsubd"           , "W:ymm, ymm, ymm/m256"                     , "RVM"  , "VEX.NDS.256.66.0F.WIG FA /r"      , "ANY AVX2"],
    ["vpsubq"           , "W:xmm, xmm, xmm/m128"                     , "RVM"  , "VEX.NDS.128.66.0F.WIG FB /r"      , "ANY AVX"],
    ["vpsubq"           , "W:ymm, ymm, ymm/m256"                     , "RVM"  , "VEX.NDS.256.66.0F.WIG FB /r"      , "ANY AVX2"],
    ["vpsubsb"          , "W:xmm, xmm, xmm/m128"                     , "RVM"  , "VEX.NDS.128.66.0F.WIG E8 /r"      , "ANY AVX"],
    ["vpsubsb"          , "W:ymm, ymm, ymm/m256"                     , "RVM"  , "VEX.NDS.256.66.0F.WIG E8 /r"      , "ANY AVX2"],
    ["vpsubsw"          , "W:xmm, xmm, xmm/m128"                     , "RVM"  , "VEX.NDS.128.66.0F.WIG E9 /r"      , "ANY AVX"],
    ["vpsubsw"          , "W:ymm, ymm, ymm/m256"                     , "RVM"  , "VEX.NDS.256.66.0F.WIG E9 /r"      , "ANY AVX2"],
    ["vpsubusb"         , "W:xmm, xmm, xmm/m128"                     , "RVM"  , "VEX.NDS.128.66.0F.WIG D8 /r"      , "ANY AVX"],
    ["vpsubusb"         , "W:ymm, ymm, ymm/m256"                     , "RVM"  , "VEX.NDS.256.66.0F.WIG D8 /r"      , "ANY AVX2"],
    ["vpsubusw"         , "W:xmm, xmm, xmm/m128"                     , "RVM"  , "VEX.NDS.128.66.0F.WIG D9 /r"      , "ANY AVX"],
    ["vpsubusw"         , "W:ymm, ymm, ymm/m256"                     , "RVM"  , "VEX.NDS.256.66.0F.WIG D9 /r"      , "ANY AVX2"],
    ["vpsubw"           , "W:xmm, xmm, xmm/m128"                     , "RVM"  , "VEX.NDS.128.66.0F.WIG F9 /r"      , "ANY AVX"],
    ["vpsubw"           , "W:ymm, ymm, ymm/m256"                     , "RVM"  , "VEX.NDS.256.66.0F.WIG F9 /r"      , "ANY AVX2"],
    ["vptest"           , "R:xmm, xmm/m128"                          , "RM"   , "VEX.128.66.0F38.WIG 17 /r"        , "ANY AVX OF=0 SF=0 ZF=W AF=0 PF=0 CF=W"],
    ["vptest"           , "R:ymm, ymm/m256"                          , "RM"   , "VEX.256.66.0F38.WIG 17 /r"        , "ANY AVX OF=0 SF=0 ZF=W AF=0 PF=0 CF=W"],
    ["vpunpckhbw"       , "W:xmm, xmm, xmm/m128"                     , "RVM"  , "VEX.NDS.128.66.0F.WIG 68 /r"      , "ANY AVX"],
    ["vpunpckhbw"       , "W:ymm, ymm, ymm/m256"                     , "RVM"  , "VEX.NDS.256.66.0F.WIG 68 /r"      , "ANY AVX2"],
    ["vpunpckhdq"       , "W:xmm, xmm, xmm/m128"                     , "RVM"  , "VEX.NDS.128.66.0F.WIG 6A /r"      , "ANY AVX"],
    ["vpunpckhdq"       , "W:ymm, ymm, ymm/m256"                     , "RVM"  , "VEX.NDS.256.66.0F.WIG 6A /r"      , "ANY AVX2"],
    ["vpunpckhqdq"      , "W:xmm, xmm, xmm/m128"                     , "RVM"  , "VEX.NDS.128.66.0F.WIG 6D /r"      , "ANY AVX"],
    ["vpunpckhqdq"      , "W:ymm, ymm, ymm/m256"                     , "RVM"  , "VEX.NDS.256.66.0F.WIG 6D /r"      , "ANY AVX2"],
    ["vpunpckhwd"       , "W:xmm, xmm, xmm/m128"                     , "RVM"  , "VEX.NDS.128.66.0F.WIG 69 /r"      , "ANY AVX"],
    ["vpunpckhwd"       , "W:ymm, ymm, ymm/m256"                     , "RVM"  , "VEX.NDS.256.66.0F.WIG 69 /r"      , "ANY AVX2"],
    ["vpunpcklbw"       , "W:xmm, xmm, xmm/m128"                     , "RVM"  , "VEX.NDS.128.66.0F.WIG 60 /r"      , "ANY AVX"],
    ["vpunpcklbw"       , "W:ymm, ymm, ymm/m256"                     , "RVM"  , "VEX.NDS.256.66.0F.WIG 60 /r"      , "ANY AVX2"],
    ["vpunpckldq"       , "W:xmm, xmm, xmm/m128"                     , "RVM"  , "VEX.NDS.128.66.0F.WIG 62 /r"      , "ANY AVX"],
    ["vpunpckldq"       , "W:ymm, ymm, ymm/m256"                     , "RVM"  , "VEX.NDS.256.66.0F.WIG 62 /r"      , "ANY AVX2"],
    ["vpunpcklqdq"      , "W:xmm, xmm, xmm/m128"                     , "RVM"  , "VEX.NDS.128.66.0F.WIG 6C /r"      , "ANY AVX"],
    ["vpunpcklqdq"      , "W:ymm, ymm, ymm/m256"                     , "RVM"  , "VEX.NDS.256.66.0F.WIG 6C /r"      , "ANY AVX2"],
    ["vpunpcklwd"       , "W:xmm, xmm, xmm/m128"                     , "RVM"  , "VEX.NDS.128.66.0F.WIG 61 /r"      , "ANY AVX"],
    ["vpunpcklwd"       , "W:ymm, ymm, ymm/m256"                     , "RVM"  , "VEX.NDS.256.66.0F.WIG 61 /r"      , "ANY AVX2"],
    ["vpxor"            , "W:xmm, xmm, xmm/m128"                     , "RVM"  , "VEX.NDS.128.66.0F.WIG EF /r"      , "ANY AVX"],
    ["vpxor"            , "W:ymm, ymm, ymm/m256"                     , "RVM"  , "VEX.NDS.256.66.0F.WIG EF /r"      , "ANY AVX2"],
    ["vrcpps"           , "W:xmm, xmm/m128"                          , "RM"   , "VEX.128.0F.WIG 53 /r"             , "ANY AVX"],
    ["vrcpps"           , "W:ymm, ymm/m256"                          , "RM"   , "VEX.256.0F.WIG 53 /r"             , "ANY AVX"],
    ["vrcpss"           , "W:xmm, xmm, xmm/m32"                      , "RVM"  , "VEX.NDS.LIG.F3.0F.WIG 53 /r"      , "ANY AVX"],
    ["vroundpd"         , "W:xmm, xmm/m128, ib"                      , "RMI"  , "VEX.128.66.0F3A.WIG 09 /r ib"     , "ANY AVX"],
    ["vroundpd"         , "W:ymm, ymm/m256, ib"                      , "RMI"  , "VEX.256.66.0F3A.WIG 09 /r ib"     , "ANY AVX"],
    ["vroundps"         , "W:xmm, xmm/m128, ib"                      , "RMI"  , "VEX.128.66.0F3A.WIG 08 /r ib"     , "ANY AVX"],
    ["vroundps"         , "W:ymm, ymm/m256, ib"                      , "RMI"  , "VEX.256.66.0F3A.WIG 08 /r ib"     , "ANY AVX"],
    ["vroundsd"         , "W:xmm, xmm, xmm/m64, ib"                  , "RVMI" , "VEX.NDS.LIG.66.0F3A.WIG 0B /r ib" , "ANY AVX"],
    ["vroundss"         , "W:xmm, xmm, xmm/m32, ib"                  , "RVMI" , "VEX.NDS.LIG.66.0F3A.WIG 0A /r ib" , "ANY AVX"],
    ["vrsqrtps"         , "W:xmm, xmm/m128"                          , "RM"   , "VEX.128.0F.WIG 52 /r"             , "ANY AVX"],
    ["vrsqrtps"         , "W:ymm, ymm/m256"                          , "RM"   , "VEX.256.0F.WIG 52 /r"             , "ANY AVX"],
    ["vrsqrtss"         , "W:xmm, xmm, xmm/m32"                      , "RVM"  , "VEX.NDS.LIG.F3.0F.WIG 52 /r"      , "ANY AVX"],
    ["vshufpd"          , "W:xmm, xmm, xmm/m128, ib"                 , "RVMI" , "VEX.NDS.128.66.0F.WIG C6 /r ib"   , "ANY AVX"],
    ["vshufpd"          , "W:ymm, ymm, ymm/m256, ib"                 , "RVMI" , "VEX.NDS.256.66.0F.WIG C6 /r ib"   , "ANY AVX"],
    ["vshufps"          , "W:xmm, xmm, xmm/m128, ib"                 , "RVMI" , "VEX.NDS.128.0F.WIG C6 /r ib"      , "ANY AVX"],
    ["vshufps"          , "W:ymm, ymm, ymm/m256, ib"                 , "RVMI" , "VEX.NDS.256.0F.WIG C6 /r ib"      , "ANY AVX"],
    ["vsqrtpd"          , "W:xmm, xmm/m128"                          , "RM"   , "VEX.128.66.0F.WIG 51 /r"          , "ANY AVX"],
    ["vsqrtpd"          , "W:ymm, ymm/m256"                          , "RM"   , "VEX.256.66.0F.WIG 51 /r"          , "ANY AVX"],
    ["vsqrtps"          , "W:xmm, xmm/m128"                          , "RM"   , "VEX.128.0F.WIG 51 /r"             , "ANY AVX"],
    ["vsqrtps"          , "W:ymm, ymm/m256"                          , "RM"   , "VEX.256.0F.WIG 51 /r"             , "ANY AVX"],
    ["vsqrtsd"          , "W:xmm, xmm, xmm/m64"                      , "RVM"  , "VEX.NDS.LIG.F2.0F.WIG 51 /r"      , "ANY AVX"],
    ["vsqrtss"          , "W:xmm, xmm, xmm/m32"                      , "RVM"  , "VEX.NDS.LIG.F3.0F.WIG 51 /r"      , "ANY AVX"],
    ["vstmxcsr"         , "W:m32"                                    , "M"    , "VEX.LZ.0F.WIG AE /3"              , "ANY AVX"],
    ["vsubpd"           , "W:xmm, xmm, xmm/m128"                     , "RVM"  , "VEX.NDS.128.66.0F.WIG 5C /r"      , "ANY AVX"],
    ["vsubpd"           , "W:ymm, ymm, ymm/m256"                     , "RVM"  , "VEX.NDS.256.66.0F.WIG 5C /r"      , "ANY AVX2"],
    ["vsubps"           , "W:xmm, xmm, xmm/m128"                     , "RVM"  , "VEX.NDS.128.0F.WIG 5C /r"         , "ANY AVX"],
    ["vsubps"           , "W:ymm, ymm, ymm/m256"                     , "RVM"  , "VEX.NDS.256.0F.WIG 5C /r"         , "ANY AVX2"],
    ["vsubsd"           , "W:xmm, xmm, xmm/m64"                      , "RVM"  , "VEX.NDS.LIG.F2.0F.WIG 5C /r"      , "ANY AVX"],
    ["vsubss"           , "W:xmm, xmm, xmm/m32"                      , "RVM"  , "VEX.NDS.LIG.F3.0F.WIG 5C /r"      , "ANY AVX"],
    ["vtestpd"          , "R:xmm, xmm/m128"                          , "RM"   , "VEX.128.66.0F38.W0 0F /r"         , "ANY AVX OF=0 SF=0 ZF=W AF=0 PF=0 CF=W"],
    ["vtestpd"          , "R:ymm, ymm/m256"                          , "RM"   , "VEX.256.66.0F38.W0 0F /r"         , "ANY AVX OF=0 SF=0 ZF=W AF=0 PF=0 CF=W"],
    ["vtestps"          , "R:xmm, xmm/m128"                          , "RM"   , "VEX.128.66.0F38.W0 0E /r"         , "ANY AVX OF=0 SF=0 ZF=W AF=0 PF=0 CF=W"],
    ["vtestps"          , "R:ymm, ymm/m256"                          , "RM"   , "VEX.256.66.0F38.W0 0E /r"         , "ANY AVX OF=0 SF=0 ZF=W AF=0 PF=0 CF=W"],
    ["vucomisd"         , "R:xmm, xmm/m64"                           , "RM"   , "VEX.LIG.66.0F.WIG 2E /r"          , "ANY AVX OF=0 SF=0 ZF=W AF=0 PF=W CF=W"],
    ["vucomiss"         , "R:xmm, xmm/m32"                           , "RM"   , "VEX.LIG.0F.WIG 2E /r"             , "ANY AVX OF=0 SF=0 ZF=W AF=0 PF=W CF=W"],
    ["vunpckhpd"        , "W:xmm, xmm, xmm/m128"                     , "RVM"  , "VEX.NDS.128.66.0F.WIG 15 /r"      , "ANY AVX"],
    ["vunpckhpd"        , "W:ymm, ymm, ymm/m256"                     , "RVM"  , "VEX.NDS.256.66.0F.WIG 15 /r"      , "ANY AVX"],
    ["vunpckhps"        , "W:xmm, xmm, xmm/m128"                     , "RVM"  , "VEX.NDS.128.0F.WIG 15 /r"         , "ANY AVX"],
    ["vunpckhps"        , "W:ymm, ymm, ymm/m256"                     , "RVM"  , "VEX.NDS.256.0F.WIG 15 /r"         , "ANY AVX"],
    ["vunpcklpd"        , "W:xmm, xmm, xmm/m128"                     , "RVM"  , "VEX.NDS.128.66.0F.WIG 14 /r"      , "ANY AVX"],
    ["vunpcklpd"        , "W:ymm, ymm, ymm/m256"                     , "RVM"  , "VEX.NDS.256.66.0F.WIG 14 /r"      , "ANY AVX"],
    ["vunpcklps"        , "W:xmm, xmm, xmm/m128"                     , "RVM"  , "VEX.NDS.128.0F.WIG 14 /r"         , "ANY AVX"],
    ["vunpcklps"        , "W:ymm, ymm, ymm/m256"                     , "RVM"  , "VEX.NDS.256.0F.WIG 14 /r"         , "ANY AVX"],
    ["vxorpd"           , "W:xmm, xmm, xmm/m128"                     , "RVM"  , "VEX.NDS.128.66.0F.WIG 57 /r"      , "ANY AVX"],
    ["vxorpd"           , "W:ymm, ymm, ymm/m256"                     , "RVM"  , "VEX.NDS.256.66.0F.WIG 57 /r"      , "ANY AVX"],
    ["vxorps"           , "W:xmm, xmm, xmm/m128"                     , "RVM"  , "VEX.NDS.128.0F.WIG 57 /r"         , "ANY AVX"],
    ["vxorps"           , "W:ymm, ymm, ymm/m256"                     , "RVM"  , "VEX.NDS.256.0F.WIG 57 /r"         , "ANY AVX"],
    ["vzeroall"         , ""                                         , "NONE" , "VEX.256.0F.WIG 77"                , "ANY AVX"],
    ["vzeroupper"       , ""                                         , "NONE" , "VEX.128.0F.WIG 77"                , "ANY AVX"],

    // F16C.
    ["vcvtph2ps"        , "W:xmm, xmm/m64"                           , "RM"   , "VEX.128.66.0F38.W0 13 /r"         , "ANY F16C"],
    ["vcvtph2ps"        , "W:ymm, xmm/m128"                          , "RM"   , "VEX.256.66.0F38.W0 13 /r"         , "ANY F16C"],
    ["vcvtps2ph"        , "W:xmm/m64, xmm, ib"                       , "MRI"  , "VEX.128.66.0F3A.W0 1D /r ib"      , "ANY F16C"],
    ["vcvtps2ph"        , "W:xmm/m128, ymm, ib"                      , "MRI"  , "VEX.256.66.0F3A.W0 1D /r ib"      , "ANY F16C"],

    // FMA.
    ["vfmadd132pd"      , "X:xmm, xmm, xmm/m128"                     , "RVM"  , "VEX.DDS.128.66.0F38.W1 98 /r"     , "ANY FMA"],
    ["vfmadd132pd"      , "X:ymm, ymm, ymm/m256"                     , "RVM"  , "VEX.DDS.256.66.0F38.W1 98 /r"     , "ANY FMA"],
    ["vfmadd132ps"      , "X:xmm, xmm, xmm/m128"                     , "RVM"  , "VEX.DDS.128.66.0F38.W0 98 /r"     , "ANY FMA"],
    ["vfmadd132ps"      , "X:ymm, ymm, ymm/m256"                     , "RVM"  , "VEX.DDS.256.66.0F38.W0 98 /r"     , "ANY FMA"],
    ["vfmadd132sd"      , "X:xmm, xmm, xmm/m64"                      , "RVM"  , "VEX.DDS.LIG.66.0F38.W1 99 /r"     , "ANY FMA"],
    ["vfmadd132ss"      , "X:xmm, xmm, xmm/m32"                      , "RVM"  , "VEX.DDS.LIG.66.0F38.W0 99 /r"     , "ANY FMA"],
    ["vfmadd213pd"      , "X:xmm, xmm, xmm/m128"                     , "RVM"  , "VEX.DDS.128.66.0F38.W1 A8 /r"     , "ANY FMA"],
    ["vfmadd213pd"      , "X:ymm, ymm, ymm/m256"                     , "RVM"  , "VEX.DDS.256.66.0F38.W1 A8 /r"     , "ANY FMA"],
    ["vfmadd213ps"      , "X:xmm, xmm, xmm/m128"                     , "RVM"  , "VEX.DDS.128.66.0F38.W0 A8 /r"     , "ANY FMA"],
    ["vfmadd213ps"      , "X:ymm, ymm, ymm/m256"                     , "RVM"  , "VEX.DDS.256.66.0F38.W0 A8 /r"     , "ANY FMA"],
    ["vfmadd213sd"      , "X:xmm, xmm, xmm/m64"                      , "RVM"  , "VEX.DDS.LIG.66.0F38.W1 A9 /r"     , "ANY FMA"],
    ["vfmadd213ss"      , "X:xmm, xmm, xmm/m32"                      , "RVM"  , "VEX.DDS.LIG.66.0F38.W0 A9 /r"     , "ANY FMA"],
    ["vfmadd231pd"      , "X:xmm, xmm, xmm/m128"                     , "RVM"  , "VEX.DDS.128.66.0F38.W1 B8 /r"     , "ANY FMA"],
    ["vfmadd231pd"      , "X:ymm, ymm, ymm/m256"                     , "RVM"  , "VEX.DDS.256.66.0F38.W1 B8 /r"     , "ANY FMA"],
    ["vfmadd231ps"      , "X:xmm, xmm, xmm/m128"                     , "RVM"  , "VEX.DDS.128.66.0F38.W0 B8 /r"     , "ANY FMA"],
    ["vfmadd231ps"      , "X:ymm, ymm, ymm/m256"                     , "RVM"  , "VEX.DDS.256.66.0F38.W0 B8 /r"     , "ANY FMA"],
    ["vfmadd231sd"      , "X:xmm, xmm, xmm/m64"                      , "RVM"  , "VEX.DDS.LIG.66.0F38.W1 B9 /r"     , "ANY FMA"],
    ["vfmadd231ss"      , "X:xmm, xmm, xmm/m32"                      , "RVM"  , "VEX.DDS.LIG.66.0F38.W0 B9 /r"     , "ANY FMA"],
    ["vfmaddsub132pd"   , "X:xmm, xmm, xmm/m128"                     , "RVM"  , "VEX.DDS.128.66.0F38.W1 96 /r"     , "ANY FMA"],
    ["vfmaddsub132pd"   , "X:ymm, ymm, ymm/m256"                     , "RVM"  , "VEX.DDS.256.66.0F38.W1 96 /r"     , "ANY FMA"],
    ["vfmaddsub132ps"   , "X:xmm, xmm, xmm/m128"                     , "RVM"  , "VEX.DDS.128.66.0F38.W0 96 /r"     , "ANY FMA"],
    ["vfmaddsub132ps"   , "X:ymm, ymm, ymm/m256"                     , "RVM"  , "VEX.DDS.256.66.0F38.W0 96 /r"     , "ANY FMA"],
    ["vfmaddsub213pd"   , "X:xmm, xmm, xmm/m128"                     , "RVM"  , "VEX.DDS.128.66.0F38.W1 A6 /r"     , "ANY FMA"],
    ["vfmaddsub213pd"   , "X:ymm, ymm, ymm/m256"                     , "RVM"  , "VEX.DDS.256.66.0F38.W1 A6 /r"     , "ANY FMA"],
    ["vfmaddsub213ps"   , "X:xmm, xmm, xmm/m128"                     , "RVM"  , "VEX.DDS.128.66.0F38.W0 A6 /r"     , "ANY FMA"],
    ["vfmaddsub213ps"   , "X:ymm, ymm, ymm/m256"                     , "RVM"  , "VEX.DDS.256.66.0F38.W0 A6 /r"     , "ANY FMA"],
    ["vfmaddsub231pd"   , "X:xmm, xmm, xmm/m128"                     , "RVM"  , "VEX.DDS.128.66.0F38.W1 B6 /r"     , "ANY FMA"],
    ["vfmaddsub231pd"   , "X:ymm, ymm, ymm/m256"                     , "RVM"  , "VEX.DDS.256.66.0F38.W1 B6 /r"     , "ANY FMA"],
    ["vfmaddsub231ps"   , "X:xmm, xmm, xmm/m128"                     , "RVM"  , "VEX.DDS.128.66.0F38.W0 B6 /r"     , "ANY FMA"],
    ["vfmaddsub231ps"   , "X:ymm, ymm, ymm/m256"                     , "RVM"  , "VEX.DDS.256.66.0F38.W0 B6 /r"     , "ANY FMA"],
    ["vfmsub132pd"      , "X:xmm, xmm, xmm/m128"                     , "RVM"  , "VEX.DDS.128.66.0F38.W1 9A /r"     , "ANY FMA"],
    ["vfmsub132pd"      , "X:ymm, ymm, ymm/m256"                     , "RVM"  , "VEX.DDS.256.66.0F38.W1 9A /r"     , "ANY FMA"],
    ["vfmsub132ps"      , "X:xmm, xmm, xmm/m128"                     , "RVM"  , "VEX.DDS.128.66.0F38.W0 9A /r"     , "ANY FMA"],
    ["vfmsub132ps"      , "X:ymm, ymm, ymm/m256"                     , "RVM"  , "VEX.DDS.256.66.0F38.W0 9A /r"     , "ANY FMA"],
    ["vfmsub132sd"      , "X:xmm, xmm, xmm/m64"                      , "RVM"  , "VEX.DDS.LIG.66.0F38.W1 9B /r"     , "ANY FMA"],
    ["vfmsub132ss"      , "X:xmm, xmm, xmm/m32"                      , "RVM"  , "VEX.DDS.LIG.66.0F38.W0 9B /r"     , "ANY FMA"],
    ["vfmsub213pd"      , "X:xmm, xmm, xmm/m128"                     , "RVM"  , "VEX.DDS.128.66.0F38.W1 AA /r"     , "ANY FMA"],
    ["vfmsub213pd"      , "X:ymm, ymm, ymm/m256"                     , "RVM"  , "VEX.DDS.256.66.0F38.W1 AA /r"     , "ANY FMA"],
    ["vfmsub213ps"      , "X:xmm, xmm, xmm/m128"                     , "RVM"  , "VEX.DDS.128.66.0F38.W0 AA /r"     , "ANY FMA"],
    ["vfmsub213ps"      , "X:ymm, ymm, ymm/m256"                     , "RVM"  , "VEX.DDS.256.66.0F38.W0 AA /r"     , "ANY FMA"],
    ["vfmsub213sd"      , "X:xmm, xmm, xmm/m64"                      , "RVM"  , "VEX.DDS.LIG.66.0F38.W1 AB /r"     , "ANY FMA"],
    ["vfmsub213ss"      , "X:xmm, xmm, xmm/m32"                      , "RVM"  , "VEX.DDS.LIG.66.0F38.W0 AB /r"     , "ANY FMA"],
    ["vfmsub231pd"      , "X:xmm, xmm, xmm/m128"                     , "RVM"  , "VEX.DDS.128.66.0F38.W1 BA /r"     , "ANY FMA"],
    ["vfmsub231pd"      , "X:ymm, ymm, ymm/m256"                     , "RVM"  , "VEX.DDS.256.66.0F38.W1 BA /r"     , "ANY FMA"],
    ["vfmsub231ps"      , "X:xmm, xmm, xmm/m128"                     , "RVM"  , "VEX.DDS.128.66.0F38.W0 BA /r"     , "ANY FMA"],
    ["vfmsub231ps"      , "X:ymm, ymm, ymm/m256"                     , "RVM"  , "VEX.DDS.256.66.0F38.W0 BA /r"     , "ANY FMA"],
    ["vfmsub231sd"      , "X:xmm, xmm, xmm/m64"                      , "RVM"  , "VEX.DDS.LIG.66.0F38.W1 BB /r"     , "ANY FMA"],
    ["vfmsub231ss"      , "X:xmm, xmm, xmm/m32"                      , "RVM"  , "VEX.DDS.LIG.66.0F38.W0 BB /r"     , "ANY FMA"],
    ["vfmsubadd132pd"   , "X:xmm, xmm, xmm/m128"                     , "RVM"  , "VEX.DDS.128.66.0F38.W1 97 /r"     , "ANY FMA"],
    ["vfmsubadd132pd"   , "X:ymm, ymm, ymm/m256"                     , "RVM"  , "VEX.DDS.256.66.0F38.W1 97 /r"     , "ANY FMA"],
    ["vfmsubadd132ps"   , "X:xmm, xmm, xmm/m128"                     , "RVM"  , "VEX.DDS.128.66.0F38.W0 97 /r"     , "ANY FMA"],
    ["vfmsubadd132ps"   , "X:ymm, ymm, ymm/m256"                     , "RVM"  , "VEX.DDS.256.66.0F38.W0 97 /r"     , "ANY FMA"],
    ["vfmsubadd213pd"   , "X:xmm, xmm, xmm/m128"                     , "RVM"  , "VEX.DDS.128.66.0F38.W1 A7 /r"     , "ANY FMA"],
    ["vfmsubadd213pd"   , "X:ymm, ymm, ymm/m256"                     , "RVM"  , "VEX.DDS.256.66.0F38.W1 A7 /r"     , "ANY FMA"],
    ["vfmsubadd213ps"   , "X:xmm, xmm, xmm/m128"                     , "RVM"  , "VEX.DDS.128.66.0F38.W0 A7 /r"     , "ANY FMA"],
    ["vfmsubadd213ps"   , "X:ymm, ymm, ymm/m256"                     , "RVM"  , "VEX.DDS.256.66.0F38.W0 A7 /r"     , "ANY FMA"],
    ["vfmsubadd231pd"   , "X:xmm, xmm, xmm/m128"                     , "RVM"  , "VEX.DDS.128.66.0F38.W1 B7 /r"     , "ANY FMA"],
    ["vfmsubadd231pd"   , "X:ymm, ymm, ymm/m256"                     , "RVM"  , "VEX.DDS.256.66.0F38.W1 B7 /r"     , "ANY FMA"],
    ["vfmsubadd231ps"   , "X:xmm, xmm, xmm/m128"                     , "RVM"  , "VEX.DDS.128.66.0F38.W0 B7 /r"     , "ANY FMA"],
    ["vfmsubadd231ps"   , "X:ymm, ymm, ymm/m256"                     , "RVM"  , "VEX.DDS.256.66.0F38.W0 B7 /r"     , "ANY FMA"],
    ["vfnmadd132pd"     , "X:xmm, xmm, xmm/m128"                     , "RVM"  , "VEX.DDS.128.66.0F38.W1 9C /r"     , "ANY FMA"],
    ["vfnmadd132pd"     , "X:ymm, ymm, ymm/m256"                     , "RVM"  , "VEX.DDS.256.66.0F38.W1 9C /r"     , "ANY FMA"],
    ["vfnmadd132ps"     , "X:xmm, xmm, xmm/m128"                     , "RVM"  , "VEX.DDS.128.66.0F38.W0 9C /r"     , "ANY FMA"],
    ["vfnmadd132ps"     , "X:ymm, ymm, ymm/m256"                     , "RVM"  , "VEX.DDS.256.66.0F38.W0 9C /r"     , "ANY FMA"],
    ["vfnmadd132sd"     , "X:xmm, xmm, xmm/m64"                      , "RVM"  , "VEX.DDS.LIG.66.0F38.W1 9D /r"     , "ANY FMA"],
    ["vfnmadd132ss"     , "X:xmm, xmm, xmm/m32"                      , "RVM"  , "VEX.DDS.LIG.66.0F38.W0 9D /r"     , "ANY FMA"],
    ["vfnmadd213pd"     , "X:xmm, xmm, xmm/m128"                     , "RVM"  , "VEX.DDS.128.66.0F38.W1 AC /r"     , "ANY FMA"],
    ["vfnmadd213pd"     , "X:ymm, ymm, ymm/m256"                     , "RVM"  , "VEX.DDS.256.66.0F38.W1 AC /r"     , "ANY FMA"],
    ["vfnmadd213ps"     , "X:xmm, xmm, xmm/m128"                     , "RVM"  , "VEX.DDS.128.66.0F38.W0 AC /r"     , "ANY FMA"],
    ["vfnmadd213ps"     , "X:ymm, ymm, ymm/m256"                     , "RVM"  , "VEX.DDS.256.66.0F38.W0 AC /r"     , "ANY FMA"],
    ["vfnmadd213sd"     , "X:xmm, xmm, xmm/m64"                      , "RVM"  , "VEX.DDS.LIG.66.0F38.W1 AD /r"     , "ANY FMA"],
    ["vfnmadd213ss"     , "X:xmm, xmm, xmm/m32"                      , "RVM"  , "VEX.DDS.LIG.66.0F38.W0 AD /r"     , "ANY FMA"],
    ["vfnmadd231pd"     , "X:xmm, xmm, xmm/m128"                     , "RVM"  , "VEX.DDS.128.66.0F38.W1 BC /r"     , "ANY FMA"],
    ["vfnmadd231pd"     , "X:ymm, ymm, ymm/m256"                     , "RVM"  , "VEX.DDS.256.66.0F38.W1 BC /r"     , "ANY FMA"],
    ["vfnmadd231ps"     , "X:xmm, xmm, xmm/m128"                     , "RVM"  , "VEX.DDS.128.66.0F38.W0 BC /r"     , "ANY FMA"],
    ["vfnmadd231ps"     , "X:ymm, ymm, ymm/m256"                     , "RVM"  , "VEX.DDS.256.66.0F38.W0 BC /r"     , "ANY FMA"],
    ["vfnmadd231sd"     , "X:xmm, xmm, xmm/m64"                      , "RVM"  , "VEX.DDS.LIG.66.0F38.W1 BD /r"     , "ANY FMA"],
    ["vfnmadd231ss"     , "X:xmm, xmm, xmm/m32"                      , "RVM"  , "VEX.DDS.LIG.66.0F38.W0 BD /r"     , "ANY FMA"],
    ["vfnmsub132pd"     , "X:xmm, xmm, xmm/m128"                     , "RVM"  , "VEX.DDS.128.66.0F38.W1 9E /r"     , "ANY FMA"],
    ["vfnmsub132pd"     , "X:ymm, ymm, ymm/m256"                     , "RVM"  , "VEX.DDS.256.66.0F38.W1 9E /r"     , "ANY FMA"],
    ["vfnmsub132ps"     , "X:xmm, xmm, xmm/m128"                     , "RVM"  , "VEX.DDS.128.66.0F38.W0 9E /r"     , "ANY FMA"],
    ["vfnmsub132ps"     , "X:ymm, ymm, ymm/m256"                     , "RVM"  , "VEX.DDS.256.66.0F38.W0 9E /r"     , "ANY FMA"],
    ["vfnmsub132sd"     , "X:xmm, xmm, xmm/m64"                      , "RVM"  , "VEX.DDS.LIG.66.0F38.W1 9F /r"     , "ANY FMA"],
    ["vfnmsub132ss"     , "X:xmm, xmm, xmm/m32"                      , "RVM"  , "VEX.DDS.LIG.66.0F38.W0 9F /r"     , "ANY FMA"],
    ["vfnmsub213pd"     , "X:xmm, xmm, xmm/m128"                     , "RVM"  , "VEX.DDS.128.66.0F38.W1 AE /r"     , "ANY FMA"],
    ["vfnmsub213pd"     , "X:ymm, ymm, ymm/m256"                     , "RVM"  , "VEX.DDS.256.66.0F38.W1 AE /r"     , "ANY FMA"],
    ["vfnmsub213ps"     , "X:xmm, xmm, xmm/m128"                     , "RVM"  , "VEX.DDS.128.66.0F38.W0 AE /r"     , "ANY FMA"],
    ["vfnmsub213ps"     , "X:ymm, ymm, ymm/m256"                     , "RVM"  , "VEX.DDS.256.66.0F38.W0 AE /r"     , "ANY FMA"],
    ["vfnmsub213sd"     , "X:xmm, xmm, xmm/m64"                      , "RVM"  , "VEX.DDS.LIG.66.0F38.W1 AF /r"     , "ANY FMA"],
    ["vfnmsub213ss"     , "X:xmm, xmm, xmm/m32"                      , "RVM"  , "VEX.DDS.LIG.66.0F38.W0 AF /r"     , "ANY FMA"],
    ["vfnmsub231pd"     , "X:xmm, xmm, xmm/m128"                     , "RVM"  , "VEX.DDS.128.66.0F38.W1 BE /r"     , "ANY FMA"],
    ["vfnmsub231pd"     , "X:ymm, ymm, ymm/m256"                     , "RVM"  , "VEX.DDS.256.66.0F38.W1 BE /r"     , "ANY FMA"],
    ["vfnmsub231ps"     , "X:xmm, xmm, xmm/m128"                     , "RVM"  , "VEX.DDS.128.66.0F38.W0 BE /r"     , "ANY FMA"],
    ["vfnmsub231ps"     , "X:ymm, ymm, ymm/m256"                     , "RVM"  , "VEX.DDS.256.66.0F38.W0 BE /r"     , "ANY FMA"],
    ["vfnmsub231sd"     , "X:xmm, xmm, xmm/m64"                      , "RVM"  , "VEX.DDS.LIG.66.0F38.W1 BF /r"     , "ANY FMA"],
    ["vfnmsub231ss"     , "X:xmm, xmm, xmm/m32"                      , "RVM"  , "VEX.DDS.LIG.66.0F38.W0 BF /r"     , "ANY FMA"],

    // FMA4.
    ["vfmaddpd"         , "W:xmm, xmm, xmm, xmm/m128"                , "RVSM" , "VEX.NDS.128.66.0F3A.W1 69 /r /is4", "ANY FMA4"],
    ["vfmaddpd"         , "W:xmm, xmm, xmm/m128, xmm"                , "RVMS" , "VEX.NDS.128.66.0F3A.W0 69 /r /is4", "ANY FMA4"],
    ["vfmaddpd"         , "W:ymm, ymm, ymm, ymm/m256"                , "RVSM" , "VEX.NDS.256.66.0F3A.W1 69 /r /is4", "ANY FMA4"],
    ["vfmaddpd"         , "W:ymm, ymm, ymm/m256, ymm"                , "RVMS" , "VEX.NDS.256.66.0F3A.W0 69 /r /is4", "ANY FMA4"],
    ["vfmaddps"         , "W:xmm, xmm, xmm, xmm/m128"                , "RVSM" , "VEX.NDS.128.66.0F3A.W1 68 /r /is4", "ANY FMA4"],
    ["vfmaddps"         , "W:xmm, xmm, xmm/m128, xmm"                , "RVMS" , "VEX.NDS.128.66.0F3A.W0 68 /r /is4", "ANY FMA4"],
    ["vfmaddps"         , "W:ymm, ymm, ymm, ymm/m256"                , "RVSM" , "VEX.NDS.256.66.0F3A.W1 68 /r /is4", "ANY FMA4"],
    ["vfmaddps"         , "W:ymm, ymm, ymm/m256, ymm"                , "RVMS" , "VEX.NDS.256.66.0F3A.W0 68 /r /is4", "ANY FMA4"],
    ["vfmaddsd"         , "W:xmm, xmm, xmm, xmm/m64"                 , "RVSM" , "VEX.NDS.128.66.0F3A.W1 6b /r /is4", "ANY FMA4"],
    ["vfmaddsd"         , "W:xmm, xmm, xmm/m64, xmm"                 , "RVMS" , "VEX.NDS.128.66.0F3A.W0 6b /r /is4", "ANY FMA4"],
    ["vfmaddss"         , "W:xmm, xmm, xmm, xmm/m32"                 , "RVSM" , "VEX.NDS.128.66.0F3A.W1 6a /r /is4", "ANY FMA4"],
    ["vfmaddss"         , "W:xmm, xmm, xmm/m32, xmm"                 , "RVMS" , "VEX.NDS.128.66.0F3A.W0 6a /r /is4", "ANY FMA4"],
    ["vfmaddsubpd"      , "W:xmm, xmm, xmm, xmm/m128"                , "RVSM" , "VEX.NDS.128.66.0F3A.W1 5D /r /is4", "ANY FMA4"],
    ["vfmaddsubpd"      , "W:xmm, xmm, xmm/m128, xmm"                , "RVMS" , "VEX.NDS.128.66.0F3A.W0 5D /r /is4", "ANY FMA4"],
    ["vfmaddsubpd"      , "W:ymm, ymm, ymm, ymm/m256"                , "RVSM" , "VEX.NDS.256.66.0F3A.W1 5D /r /is4", "ANY FMA4"],
    ["vfmaddsubpd"      , "W:ymm, ymm, ymm/m256, ymm"                , "RVMS" , "VEX.NDS.256.66.0F3A.W0 5D /r /is4", "ANY FMA4"],
    ["vfmaddsubps"      , "W:xmm, xmm, xmm, xmm/m128"                , "RVSM" , "VEX.NDS.128.66.0F3A.W1 5C /r /is4", "ANY FMA4"],
    ["vfmaddsubps"      , "W:xmm, xmm, xmm/m128, xmm"                , "RVMS" , "VEX.NDS.128.66.0F3A.W0 5C /r /is4", "ANY FMA4"],
    ["vfmaddsubps"      , "W:ymm, ymm, ymm, ymm/m256"                , "RVSM" , "VEX.NDS.256.66.0F3A.W1 5C /r /is4", "ANY FMA4"],
    ["vfmaddsubps"      , "W:ymm, ymm, ymm/m256, ymm"                , "RVMS" , "VEX.NDS.256.66.0F3A.W0 5C /r /is4", "ANY FMA4"],
    ["vfmsubaddpd"      , "W:xmm, xmm, xmm, xmm/m128"                , "RVSM" , "VEX.NDS.128.66.0F3A.W1 5F /r /is4", "ANY FMA4"],
    ["vfmsubaddpd"      , "W:xmm, xmm, xmm/m128, xmm"                , "RVMS" , "VEX.NDS.128.66.0F3A.W0 5F /r /is4", "ANY FMA4"],
    ["vfmsubaddpd"      , "W:ymm, ymm, ymm, ymm/m256"                , "RVSM" , "VEX.NDS.256.66.0F3A.W1 5F /r /is4", "ANY FMA4"],
    ["vfmsubaddpd"      , "W:ymm, ymm, ymm/m256, ymm"                , "RVMS" , "VEX.NDS.256.66.0F3A.W0 5F /r /is4", "ANY FMA4"],
    ["vfmsubaddps"      , "W:xmm, xmm, xmm, xmm/m128"                , "RVSM" , "VEX.NDS.128.66.0F3A.W1 5E /r /is4", "ANY FMA4"],
    ["vfmsubaddps"      , "W:xmm, xmm, xmm/m128, xmm"                , "RVMS" , "VEX.NDS.128.66.0F3A.W0 5E /r /is4", "ANY FMA4"],
    ["vfmsubaddps"      , "W:ymm, ymm, ymm, ymm/m256"                , "RVSM" , "VEX.NDS.256.66.0F3A.W1 5E /r /is4", "ANY FMA4"],
    ["vfmsubaddps"      , "W:ymm, ymm, ymm/m256, ymm"                , "RVMS" , "VEX.NDS.256.66.0F3A.W0 5E /r /is4", "ANY FMA4"],
    ["vfmsubpd"         , "W:xmm, xmm, xmm, xmm/m128"                , "RVSM" , "VEX.NDS.128.66.0F3A.W1 6D /r /is4", "ANY FMA4"],
    ["vfmsubpd"         , "W:xmm, xmm, xmm/m128, xmm"                , "RVMS" , "VEX.NDS.128.66.0F3A.W0 6D /r /is4", "ANY FMA4"],
    ["vfmsubpd"         , "W:ymm, ymm, ymm, ymm/m256"                , "RVSM" , "VEX.NDS.256.66.0F3A.W1 6D /r /is4", "ANY FMA4"],
    ["vfmsubpd"         , "W:ymm, ymm, ymm/m256, ymm"                , "RVMS" , "VEX.NDS.256.66.0F3A.W0 6D /r /is4", "ANY FMA4"],
    ["vfmsubps"         , "W:xmm, xmm, xmm, xmm/m128"                , "RVSM" , "VEX.NDS.128.66.0F3A.W1 6C /r /is4", "ANY FMA4"],
    ["vfmsubps"         , "W:xmm, xmm, xmm/m128, xmm"                , "RVMS" , "VEX.NDS.128.66.0F3A.W0 6C /r /is4", "ANY FMA4"],
    ["vfmsubps"         , "W:ymm, ymm, ymm, ymm/m256"                , "RVSM" , "VEX.NDS.256.66.0F3A.W1 6C /r /is4", "ANY FMA4"],
    ["vfmsubps"         , "W:ymm, ymm, ymm/m256, ymm"                , "RVMS" , "VEX.NDS.256.66.0F3A.W0 6C /r /is4", "ANY FMA4"],
    ["vfmsubsd"         , "W:xmm, xmm, xmm, xmm/m64"                 , "RVSM" , "VEX.NDS.128.66.0F3A.W1 6F /r /is4", "ANY FMA4"],
    ["vfmsubsd"         , "W:xmm, xmm, xmm/m64, xmm"                 , "RVMS" , "VEX.NDS.128.66.0F3A.W0 6F /r /is4", "ANY FMA4"],
    ["vfmsubss"         , "W:xmm, xmm, xmm, xmm/m32"                 , "RVSM" , "VEX.NDS.128.66.0F3A.W1 6E /r /is4", "ANY FMA4"],
    ["vfmsubss"         , "W:xmm, xmm, xmm/m32, xmm"                 , "RVMS" , "VEX.NDS.128.66.0F3A.W0 6E /r /is4", "ANY FMA4"],
    ["vfnmaddpd"        , "W:xmm, xmm, xmm, xmm/m128"                , "RVSM" , "VEX.NDS.128.66.0F3A.W1 79 /r /is4", "ANY FMA4"],
    ["vfnmaddpd"        , "W:xmm, xmm, xmm/m128, xmm"                , "RVMS" , "VEX.NDS.128.66.0F3A.W0 79 /r /is4", "ANY FMA4"],
    ["vfnmaddpd"        , "W:ymm, ymm, ymm, ymm/m256"                , "RVSM" , "VEX.NDS.256.66.0F3A.W1 79 /r /is4", "ANY FMA4"],
    ["vfnmaddpd"        , "W:ymm, ymm, ymm/m256, ymm"                , "RVMS" , "VEX.NDS.256.66.0F3A.W0 79 /r /is4", "ANY FMA4"],
    ["vfnmaddps"        , "W:xmm, xmm, xmm, xmm/m128"                , "RVSM" , "VEX.NDS.128.66.0F3A.W1 78 /r /is4", "ANY FMA4"],
    ["vfnmaddps"        , "W:xmm, xmm, xmm/m128, xmm"                , "RVMS" , "VEX.NDS.128.66.0F3A.W0 78 /r /is4", "ANY FMA4"],
    ["vfnmaddps"        , "W:ymm, ymm, ymm, ymm/m256"                , "RVSM" , "VEX.NDS.256.66.0F3A.W1 78 /r /is4", "ANY FMA4"],
    ["vfnmaddps"        , "W:ymm, ymm, ymm/m256, ymm"                , "RVMS" , "VEX.NDS.256.66.0F3A.W0 78 /r /is4", "ANY FMA4"],
    ["vfnmaddsd"        , "W:xmm, xmm, xmm, xmm/m64"                 , "RVSM" , "VEX.NDS.128.66.0F3A.W1 7B /r /is4", "ANY FMA4"],
    ["vfnmaddsd"        , "W:xmm, xmm, xmm/m64, xmm"                 , "RVMS" , "VEX.NDS.128.66.0F3A.W0 7B /r /is4", "ANY FMA4"],
    ["vfnmaddss"        , "W:xmm, xmm, xmm, xmm/m32"                 , "RVSM" , "VEX.NDS.128.66.0F3A.W1 7A /r /is4", "ANY FMA4"],
    ["vfnmaddss"        , "W:xmm, xmm, xmm/m32, xmm"                 , "RVMS" , "VEX.NDS.128.66.0F3A.W0 7A /r /is4", "ANY FMA4"],
    ["vfnmsubpd"        , "W:xmm, xmm, xmm, xmm/m128"                , "RVSM" , "VEX.NDS.128.66.0F3A.W1 7D /r /is4", "ANY FMA4"],
    ["vfnmsubpd"        , "W:xmm, xmm, xmm/m128, xmm"                , "RVMS" , "VEX.NDS.128.66.0F3A.W0 7D /r /is4", "ANY FMA4"],
    ["vfnmsubpd"        , "W:ymm, ymm, ymm, ymm/m256"                , "RVSM" , "VEX.NDS.256.66.0F3A.W1 7D /r /is4", "ANY FMA4"],
    ["vfnmsubpd"        , "W:ymm, ymm, ymm/m256, ymm"                , "RVMS" , "VEX.NDS.256.66.0F3A.W0 7D /r /is4", "ANY FMA4"],
    ["vfnmsubps"        , "W:xmm, xmm, xmm, xmm/m128"                , "RVSM" , "VEX.NDS.128.66.0F3A.W1 7C /r /is4", "ANY FMA4"],
    ["vfnmsubps"        , "W:xmm, xmm, xmm/m128, xmm"                , "RVMS" , "VEX.NDS.128.66.0F3A.W0 7C /r /is4", "ANY FMA4"],
    ["vfnmsubps"        , "W:ymm, ymm, ymm, ymm/m256"                , "RVSM" , "VEX.NDS.256.66.0F3A.W1 7C /r /is4", "ANY FMA4"],
    ["vfnmsubps"        , "W:ymm, ymm, ymm/m256, ymm"                , "RVMS" , "VEX.NDS.256.66.0F3A.W0 7C /r /is4", "ANY FMA4"],
    ["vfnmsubsd"        , "W:xmm, xmm, xmm, xmm/m64"                 , "RVSM" , "VEX.NDS.128.66.0F3A.W1 7F /r /is4", "ANY FMA4"],
    ["vfnmsubsd"        , "W:xmm, xmm, xmm/m64, xmm"                 , "RVMS" , "VEX.NDS.128.66.0F3A.W0 7F /r /is4", "ANY FMA4"],
    ["vfnmsubss"        , "W:xmm, xmm, xmm, xmm/m32"                 , "RVSM" , "VEX.NDS.128.66.0F3A.W1 7E /r /is4", "ANY FMA4"],
    ["vfnmsubss"        , "W:xmm, xmm, xmm/m32, xmm"                 , "RVMS" , "VEX.NDS.128.66.0F3A.W0 7E /r /is4", "ANY FMA4"],

    // XOP.
    ["vfrczpd"          , "W:xmm, xmm/m128"                          , "RM"   , "XOP.L0.P0.M9.W0 81 /r"            , "ANY XOP"],
    ["vfrczpd"          , "W:ymm, ymm/m256"                          , "RM"   , "XOP.L1.P0.M9.W0 81 /r"            , "ANY XOP"],
    ["vfrczps"          , "W:xmm, xmm/m128"                          , "RM"   , "XOP.L0.P0.M9.W0 80 /r"            , "ANY XOP"],
    ["vfrczps"          , "W:ymm, ymm/m256"                          , "RM"   , "XOP.L1.P0.M9.W0 80 /r"            , "ANY XOP"],
    ["vfrczsd"          , "W:xmm, xmm/m64"                           , "RM"   , "XOP.L0.P0.M9.W0 83 /r"            , "ANY XOP"],
    ["vfrczss"          , "W:xmm, xmm/m32"                           , "RM"   , "XOP.L0.P0.M9.W0 82 /r"            , "ANY XOP"],
    ["vpcmov"           , "W:xmm, xmm, xmm, xmm/m128"                , "RVSM" , "XOP.NDS.L0.P0.M8.W1 A2 /r /is4"   , "ANY XOP"],
    ["vpcmov"           , "W:xmm, xmm, xmm/m128, xmm"                , "RVMS" , "XOP.NDS.L0.P0.M8.W0 A2 /r /is4"   , "ANY XOP"],
    ["vpcmov"           , "W:ymm, ymm, ymm, ymm/m256"                , "RVSM" , "XOP.NDS.L1.P0.M8.W1 A2 /r /is4"   , "ANY XOP"],
    ["vpcmov"           , "W:ymm, ymm, ymm/m256, ymm"                , "RVMS" , "XOP.NDS.L1.P0.M8.W0 A2 /r /is4"   , "ANY XOP"],
    ["vpcomb"           , "W:xmm, xmm, xmm/m128, ib"                 , "RVMI" , "XOP.NDS.L0.P0.M8.W0 CC /r ib"     , "ANY XOP"],
    ["vpcomd"           , "W:xmm, xmm, xmm/m128, ib"                 , "RVMI" , "XOP.NDS.L0.P0.M8.W0 CE /r ib"     , "ANY XOP"],
    ["vpcomq"           , "W:xmm, xmm, xmm/m128, ib"                 , "RVMI" , "XOP.NDS.L0.P0.M8.W0 CF /r ib"     , "ANY XOP"],
    ["vpcomub"          , "W:xmm, xmm, xmm/m128, ib"                 , "RVMI" , "XOP.NDS.L0.P0.M8.W0 EC /r ib"     , "ANY XOP"],
    ["vpcomud"          , "W:xmm, xmm, xmm/m128, ib"                 , "RVMI" , "XOP.NDS.L0.P0.M8.W0 EE /r ib"     , "ANY XOP"],
    ["vpcomuq"          , "W:xmm, xmm, xmm/m128, ib"                 , "RVMI" , "XOP.NDS.L0.P0.M8.W0 EF /r ib"     , "ANY XOP"],
    ["vpcomuw"          , "W:xmm, xmm, xmm/m128, ib"                 , "RVMI" , "XOP.NDS.L0.P0.M8.W0 ED /r ib"     , "ANY XOP"],
    ["vpcomw"           , "W:xmm, xmm, xmm/m128, ib"                 , "RVMI" , "XOP.NDS.L0.P0.M8.W0 CD /r ib"     , "ANY XOP"],
    ["vpermil2pd"       , "W:xmm1, xmm2, xmm3/m128, xmm4, i4"        , "RVMSI", "VEX.L0.66.0F3A.W0 49 /r /is4"     , "ANY XOP"],
    ["vpermil2pd"       , "W:xmm1, xmm2, xmm3, xmm4/m128, i4"        , "RVSMI", "VEX.L0.66.0F3A.W1 49 /r /is4"     , "ANY XOP"],
    ["vpermil2pd"       , "W:ymm1, ymm2, ymm3/m256, ymm4, i4"        , "RVMSI", "VEX.L1.66.0F3A.W0 49 /r /is4"     , "ANY XOP"],
    ["vpermil2pd"       , "W:ymm1, ymm2, ymm3, ymm4/m256, i4"        , "RVSMI", "VEX.L1.66.0F3A.W1 49 /r /is4"     , "ANY XOP"],
    ["vpermil2ps"       , "W:xmm1, xmm2, xmm3/m128, xmm4, i4"        , "RVMSI", "VEX.L0.66.0F3A.W0 48 /r /is4"     , "ANY XOP"],
    ["vpermil2ps"       , "W:xmm1, xmm2, xmm3, xmm4/m128, i4"        , "RVSMI", "VEX.L0.66.0F3A.W1 48 /r /is4"     , "ANY XOP"],
    ["vpermil2ps"       , "W:ymm1, ymm2, ymm3/m256, ymm4, i4"        , "RVMSI", "VEX.L1.66.0F3A.W0 48 /r /is4"     , "ANY XOP"],
    ["vpermil2ps"       , "W:ymm1, ymm2, ymm3, ymm4/m256, i4"        , "RVSMI", "VEX.L1.66.0F3A.W1 48 /r /is4"     , "ANY XOP"],
    ["vphaddbd"         , "W:xmm, xmm/m128"                          , "RM"   , "XOP.L0.P0.M9.W0 C2 /r"            , "ANY XOP"],
    ["vphaddbq"         , "W:xmm, xmm/m128"                          , "RM"   , "XOP.L0.P0.M9.W0 C3 /r"            , "ANY XOP"],
    ["vphaddbw"         , "W:xmm, xmm/m128"                          , "RM"   , "XOP.L0.P0.M9.W0 C1 /r"            , "ANY XOP"],
    ["vphadddq"         , "W:xmm, xmm/m128"                          , "RM"   , "XOP.L0.P0.M9.W0 CB /r"            , "ANY XOP"],
    ["vphaddubd"        , "W:xmm, xmm/m128"                          , "RM"   , "XOP.L0.P0.M9.W0 D2 /r"            , "ANY XOP"],
    ["vphaddubq"        , "W:xmm, xmm/m128"                          , "RM"   , "XOP.L0.P0.M9.W0 D3 /r"            , "ANY XOP"],
    ["vphaddubw"        , "W:xmm, xmm/m128"                          , "RM"   , "XOP.L0.P0.M9.W0 D1 /r"            , "ANY XOP"],
    ["vphaddudq"        , "W:xmm, xmm/m128"                          , "RM"   , "XOP.L0.P0.M9.W0 DB /r"            , "ANY XOP"],
    ["vphadduwd"        , "W:xmm, xmm/m128"                          , "RM"   , "XOP.L0.P0.M9.W0 D6 /r"            , "ANY XOP"],
    ["vphadduwq"        , "W:xmm, xmm/m128"                          , "RM"   , "XOP.L0.P0.M9.W0 D7 /r"            , "ANY XOP"],
    ["vphaddwd"         , "W:xmm, xmm/m128"                          , "RM"   , "XOP.L0.P0.M9.W0 C6 /r"            , "ANY XOP"],
    ["vphaddwq"         , "W:xmm, xmm/m128"                          , "RM"   , "XOP.L0.P0.M9.W0 C7 /r"            , "ANY XOP"],
    ["vphsubbw"         , "W:xmm, xmm/m128"                          , "RM"   , "XOP.L0.P0.M9.W0 E1 /r"            , "ANY XOP"],
    ["vphsubdq"         , "W:xmm, xmm/m128"                          , "RM"   , "XOP.L0.P0.M9.W0 E3 /r"            , "ANY XOP"],
    ["vphsubwd"         , "W:xmm, xmm/m128"                          , "RM"   , "XOP.L0.P0.M9.W0 E2 /r"            , "ANY XOP"],
    ["vpmacsdd"         , "W:xmm, xmm, xmm/m128, xmm"                , "RVMS" , "XOP.NDS.L0.P0.M8.W0 9E /r /is4"   , "ANY XOP"],
    ["vpmacsdqh"        , "W:xmm, xmm, xmm/m128, xmm"                , "RVMS" , "XOP.NDS.L0.P0.M8.W0 9F /r /is4"   , "ANY XOP"],
    ["vpmacsdql"        , "W:xmm, xmm, xmm/m128, xmm"                , "RVMS" , "XOP.NDS.L0.P0.M8.W0 97 /r /is4"   , "ANY XOP"],
    ["vpmacssdd"        , "W:xmm, xmm, xmm/m128, xmm"                , "RVMS" , "XOP.NDS.L0.P0.M8.W0 8E /r /is4"   , "ANY XOP"],
    ["vpmacssdqh"       , "W:xmm, xmm, xmm/m128, xmm"                , "RVMS" , "XOP.NDS.L0.P0.M8.W0 8F /r /is4"   , "ANY XOP"],
    ["vpmacssdql"       , "W:xmm, xmm, xmm/m128, xmm"                , "RVMS" , "XOP.NDS.L0.P0.M8.W0 87 /r /is4"   , "ANY XOP"],
    ["vpmacsswd"        , "W:xmm, xmm, xmm/m128, xmm"                , "RVMS" , "XOP.NDS.L0.P0.M8.W0 86 /r /is4"   , "ANY XOP"],
    ["vpmacssww"        , "W:xmm, xmm, xmm/m128, xmm"                , "RVMS" , "XOP.NDS.L0.P0.M8.W0 85 /r /is4"   , "ANY XOP"],
    ["vpmacswd"         , "W:xmm, xmm, xmm/m128, xmm"                , "RVMS" , "XOP.NDS.L0.P0.M8.W0 96 /r /is4"   , "ANY XOP"],
    ["vpmacsww"         , "W:xmm, xmm, xmm/m128, xmm"                , "RVMS" , "XOP.NDS.L0.P0.M8.W0 95 /r /is4"   , "ANY XOP"],
    ["vpmadcsswd"       , "W:xmm, xmm, xmm/m128, xmm"                , "RVMS" , "XOP.NDS.L0.P0.M8.W0 A6 /r /is4"   , "ANY XOP"],
    ["vpmadcswd"        , "W:xmm, xmm, xmm/m128, xmm"                , "RVMS" , "XOP.NDS.L0.P0.M8.W0 B6 /r /is4"   , "ANY XOP"],
    ["vpperm"           , "W:xmm, xmm, xmm, xmm/m128"                , "RVSM" , "XOP.NDS.L0.P0.M8.W1 A3 /r /is4"   , "ANY XOP"],
    ["vpperm"           , "W:xmm, xmm, xmm/m128, xmm"                , "RVMS" , "XOP.NDS.L0.P0.M8.W0 A3 /r /is4"   , "ANY XOP"],
    ["vprotb"           , "W:xmm, xmm, xmm/m128"                     , "RVM"  , "XOP.NDS.L0.P0.M9.W1 90 /r"        , "ANY XOP"],
    ["vprotb"           , "W:xmm, xmm/m128, ib"                      , "RMI"  , "XOP.L0.P0.M8.W0 C0 /r ib"         , "ANY XOP"],
    ["vprotb"           , "W:xmm, xmm/m128, xmm"                     , "RMV"  , "XOP.NDS.L0.P0.M9.W0 90 /r"        , "ANY XOP"],
    ["vprotd"           , "W:xmm, xmm, xmm/m128"                     , "RVM"  , "XOP.NDS.L0.P0.M9.W1 92 /r"        , "ANY XOP"],
    ["vprotd"           , "W:xmm, xmm/m128, ib"                      , "RMI"  , "XOP.L0.P0.M8.W0 C2 /r ib"         , "ANY XOP"],
    ["vprotd"           , "W:xmm, xmm/m128, xmm"                     , "RMV"  , "XOP.NDS.L0.P0.M9.W0 92 /r"        , "ANY XOP"],
    ["vprotq"           , "W:xmm, xmm, xmm/m128"                     , "RVM"  , "XOP.NDS.L0.P0.M9.W1 93 /r"        , "ANY XOP"],
    ["vprotq"           , "W:xmm, xmm/m128, ib"                      , "RMI"  , "XOP.L0.P0.M8.W0 C3 /r ib"         , "ANY XOP"],
    ["vprotq"           , "W:xmm, xmm/m128, xmm"                     , "RMV"  , "XOP.NDS.L0.P0.M9.W0 93 /r"        , "ANY XOP"],
    ["vprotw"           , "W:xmm, xmm, xmm/m128"                     , "RVM"  , "XOP.NDS.L0.P0.M9.W1 91 /r"        , "ANY XOP"],
    ["vprotw"           , "W:xmm, xmm/m128, ib"                      , "RMI"  , "XOP.L0.P0.M8.W0 C1 /r ib"         , "ANY XOP"],
    ["vprotw"           , "W:xmm, xmm/m128, xmm"                     , "RMV"  , "XOP.NDS.L0.P0.M9.W0 91 /r"        , "ANY XOP"],
    ["vpshab"           , "W:xmm, xmm, xmm/m128"                     , "RVM"  , "XOP.NDS.L0.P0.M9.W1 98 /r"        , "ANY XOP"],
    ["vpshab"           , "W:xmm, xmm/m128, xmm"                     , "RMV"  , "XOP.NDS.L0.P0.M9.W0 98 /r"        , "ANY XOP"],
    ["vpshad"           , "W:xmm, xmm, xmm/m128"                     , "RVM"  , "XOP.NDS.L0.P0.M9.W1 9A /r"        , "ANY XOP"],
    ["vpshad"           , "W:xmm, xmm/m128, xmm"                     , "RMV"  , "XOP.NDS.L0.P0.M9.W0 9A /r"        , "ANY XOP"],
    ["vpshaq"           , "W:xmm, xmm, xmm/m128"                     , "RVM"  , "XOP.NDS.L0.P0.M9.W1 9B /r"        , "ANY XOP"],
    ["vpshaq"           , "W:xmm, xmm/m128, xmm"                     , "RMV"  , "XOP.NDS.L0.P0.M9.W0 9B /r"        , "ANY XOP"],
    ["vpshaw"           , "W:xmm, xmm, xmm/m128"                     , "RVM"  , "XOP.NDS.L0.P0.M9.W1 99 /r"        , "ANY XOP"],
    ["vpshaw"           , "W:xmm, xmm/m128, xmm"                     , "RMV"  , "XOP.NDS.L0.P0.M9.W0 99 /r"        , "ANY XOP"],
    ["vpshlb"           , "W:xmm, xmm, xmm/m128"                     , "RVM"  , "XOP.NDS.L0.P0.M9.W1 94 /r"        , "ANY XOP"],
    ["vpshlb"           , "W:xmm, xmm/m128, xmm"                     , "RMV"  , "XOP.NDS.L0.P0.M9.W0 94 /r"        , "ANY XOP"],
    ["vpshld"           , "W:xmm, xmm, xmm/m128"                     , "RVM"  , "XOP.NDS.L0.P0.M9.W1 96 /r"        , "ANY XOP"],
    ["vpshld"           , "W:xmm, xmm/m128, xmm"                     , "RMV"  , "XOP.NDS.L0.P0.M9.W0 96 /r"        , "ANY XOP"],
    ["vpshlq"           , "W:xmm, xmm, xmm/m128"                     , "RVM"  , "XOP.NDS.L0.P0.M9.W1 97 /r"        , "ANY XOP"],
    ["vpshlq"           , "W:xmm, xmm/m128, xmm"                     , "RMV"  , "XOP.NDS.L0.P0.M9.W0 97 /r"        , "ANY XOP"],
    ["vpshlw"           , "W:xmm, xmm, xmm/m128"                     , "RVM"  , "XOP.NDS.L0.P0.M9.W1 95 /r"        , "ANY XOP"],
    ["vpshlw"           , "W:xmm, xmm/m128, xmm"                     , "RMV"  , "XOP.NDS.L0.P0.M9.W0 95 /r"        , "ANY XOP"],

    // AVX512.
    ["kaddb"            , "W:k, k, k"                                , "RVM"  , "VEX.L1.66.0F.W0 4A /r"            , "ANY AVX512DQ"],
    ["kaddd"            , "W:k, k, k"                                , "RVM"  , "VEX.L1.66.0F.W1 4A /r"            , "ANY AVX512BW"],
    ["kaddq"            , "W:k, k, k"                                , "RVM"  , "VEX.L1.0F.W1 4A /r"               , "ANY AVX512BW"],
    ["kaddw"            , "W:k, k, k"                                , "RVM"  , "VEX.L1.0F.W0 4A /r"               , "ANY AVX512DQ"],
    ["kandb"            , "W:k, k, k"                                , "RVM"  , "VEX.L1.66.0F.W0 41 /r"            , "ANY AVX512DQ"],
    ["kandd"            , "W:k, k, k"                                , "RVM"  , "VEX.L1.66.0F.W1 41 /r"            , "ANY AVX512BW"],
    ["kandnb"           , "W:k, k, k"                                , "RVM"  , "VEX.L1.66.0F.W0 42 /r"            , "ANY AVX512DQ"],
    ["kandnd"           , "W:k, k, k"                                , "RVM"  , "VEX.L1.66.0F.W1 42 /r"            , "ANY AVX512BW"],
    ["kandnq"           , "W:k, k, k"                                , "RVM"  , "VEX.L1.0F.W1 42 /r"               , "ANY AVX512BW"],
    ["kandnw"           , "W:k, k, k"                                , "RVM"  , "VEX.NDS.L1.0F.W0 42 /r"           , "ANY AVX512F"],
    ["kandq"            , "W:k, k, k"                                , "RVM"  , "VEX.L1.0F.W1 41 /r"               , "ANY AVX512BW"],
    ["kandw"            , "W:k, k, k"                                , "RVM"  , "VEX.NDS.L1.0F.W0 41 /r"           , "ANY AVX512F"],
    ["kmovb"            , "W:k, k/m8"                                , "RM"   , "VEX.L0.66.0F.W0 90 /r"            , "ANY AVX512DQ"],
    ["kmovb"            , "W:k, r32"                                 , "RM"   , "VEX.L0.66.0F.W0 92 /r"            , "ANY AVX512DQ"],
    ["kmovb"            , "W:m8, k"                                  , "MR"   , "VEX.L0.66.0F.W0 91 /r"            , "ANY AVX512DQ"],
    ["kmovb"            , "W:r32, k"                                 , "RM"   , "VEX.L0.66.0F.W0 93 /r"            , "ANY AVX512DQ"],
    ["kmovd"            , "W:k, k/m32"                               , "RM"   , "VEX.L0.66.0F.W1 90 /r"            , "ANY AVX512BW"],
    ["kmovd"            , "W:k, r32"                                 , "RM"   , "VEX.L0.F2.0F.W0 92 /r"            , "ANY AVX512BW"],
    ["kmovd"            , "W:m32, k"                                 , "MR"   , "VEX.L0.66.0F.W1 91 /r"            , "ANY AVX512BW"],
    ["kmovd"            , "W:r32, k"                                 , "RM"   , "VEX.L0.F2.0F.W0 93 /r"            , "ANY AVX512BW"],
    ["kmovq"            , "W:k, k/m64"                               , "RM"   , "VEX.L0.0F.W1 90 /r"               , "ANY AVX512BW"],
    ["kmovq"            , "W:k, r64"                                 , "RM"   , "VEX.L0.F2.0F.W1 92 /r"            , "X64 AVX512BW"],
    ["kmovq"            , "W:m64, k"                                 , "MR"   , "VEX.L0.0F.W1 91 /r"               , "ANY AVX512BW"],
    ["kmovq"            , "W:r64, k"                                 , "RM"   , "VEX.L0.F2.0F.W1 93 /r"            , "X64 AVX512BW"],
    ["kmovw"            , "W:k, k/m16"                               , "RM"   , "VEX.L0.0F.W0 90 /r"               , "ANY AVX512F"],
    ["kmovw"            , "W:k, r32"                                 , "RM"   , "VEX.L0.0F.W0 92 /r"               , "ANY AVX512F"],
    ["kmovw"            , "W:m16, k"                                 , "MR"   , "VEX.L0.0F.W0 91 /r"               , "ANY AVX512F"],
    ["kmovw"            , "W:r32, k"                                 , "RM"   , "VEX.L0.0F.W0 93 /r"               , "ANY AVX512F"],
    ["knotb"            , "W:k, k"                                   , "RM"   , "VEX.L0.66.0F.W0 44 /r"            , "ANY AVX512DQ"],
    ["knotd"            , "W:k, k"                                   , "RM"   , "VEX.L0.66.0F.W1 44 /r"            , "ANY AVX512BW"],
    ["knotq"            , "W:k, k"                                   , "RM"   , "VEX.L0.0F.W1 44 /r"               , "ANY AVX512BW"],
    ["knotw"            , "W:k, k"                                   , "RM"   , "VEX.L0.0F.W0 44 /r"               , "ANY AVX512F"],
    ["korb"             , "W:k, k, k"                                , "RVM"  , "VEX.L1.66.0F.W0 45 /r"            , "ANY AVX512DQ"],
    ["kord"             , "W:k, k, k"                                , "RVM"  , "VEX.L1.66.0F.W1 45 /r"            , "ANY AVX512BW"],
    ["korq"             , "W:k, k, k"                                , "RVM"  , "VEX.L1.0F.W1 45 /r"               , "ANY AVX512BW"],
    ["kortestb"         , "R:k, k"                                   , "RM"   , "VEX.L0.66.0F.W0 98 /r"            , "ANY AVX512DQ OF=0 SF=0 ZF=W AF=0 PF=0 CF=W"],
    ["kortestd"         , "R:k, k"                                   , "RM"   , "VEX.L0.66.0F.W1 98 /r"            , "ANY AVX512BW OF=0 SF=0 ZF=W AF=0 PF=0 CF=W"],
    ["kortestq"         , "R:k, k"                                   , "RM"   , "VEX.L0.0F.W1 98 /r"               , "ANY AVX512BW OF=0 SF=0 ZF=W AF=0 PF=0 CF=W"],
    ["kortestw"         , "R:k, k"                                   , "RM"   , "VEX.L0.0F.W0 98 /r"               , "ANY AVX512F  OF=0 SF=0 ZF=W AF=0 PF=0 CF=W"],
    ["korw"             , "W:k, k, k"                                , "RVM"  , "VEX.NDS.L1.0F.W0 45 /r"           , "ANY AVX512F"],
    ["kshiftlb"         , "W:k, k, ib"                               , "RMI"  , "VEX.L0.66.0F3A.W0 32 /r ib"       , "ANY AVX512DQ"],
    ["kshiftld"         , "W:k, k, ib"                               , "RMI"  , "VEX.L0.66.0F3A.W0 33 /r ib"       , "ANY AVX512BW"],
    ["kshiftlq"         , "W:k, k, ib"                               , "RMI"  , "VEX.L0.66.0F3A.W1 33 /r ib"       , "ANY AVX512BW"],
    ["kshiftlw"         , "W:k, k, ib"                               , "RMI"  , "VEX.L0.66.0F3A.W1 32 /r ib"       , "ANY AVX512F"],
    ["kshiftrb"         , "W:k, k, ib"                               , "RMI"  , "VEX.L0.66.0F3A.W0 30 /r ib"       , "ANY AVX512DQ"],
    ["kshiftrd"         , "W:k, k, ib"                               , "RMI"  , "VEX.L0.66.0F3A.W0 31 /r ib"       , "ANY AVX512BW"],
    ["kshiftrq"         , "W:k, k, ib"                               , "RMI"  , "VEX.L0.66.0F3A.W1 31 /r ib"       , "ANY AVX512BW"],
    ["kshiftrw"         , "W:k, k, ib"                               , "RMI"  , "VEX.L0.66.0F3A.W1 30 /r ib"       , "ANY AVX512F"],
    ["ktestb"           , "R:k, k"                                   , "RM"   , "VEX.L0.66.0F.W0 99 /r"            , "ANY AVX512DQ OF=0 SF=0 ZF=W AF=0 PF=0 CF=W"],
    ["ktestd"           , "R:k, k"                                   , "RM"   , "VEX.L0.66.0F.W1 99 /r"            , "ANY AVX512BW OF=0 SF=0 ZF=W AF=0 PF=0 CF=W"],
    ["ktestq"           , "R:k, k"                                   , "RM"   , "VEX.L0.0F.W1 99 /r"               , "ANY AVX512BW OF=0 SF=0 ZF=W AF=0 PF=0 CF=W"],
    ["ktestw"           , "R:k, k"                                   , "RM"   , "VEX.L0.0F.W0 99 /r"               , "ANY AVX512DQ OF=0 SF=0 ZF=W AF=0 PF=0 CF=W"],
    ["kunpckbw"         , "W:k, k, k"                                , "RVM"  , "VEX.NDS.L1.66.0F.W0 4B /r"        , "ANY AVX512F"],
    ["kunpckdq"         , "W:k, k, k"                                , "RVM"  , "VEX.NDS.L1.0F.W1 4B /r"           , "ANY AVX512BW"],
    ["kunpckwd"         , "W:k, k, k"                                , "RVM"  , "VEX.NDS.L1.0F.W0 4B /r"           , "ANY AVX512BW"],
    ["kxnorb"           , "W:k, k, k"                                , "RVM"  , "VEX.L1.66.0F.W0 46 /r"            , "ANY AVX512DQ"],
    ["kxnord"           , "W:k, k, k"                                , "RVM"  , "VEX.L1.66.0F.W1 46 /r"            , "ANY AVX512BW"],
    ["kxnorq"           , "W:k, k, k"                                , "RVM"  , "VEX.L1.0F.W1 46 /r"               , "ANY AVX512BW"],
    ["kxnorw"           , "W:k, k, k"                                , "RVM"  , "VEX.NDS.L1.0F.W0 46 /r"           , "ANY AVX512F"],
    ["kxorb"            , "W:k, k, k"                                , "RVM"  , "VEX.L1.66.0F.W0 47 /r"            , "ANY AVX512DQ"],
    ["kxord"            , "W:k, k, k"                                , "RVM"  , "VEX.L1.66.0F.W1 47 /r"            , "ANY AVX512BW"],
    ["kxorq"            , "W:k, k, k"                                , "RVM"  , "VEX.L1.0F.W1 47 /r"               , "ANY AVX512BW"],
    ["kxorw"            , "W:k, k, k"                                , "RVM"  , "VEX.NDS.L1.0F.W0 47 /r"           , "ANY AVX512F"],

    ["vaddpd"           , "W:xmm {kz}, xmm, xmm/m128/b64"            , "RVM"  , "EVEX.NDS.128.66.0F.W1 58 /r"      , "ANY AVX512F-VL"],
    ["vaddpd"           , "W:ymm {kz}, ymm, ymm/m256/b64"            , "RVM"  , "EVEX.NDS.256.66.0F.W1 58 /r"      , "ANY AVX512F-VL"],
    ["vaddpd"           , "W:zmm {kz}, zmm, zmm/m512/b64 {er}"       , "RVM"  , "EVEX.NDS.512.66.0F.W1 58 /r"      , "ANY AVX512F"],
    ["vaddps"           , "W:xmm {kz}, xmm, xmm/m128/b32"            , "RVM"  , "EVEX.NDS.128.0F.W0 58 /r"         , "ANY AVX512F-VL"],
    ["vaddps"           , "W:ymm {kz}, ymm, ymm/m256/b32"            , "RVM"  , "EVEX.NDS.256.0F.W0 58 /r"         , "ANY AVX512F-VL"],
    ["vaddps"           , "W:zmm {kz}, zmm, zmm/m512/b32 {er}"       , "RVM"  , "EVEX.NDS.512.0F.W0 58 /r"         , "ANY AVX512F"],
    ["vaddsd"           , "W:xmm {kz}, xmm, xmm/m64 {er}"            , "RVM"  , "EVEX.NDS.LIG.F2.0F.W1 58 /r"      , "ANY AVX512F"],
    ["vaddss"           , "W:xmm {kz}, xmm, xmm/m32 {er}"            , "RVM"  , "EVEX.NDS.LIG.F3.0F.W0 58 /r"      , "ANY AVX512F"],
    ["valignd"          , "W:xmm {kz}, xmm, xmm/m128/b32, ib"        , "RVMI" , "EVEX.NDS.128.66.0F3A.W0 03 /r ib" , "ANY AVX512F-VL"],
    ["valignd"          , "W:ymm {kz}, ymm, ymm/m256/b32, ib"        , "RVMI" , "EVEX.NDS.256.66.0F3A.W0 03 /r ib" , "ANY AVX512F-VL"],
    ["valignd"          , "W:zmm {kz}, zmm, zmm/m512/b32, ib"        , "RVMI" , "EVEX.NDS.512.66.0F3A.W0 03 /r ib" , "ANY AVX512F"],
    ["valignq"          , "W:xmm {kz}, xmm, xmm/m128/b64, ib"        , "RVMI" , "EVEX.NDS.128.66.0F3A.W1 03 /r ib" , "ANY AVX512F-VL"],
    ["valignq"          , "W:ymm {kz}, ymm, ymm/m256/b64, ib"        , "RVMI" , "EVEX.NDS.256.66.0F3A.W1 03 /r ib" , "ANY AVX512F-VL"],
    ["valignq"          , "W:zmm {kz}, zmm, zmm/m512/b64, ib"        , "RVMI" , "EVEX.NDS.512.66.0F3A.W1 03 /r ib" , "ANY AVX512F"],
    ["vandnpd"          , "W:xmm {kz}, xmm, xmm/m128/b64"            , "RVM"  , "EVEX.NDS.128.66.0F.W1 55 /r"      , "ANY AVX512DQ-VL"],
    ["vandnpd"          , "W:ymm {kz}, ymm, ymm/m256/b64"            , "RVM"  , "EVEX.NDS.256.66.0F.W1 55 /r"      , "ANY AVX512DQ-VL"],
    ["vandnpd"          , "W:zmm {kz}, zmm, zmm/m512/b64"            , "RVM"  , "EVEX.NDS.512.66.0F.W1 55 /r"      , "ANY AVX512DQ"],
    ["vandnps"          , "W:xmm {kz}, xmm, xmm/m128/b32"            , "RVM"  , "EVEX.NDS.128.66.W0 55 /r"         , "ANY AVX512DQ-VL"],
    ["vandnps"          , "W:ymm {kz}, ymm, ymm/m256/b32"            , "RVM"  , "EVEX.NDS.256.66.W0 55 /r"         , "ANY AVX512DQ-VL"],
    ["vandnps"          , "W:zmm {kz}, zmm, zmm/m512/b32"            , "RVM"  , "EVEX.NDS.512.66.W0 55 /r"         , "ANY AVX512DQ"],
    ["vandpd"           , "W:xmm {kz}, xmm, xmm/m128/b64"            , "RVM"  , "EVEX.NDS.128.66.0F.W1 54 /r"      , "ANY AVX512DQ-VL"],
    ["vandpd"           , "W:ymm {kz}, ymm, ymm/m256/b64"            , "RVM"  , "EVEX.NDS.256.66.0F.W1 54 /r"      , "ANY AVX512DQ-VL"],
    ["vandpd"           , "W:zmm {kz}, zmm, zmm/m512/b64"            , "RVM"  , "EVEX.NDS.512.66.0F.W1 54 /r"      , "ANY AVX512DQ"],
    ["vandps"           , "W:xmm {kz}, xmm, xmm/m128/b32"            , "RVM"  , "EVEX.NDS.128.0F.W0 54 /r"         , "ANY AVX512DQ-VL"],
    ["vandps"           , "W:ymm {kz}, ymm, ymm/m256/b32"            , "RVM"  , "EVEX.NDS.256.0F.W0 54 /r"         , "ANY AVX512DQ-VL"],
    ["vandps"           , "W:zmm {kz}, zmm, zmm/m512/b32"            , "RVM"  , "EVEX.NDS.512.0F.W0 54 /r"         , "ANY AVX512DQ"],
    ["vblendmb"         , "W:xmm {kz}, xmm, xmm/m128"                , "RVM"  , "EVEX.NDS.128.66.0F38.W0 66 /r"    , "ANY AVX512BW-VL"],
    ["vblendmb"         , "W:ymm {kz}, ymm, ymm/m256"                , "RVM"  , "EVEX.NDS.256.66.0F38.W0 66 /r"    , "ANY AVX512BW-VL"],
    ["vblendmb"         , "W:zmm {kz}, zmm, zmm/m512"                , "RVM"  , "EVEX.NDS.512.66.0F38.W0 66 /r"    , "ANY AVX512BW"],
    ["vblendmd"         , "W:xmm {kz}, xmm, xmm/m128/b32"            , "RVM"  , "EVEX.NDS.128.66.0F38.W0 64 /r"    , "ANY AVX512F-VL"],
    ["vblendmd"         , "W:ymm {kz}, ymm, ymm/m256/b32"            , "RVM"  , "EVEX.NDS.256.66.0F38.W0 64 /r"    , "ANY AVX512F-VL"],
    ["vblendmd"         , "W:zmm {kz}, zmm, zmm/m512/b32"            , "RVM"  , "EVEX.NDS.512.66.0F38.W0 64 /r"    , "ANY AVX512F"],
    ["vblendmpd"        , "W:xmm {kz}, xmm, xmm/m128/b64"            , "RVM"  , "EVEX.NDS.128.66.0F38.W1 65 /r"    , "ANY AVX512F-VL"],
    ["vblendmpd"        , "W:ymm {kz}, ymm, ymm/m256/b64"            , "RVM"  , "EVEX.NDS.256.66.0F38.W1 65 /r"    , "ANY AVX512F-VL"],
    ["vblendmpd"        , "W:zmm {kz}, zmm, zmm/m512/b64"            , "RVM"  , "EVEX.NDS.512.66.0F38.W1 65 /r"    , "ANY AVX512F"],
    ["vblendmps"        , "W:xmm {kz}, xmm, xmm/m128/b32"            , "RVM"  , "EVEX.NDS.128.66.0F38.W0 65 /r"    , "ANY AVX512F-VL"],
    ["vblendmps"        , "W:ymm {kz}, ymm, ymm/m256/b32"            , "RVM"  , "EVEX.NDS.256.66.0F38.W0 65 /r"    , "ANY AVX512F-VL"],
    ["vblendmps"        , "W:zmm {kz}, zmm, zmm/m512/b32"            , "RVM"  , "EVEX.NDS.512.66.0F38.W0 65 /r"    , "ANY AVX512F"],
    ["vblendmq"         , "W:xmm {kz}, xmm, xmm/m128/b64"            , "RVM"  , "EVEX.NDS.128.66.0F38.W1 64 /r"    , "ANY AVX512F-VL"],
    ["vblendmq"         , "W:ymm {kz}, ymm, ymm/m256/b64"            , "RVM"  , "EVEX.NDS.256.66.0F38.W1 64 /r"    , "ANY AVX512F-VL"],
    ["vblendmq"         , "W:zmm {kz}, zmm, zmm/m512/b64"            , "RVM"  , "EVEX.NDS.512.66.0F38.W1 64 /r"    , "ANY AVX512F"],
    ["vblendmw"         , "W:xmm {kz}, xmm, xmm/m128"                , "RVM"  , "EVEX.NDS.128.66.0F38.W1 66 /r"    , "ANY AVX512BW-VL"],
    ["vblendmw"         , "W:ymm {kz}, ymm, ymm/m256"                , "RVM"  , "EVEX.NDS.256.66.0F38.W1 66 /r"    , "ANY AVX512BW-VL"],
    ["vblendmw"         , "W:zmm {kz}, zmm, zmm/m512"                , "RVM"  , "EVEX.NDS.512.66.0F38.W1 66 /r"    , "ANY AVX512BW"],
    ["vbroadcastf32x2"  , "W:ymm {kz}, xmm/m64"                      , "RM"   , "EVEX.256.66.0F38.W0 19 /r"        , "ANY AVX512DQ-VL"],
    ["vbroadcastf32x2"  , "W:zmm {kz}, xmm/m64"                      , "RM"   , "EVEX.512.66.0F38.W0 19 /r"        , "ANY AVX512DQ"],
    ["vbroadcastf32x4"  , "W:ymm {kz}, m128"                         , "RM"   , "EVEX.256.66.0F38.W0 1A /r"        , "ANY AVX512F"],
    ["vbroadcastf32x4"  , "W:zmm {kz}, m128"                         , "RM"   , "EVEX.512.66.0F38.W0 1A /r"        , "ANY AVX512F"],
    ["vbroadcastf32x8"  , "W:zmm {kz}, m256"                         , "RM"   , "EVEX.512.66.0F38.W0 1B /r"        , "ANY AVX512DQ"],
    ["vbroadcastf64x2"  , "W:ymm {kz}, m128"                         , "RM"   , "EVEX.256.66.0F38.W1 1A /r"        , "ANY AVX512DQ-VL"],
    ["vbroadcastf64x2"  , "W:zmm {kz}, m128"                         , "RM"   , "EVEX.512.66.0F38.W1 1A /r"        , "ANY AVX512DQ"],
    ["vbroadcastf64x4"  , "W:zmm {kz}, m256"                         , "RM"   , "EVEX.512.66.0F38.W1 1B /r"        , "ANY AVX512F"],
    ["vbroadcasti32x2"  , "W:xmm {kz}, xmm/m64"                      , "RM"   , "EVEX.128.66.0F38.W0 59 /r"        , "ANY AVX512DQ-VL"],
    ["vbroadcasti32x2"  , "W:ymm {kz}, xmm/m64"                      , "RM"   , "EVEX.256.66.0F38.W0 59 /r"        , "ANY AVX512DQ-VL"],
    ["vbroadcasti32x2"  , "W:zmm {kz}, xmm/m64"                      , "RM"   , "EVEX.512.66.0F38.W0 59 /r"        , "ANY AVX512DQ"],
    ["vbroadcasti32x4"  , "W:ymm {kz}, xmm/m64"                      , "RM"   , "EVEX.256.66.0F38.W0 5A /r"        , "ANY AVX512F-VL"],
    ["vbroadcasti32x4"  , "W:zmm {kz}, xmm/m64"                      , "RM"   , "EVEX.512.66.0F38.W0 5A /r"        , "ANY AVX512F"],
    ["vbroadcasti32x8"  , "W:zmm {kz}, xmm/m64"                      , "RM"   , "EVEX.512.66.0F38.W1 5B /r"        , "ANY AVX512DQ"],
    ["vbroadcasti64x2"  , "W:ymm {kz}, xmm/m64"                      , "RM"   , "EVEX.256.66.0F38.W1 5A /r"        , "ANY AVX512DQ-VL"],
    ["vbroadcasti64x2"  , "W:zmm {kz}, xmm/m64"                      , "RM"   , "EVEX.512.66.0F38.W1 5A /r"        , "ANY AVX512DQ"],
    ["vbroadcasti64x4"  , "W:zmm {kz}, xmm/m64"                      , "RM"   , "EVEX.512.66.0F38.W1 5B /r"        , "ANY AVX512F"],
    ["vbroadcastsd"     , "W:ymm {kz}, xmm/m64"                      , "RM"   , "EVEX.256.66.0F38.W1 19 /r"        , "ANY AVX512F-VL"],
    ["vbroadcastsd"     , "W:zmm {kz}, xmm/m64"                      , "RM"   , "EVEX.512.66.0F38.W1 19 /r"        , "ANY AVX512F"],
    ["vbroadcastss"     , "W:xmm {kz}, xmm/m32"                      , "RM"   , "EVEX.128.66.0F38.W0 18 /r"        , "ANY AVX512F-VL"],
    ["vbroadcastss"     , "W:ymm {kz}, xmm/m32"                      , "RM"   , "EVEX.256.66.0F38.W0 18 /r"        , "ANY AVX512F"],
    ["vbroadcastss"     , "W:zmm {kz}, xmm/m32"                      , "RM"   , "EVEX.512.66.0F38.W0 18 /r"        , "ANY AVX512F-VL"],
    ["vcmppd"           , "W:k {kz}, xmm, xmm/m128/b64, ib"          , "RVMI" , "EVEX.NDS.128.66.0F.W1 C2 /r ib"   , "ANY AVX512F-VL"],
    ["vcmppd"           , "W:k {kz}, ymm, ymm/m256/b64, ib"          , "RVMI" , "EVEX.NDS.256.66.0F.W1 C2 /r ib"   , "ANY AVX512F-VL"],
    ["vcmppd"           , "W:k {kz}, zmm, zmm/m512/b64, ib {sae}"    , "RVMI" , "EVEX.NDS.512.66.0F.W1 C2 /r ib"   , "ANY AVX512F"],
    ["vcmpps"           , "W:k {kz}, xmm, xmm/m128/b32, ib"          , "RVMI" , "EVEX.NDS.128.0F.W0 C2 /r ib"      , "ANY AVX512F-VL"],
    ["vcmpps"           , "W:k {kz}, ymm, ymm/m256/b32, ib"          , "RVMI" , "EVEX.NDS.256.0F.W0 C2 /r ib"      , "ANY AVX512F-VL"],
    ["vcmpps"           , "W:k {kz}, zmm, zmm/m512/b32, ib {sae}"    , "RVMI" , "EVEX.NDS.512.0F.W0 C2 /r ib"      , "ANY AVX512F"],
    ["vcmpsd"           , "W:k {kz}, xmm, xmm/m64, ib {sae}"         , "RVMI" , "EVEX.NDS.LIG.F2.0F.W1 C2 /r ib"   , "ANY AVX512F"],
    ["vcmpss"           , "W:k {kz}, xmm, xmm/m32, ib {sae}"         , "RVMI" , "EVEX.NDS.LIG.F3.0F.W0 C2 /r ib"   , "ANY AVX512F"],
    ["vcomisd"          , "R:xmm, xmm/m64 {sae}"                     , "RM"   , "EVEX.LIG.66.0F.W1 2F /r"          , "ANY AVX512F OF=0 SF=0 ZF=W AF=0 PF=W CF=W"],
    ["vcomiss"          , "R:xmm, xmm/m32 {sae}"                     , "RM"   , "EVEX.LIG.0F.W0 2F /r"             , "ANY AVX512F OF=0 SF=0 ZF=W AF=0 PF=W CF=W"],
    ["vcompresspd"      , "W:xmm/m128 {kz}, xmm"                     , "MR"   , "EVEX.128.66.0F38.W1 8A /r"        , "ANY AVX512F-VL"],
    ["vcompresspd"      , "W:ymm/m256 {kz}, ymm"                     , "MR"   , "EVEX.256.66.0F38.W1 8A /r"        , "ANY AVX512F-VL"],
    ["vcompresspd"      , "W:zmm/m512 {kz}, zmm"                     , "MR"   , "EVEX.512.66.0F38.W1 8A /r"        , "ANY AVX512F"],
    ["vcompressps"      , "W:xmm/m128 {kz}, xmm"                     , "RVM"  , "EVEX.128.66.0F38.W0 8A /r"        , "ANY AVX512F-VL"],
    ["vcompressps"      , "W:ymm/m256 {kz}, ymm"                     , "RVM"  , "EVEX.256.66.0F38.W0 8A /r"        , "ANY AVX512F-VL"],
    ["vcompressps"      , "W:zmm/m512 {kz}, zmm"                     , "RVM"  , "EVEX.512.66.0F38.W0 8A /r"        , "ANY AVX512F"],
    ["vcvtdq2pd"        , "W:xmm {kz}, xmm/m64 /b32"                 , "RM"   , "EVEX.128.F3.0F.W0 E6 /r"          , "ANY AVX512F-VL"],
    ["vcvtdq2pd"        , "W:ymm {kz}, xmm/m128/b32"                 , "RM"   , "EVEX.256.F3.0F.W0 E6 /r"          , "ANY AVX512F-VL"],
    ["vcvtdq2pd"        , "W:zmm {kz}, ymm/m256/b32"                 , "RM"   , "EVEX.512.F3.0F.W0 E6 /r"          , "ANY AVX512F"],
    ["vcvtdq2ps"        , "W:xmm {kz}, xmm/m128/b32"                 , "RM"   , "EVEX.128.0F.W0 5B /r"             , "ANY AVX512F-VL"],
    ["vcvtdq2ps"        , "W:ymm {kz}, ymm/m256/b32"                 , "RM"   , "EVEX.256.0F.W0 5B /r"             , "ANY AVX512F-VL"],
    ["vcvtdq2ps"        , "W:zmm {kz}, zmm/m512/b32 {er}"            , "RM"   , "EVEX.512.0F.W0 5B /r"             , "ANY AVX512F"],
    ["vcvtpd2dq"        , "W:xmm {kz}, xmm/m128/b64"                 , "RM"   , "EVEX.128.F2.0F.W1 E6 /r"          , "ANY AVX512F-VL"],
    ["vcvtpd2dq"        , "W:xmm {kz}, ymm/m256/b64"                 , "RM"   , "EVEX.256.F2.0F.W1 E6 /r"          , "ANY AVX512F-VL"],
    ["vcvtpd2dq"        , "W:ymm {kz}, zmm/m512/b64 {er}"            , "RM"   , "EVEX.512.F2.0F.W1 E6 /r"          , "ANY AVX512F"],
    ["vcvtpd2qq"        , "W:xmm {kz}, xmm/m128/b64"                 , "RM"   , "EVEX.128.66.0F.W1 7B /r"          , "ANY AVX512DQ-VL"],
    ["vcvtpd2qq"        , "W:ymm {kz}, ymm/m256/b64"                 , "RM"   , "EVEX.256.66.0F.W1 7B /r"          , "ANY AVX512DQ-VL"],
    ["vcvtpd2qq"        , "W:zmm {kz}, zmm/m512/b64 {er}"            , "RM"   , "EVEX.512.66.0F.W1 7B /r"          , "ANY AVX512DQ"],
    ["vcvtpd2udq"       , "W:xmm {kz}, xmm/m128/b64"                 , "RM"   , "EVEX.128.0F.W1 79 /r"             , "ANY AVX512F-VL"],
    ["vcvtpd2udq"       , "W:xmm {kz}, ymm/m256/b64"                 , "RM"   , "EVEX.256.0F.W1 79 /r"             , "ANY AVX512F-VL"],
    ["vcvtpd2udq"       , "W:ymm {kz}, zmm/m512/b64 {er}"            , "RM"   , "EVEX.512.0F.W1 79 /r"             , "ANY AVX512F"],
    ["vcvtpd2uqq"       , "W:xmm {kz}, xmm/m128/b64"                 , "RM"   , "EVEX.128.66.0F.W1 79 /r"          , "ANY AVX512DQ-VL"],
    ["vcvtpd2uqq"       , "W:ymm {kz}, ymm/m256/b64"                 , "RM"   , "EVEX.256.66.0F.W1 79 /r"          , "ANY AVX512DQ-VL"],
    ["vcvtpd2uqq"       , "W:zmm {kz}, zmm/m512/b64 {er}"            , "RM"   , "EVEX.512.66.0F.W1 79 /r"          , "ANY AVX512DQ"],
    ["vcvtph2ps"        , "W:xmm {kz}, xmm/m64"                      , "RM"   , "EVEX.128.66.0F38.W0 13 /r"        , "ANY AVX512F-VL"],
    ["vcvtph2ps"        , "W:ymm {kz}, xmm/m128"                     , "RM"   , "EVEX.256.66.0F38.W0 13 /r"        , "ANY AVX512F-VL"],
    ["vcvtph2ps"        , "W:zmm {kz}, ymm/m256 {sae}"               , "RM"   , "EVEX.512.66.0F38.W0 13 /r"        , "ANY AVX512F"],
    ["vcvtps2dq"        , "W:xmm {kz}, xmm/m128/b32"                 , "RM"   , "EVEX.128.66.0F.W0 5B /r"          , "ANY AVX512F-VL"],
    ["vcvtps2dq"        , "W:ymm {kz}, ymm/m256/b32"                 , "RM"   , "EVEX.256.66.0F.W0 5B /r"          , "ANY AVX512F-VL"],
    ["vcvtps2dq"        , "W:zmm {kz}, zmm/m512/b32 {er}"            , "RM"   , "EVEX.512.66.0F.W0 5B /r"          , "ANY AVX512F"],
    ["vcvtps2pd"        , "W:xmm {kz}, xmm/m64 /b32"                 , "RM"   , "EVEX.128.0F.W0 5A /r"             , "ANY AVX512F-VL"],
    ["vcvtps2pd"        , "W:ymm {kz}, xmm/m128/b32"                 , "RM"   , "EVEX.256.0F.W0 5A /r"             , "ANY AVX512F-VL"],
    ["vcvtps2pd"        , "W:zmm {kz}, ymm/m256/b32 {er}"            , "RM"   , "EVEX.512.0F.W0 5A /r"             , "ANY AVX512F"],
    ["vcvtps2ph"        , "W:xmm/m64  {kz}, xmm, ib"                 , "MRI"  , "EVEX.128.66.0F3A.W0 1D /r ib"     , "ANY AVX512F-VL"],
    ["vcvtps2ph"        , "W:xmm/m128 {kz}, ymm, ib"                 , "MRI"  , "EVEX.256.66.0F3A.W0 1D /r ib"     , "ANY AVX512F-VL"],
    ["vcvtps2ph"        , "W:ymm/m256 {kz}, zmm, ib {sae}"           , "MRI"  , "EVEX.512.66.0F3A.W0 1D /r ib"     , "ANY AVX512F"],
    ["vcvtps2qq"        , "W:xmm {kz}, xmm/m64 /b32"                 , "RM"   , "EVEX.128.66.0F.W0 7B /r"          , "ANY AVX512DQ-VL"],
    ["vcvtps2qq"        , "W:ymm {kz}, xmm/m128/b32"                 , "RM"   , "EVEX.256.66.0F.W0 7B /r"          , "ANY AVX512DQ-VL"],
    ["vcvtps2qq"        , "W:zmm {kz}, ymm/m256/b32 {er}"            , "RM"   , "EVEX.512.66.0F.W0 7B /r"          , "ANY AVX512DQ"],
    ["vcvtps2udq"       , "W:xmm {kz}, xmm/m128/b32"                 , "RM"   , "EVEX.128.0F.W0 79 /r"             , "ANY AVX512F-VL"],
    ["vcvtps2udq"       , "W:ymm {kz}, ymm/m256/b32"                 , "RM"   , "EVEX.256.0F.W0 79 /r"             , "ANY AVX512F-VL"],
    ["vcvtps2udq"       , "W:zmm {kz}, zmm/m512/b32 {er}"            , "RM"   , "EVEX.512.0F.W0 79 /r"             , "ANY AVX512F"],
    ["vcvtps2uqq"       , "W:xmm {kz}, xmm/m64 /b32"                 , "RM"   , "EVEX.128.66.0F.W0 79 /r"          , "ANY AVX512DQ-VL"],
    ["vcvtps2uqq"       , "W:ymm {kz}, xmm/m128/b32"                 , "RM"   , "EVEX.256.66.0F.W0 79 /r"          , "ANY AVX512DQ-VL"],
    ["vcvtps2uqq"       , "W:zmm {kz}, ymm/m256/b32 {er}"            , "RM"   , "EVEX.512.66.0F.W0 79 /r"          , "ANY AVX512DQ"],
    ["vcvtqq2pd"        , "W:xmm {kz}, xmm/m128/b64"                 , "RM"   , "EVEX.128.F3.0F.W1 E6 /r"          , "ANY AVX512DQ-VL"],
    ["vcvtqq2pd"        , "W:ymm {kz}, ymm/m256/b64"                 , "RM"   , "EVEX.256.F3.0F.W1 E6 /r"          , "ANY AVX512DQ-VL"],
    ["vcvtqq2pd"        , "W:zmm {kz}, zmm/m512/b64 {er}"            , "RM"   , "EVEX.512.F3.0F.W1 E6 /r"          , "ANY AVX512DQ"],
    ["vcvtqq2ps"        , "W:xmm {kz}, xmm/m128/b64"                 , "RM"   , "EVEX.128.0F.W1 5B /r"             , "ANY AVX512DQ-VL"],
    ["vcvtqq2ps"        , "W:xmm {kz}, ymm/m256/b64"                 , "RM"   , "EVEX.256.0F.W1 5B /r"             , "ANY AVX512DQ-VL"],
    ["vcvtqq2ps"        , "W:ymm {kz}, zmm/m512/b64 {er}"            , "RM"   , "EVEX.512.0F.W1 5B /r"             , "ANY AVX512DQ"],
    ["vcvtsd2si"        , "W:r32, xmm/m64 {er}"                      , "RM"   , "EVEX.LIG.F2.0F.W0 2D /r"          , "ANY AVX512F"],
    ["vcvtsd2si"        , "W:r64, xmm/m64 {er}"                      , "RM"   , "EVEX.LIG.F2.0F.W1 2D /r"          , "X64 AVX512F"],
    ["vcvtsd2ss"        , "W:xmm {kz}, xmm, xmm/m64 {er}"            , "RVM"  , "EVEX.NDS.LIG.F2.0F.W1 5A /r"      , "ANY AVX512F"],
    ["vcvtsd2usi"       , "W:r32, xmm/m64 {er}"                      , "RM"   , "EVEX.LIG.F2.0F.W0 79 /r"          , "ANY AVX512F"],
    ["vcvtsd2usi"       , "W:r64, xmm/m64 {er}"                      , "RM"   , "EVEX.LIG.F2.0F.W1 79 /r"          , "X64 AVX512F"],
    ["vcvtsi2sd"        , "W:xmm, xmm, r32/m32 {er}"                 , "RVM"  , "EVEX.NDS.LIG.F2.0F.W0 2A /r"      , "ANY AVX512F"],
    ["vcvtsi2sd"        , "W:xmm, xmm, r64/m64 {er}"                 , "RVM"  , "EVEX.NDS.LIG.F2.0F.W1 2A /r"      , "X64 AVX512F"],
    ["vcvtsi2ss"        , "W:xmm, xmm, r32/m32 {er}"                 , "RVM"  , "EVEX.NDS.LIG.F3.0F.W0 2A /r"      , "ANY AVX512F"],
    ["vcvtsi2ss"        , "W:xmm, xmm, r64/m64 {er}"                 , "RVM"  , "EVEX.NDS.LIG.F3.0F.W1 2A /r"      , "X64 AVX512F"],
    ["vcvtss2sd"        , "W:xmm {kz}, xmm, xmm/m32 {sae}"           , "RVM"  , "EVEX.NDS.LIG.F3.0F.W0 5A /r"      , "ANY AVX512F"],
    ["vcvtss2si"        , "W:r32, xmm/m32 {er}"                      , "RM"   , "EVEX.LIG.F3.0F.W0 2D /r"          , "ANY AVX512F"],
    ["vcvtss2si"        , "W:r64, xmm/m64 {er}"                      , "RM"   , "EVEX.LIG.F3.0F.W1 2D /r"          , "X64 AVX512F"],
    ["vcvtss2usi"       , "W:r32, xmm/m32 {er}"                      , "RM"   , "EVEX.LIG.F3.0F.W0 79 /r"          , "ANY AVX512F"],
    ["vcvtss2usi"       , "W:r64, xmm/m64 {er}"                      , "RM"   , "EVEX.LIG.F3.0F.W1 79 /r"          , "X64 AVX512F"],
    ["vcvttpd2dq"       , "W:xmm {kz}, xmm/m128/b64"                 , "RM"   , "EVEX.128.66.0F.W1 E6 /r"          , "ANY AVX512F-VL"],
    ["vcvttpd2dq"       , "W:xmm {kz}, ymm/m256/b64"                 , "RM"   , "EVEX.256.66.0F.W1 E6 /r"          , "ANY AVX512F-VL"],
    ["vcvttpd2dq"       , "W:ymm {kz}, zmm/m512/b64 {sae}"           , "RM"   , "EVEX.512.66.0F.W1 E6 /r"          , "ANY AVX512F"],
    ["vcvttpd2qq"       , "W:xmm {kz}, xmm/m128/b64"                 , "RM"   , "EVEX.128.66.0F.W1 7A /r"          , "ANY AVX512F-VL"],
    ["vcvttpd2qq"       , "W:ymm {kz}, ymm/m256/b64"                 , "RM"   , "EVEX.256.66.0F.W1 7A /r"          , "ANY AVX512F-VL"],
    ["vcvttpd2qq"       , "W:zmm {kz}, zmm/m512/b64 {sae}"           , "RM"   , "EVEX.512.66.0F.W1 7A /r"          , "ANY AVX512F"],
    ["vcvttpd2udq"      , "W:xmm {kz}, xmm/m128/b64"                 , "RM"   , "EVEX.128.0F.W1 78 /r"             , "ANY AVX512F-VL"],
    ["vcvttpd2udq"      , "W:xmm {kz}, ymm/m256/b64"                 , "RM"   , "EVEX.256.0F.W1 78 /r"             , "ANY AVX512F-VL"],
    ["vcvttpd2udq"      , "W:ymm {kz}, zmm/m512/b64 {sae}"           , "RM"   , "EVEX.512.0F.W1 78 /r"             , "ANY AVX512F"],
    ["vcvttpd2uqq"      , "W:xmm {kz}, xmm/m128/b64"                 , "RM"   , "EVEX.128.66.0F.W1 78 /r"          , "ANY AVX512DQ-VL"],
    ["vcvttpd2uqq"      , "W:ymm {kz}, ymm/m256/b64"                 , "RM"   , "EVEX.256.66.0F.W1 78 /r"          , "ANY AVX512DQ-VL"],
    ["vcvttpd2uqq"      , "W:zmm {kz}, zmm/m512/b64 {sae}"           , "RM"   , "EVEX.512.66.0F.W1 78 /r"          , "ANY AVX512DQ"],
    ["vcvttps2dq"       , "W:xmm {kz}, xmm/m128/b32"                 , "RM"   , "EVEX.128.F3.0F.W0 5B /r"          , "ANY AVX512F-VL"],
    ["vcvttps2dq"       , "W:ymm {kz}, ymm/m256/b32"                 , "RM"   , "EVEX.256.F3.0F.W0 5B /r"          , "ANY AVX512F-VL"],
    ["vcvttps2dq"       , "W:zmm {kz}, zmm/m512/b32 {sae}"           , "RM"   , "EVEX.512.F3.0F.W0 5B /r"          , "ANY AVX512F"],
    ["vcvttps2qq"       , "W:xmm {kz}, xmm/m64 /b32"                 , "RM"   , "EVEX.128.66.0F.W0 7A /r"          , "ANY AVX512DQ-VL"],
    ["vcvttps2qq"       , "W:ymm {kz}, xmm/m128/b32"                 , "RM"   , "EVEX.256.66.0F.W0 7A /r"          , "ANY AVX512DQ-VL"],
    ["vcvttps2qq"       , "W:zmm {kz}, ymm/m256/b32 {sae}"           , "RM"   , "EVEX.512.66.0F.W0 7A /r"          , "ANY AVX512DQ"],
    ["vcvttps2udq"      , "W:xmm {kz}, xmm/m128/b32"                 , "RM"   , "EVEX.128.0F.W0 78 /r"             , "ANY AVX512F-VL"],
    ["vcvttps2udq"      , "W:ymm {kz}, ymm/m256/b32"                 , "RM"   , "EVEX.256.0F.W0 78 /r"             , "ANY AVX512F-VL"],
    ["vcvttps2udq"      , "W:zmm {kz}, zmm/m512/b32 {sae}"           , "RM"   , "EVEX.512.0F.W0 78 /r"             , "ANY AVX512F"],
    ["vcvttps2uqq"      , "W:xmm {kz}, xmm/m64 /b32"                 , "RM"   , "EVEX.128.66.0F.W0 78 /r"          , "ANY AVX512DQ-VL"],
    ["vcvttps2uqq"      , "W:ymm {kz}, xmm/m128/b32"                 , "RM"   , "EVEX.256.66.0F.W0 78 /r"          , "ANY AVX512DQ-VL"],
    ["vcvttps2uqq"      , "W:zmm {kz}, ymm/m256/b32 {sae}"           , "RM"   , "EVEX.512.66.0F.W0 78 /r"          , "ANY AVX512DQ"],
    ["vcvttsd2si"       , "W:r32, xmm/m64 {sae}"                     , "RM"   , "EVEX.LIG.F2.0F.W0 2C /r"          , "ANY AVX512F"],
    ["vcvttsd2si"       , "W:r64, xmm/m64 {sae}"                     , "RM"   , "EVEX.LIG.F2.0F.W1 2C /r"          , "X64 AVX512F"],
    ["vcvttsd2usi"      , "W:r32, xmm/m64 {sae}"                     , "RM"   , "EVEX.LIG.F2.0F.W0 78 /r"          , "ANY AVX512F"],
    ["vcvttsd2usi"      , "W:r64, xmm/m64 {sae}"                     , "RM"   , "EVEX.LIG.F2.0F.W1 78 /r"          , "X64 AVX512F"],
    ["vcvttss2si"       , "W:r32, xmm/m32 {sae}"                     , "RM"   , "EVEX.LIG.F3.0F.W0 2C /r"          , "ANY AVX512F"],
    ["vcvttss2si"       , "W:r64, xmm/m64 {sae}"                     , "RM"   , "EVEX.LIG.F3.0F.W1 2C /r"          , "X64 AVX512F"],
    ["vcvttss2usi"      , "W:r32, xmm/m32 {sae}"                     , "RM"   , "EVEX.LIG.F3.0F.W0 78 /r"          , "ANY AVX512F"],
    ["vcvttss2usi"      , "W:r64, xmm/m64 {sae}"                     , "RM"   , "EVEX.LIG.F3.0F.W1 78 /r"          , "X64 AVX512F"],
    ["vcvtudq2pd"       , "W:xmm {kz}, xmm/m64 /b32"                 , "RM"   , "EVEX.128.F3.0F.W0 7A /r"          , "ANY AVX512F-VL"],
    ["vcvtudq2pd"       , "W:ymm {kz}, xmm/m128/b32"                 , "RM"   , "EVEX.256.F3.0F.W0 7A /r"          , "ANY AVX512F-VL"],
    ["vcvtudq2pd"       , "W:zmm {kz}, ymm/m256/b32"                 , "RM"   , "EVEX.512.F3.0F.W0 7A /r"          , "ANY AVX512F"],
    ["vcvtudq2ps"       , "W:xmm {kz}, xmm/m128/b32"                 , "RM"   , "EVEX.128.F2.0F.W0 7A /r"          , "ANY AVX512F-VL"],
    ["vcvtudq2ps"       , "W:ymm {kz}, ymm/m256/b32"                 , "RM"   , "EVEX.256.F2.0F.W0 7A /r"          , "ANY AVX512F-VL"],
    ["vcvtudq2ps"       , "W:zmm {kz}, zmm/m512/b32 {er}"            , "RM"   , "EVEX.512.F2.0F.W0 7A /r"          , "ANY AVX512F"],
    ["vcvtuqq2pd"       , "W:xmm {kz}, xmm/m128/b64"                 , "RM"   , "EVEX.128.F3.0F.W1 7A /r"          , "ANY AVX512DQ-VL"],
    ["vcvtuqq2pd"       , "W:ymm {kz}, ymm/m256/b64"                 , "RM"   , "EVEX.256.F3.0F.W1 7A /r"          , "ANY AVX512DQ-VL"],
    ["vcvtuqq2pd"       , "W:zmm {kz}, zmm/m512/b64 {er}"            , "RM"   , "EVEX.512.F3.0F.W1 7A /r"          , "ANY AVX512DQ"],
    ["vcvtuqq2ps"       , "W:xmm {kz}, xmm/m128/b64"                 , "RM"   , "EVEX.128.F2.0F.W1 7A /r"          , "ANY AVX512DQ-VL"],
    ["vcvtuqq2ps"       , "W:xmm {kz}, ymm/m256/b64"                 , "RM"   , "EVEX.256.F2.0F.W1 7A /r"          , "ANY AVX512DQ-VL"],
    ["vcvtuqq2ps"       , "W:ymm {kz}, zmm/m512/b64 {er}"            , "RM"   , "EVEX.512.F2.0F.W1 7A /r"          , "ANY AVX512DQ"],
    ["vcvtusi2sd"       , "W:xmm, xmm, r32/m32 {er}"                 , "RVM"  , "EVEX.NDS.LIG.F2.0F.W0 7B /r"      , "ANY AVX512F"],
    ["vcvtusi2sd"       , "W:xmm, xmm, r64/m64 {er}"                 , "RVM"  , "EVEX.NDS.LIG.F2.0F.W1 7B /r"      , "X64 AVX512F"],
    ["vcvtusi2ss"       , "W:xmm, xmm, r32/m32 {er}"                 , "RVM"  , "EVEX.NDS.LIG.F3.0F.W0 7B /r"      , "ANY AVX512F"],
    ["vcvtusi2ss"       , "W:xmm, xmm, r64/m64 {er}"                 , "RVM"  , "EVEX.NDS.LIG.F3.0F.W1 7B /r"      , "X64 AVX512F"],
    ["vdbpsadbw"        , "W:xmm {kz}, xmm, xmm/m128, ib"            , "RVMI" , "EVEX.NDS.128.66.0F3A.W0 42 /r ib" , "ANY AVX512BW-VL"],
    ["vdbpsadbw"        , "W:ymm {kz}, ymm, ymm/m256, ib"            , "RVMI" , "EVEX.NDS.256.66.0F3A.W0 42 /r ib" , "ANY AVX512BW-VL"],
    ["vdbpsadbw"        , "W:zmm {kz}, zmm, zmm/m512, ib"            , "RVMI" , "EVEX.NDS.512.66.0F3A.W0 42 /r ib" , "ANY AVX512BW"],
    ["vdivpd"           , "W:xmm {kz}, xmm, xmm/m128/b64"            , "RVM"  , "EVEX.NDS.128.66.0F.W1 5E /r"      , "ANY AVX512F-VL"],
    ["vdivpd"           , "W:ymm {kz}, ymm, ymm/m256/b64"            , "RVM"  , "EVEX.NDS.256.66.0F.W1 5E /r"      , "ANY AVX512F-VL"],
    ["vdivpd"           , "W:zmm {kz}, zmm, zmm/m512/b64 {er}"       , "RVM"  , "EVEX.NDS.512.66.0F.W1 5E /r"      , "ANY AVX512F"],
    ["vdivps"           , "W:xmm {kz}, xmm, xmm/m128/b32"            , "RVM"  , "EVEX.NDS.128.0F.W0 5E /r"         , "ANY AVX512F-VL"],
    ["vdivps"           , "W:ymm {kz}, ymm, ymm/m256/b32"            , "RVM"  , "EVEX.NDS.256.0F.W0 5E /r"         , "ANY AVX512F-VL"],
    ["vdivps"           , "W:zmm {kz}, zmm, zmm/m512/b32 {er}"       , "RVM"  , "EVEX.NDS.512.0F.W0 5E /r"         , "ANY AVX512F"],
    ["vdivsd"           , "W:xmm {kz}, xmm, xmm/m64 {er}"            , "RVM"  , "EVEX.NDS.LIG.F2.0F.W1 5E /r"      , "ANY AVX512F"],
    ["vdivss"           , "W:xmm {kz}, xmm, xmm/m32 {er}"            , "RVM"  , "EVEX.NDS.LIG.F3.0F.W0 5E /r"      , "ANY AVX512F"],
    ["vexp2pd"          , "W:zmm {kz}, zmm/m512/b64 {sae}"           , "RM"   , "EVEX.512.66.0F38.W1 C8 /r"        , "ANY AVX512ER"],
    ["vexp2ps"          , "W:zmm {kz}, zmm/m512/b32 {sae}"           , "RM"   , "EVEX.512.66.0F38.W0 C8 /r"        , "ANY AVX512ER"],
    ["vexpandd"         , "W:xmm {kz}, xmm/m128"                     , "RM"   , "EVEX.128.66.0F38.W0 89 /r"        , "ANY AVX512F-VL"],
    ["vexpandd"         , "W:ymm {kz}, ymm/m256"                     , "RM"   , "EVEX.256.66.0F38.W0 89 /r"        , "ANY AVX512F-VL"],
    ["vexpandd"         , "W:zmm {kz}, zmm/m512"                     , "RM"   , "EVEX.512.66.0F38.W0 89 /r"        , "ANY AVX512F"],
    ["vexpandpd"        , "W:xmm {kz}, xmm/m128"                     , "RM"   , "EVEX.128.66.0F38.W1 88 /r"        , "ANY AVX512F-VL"],
    ["vexpandpd"        , "W:ymm {kz}, ymm/m256"                     , "RM"   , "EVEX.256.66.0F38.W1 88 /r"        , "ANY AVX512F-VL"],
    ["vexpandpd"        , "W:zmm {kz}, zmm/m512"                     , "RM"   , "EVEX.512.66.0F38.W1 88 /r"        , "ANY AVX512F"],
    ["vexpandps"        , "W:xmm {kz}, xmm/m128"                     , "RM"   , "EVEX.128.66.0F38.W0 88 /r"        , "ANY AVX512F-VL"],
    ["vexpandps"        , "W:ymm {kz}, ymm/m256"                     , "RM"   , "EVEX.256.66.0F38.W0 88 /r"        , "ANY AVX512F-VL"],
    ["vexpandps"        , "W:zmm {kz}, zmm/m512"                     , "RM"   , "EVEX.512.66.0F38.W0 88 /r"        , "ANY AVX512F"],
    ["vexpandq"         , "W:xmm {kz}, xmm/m128"                     , "RM"   , "EVEX.128.66.0F38.W1 89 /r"        , "ANY AVX512F-VL"],
    ["vexpandq"         , "W:ymm {kz}, ymm/m256"                     , "RM"   , "EVEX.256.66.0F38.W1 89 /r"        , "ANY AVX512F-VL"],
    ["vexpandq"         , "W:zmm {kz}, zmm/m512"                     , "RM"   , "EVEX.512.66.0F38.W1 89 /r"        , "ANY AVX512F"],
    ["vextractf32x4"    , "W:xmm/m128 {kz}, ymm, ib"                 , "MRI"  , "EVEX.256.66.0F3A.W0 19 /r ib"     , "ANY AVX512F-VL"],
    ["vextractf32x4"    , "W:xmm/m128 {kz}, zmm, ib"                 , "MRI"  , "EVEX.512.66.0F3A.W0 19 /r ib"     , "ANY AVX512F"],
    ["vextractf32x8"    , "W:ymm/m256 {kz}, zmm, ib"                 , "MRI"  , "EVEX.512.66.0F3A.W0 1B /r ib"     , "ANY AVX512DQ"],
    ["vextractf64x2"    , "W:xmm/m128 {kz}, ymm, ib"                 , "MRI"  , "EVEX.256.66.0F3A.W1 19 /r ib"     , "ANY AVX512DQ-VL"],
    ["vextractf64x2"    , "W:xmm/m128 {kz}, zmm, ib"                 , "MRI"  , "EVEX.512.66.0F3A.W1 19 /r ib"     , "ANY AVX512DQ"],
    ["vextractf64x4"    , "W:ymm/m256 {kz}, zmm, ib"                 , "MRI"  , "EVEX.512.66.0F3A.W1 1B /r ib"     , "ANY AVX512F"],
    ["vextracti32x4"    , "W:xmm/m128 {kz}, ymm, ib"                 , "MRI"  , "EVEX.256.66.0F3A.W0 39 /r ib"     , "ANY AVX512F-VL"],
    ["vextracti32x4"    , "W:xmm/m128 {kz}, zmm, ib"                 , "MRI"  , "EVEX.512.66.0F3A.W0 39 /r ib"     , "ANY AVX512F"],
    ["vextracti32x8"    , "W:ymm/m256 {kz}, zmm, ib"                 , "MRI"  , "EVEX.512.66.0F3A.W0 3B /r ib"     , "ANY AVX512DQ"],
    ["vextracti64x2"    , "W:xmm/m128 {kz}, ymm, ib"                 , "MRI"  , "EVEX.256.66.0F3A.W1 39 /r ib"     , "ANY AVX512DQ-VL"],
    ["vextracti64x2"    , "W:xmm/m128 {kz}, zmm, ib"                 , "MRI"  , "EVEX.512.66.0F3A.W1 39 /r ib"     , "ANY AVX512DQ"],
    ["vextracti64x4"    , "W:ymm/m256 {kz}, zmm, ib"                 , "MRI"  , "EVEX.512.66.0F3A.W1 3B /r ib"     , "ANY AVX512F"],
    ["vextractps"       , "W:r32/m32, xmm, ib"                       , "MRI"  , "EVEX.128.66.0F3A.WIG 17 /r ib"    , "ANY AVX512F"],
    ["vfixupimmpd"      , "X:xmm {kz}, xmm, xmm/m128/b64, ib"        , "RVMI" , "EVEX.NDS.128.66.0F3A.W1 54 /r ib" , "ANY AVX512F-VL"],
    ["vfixupimmpd"      , "X:ymm {kz}, ymm, ymm/m256/b64, ib"        , "RVMI" , "EVEX.NDS.256.66.0F3A.W1 54 /r ib" , "ANY AVX512F-VL"],
    ["vfixupimmpd"      , "X:zmm {kz}, zmm, zmm/m512/b64, ib {sae}"  , "RVMI" , "EVEX.NDS.512.66.0F3A.W1 54 /r ib" , "ANY AVX512F"],
    ["vfixupimmps"      , "X:xmm {kz}, xmm, xmm/m128/b32, ib"        , "RVMI" , "EVEX.NDS.128.66.0F3A.W0 54 /r ib" , "ANY AVX512F-VL"],
    ["vfixupimmps"      , "X:ymm {kz}, ymm, ymm/m256/b32, ib"        , "RVMI" , "EVEX.NDS.256.66.0F3A.W0 54 /r ib" , "ANY AVX512F-VL"],
    ["vfixupimmps"      , "X:zmm {kz}, zmm, zmm/m512/b32, ib {sae}"  , "RVMI" , "EVEX.NDS.512.66.0F3A.W0 54 /r ib" , "ANY AVX512F"],
    ["vfixupimmsd"      , "X:xmm {kz}, xmm, xmm/m64, ib {sae}"       , "RVMI" , "EVEX.NDS.LIG.66.0F3A.W1 55 /r ib" , "ANY AVX512F"],
    ["vfixupimmss"      , "X:xmm {kz}, xmm, xmm/m32, ib {sae}"       , "RVMI" , "EVEX.NDS.LIG.66.0F3A.W0 55 /r ib" , "ANY AVX512F"],
    ["vfmadd132pd"      , "X:xmm {kz}, xmm, xmm/m128/b64"            , "RVM"  , "EVEX.NDS.128.66.0F38.W1 98 /r"    , "ANY AVX512F-VL"],
    ["vfmadd132pd"      , "X:ymm {kz}, ymm, ymm/m256/b64"            , "RVM"  , "EVEX.NDS.256.66.0F38.W1 98 /r"    , "ANY AVX512F-VL"],
    ["vfmadd132pd"      , "X:zmm {kz}, zmm, zmm/m512/b64 {er}"       , "RVM"  , "EVEX.NDS.512.66.0F38.W1 98 /r"    , "ANY AVX512F"],
    ["vfmadd132ps"      , "X:xmm {kz}, xmm, xmm/m128/b32"            , "RVM"  , "EVEX.NDS.128.66.0F38.W0 98 /r"    , "ANY AVX512F-VL"],
    ["vfmadd132ps"      , "X:ymm {kz}, ymm, ymm/m256/b32"            , "RVM"  , "EVEX.NDS.256.66.0F38.W0 98 /r"    , "ANY AVX512F-VL"],
    ["vfmadd132ps"      , "X:zmm {kz}, zmm, zmm/m512/b32 {er}"       , "RVM"  , "EVEX.NDS.512.66.0F38.W0 98 /r"    , "ANY AVX512F"],
    ["vfmadd132sd"      , "X:xmm {kz}, xmm, xmm/m64 {er}"            , "RVM"  , "EVEX.DDS.LIG.66.0F38.W1 99 /r"    , "ANY AVX512F"],
    ["vfmadd132ss"      , "X:xmm {kz}, xmm, xmm/m32 {er}"            , "RVM"  , "EVEX.DDS.LIG.66.0F38.W0 99 /r"    , "ANY AVX512F"],
    ["vfmadd213pd"      , "X:xmm {kz}, xmm, xmm/m128/b64"            , "RVM"  , "EVEX.NDS.128.66.0F38.W1 A8 /r"    , "ANY AVX512F-VL"],
    ["vfmadd213pd"      , "X:ymm {kz}, ymm, ymm/m256/b64"            , "RVM"  , "EVEX.NDS.256.66.0F38.W1 A8 /r"    , "ANY AVX512F-VL"],
    ["vfmadd213pd"      , "X:zmm {kz}, zmm, zmm/m512/b64 {er}"       , "RVM"  , "EVEX.NDS.512.66.0F38.W1 A8 /r"    , "ANY AVX512F"],
    ["vfmadd213ps"      , "X:xmm {kz}, xmm, xmm/m128/b32"            , "RVM"  , "EVEX.NDS.128.66.0F38.W0 A8 /r"    , "ANY AVX512F-VL"],
    ["vfmadd213ps"      , "X:ymm {kz}, ymm, ymm/m256/b32"            , "RVM"  , "EVEX.NDS.256.66.0F38.W0 A8 /r"    , "ANY AVX512F-VL"],
    ["vfmadd213ps"      , "X:zmm {kz}, zmm, zmm/m512/b32 {er}"       , "RVM"  , "EVEX.NDS.512.66.0F38.W0 A8 /r"    , "ANY AVX512F"],
    ["vfmadd213sd"      , "X:xmm {kz}, xmm, xmm/m64 {er}"            , "RVM"  , "EVEX.DDS.LIG.66.0F38.W1 A9 /r"    , "ANY AVX512F"],
    ["vfmadd213ss"      , "X:xmm {kz}, xmm, xmm/m32 {er}"            , "RVM"  , "EVEX.DDS.LIG.66.0F38.W0 A9 /r"    , "ANY AVX512F"],
    ["vfmadd231pd"      , "X:xmm {kz}, xmm, xmm/m128/b64"            , "RVM"  , "EVEX.NDS.128.66.0F38.W1 B8 /r"    , "ANY AVX512F-VL"],
    ["vfmadd231pd"      , "X:ymm {kz}, ymm, ymm/m256/b64"            , "RVM"  , "EVEX.NDS.256.66.0F38.W1 B8 /r"    , "ANY AVX512F-VL"],
    ["vfmadd231pd"      , "X:zmm {kz}, zmm, zmm/m512/b64 {er}"       , "RVM"  , "EVEX.NDS.512.66.0F38.W1 B8 /r"    , "ANY AVX512F"],
    ["vfmadd231ps"      , "X:xmm {kz}, xmm, xmm/m128/b32"            , "RVM"  , "EVEX.NDS.128.66.0F38.W0 B8 /r"    , "ANY AVX512F-VL"],
    ["vfmadd231ps"      , "X:ymm {kz}, ymm, ymm/m256/b32"            , "RVM"  , "EVEX.NDS.256.66.0F38.W0 B8 /r"    , "ANY AVX512F-VL"],
    ["vfmadd231ps"      , "X:zmm {kz}, zmm, zmm/m512/b32 {er}"       , "RVM"  , "EVEX.NDS.512.66.0F38.W0 B8 /r"    , "ANY AVX512F"],
    ["vfmadd231sd"      , "X:xmm {kz}, xmm, xmm/m64 {er}"            , "RVM"  , "EVEX.DDS.LIG.66.0F38.W1 B9 /r"    , "ANY AVX512F"],
    ["vfmadd231ss"      , "X:xmm {kz}, xmm, xmm/m32 {er}"            , "RVM"  , "EVEX.DDS.LIG.66.0F38.W0 B9 /r"    , "ANY AVX512F"],
    ["vfmaddsub132pd"   , "X:xmm {kz}, xmm, xmm/m128/b64"            , "RVM"  , "EVEX.DDS.128.66.0F38.W1 96 /r"    , "ANY AVX512F-VL"],
    ["vfmaddsub132pd"   , "X:ymm {kz}, ymm, ymm/m256/b64"            , "RVM"  , "EVEX.DDS.256.66.0F38.W1 96 /r"    , "ANY AVX512F-VL"],
    ["vfmaddsub132pd"   , "X:zmm {kz}, zmm, zmm/m512/b64 {er}"       , "RVM"  , "EVEX.DDS.512.66.0F38.W1 96 /r"    , "ANY AVX512F"],
    ["vfmaddsub132ps"   , "X:xmm {kz}, xmm, xmm/m128/b32"            , "RVM"  , "EVEX.DDS.128.66.0F38.W0 96 /r"    , "ANY AVX512F-VL"],
    ["vfmaddsub132ps"   , "X:ymm {kz}, ymm, ymm/m256/b32"            , "RVM"  , "EVEX.DDS.256.66.0F38.W0 96 /r"    , "ANY AVX512F-VL"],
    ["vfmaddsub132ps"   , "X:zmm {kz}, zmm, zmm/m512/b32 {er}"       , "RVM"  , "EVEX.DDS.512.66.0F38.W0 96 /r"    , "ANY AVX512F"],
    ["vfmaddsub213pd"   , "X:xmm {kz}, xmm, xmm/m128/b64"            , "RVM"  , "EVEX.DDS.128.66.0F38.W1 A6 /r"    , "ANY AVX512F-VL"],
    ["vfmaddsub213pd"   , "X:ymm {kz}, ymm, ymm/m256/b64"            , "RVM"  , "EVEX.DDS.256.66.0F38.W1 A6 /r"    , "ANY AVX512F-VL"],
    ["vfmaddsub213pd"   , "X:zmm {kz}, zmm, zmm/m512/b64 {er}"       , "RVM"  , "EVEX.DDS.512.66.0F38.W1 A6 /r"    , "ANY AVX512F"],
    ["vfmaddsub213ps"   , "X:xmm {kz}, xmm, xmm/m128/b32"            , "RVM"  , "EVEX.DDS.128.66.0F38.W0 A6 /r"    , "ANY AVX512F-VL"],
    ["vfmaddsub213ps"   , "X:ymm {kz}, ymm, ymm/m256/b32"            , "RVM"  , "EVEX.DDS.256.66.0F38.W0 A6 /r"    , "ANY AVX512F-VL"],
    ["vfmaddsub213ps"   , "X:zmm {kz}, zmm, zmm/m512/b32 {er}"       , "RVM"  , "EVEX.DDS.512.66.0F38.W0 A6 /r"    , "ANY AVX512F"],
    ["vfmaddsub231pd"   , "X:xmm {kz}, xmm, xmm/m128/b64"            , "RVM"  , "EVEX.DDS.128.66.0F38.W1 B6 /r"    , "ANY AVX512F-VL"],
    ["vfmaddsub231pd"   , "X:ymm {kz}, ymm, ymm/m256/b64"            , "RVM"  , "EVEX.DDS.256.66.0F38.W1 B6 /r"    , "ANY AVX512F-VL"],
    ["vfmaddsub231pd"   , "X:zmm {kz}, zmm, zmm/m512/b64 {er}"       , "RVM"  , "EVEX.DDS.512.66.0F38.W1 B6 /r"    , "ANY AVX512F"],
    ["vfmaddsub231ps"   , "X:xmm {kz}, xmm, xmm/m128/b32"            , "RVM"  , "EVEX.DDS.128.66.0F38.W0 B6 /r"    , "ANY AVX512F-VL"],
    ["vfmaddsub231ps"   , "X:ymm {kz}, ymm, ymm/m256/b32"            , "RVM"  , "EVEX.DDS.256.66.0F38.W0 B6 /r"    , "ANY AVX512F-VL"],
    ["vfmaddsub231ps"   , "X:zmm {kz}, zmm, zmm/m512/b32 {er}"       , "RVM"  , "EVEX.DDS.512.66.0F38.W0 B6 /r"    , "ANY AVX512F"],
    ["vfmsub132pd"      , "X:xmm {kz}, xmm, xmm/m128/b64"            , "RVM"  , "EVEX.NDS.128.66.0F38.W1 9A /r"    , "ANY AVX512F-VL"],
    ["vfmsub132pd"      , "X:ymm {kz}, ymm, ymm/m256/b64"            , "RVM"  , "EVEX.NDS.256.66.0F38.W1 9A /r"    , "ANY AVX512F-VL"],
    ["vfmsub132pd"      , "X:zmm {kz}, zmm, zmm/m512/b64 {er}"       , "RVM"  , "EVEX.NDS.512.66.0F38.W1 9A /r"    , "ANY AVX512F"],
    ["vfmsub132ps"      , "X:xmm {kz}, xmm, xmm/m128/b32"            , "RVM"  , "EVEX.NDS.128.66.0F38.W0 9A /r"    , "ANY AVX512F-VL"],
    ["vfmsub132ps"      , "X:ymm {kz}, ymm, ymm/m256/b32"            , "RVM"  , "EVEX.NDS.256.66.0F38.W0 9A /r"    , "ANY AVX512F-VL"],
    ["vfmsub132ps"      , "X:zmm {kz}, zmm, zmm/m512/b32 {er}"       , "RVM"  , "EVEX.NDS.512.66.0F38.W0 9A /r"    , "ANY AVX512F"],
    ["vfmsub132sd"      , "X:xmm {kz}, xmm, xmm/m64 {er}"            , "RVM"  , "EVEX.DDS.LIG.66.0F38.W1 9B /r"    , "ANY AVX512F"],
    ["vfmsub132ss"      , "X:xmm {kz}, xmm, xmm/m32 {er}"            , "RVM"  , "EVEX.DDS.LIG.66.0F38.W0 9B /r"    , "ANY AVX512F"],
    ["vfmsub213pd"      , "X:xmm {kz}, xmm, xmm/m128/b64"            , "RVM"  , "EVEX.NDS.128.66.0F38.W1 AA /r"    , "ANY AVX512F-VL"],
    ["vfmsub213pd"      , "X:ymm {kz}, ymm, ymm/m256/b64"            , "RVM"  , "EVEX.NDS.256.66.0F38.W1 AA /r"    , "ANY AVX512F-VL"],
    ["vfmsub213pd"      , "X:zmm {kz}, zmm, zmm/m512/b64 {er}"       , "RVM"  , "EVEX.NDS.512.66.0F38.W1 AA /r"    , "ANY AVX512F"],
    ["vfmsub213ps"      , "X:xmm {kz}, xmm, xmm/m128/b32"            , "RVM"  , "EVEX.NDS.128.66.0F38.W0 AA /r"    , "ANY AVX512F-VL"],
    ["vfmsub213ps"      , "X:ymm {kz}, ymm, ymm/m256/b32"            , "RVM"  , "EVEX.NDS.256.66.0F38.W0 AA /r"    , "ANY AVX512F-VL"],
    ["vfmsub213ps"      , "X:zmm {kz}, zmm, zmm/m512/b32 {er}"       , "RVM"  , "EVEX.NDS.512.66.0F38.W0 AA /r"    , "ANY AVX512F"],
    ["vfmsub213sd"      , "X:xmm {kz}, xmm, xmm/m64 {er}"            , "RVM"  , "EVEX.DDS.LIG.66.0F38.W1 AB /r"    , "ANY AVX512F"],
    ["vfmsub213ss"      , "X:xmm {kz}, xmm, xmm/m32 {er}"            , "RVM"  , "EVEX.DDS.LIG.66.0F38.W0 AB /r"    , "ANY AVX512F"],
    ["vfmsub231pd"      , "X:xmm {kz}, xmm, xmm/m128/b64"            , "RVM"  , "EVEX.NDS.128.66.0F38.W1 BA /r"    , "ANY AVX512F-VL"],
    ["vfmsub231pd"      , "X:ymm {kz}, ymm, ymm/m256/b64"            , "RVM"  , "EVEX.NDS.256.66.0F38.W1 BA /r"    , "ANY AVX512F-VL"],
    ["vfmsub231pd"      , "X:zmm {kz}, zmm, zmm/m512/b64 {er}"       , "RVM"  , "EVEX.NDS.512.66.0F38.W1 BA /r"    , "ANY AVX512F"],
    ["vfmsub231ps"      , "X:xmm {kz}, xmm, xmm/m128/b32"            , "RVM"  , "EVEX.NDS.128.66.0F38.W0 BA /r"    , "ANY AVX512F-VL"],
    ["vfmsub231ps"      , "X:ymm {kz}, ymm, ymm/m256/b32"            , "RVM"  , "EVEX.NDS.256.66.0F38.W0 BA /r"    , "ANY AVX512F-VL"],
    ["vfmsub231ps"      , "X:zmm {kz}, zmm, zmm/m512/b32 {er}"       , "RVM"  , "EVEX.NDS.512.66.0F38.W0 BA /r"    , "ANY AVX512F"],
    ["vfmsub231sd"      , "X:xmm {kz}, xmm, xmm/m64 {er}"            , "RVM"  , "EVEX.DDS.LIG.66.0F38.W1 BB /r"    , "ANY AVX512F"],
    ["vfmsub231ss"      , "X:xmm {kz}, xmm, xmm/m32 {er}"            , "RVM"  , "EVEX.DDS.LIG.66.0F38.W0 BB /r"    , "ANY AVX512F"],
    ["vfmsubadd132pd"   , "X:xmm {kz}, xmm, xmm/m128/b64"            , "RVM"  , "EVEX.DDS.128.66.0F38.W1 97 /r"    , "ANY AVX512F-VL"],
    ["vfmsubadd132pd"   , "X:ymm {kz}, ymm, ymm/m256/b64"            , "RVM"  , "EVEX.DDS.256.66.0F38.W1 97 /r"    , "ANY AVX512F-VL"],
    ["vfmsubadd132pd"   , "X:zmm {kz}, zmm, zmm/m512/b64 {er}"       , "RVM"  , "EVEX.DDS.512.66.0F38.W1 97 /r"    , "ANY AVX512F"],
    ["vfmsubadd132ps"   , "X:xmm {kz}, xmm, xmm/m128/b32"            , "RVM"  , "EVEX.DDS.128.66.0F38.W0 97 /r"    , "ANY AVX512F-VL"],
    ["vfmsubadd132ps"   , "X:ymm {kz}, ymm, ymm/m256/b32"            , "RVM"  , "EVEX.DDS.256.66.0F38.W0 97 /r"    , "ANY AVX512F-VL"],
    ["vfmsubadd132ps"   , "X:zmm {kz}, zmm, zmm/m512/b32 {er}"       , "RVM"  , "EVEX.DDS.512.66.0F38.W0 97 /r"    , "ANY AVX512F"],
    ["vfmsubadd213pd"   , "X:xmm {kz}, xmm, xmm/m128/b64"            , "RVM"  , "EVEX.DDS.128.66.0F38.W1 A7 /r"    , "ANY AVX512F-VL"],
    ["vfmsubadd213pd"   , "X:ymm {kz}, ymm, ymm/m256/b64"            , "RVM"  , "EVEX.DDS.256.66.0F38.W1 A7 /r"    , "ANY AVX512F-VL"],
    ["vfmsubadd213pd"   , "X:zmm {kz}, zmm, zmm/m512/b64 {er}"       , "RVM"  , "EVEX.DDS.512.66.0F38.W1 A7 /r"    , "ANY AVX512F"],
    ["vfmsubadd213ps"   , "X:xmm {kz}, xmm, xmm/m128/b32"            , "RVM"  , "EVEX.DDS.128.66.0F38.W0 A7 /r"    , "ANY AVX512F-VL"],
    ["vfmsubadd213ps"   , "X:ymm {kz}, ymm, ymm/m256/b32"            , "RVM"  , "EVEX.DDS.256.66.0F38.W0 A7 /r"    , "ANY AVX512F-VL"],
    ["vfmsubadd213ps"   , "X:zmm {kz}, zmm, zmm/m512/b32 {er}"       , "RVM"  , "EVEX.DDS.512.66.0F38.W0 A7 /r"    , "ANY AVX512F"],
    ["vfmsubadd231pd"   , "X:xmm {kz}, xmm, xmm/m128/b64"            , "RVM"  , "EVEX.DDS.128.66.0F38.W1 B7 /r"    , "ANY AVX512F-VL"],
    ["vfmsubadd231pd"   , "X:ymm {kz}, ymm, ymm/m256/b64"            , "RVM"  , "EVEX.DDS.256.66.0F38.W1 B7 /r"    , "ANY AVX512F-VL"],
    ["vfmsubadd231pd"   , "X:zmm {kz}, zmm, zmm/m512/b64 {er}"       , "RVM"  , "EVEX.DDS.512.66.0F38.W1 B7 /r"    , "ANY AVX512F"],
    ["vfmsubadd231ps"   , "X:xmm {kz}, xmm, xmm/m128/b32"            , "RVM"  , "EVEX.DDS.128.66.0F38.W0 B7 /r"    , "ANY AVX512F-VL"],
    ["vfmsubadd231ps"   , "X:ymm {kz}, ymm, ymm/m256/b32"            , "RVM"  , "EVEX.DDS.256.66.0F38.W0 B7 /r"    , "ANY AVX512F-VL"],
    ["vfmsubadd231ps"   , "X:zmm {kz}, zmm, zmm/m512/b32 {er}"       , "RVM"  , "EVEX.DDS.512.66.0F38.W0 B7 /r"    , "ANY AVX512F"],
    ["vfnmadd132pd"     , "X:xmm {kz}, xmm, xmm/m128/b64"            , "RVM"  , "EVEX.NDS.128.66.0F38.W1 9C /r"    , "ANY AVX512F-VL"],
    ["vfnmadd132pd"     , "X:ymm {kz}, ymm, ymm/m256/b64"            , "RVM"  , "EVEX.NDS.256.66.0F38.W1 9C /r"    , "ANY AVX512F-VL"],
    ["vfnmadd132pd"     , "X:zmm {kz}, zmm, zmm/m512/b64 {er}"       , "RVM"  , "EVEX.NDS.512.66.0F38.W1 9C /r"    , "ANY AVX512F"],
    ["vfnmadd132ps"     , "X:xmm {kz}, xmm, xmm/m128/b32"            , "RVM"  , "EVEX.NDS.128.66.0F38.W0 9C /r"    , "ANY AVX512F-VL"],
    ["vfnmadd132ps"     , "X:ymm {kz}, ymm, ymm/m256/b32"            , "RVM"  , "EVEX.NDS.256.66.0F38.W0 9C /r"    , "ANY AVX512F-VL"],
    ["vfnmadd132ps"     , "X:zmm {kz}, zmm, zmm/m512/b32 {er}"       , "RVM"  , "EVEX.NDS.512.66.0F38.W0 9C /r"    , "ANY AVX512F"],
    ["vfnmadd132sd"     , "X:xmm {kz}, xmm, xmm/m64 {er}"            , "RVM"  , "EVEX.DDS.LIG.66.0F38.W1 9D /r"    , "ANY AVX512F"],
    ["vfnmadd132ss"     , "X:xmm {kz}, xmm, xmm/m32 {er}"            , "RVM"  , "EVEX.DDS.LIG.66.0F38.W0 9D /r"    , "ANY AVX512F"],
    ["vfnmadd213pd"     , "X:xmm {kz}, xmm, xmm/m128/b64"            , "RVM"  , "EVEX.NDS.128.66.0F38.W1 AC /r"    , "ANY AVX512F-VL"],
    ["vfnmadd213pd"     , "X:ymm {kz}, ymm, ymm/m256/b64"            , "RVM"  , "EVEX.NDS.256.66.0F38.W1 AC /r"    , "ANY AVX512F-VL"],
    ["vfnmadd213pd"     , "X:zmm {kz}, zmm, zmm/m512/b64 {er}"       , "RVM"  , "EVEX.NDS.512.66.0F38.W1 AC /r"    , "ANY AVX512F"],
    ["vfnmadd213ps"     , "X:xmm {kz}, xmm, xmm/m128/b32"            , "RVM"  , "EVEX.NDS.128.66.0F38.W0 AC /r"    , "ANY AVX512F-VL"],
    ["vfnmadd213ps"     , "X:ymm {kz}, ymm, ymm/m256/b32"            , "RVM"  , "EVEX.NDS.256.66.0F38.W0 AC /r"    , "ANY AVX512F-VL"],
    ["vfnmadd213ps"     , "X:zmm {kz}, zmm, zmm/m512/b32 {er}"       , "RVM"  , "EVEX.NDS.512.66.0F38.W0 AC /r"    , "ANY AVX512F"],
    ["vfnmadd213sd"     , "X:xmm {kz}, xmm, xmm/m64 {er}"            , "RVM"  , "EVEX.DDS.LIG.66.0F38.W1 AD /r"    , "ANY AVX512F"],
    ["vfnmadd213ss"     , "X:xmm {kz}, xmm, xmm/m32 {er}"            , "RVM"  , "EVEX.DDS.LIG.66.0F38.W0 AD /r"    , "ANY AVX512F"],
    ["vfnmadd231pd"     , "X:xmm {kz}, xmm, xmm/m128/b64"            , "RVM"  , "EVEX.NDS.128.66.0F38.W1 BC /r"    , "ANY AVX512F-VL"],
    ["vfnmadd231pd"     , "X:ymm {kz}, ymm, ymm/m256/b64"            , "RVM"  , "EVEX.NDS.256.66.0F38.W1 BC /r"    , "ANY AVX512F-VL"],
    ["vfnmadd231pd"     , "X:zmm {kz}, zmm, zmm/m512/b64 {er}"       , "RVM"  , "EVEX.NDS.512.66.0F38.W1 BC /r"    , "ANY AVX512F"],
    ["vfnmadd231ps"     , "X:xmm {kz}, xmm, xmm/m128/b32"            , "RVM"  , "EVEX.NDS.128.66.0F38.W0 BC /r"    , "ANY AVX512F-VL"],
    ["vfnmadd231ps"     , "X:ymm {kz}, ymm, ymm/m256/b32"            , "RVM"  , "EVEX.NDS.256.66.0F38.W0 BC /r"    , "ANY AVX512F-VL"],
    ["vfnmadd231ps"     , "X:zmm {kz}, zmm, zmm/m512/b32 {er}"       , "RVM"  , "EVEX.NDS.512.66.0F38.W0 BC /r"    , "ANY AVX512F"],
    ["vfnmadd231sd"     , "X:xmm {kz}, xmm, xmm/m64 {er}"            , "RVM"  , "EVEX.DDS.LIG.66.0F38.W1 BD /r"    , "ANY AVX512F"],
    ["vfnmadd231ss"     , "X:xmm {kz}, xmm, xmm/m32 {er}"            , "RVM"  , "EVEX.DDS.LIG.66.0F38.W0 BD /r"    , "ANY AVX512F"],
    ["vfnmsub132pd"     , "X:xmm {kz}, xmm, xmm/m128/b64"            , "RVM"  , "EVEX.NDS.128.66.0F38.W1 9E /r"    , "ANY AVX512F-VL"],
    ["vfnmsub132pd"     , "X:ymm {kz}, ymm, ymm/m256/b64"            , "RVM"  , "EVEX.NDS.256.66.0F38.W1 9E /r"    , "ANY AVX512F-VL"],
    ["vfnmsub132pd"     , "X:zmm {kz}, zmm, zmm/m512/b64 {er}"       , "RVM"  , "EVEX.NDS.512.66.0F38.W1 9E /r"    , "ANY AVX512F"],
    ["vfnmsub132ps"     , "X:xmm {kz}, xmm, xmm/m128/b32"            , "RVM"  , "EVEX.NDS.128.66.0F38.W0 9E /r"    , "ANY AVX512F-VL"],
    ["vfnmsub132ps"     , "X:ymm {kz}, ymm, ymm/m256/b32"            , "RVM"  , "EVEX.NDS.256.66.0F38.W0 9E /r"    , "ANY AVX512F-VL"],
    ["vfnmsub132ps"     , "X:zmm {kz}, zmm, zmm/m512/b32 {er}"       , "RVM"  , "EVEX.NDS.512.66.0F38.W0 9E /r"    , "ANY AVX512F"],
    ["vfnmsub132sd"     , "X:xmm {kz}, xmm, xmm/m64 {er}"            , "RVM"  , "EVEX.DDS.LIG.66.0F38.W1 9F /r"    , "ANY AVX512F"],
    ["vfnmsub132ss"     , "X:xmm {kz}, xmm, xmm/m32 {er}"            , "RVM"  , "EVEX.DDS.LIG.66.0F38.W0 9F /r"    , "ANY AVX512F"],
    ["vfnmsub213pd"     , "X:xmm {kz}, xmm, xmm/m128/b64"            , "RVM"  , "EVEX.NDS.128.66.0F38.W1 AE /r"    , "ANY AVX512F-VL"],
    ["vfnmsub213pd"     , "X:ymm {kz}, ymm, ymm/m256/b64"            , "RVM"  , "EVEX.NDS.256.66.0F38.W1 AE /r"    , "ANY AVX512F-VL"],
    ["vfnmsub213pd"     , "X:zmm {kz}, zmm, zmm/m512/b64 {er}"       , "RVM"  , "EVEX.NDS.512.66.0F38.W1 AE /r"    , "ANY AVX512F"],
    ["vfnmsub213ps"     , "X:xmm {kz}, xmm, xmm/m128/b32"            , "RVM"  , "EVEX.NDS.128.66.0F38.W0 AE /r"    , "ANY AVX512F-VL"],
    ["vfnmsub213ps"     , "X:ymm {kz}, ymm, ymm/m256/b32"            , "RVM"  , "EVEX.NDS.256.66.0F38.W0 AE /r"    , "ANY AVX512F-VL"],
    ["vfnmsub213ps"     , "X:zmm {kz}, zmm, zmm/m512/b32 {er}"       , "RVM"  , "EVEX.NDS.512.66.0F38.W0 AE /r"    , "ANY AVX512F"],
    ["vfnmsub213sd"     , "X:xmm {kz}, xmm, xmm/m64 {er}"            , "RVM"  , "EVEX.DDS.LIG.66.0F38.W1 AF /r"    , "ANY AVX512F"],
    ["vfnmsub213ss"     , "X:xmm {kz}, xmm, xmm/m32 {er}"            , "RVM"  , "EVEX.DDS.LIG.66.0F38.W0 AF /r"    , "ANY AVX512F"],
    ["vfnmsub231pd"     , "X:xmm {kz}, xmm, xmm/m128/b64"            , "RVM"  , "EVEX.NDS.128.66.0F38.W1 BE /r"    , "ANY AVX512F-VL"],
    ["vfnmsub231pd"     , "X:ymm {kz}, ymm, ymm/m256/b64"            , "RVM"  , "EVEX.NDS.256.66.0F38.W1 BE /r"    , "ANY AVX512F-VL"],
    ["vfnmsub231pd"     , "X:zmm {kz}, zmm, zmm/m512/b64 {er}"       , "RVM"  , "EVEX.NDS.512.66.0F38.W1 BE /r"    , "ANY AVX512F"],
    ["vfnmsub231ps"     , "X:xmm {kz}, xmm, xmm/m128/b32"            , "RVM"  , "EVEX.NDS.128.66.0F38.W0 BE /r"    , "ANY AVX512F-VL"],
    ["vfnmsub231ps"     , "X:ymm {kz}, ymm, ymm/m256/b32"            , "RVM"  , "EVEX.NDS.256.66.0F38.W0 BE /r"    , "ANY AVX512F-VL"],
    ["vfnmsub231ps"     , "X:zmm {kz}, zmm, zmm/m512/b32 {er}"       , "RVM"  , "EVEX.NDS.512.66.0F38.W0 BE /r"    , "ANY AVX512F"],
    ["vfnmsub231sd"     , "X:xmm {kz}, xmm, xmm/m64 {er}"            , "RVM"  , "EVEX.DDS.LIG.66.0F38.W1 BF /r"    , "ANY AVX512F"],
    ["vfnmsub231ss"     , "X:xmm {kz}, xmm, xmm/m32 {er}"            , "RVM"  , "EVEX.DDS.LIG.66.0F38.W0 BF /r"    , "ANY AVX512F"],
    ["vfpclasspd"       , "W:k {k}, xmm/m128/b64, ib"                , "RMI"  , "EVEX.128.66.0F3A.W1 66 /r ib"     , "ANY AVX512DQ-VL"],
    ["vfpclasspd"       , "W:k {k}, ymm/m256/b64, ib"                , "RMI"  , "EVEX.256.66.0F3A.W1 66 /r ib"     , "ANY AVX512DQ-VL"],
    ["vfpclasspd"       , "W:k {k}, zmm/m512/b64, ib"                , "RMI"  , "EVEX.512.66.0F3A.W1 66 /r ib"     , "ANY AVX512DQ"],
    ["vfpclassps"       , "W:k {k}, xmm/m128/b32, ib"                , "RMI"  , "EVEX.128.66.0F3A.W0 66 /r ib"     , "ANY AVX512DQ-VL"],
    ["vfpclassps"       , "W:k {k}, ymm/m256/b32, ib"                , "RMI"  , "EVEX.256.66.0F3A.W0 66 /r ib"     , "ANY AVX512DQ-VL"],
    ["vfpclassps"       , "W:k {k}, zmm/m512/b32, ib"                , "RMI"  , "EVEX.512.66.0F3A.W0 66 /r ib"     , "ANY AVX512DQ"],
    ["vfpclasssd"       , "W:k {k}, xmm/m64, ib"                     , "RMI"  , "EVEX.LIG.66.0F3A.W1 67 /r ib"     , "ANY AVX512DQ"],
    ["vfpclassss"       , "W:k {k}, xmm/m32, ib"                     , "RMI"  , "EVEX.LIG.66.0F3A.W0 67 /r ib"     , "ANY AVX512DQ"],
    ["vgatherpf0dpd"    , "R:vm32y {k}"                              , "M"    , "EVEX.512.66.0F38.W1 C6 /1"        , "ANY AVX512PF"],
    ["vgatherpf0dps"    , "R:vm32z {k}"                              , "M"    , "EVEX.512.66.0F38.W0 C6 /1"        , "ANY AVX512PF"],
    ["vgatherpf0qpd"    , "R:vm64z {k}"                              , "M"    , "EVEX.512.66.0F38.W1 C7 /1"        , "ANY AVX512PF"],
    ["vgatherpf0qps"    , "R:vm64z {k}"                              , "M"    , "EVEX.512.66.0F38.W0 C7 /1"        , "ANY AVX512PF"],
    ["vgatherpf1dpd"    , "R:vm32y {k}"                              , "M"    , "EVEX.512.66.0F38.W1 C6 /2"        , "ANY AVX512PF"],
    ["vgatherpf1dps"    , "R:vm32z {k}"                              , "M"    , "EVEX.512.66.0F38.W0 C6 /2"        , "ANY AVX512PF"],
    ["vgatherpf1qpd"    , "R:vm64z {k}"                              , "M"    , "EVEX.512.66.0F38.W1 C7 /2"        , "ANY AVX512PF"],
    ["vgatherpf1qps"    , "R:vm64z {k}"                              , "M"    , "EVEX.512.66.0F38.W0 C7 /2"        , "ANY AVX512PF"],
    ["vgetexppd"        , "W:xmm {kz}, xmm/m128/b64"                 , "RM"   , "EVEX.128.66.0F38.W1 42 /r"        , "ANY AVX512F-VL"],
    ["vgetexppd"        , "W:ymm {kz}, ymm/m256/b64"                 , "RM"   , "EVEX.256.66.0F38.W1 42 /r"        , "ANY AVX512F-VL"],
    ["vgetexppd"        , "W:zmm {kz}, zmm/m512/b64 {sae}"           , "RM"   , "EVEX.512.66.0F38.W1 42 /r"        , "ANY AVX512F"],
    ["vgetexpps"        , "W:xmm {kz}, xmm/m128/b32"                 , "RM"   , "EVEX.128.66.0F38.W0 42 /r"        , "ANY AVX512F-VL"],
    ["vgetexpps"        , "W:ymm {kz}, ymm/m256/b32"                 , "RM"   , "EVEX.256.66.0F38.W0 42 /r"        , "ANY AVX512F-VL"],
    ["vgetexpps"        , "W:zmm {kz}, zmm/m512/b32 {sae}"           , "RM"   , "EVEX.512.66.0F38.W0 42 /r"        , "ANY AVX512F"],
    ["vgetexpsd"        , "W:xmm {kz}, xmm/m64 {sae}"                , "RM"   , "EVEX.NDS.LIG.66.0F38.W1 43 /r"    , "ANY AVX512F"],
    ["vgetexpss"        , "W:xmm {kz}, xmm/m32 {sae}"                , "RM"   , "EVEX.NDS.LIG.66.0F38.W0 43 /r"    , "ANY AVX512F"],
    ["vgetmantpd"       , "W:xmm {kz}, xmm/m128/b64, ib"             , "RMI"  , "EVEX.128.66.0F3A.W1 26 /r ib"     , "ANY AVX512F-VL"],
    ["vgetmantpd"       , "W:ymm {kz}, ymm/m256/b64, ib"             , "RMI"  , "EVEX.256.66.0F3A.W1 26 /r ib"     , "ANY AVX512F-VL"],
    ["vgetmantpd"       , "W:zmm {kz}, zmm/m512/b64, ib {sae}"       , "RMI"  , "EVEX.512.66.0F3A.W1 26 /r ib"     , "ANY AVX512F"],
    ["vgetmantps"       , "W:xmm {kz}, xmm/m128/b32, ib"             , "RMI"  , "EVEX.128.66.0F3A.W0 26 /r ib"     , "ANY AVX512F-VL"],
    ["vgetmantps"       , "W:ymm {kz}, ymm/m256/b32, ib"             , "RMI"  , "EVEX.256.66.0F3A.W0 26 /r ib"     , "ANY AVX512F-VL"],
    ["vgetmantps"       , "W:zmm {kz}, zmm/m512/b32, ib {sae}"       , "RMI"  , "EVEX.512.66.0F3A.W0 26 /r ib"     , "ANY AVX512F"],
    ["vgetmantsd"       , "W:xmm {kz}, xmm/m64, ib {sae}"            , "RMI"  , "EVEX.NDS.LIG.66.0F3A.W1 27 /r ib" , "ANY AVX512F"],
    ["vgetmantss"       , "W:xmm {kz}, xmm/m32, ib {sae}"            , "RMI"  , "EVEX.NDS.LIG.66.0F3A.W0 27 /r ib" , "ANY AVX512F"],
    ["vinsertf32x4"     , "W:ymm {kz}, ymm, xmm/m128, ib"            , "RVMI" , "EVEX.NDS.256.66.0F3A.W0 18 /r ib" , "ANY AVX512F-VL"],
    ["vinsertf32x4"     , "W:zmm {kz}, zmm, xmm/m128, ib"            , "RVMI" , "EVEX.NDS.512.66.0F3A.W0 18 /r ib" , "ANY AVX512F"],
    ["vinsertf32x8"     , "W:zmm {kz}, zmm, ymm/m256, ib"            , "RVMI" , "EVEX.NDS.512.66.0F3A.W0 1A /r ib" , "ANY AVX512DQ"],
    ["vinsertf64x2"     , "W:ymm {kz}, ymm, xmm/m128, ib"            , "RVMI" , "EVEX.NDS.256.66.0F3A.W1 18 /r ib" , "ANY AVX512DQ-VL"],
    ["vinsertf64x2"     , "W:zmm {kz}, zmm, xmm/m128, ib"            , "RVMI" , "EVEX.NDS.512.66.0F3A.W1 18 /r ib" , "ANY AVX512DQ"],
    ["vinsertf64x4"     , "W:zmm {kz}, zmm, ymm/m256, ib"            , "RVMI" , "EVEX.NDS.512.66.0F3A.W1 1A /r ib" , "ANY AVX512F"],
    ["vinserti32x4"     , "W:ymm {kz}, ymm, xmm/m128, ib"            , "RVMI" , "EVEX.NDS.256.66.0F3A.W0 38 /r ib" , "ANY AVX512F-VL"],
    ["vinserti32x4"     , "W:zmm {kz}, zmm, xmm/m128, ib"            , "RVMI" , "EVEX.NDS.512.66.0F3A.W0 38 /r ib" , "ANY AVX512F"],
    ["vinserti32x8"     , "W:zmm {kz}, zmm, ymm/m256, ib"            , "RVMI" , "EVEX.NDS.512.66.0F3A.W0 3A /r ib" , "ANY AVX512DQ"],
    ["vinserti64x2"     , "W:ymm {kz}, ymm, xmm/m128, ib"            , "RVMI" , "EVEX.NDS.256.66.0F3A.W1 38 /r ib" , "ANY AVX512DQ-VL"],
    ["vinserti64x2"     , "W:zmm {kz}, zmm, xmm/m128, ib"            , "RVMI" , "EVEX.NDS.512.66.0F3A.W1 38 /r ib" , "ANY AVX512DQ"],
    ["vinserti64x4"     , "W:zmm {kz}, zmm, ymm/m256, ib"            , "RVMI" , "EVEX.NDS.512.66.0F3A.W1 3A /r ib" , "ANY AVX512F"],
    ["vinsertps"        , "W:xmm, xmm, xmm/m32, ib"                  , "RVMI" , "EVEX.NDS.128.66.0F3A.W0 21 /r ib" , "ANY AVX512F"],
    ["vmaxpd"           , "W:xmm {kz}, xmm, xmm/m128/b64"            , "RVM"  , "EVEX.NDS.128.66.0F.W1 5F /r"      , "ANY AVX512F-VL"],
    ["vmaxpd"           , "W:ymm {kz}, ymm, ymm/m256/b64"            , "RVM"  , "EVEX.NDS.256.66.0F.W1 5F /r"      , "ANY AVX512F-VL"],
    ["vmaxpd"           , "W:zmm {kz}, zmm, zmm/m512/b64 {sae}"      , "RVM"  , "EVEX.NDS.512.66.0F.W1 5F /r"      , "ANY AVX512F"],
    ["vmaxps"           , "W:xmm {kz}, xmm, xmm/m128/b32"            , "RVM"  , "EVEX.NDS.128.0F.W0 5F /r"         , "ANY AVX512F-VL"],
    ["vmaxps"           , "W:ymm {kz}, ymm, ymm/m256/b32"            , "RVM"  , "EVEX.NDS.256.0F.W0 5F /r"         , "ANY AVX512F-VL"],
    ["vmaxps"           , "W:zmm {kz}, zmm, zmm/m512/b32 {sae}"      , "RVM"  , "EVEX.NDS.512.0F.W0 5F /r"         , "ANY AVX512F"],
    ["vmaxsd"           , "W:xmm {kz}, xmm, xmm/m64 {sae}"           , "RVM"  , "EVEX.NDS.LIG.F2.0F.W1 5F /r"      , "ANY AVX512F-VL"],
    ["vmaxss"           , "W:xmm {kz}, xmm, xmm/m32 {sae}"           , "RVM"  , "EVEX.NDS.LIG.F3.0F.W0 5F /r"      , "ANY AVX512F-VL"],
    ["vminpd"           , "W:xmm {kz}, xmm, xmm/m128/b64"            , "RVM"  , "EVEX.NDS.128.66.0F.W1 5D /r"      , "ANY AVX512F-VL"],
    ["vminpd"           , "W:ymm {kz}, ymm, ymm/m256/b64"            , "RVM"  , "EVEX.NDS.256.66.0F.W1 5D /r"      , "ANY AVX512F-VL"],
    ["vminpd"           , "W:zmm {kz}, zmm, zmm/m512/b64 {sae}"      , "RVM"  , "EVEX.NDS.512.66.0F.W1 5D /r"      , "ANY AVX512F"],
    ["vminps"           , "W:xmm {kz}, xmm, xmm/m128/b32"            , "RVM"  , "EVEX.NDS.128.0F.W0 5D /r"         , "ANY AVX512F-VL"],
    ["vminps"           , "W:ymm {kz}, ymm, ymm/m256/b32"            , "RVM"  , "EVEX.NDS.256.0F.W0 5D /r"         , "ANY AVX512F-VL"],
    ["vminps"           , "W:zmm {kz}, zmm, zmm/m512/b32 {sae}"      , "RVM"  , "EVEX.NDS.512.0F.W0 5D /r"         , "ANY AVX512F"],
    ["vminsd"           , "W:xmm {kz}, xmm, xmm/m64 {sae}"           , "RVM"  , "EVEX.NDS.LIG.F2.0F.W1 5D /r"      , "ANY AVX512F-VL"],
    ["vminss"           , "W:xmm {kz}, xmm, xmm/m32 {sae}"           , "RVM"  , "EVEX.NDS.LIG.F3.0F.W0 5D /r"      , "ANY AVX512F-VL"],
    ["vmovapd"          , "W:xmm {kz}, xmm/m128"                     , "RM"   , "EVEX.128.66.0F.W1 28 /r"          , "ANY AVX512F-VL"],
    ["vmovapd"          , "W:xmm/m128 {kz}, xmm"                     , "MR"   , "EVEX.128.66.0F.W1 29 /r"          , "ANY AVX512F-VL"],
    ["vmovapd"          , "W:ymm {kz}, ymm/m256"                     , "RM"   , "EVEX.256.66.0F.W1 28 /r"          , "ANY AVX512F-VL"],
    ["vmovapd"          , "W:ymm/m256 {kz}, ymm"                     , "MR"   , "EVEX.256.66.0F.W1 29 /r"          , "ANY AVX512F-VL"],
    ["vmovapd"          , "W:zmm {kz}, zmm/m512"                     , "RM"   , "EVEX.512.66.0F.W1 28 /r"          , "ANY AVX512F"],
    ["vmovapd"          , "W:zmm/m512 {kz}, zmm"                     , "MR"   , "EVEX.512.66.0F.W1 29 /r"          , "ANY AVX512F"],
    ["vmovaps"          , "W:xmm {kz}, xmm/m128"                     , "RM"   , "EVEX.128.0F.W0 28 /r"             , "ANY AVX512F-VL"],
    ["vmovaps"          , "W:xmm/m128 {kz}, xmm"                     , "MR"   , "EVEX.128.0F.W0 29 /r"             , "ANY AVX512F-VL"],
    ["vmovaps"          , "W:ymm {kz}, ymm/m256"                     , "RM"   , "EVEX.256.0F.W0 28 /r"             , "ANY AVX512F-VL"],
    ["vmovaps"          , "W:ymm/m256 {kz}, ymm"                     , "MR"   , "EVEX.256.0F.W0 29 /r"             , "ANY AVX512F-VL"],
    ["vmovaps"          , "W:zmm {kz}, zmm/m512"                     , "RM"   , "EVEX.512.0F.W0 28 /r"             , "ANY AVX512F"],
    ["vmovaps"          , "W:zmm/m512 {kz}, zmm"                     , "MR"   , "EVEX.512.0F.W0 29 /r"             , "ANY AVX512F"],
    ["vmovd"            , "W:r32/m32, xmm"                           , "MR"   , "EVEX.128.66.0F.W0 7E /r"          , "ANY AVX512F"],
    ["vmovd"            , "W:xmm, r32/m32"                           , "RM"   , "EVEX.128.66.0F.W0 6E /r"          , "ANY AVX512F"],
    ["vmovddup"         , "W:xmm {kz}, xmm/m64"                      , "RM"   , "EVEX.128.F2.0F.W1 12 /r"          , "ANY AVX512F-VL"],
    ["vmovddup"         , "W:ymm {kz}, ymm/m256"                     , "RM"   , "EVEX.256.F2.0F.W1 12 /r"          , "ANY AVX512F-VL"],
    ["vmovddup"         , "W:zmm {kz}, zmm/m512"                     , "RM"   , "EVEX.512.F2.0F.W1 12 /r"          , "ANY AVX512F"],
    ["vmovdqa32"        , "W:xmm {kz}, xmm/m128"                     , "RM"   , "EVEX.128.66.0F.W0 6F /r"          , "ANY AVX512F-VL"],
    ["vmovdqa32"        , "W:xmm/m128 {kz}, xmm"                     , "MR"   , "EVEX.128.66.0F.W0 7F /r"          , "ANY AVX512F-VL"],
    ["vmovdqa32"        , "W:ymm {kz}, ymm/m256"                     , "RM"   , "EVEX.256.66.0F.W0 6F /r"          , "ANY AVX512F-VL"],
    ["vmovdqa32"        , "W:ymm/m256 {kz}, ymm"                     , "MR"   , "EVEX.256.66.0F.W0 7F /r"          , "ANY AVX512F-VL"],
    ["vmovdqa32"        , "W:zmm {kz}, zmm/m512"                     , "RM"   , "EVEX.512.66.0F.W0 6F /r"          , "ANY AVX512F"],
    ["vmovdqa32"        , "W:zmm/m512 {kz}, zmm"                     , "MR"   , "EVEX.512.66.0F.W0 7F /r"          , "ANY AVX512F"],
    ["vmovdqa64"        , "W:xmm {kz}, xmm/m128"                     , "RM"   , "EVEX.128.66.0F.W1 6F /r"          , "ANY AVX512F-VL"],
    ["vmovdqa64"        , "W:xmm/m128 {kz}, xmm"                     , "MR"   , "EVEX.128.66.0F.W1 7F /r"          , "ANY AVX512F-VL"],
    ["vmovdqa64"        , "W:ymm {kz}, ymm/m256"                     , "RM"   , "EVEX.256.66.0F.W1 6F /r"          , "ANY AVX512F-VL"],
    ["vmovdqa64"        , "W:ymm/m256 {kz}, ymm"                     , "MR"   , "EVEX.256.66.0F.W1 7F /r"          , "ANY AVX512F-VL"],
    ["vmovdqa64"        , "W:zmm {kz}, zmm/m512"                     , "RM"   , "EVEX.512.66.0F.W1 6F /r"          , "ANY AVX512F"],
    ["vmovdqa64"        , "W:zmm/m512 {kz}, zmm"                     , "MR"   , "EVEX.512.66.0F.W1 7F /r"          , "ANY AVX512F"],
    ["vmovdqu16"        , "W:xmm {kz}, xmm/m128"                     , "RM"   , "EVEX.128.F2.0F.W1 6F /r"          , "ANY AVX512BW-VL"],
    ["vmovdqu16"        , "W:xmm/m128 {kz}, xmm"                     , "MR"   , "EVEX.128.F2.0F.W1 7F /r"          , "ANY AVX512BW-VL"],
    ["vmovdqu16"        , "W:ymm {kz}, ymm/m256"                     , "RM"   , "EVEX.256.F2.0F.W1 6F /r"          , "ANY AVX512BW-VL"],
    ["vmovdqu16"        , "W:ymm/m256 {kz}, ymm"                     , "MR"   , "EVEX.256.F2.0F.W1 7F /r"          , "ANY AVX512BW-VL"],
    ["vmovdqu16"        , "W:zmm {kz}, zmm/m512"                     , "RM"   , "EVEX.512.F2.0F.W1 6F /r"          , "ANY AVX512BW"],
    ["vmovdqu16"        , "W:zmm/m512 {kz}, zmm"                     , "MR"   , "EVEX.512.F2.0F.W1 7F /r"          , "ANY AVX512BW"],
    ["vmovdqu32"        , "W:xmm {kz}, xmm/m128"                     , "RM"   , "EVEX.128.F3.0F.W0 6F /r"          , "ANY AVX512F-VL"],
    ["vmovdqu32"        , "W:xmm/m128 {kz}, xmm"                     , "MR"   , "EVEX.128.F3.0F.W0 7F /r"          , "ANY AVX512F-VL"],
    ["vmovdqu32"        , "W:ymm {kz}, ymm/m256"                     , "RM"   , "EVEX.256.F3.0F.W0 6F /r"          , "ANY AVX512F-VL"],
    ["vmovdqu32"        , "W:ymm/m256 {kz}, ymm"                     , "MR"   , "EVEX.256.F3.0F.W0 7F /r"          , "ANY AVX512F-VL"],
    ["vmovdqu32"        , "W:zmm {kz}, zmm/m512"                     , "RM"   , "EVEX.512.F3.0F.W0 6F /r"          , "ANY AVX512F"],
    ["vmovdqu32"        , "W:zmm/m512 {kz}, zmm"                     , "MR"   , "EVEX.512.F3.0F.W0 7F /r"          , "ANY AVX512F"],
    ["vmovdqu64"        , "W:xmm {kz}, xmm/m128"                     , "RM"   , "EVEX.128.F3.0F.W1 6F /r"          , "ANY AVX512F-VL"],
    ["vmovdqu64"        , "W:xmm/m128 {kz}, xmm"                     , "MR"   , "EVEX.128.F3.0F.W1 7F /r"          , "ANY AVX512F-VL"],
    ["vmovdqu64"        , "W:ymm {kz}, ymm/m256"                     , "RM"   , "EVEX.256.F3.0F.W1 6F /r"          , "ANY AVX512F-VL"],
    ["vmovdqu64"        , "W:ymm/m256 {kz}, ymm"                     , "MR"   , "EVEX.256.F3.0F.W1 7F /r"          , "ANY AVX512F-VL"],
    ["vmovdqu64"        , "W:zmm {kz}, zmm/m512"                     , "RM"   , "EVEX.512.F3.0F.W1 6F /r"          , "ANY AVX512F"],
    ["vmovdqu64"        , "W:zmm/m512 {kz}, zmm"                     , "MR"   , "EVEX.512.F3.0F.W1 7F /r"          , "ANY AVX512F"],
    ["vmovdqu8"         , "W:xmm {kz}, xmm/m128"                     , "RM"   , "EVEX.128.F2.0F.W0 6F /r"          , "ANY AVX512BW-VL"],
    ["vmovdqu8"         , "W:xmm/m128 {kz}, xmm"                     , "MR"   , "EVEX.128.F2.0F.W0 7F /r"          , "ANY AVX512BW-VL"],
    ["vmovdqu8"         , "W:ymm {kz}, ymm/m256"                     , "RM"   , "EVEX.256.F2.0F.W0 6F /r"          , "ANY AVX512BW-VL"],
    ["vmovdqu8"         , "W:ymm/m256 {kz}, ymm"                     , "MR"   , "EVEX.256.F2.0F.W0 7F /r"          , "ANY AVX512BW-VL"],
    ["vmovdqu8"         , "W:zmm {kz}, zmm/m512"                     , "RM"   , "EVEX.512.F2.0F.W0 6F /r"          , "ANY AVX512BW"],
    ["vmovdqu8"         , "W:zmm/m512 {kz}, zmm"                     , "MR"   , "EVEX.512.F2.0F.W0 7F /r"          , "ANY AVX512BW"],
    ["vmovhlps"         , "W:xmm, xmm, xmm"                          , "RVM"  , "EVEX.NDS.128.0F.W0 12 /r"         , "ANY AVX512F"],
    ["vmovhpd"          , "W:m64, xmm"                               , "MR"   , "EVEX.128.66.0F.W1 17 /r"          , "ANY AVX512F"],
    ["vmovhpd"          , "W:xmm, xmm, m64"                          , "RVM"  , "EVEX.NDS.128.66.0F.W1 16 /r"      , "ANY AVX512F"],
    ["vmovhps"          , "W:m64, xmm"                               , "MR"   , "EVEX.128.0F.W0 17 /r"             , "ANY AVX512F"],
    ["vmovhps"          , "W:xmm, xmm, m64"                          , "RVM"  , "EVEX.NDS.128.0F.W0 16 /r"         , "ANY AVX512F"],
    ["vmovlhps"         , "W:xmm, xmm, xmm"                          , "RVM"  , "EVEX.NDS.128.0F.W0 16 /r"         , "ANY AVX512F"],
    ["vmovlpd"          , "W:m64, xmm"                               , "MR"   , "EVEX.128.66.0F.W1 13 /r"          , "ANY AVX512F"],
    ["vmovlpd"          , "W:xmm, xmm, m64"                          , "RVM"  , "EVEX.NDS.128.66.0F.W1 12 /r"      , "ANY AVX512F"],
    ["vmovlps"          , "W:m64, xmm"                               , "MR"   , "EVEX.128.0F.W0 13 /r"             , "ANY AVX512F"],
    ["vmovlps"          , "W:xmm, xmm, m64"                          , "RVM"  , "EVEX.NDS.128.0F.W0 12 /r"         , "ANY AVX512F"],
    ["vmovntdq"         , "W:m128, xmm"                              , "MR"   , "EVEX.128.66.0F.W0 E7 /r"          , "ANY AVX512F-VL"],
    ["vmovntdq"         , "W:m256, ymm"                              , "MR"   , "EVEX.256.66.0F.W0 E7 /r"          , "ANY AVX512F-VL"],
    ["vmovntdq"         , "W:m512, zmm"                              , "MR"   , "EVEX.512.66.0F.W0 E7 /r"          , "ANY AVX512F"],
    ["vmovntdqa"        , "W:xmm, m128"                              , "RM"   , "EVEX.128.66.0F38.W0 2A /r"        , "ANY AVX512F-VL"],
    ["vmovntdqa"        , "W:ymm, m256"                              , "RM"   , "EVEX.256.66.0F38.W0 2A /r"        , "ANY AVX512F-VL"],
    ["vmovntdqa"        , "W:zmm, m512"                              , "RM"   , "EVEX.512.66.0F38.W0 2A /r"        , "ANY AVX512F"],
    ["vmovntpd"         , "W:m128, xmm"                              , "MR"   , "EVEX.128.66.0F.W1 2B /r"          , "ANY AVX512F-VL"],
    ["vmovntpd"         , "W:m256, ymm"                              , "MR"   , "EVEX.256.66.0F.W1 2B /r"          , "ANY AVX512F-VL"],
    ["vmovntpd"         , "W:m512, zmm"                              , "MR"   , "EVEX.512.66.0F.W1 2B /r"          , "ANY AVX512F"],
    ["vmovntps"         , "W:m128, xmm"                              , "MR"   , "EVEX.128.66.0F.W0 2B /r"          , "ANY AVX512F-VL"],
    ["vmovntps"         , "W:m256, ymm"                              , "MR"   , "EVEX.256.66.0F.W0 2B /r"          , "ANY AVX512F-VL"],
    ["vmovntps"         , "W:m512, zmm"                              , "MR"   , "EVEX.512.0F.W0 2B /r"             , "ANY AVX512F"],
    ["vmovq"            , "W:r64/m64, xmm"                           , "MR"   , "EVEX.128.66.0F.W1 7E /r"          , "X64 AVX512F"],
    ["vmovq"            , "W:xmm, r64/m64"                           , "RM"   , "EVEX.128.66.0F.W1 6E /r"          , "X64 AVX512F"],
    ["vmovq"            , "W:xmm, xmm/m64"                           , "RM"   , "EVEX.128.F3.0F.W1 7E /r"          , "ANY AVX512F"],
    ["vmovq"            , "W:xmm/m64, xmm"                           , "MR"   , "EVEX.128.66.0F.W1 D6 /r"          , "ANY AVX512F"],
    ["vmovsd"           , "W:m64, xmm"                               , "MR"   , "EVEX.LIG.F2.0F.W1 11 /r"          , "ANY AVX512F"],
    ["vmovsd"           , "W:xmm {kz}, m64"                          , "MR"   , "EVEX.LIG.F2.0F.W1 10 /r"          , "ANY AVX512F"],
    ["vmovsd"           , "W:xmm {kz}, xmm, xmm"                     , "MVR"  , "EVEX.NDS.LIG.F2.0F.W1 11 /r"      , "ANY AVX512F"],
    ["vmovsd"           , "W:xmm {kz}, xmm, xmm"                     , "RVM"  , "EVEX.NDS.LIG.F2.0F.W1 10 /r"      , "ANY AVX512F"],
    ["vmovshdup"        , "W:xmm {kz}, xmm/m128"                     , "RVM"  , "EVEX.128.F3.0F.W0 16 /r"          , "ANY AVX512F-VL"],
    ["vmovshdup"        , "W:ymm {kz}, ymm/m256"                     , "RVM"  , "EVEX.256.F3.0F.W0 16 /r"          , "ANY AVX512F-VL"],
    ["vmovshdup"        , "W:zmm {kz}, zmm/m512"                     , "RVM"  , "EVEX.512.F3.0F.W0 16 /r"          , "ANY AVX512F"],
    ["vmovsldup"        , "W:xmm {kz}, xmm/m128"                     , "RVM"  , "EVEX.128.F3.0F.W0 12 /r"          , "ANY AVX512F-VL"],
    ["vmovsldup"        , "W:ymm {kz}, ymm/m256"                     , "RVM"  , "EVEX.256.F3.0F.W0 12 /r"          , "ANY AVX512F-VL"],
    ["vmovsldup"        , "W:zmm {kz}, zmm/m512"                     , "RVM"  , "EVEX.512.F3.0F.W0 12 /r"          , "ANY AVX512F"],
    ["vmovss"           , "W:m64, xmm"                               , "MR"   , "EVEX.LIG.F3.0F.W0 11 /r"          , "ANY AVX512F"],
    ["vmovss"           , "W:xmm {kz}, m64"                          , "MR"   , "EVEX.LIG.F3.0F.W0 10 /r"          , "ANY AVX512F"],
    ["vmovss"           , "W:xmm {kz}, xmm, xmm"                     , "MVR"  , "EVEX.NDS.LIG.F3.0F.W0 11 /r"      , "ANY AVX512F"],
    ["vmovss"           , "W:xmm {kz}, xmm, xmm"                     , "RVM"  , "EVEX.NDS.LIG.F3.0F.W0 10 /r"      , "ANY AVX512F"],
    ["vmovupd"          , "W:xmm {kz}, xmm/m128"                     , "RM"   , "EVEX.128.66.0F.W1 10 /r"          , "ANY AVX512F-VL"],
    ["vmovupd"          , "W:xmm/m128 {kz}, xmm"                     , "MR"   , "EVEX.128.66.0F.W1 11 /r"          , "ANY AVX512F-VL"],
    ["vmovupd"          , "W:ymm {kz}, ymm/m256"                     , "RM"   , "EVEX.256.66.0F.W1 11 /r"          , "ANY AVX512F-VL"],
    ["vmovupd"          , "W:ymm/m256 {kz}, ymm"                     , "MR"   , "EVEX.256.66.0F.W1 10 /r"          , "ANY AVX512F-VL"],
    ["vmovupd"          , "W:zmm {kz}, zmm/m512"                     , "RM"   , "EVEX.512.66.0F.W1 10 /r"          , "ANY AVX512F"],
    ["vmovupd"          , "W:zmm/m512 {kz}, zmm"                     , "MR"   , "EVEX.512.66.0F.W1 11 /r"          , "ANY AVX512F"],
    ["vmovups"          , "W:xmm {kz}, xmm/m128"                     , "RM"   , "EVEX.128.0F.W0 10 /r"             , "ANY AVX512F-VL"],
    ["vmovups"          , "W:xmm/m128 {kz}, xmm"                     , "MR"   , "EVEX.128.0F.W0 11 /r"             , "ANY AVX512F-VL"],
    ["vmovups"          , "W:ymm {kz}, ymm/m256"                     , "RM"   , "EVEX.256.0F.W0 10 /r"             , "ANY AVX512F-VL"],
    ["vmovups"          , "W:ymm/m256 {kz}, ymm"                     , "MR"   , "EVEX.256.0F.W0 11 /r"             , "ANY AVX512F-VL"],
    ["vmovups"          , "W:zmm {kz}, zmm/m512"                     , "RM"   , "EVEX.512.0F.W0 10 /r"             , "ANY AVX512F"],
    ["vmovups"          , "W:zmm/m512 {kz}, zmm"                     , "MR"   , "EVEX.512.0F.W0 11 /r"             , "ANY AVX512F"],
    ["vmulpd"           , "W:xmm {kz}, xmm, xmm/m128/b64"            , "RVM"  , "EVEX.NDS.128.66.0F.W1 59 /r"      , "ANY AVX512F-VL"],
    ["vmulpd"           , "W:ymm {kz}, ymm, ymm/m256/b64"            , "RVM"  , "EVEX.NDS.256.66.0F.W1 59 /r"      , "ANY AVX512F-VL"],
    ["vmulpd"           , "W:zmm {kz}, zmm, zmm/m512/b64 {er}"       , "RVM"  , "EVEX.NDS.512.66.0F.W1 59 /r"      , "ANY AVX512F"],
    ["vmulps"           , "W:xmm {kz}, xmm, xmm/m128/b32"            , "RVM"  , "EVEX.NDS.128.0F.W0 59 /r"         , "ANY AVX512F-VL"],
    ["vmulps"           , "W:ymm {kz}, ymm, ymm/m256/b32"            , "RVM"  , "EVEX.NDS.256.0F.W0 59 /r"         , "ANY AVX512F-VL"],
    ["vmulps"           , "W:zmm {kz}, zmm, zmm/m512/b32 {er}"       , "RVM"  , "EVEX.NDS.512.0F.W0 59 /r"         , "ANY AVX512F"],
    ["vmulsd"           , "W:xmm {kz}, xmm, xmm/m64 {er}"            , "RVM"  , "EVEX.NDS.LIG.F2.0F.W1 59 /r"      , "ANY AVX512F"],
    ["vmulss"           , "W:xmm {kz}, xmm, xmm/m32 {er}"            , "RVM"  , "EVEX.NDS.LIG.F3.0F.W0 59 /r"      , "ANY AVX512F"],
    ["vorpd"            , "W:xmm {kz}, xmm, xmm/m128/b64"            , "RVM"  , "EVEX.NDS.128.66.0F.W1 56 /r"      , "ANY AVX512DQ-VL"],
    ["vorpd"            , "W:ymm {kz}, ymm, ymm/m256/b64"            , "RVM"  , "EVEX.NDS.256.66.0F.W1 56 /r"      , "ANY AVX512DQ-VL"],
    ["vorpd"            , "W:zmm {kz}, zmm, zmm/m512/b64"            , "RVM"  , "EVEX.NDS.512.66.0F.W1 56 /r"      , "ANY AVX512DQ"],
    ["vorps"            , "W:xmm {kz}, xmm, xmm/m128/b32"            , "RVM"  , "EVEX.NDS.128.0F.W0 56 /r"         , "ANY AVX512F-VL"],
    ["vorps"            , "W:ymm {kz}, ymm, ymm/m256/b32"            , "RVM"  , "EVEX.NDS.256.0F.W0 56 /r"         , "ANY AVX512F-VL"],
    ["vorps"            , "W:zmm {kz}, zmm, zmm/m512/b32"            , "RVM"  , "EVEX.NDS.512.0F.W0 56 /r"         , "ANY AVX512F"],
    ["vpabsb"           , "W:xmm {kz}, xmm/m128"                     , "RM"   , "EVEX.128.66.0F38 1C /r"           , "ANY AVX512BW-VL"],
    ["vpabsb"           , "W:ymm {kz}, ymm/m256"                     , "RM"   , "EVEX.256.66.0F38 1C /r"           , "ANY AVX512BW-VL"],
    ["vpabsb"           , "W:zmm {kz}, zmm/m512"                     , "RM"   , "EVEX.512.66.0F38 1C /r"           , "ANY AVX512BW"],
    ["vpabsd"           , "W:xmm {kz}, xmm/m128"                     , "RM"   , "EVEX.128.66.0F38.W0 1E /r"        , "ANY AVX512F-VL"],
    ["vpabsd"           , "W:ymm {kz}, ymm/m256"                     , "RM"   , "EVEX.256.66.0F38.W0 1E /r"        , "ANY AVX512F-VL"],
    ["vpabsd"           , "W:zmm {kz}, zmm/m512"                     , "RM"   , "EVEX.512.66.0F38.W0 1E /r"        , "ANY AVX512F"],
    ["vpabsq"           , "W:xmm {kz}, xmm/m128"                     , "RM"   , "EVEX.128.66.0F38.W1 1F /r"        , "ANY AVX512F-VL"],
    ["vpabsq"           , "W:ymm {kz}, ymm/m256"                     , "RM"   , "EVEX.256.66.0F38.W1 1F /r"        , "ANY AVX512F-VL"],
    ["vpabsq"           , "W:zmm {kz}, zmm/m512"                     , "RM"   , "EVEX.512.66.0F38.W1 1F /r"        , "ANY AVX512F"],
    ["vpabsw"           , "W:xmm {kz}, xmm/m128"                     , "RM"   , "EVEX.128.66.0F38 1D /r"           , "ANY AVX512BW-VL"],
    ["vpabsw"           , "W:ymm {kz}, ymm/m256"                     , "RM"   , "EVEX.256.66.0F38 1D /r"           , "ANY AVX512BW-VL"],
    ["vpabsw"           , "W:zmm {kz}, zmm/m512"                     , "RM"   , "EVEX.512.66.0F38 1D /r"           , "ANY AVX512BW"],
    ["vpackssdw"        , "W:xmm {kz}, xmm, xmm/m128/b32"            , "RVM"  , "EVEX.NDS.128.66.0F.W0 6B /r"      , "ANY AVX512BW-VL"],
    ["vpackssdw"        , "W:ymm {kz}, ymm, ymm/m256/b32"            , "RVM"  , "EVEX.NDS.256.66.0F.W0 6B /r"      , "ANY AVX512BW-VL"],
    ["vpackssdw"        , "W:zmm {kz}, zmm, zmm/m512/b32"            , "RVM"  , "EVEX.NDS.512.66.0F.W0 6B /r"      , "ANY AVX512BW"],
    ["vpacksswb"        , "W:xmm {kz}, xmm, xmm/m128"                , "RVM"  , "EVEX.NDS.128.66.0F.WIG 63 /r"     , "ANY AVX512BW-VL"],
    ["vpacksswb"        , "W:ymm {kz}, ymm, ymm/m256"                , "RVM"  , "EVEX.NDS.256.66.0F.WIG 63 /r"     , "ANY AVX512BW-VL"],
    ["vpacksswb"        , "W:zmm {kz}, zmm, zmm/m512"                , "RVM"  , "EVEX.NDS.512.66.0F.WIG 63 /r"     , "ANY AVX512BW"],
    ["vpackusdw"        , "W:xmm {kz}, xmm, xmm/m128/b32"            , "RVM"  , "EVEX.NDS.128.66.0F38.W0 2B /r"    , "ANY AVX512BW-VL"],
    ["vpackusdw"        , "W:ymm {kz}, ymm, ymm/m256/b32"            , "RVM"  , "EVEX.NDS.256.66.0F38.W0 2B /r"    , "ANY AVX512BW-VL"],
    ["vpackusdw"        , "W:zmm {kz}, zmm, zmm/m512/b32"            , "RVM"  , "EVEX.NDS.512.66.0F38.W0 2B /r"    , "ANY AVX512BW"],
    ["vpackuswb"        , "W:xmm {kz}, xmm, xmm/m128"                , "RVM"  , "EVEX.NDS.128.66.0F.WIG 67 /r"     , "ANY AVX512BW-VL"],
    ["vpackuswb"        , "W:ymm {kz}, ymm, ymm/m256"                , "RVM"  , "EVEX.NDS.256.66.0F.WIG 67 /r"     , "ANY AVX512BW-VL"],
    ["vpackuswb"        , "W:zmm {kz}, zmm, zmm/m512"                , "RVM"  , "EVEX.NDS.512.66.0F.WIG 67 /r"     , "ANY AVX512BW"],
    ["vpaddb"           , "W:xmm {kz}, xmm, xmm/m128"                , "RVM"  , "EVEX.NDS.128.66.0F.WIG FC /r"     , "ANY AVX512BW-VL"],
    ["vpaddb"           , "W:ymm {kz}, ymm, ymm/m256"                , "RVM"  , "EVEX.NDS.256.66.0F.WIG FC /r"     , "ANY AVX512BW-VL"],
    ["vpaddb"           , "W:zmm {kz}, zmm, zmm/m512"                , "RVM"  , "EVEX.NDS.512.66.0F.WIG FC /r"     , "ANY AVX512BW"],
    ["vpaddd"           , "W:xmm {kz}, xmm, xmm/m128/b32"            , "RVM"  , "EVEX.NDS.128.66.0F.W0 FE /r"      , "ANY AVX512F-VL"],
    ["vpaddd"           , "W:ymm {kz}, ymm, ymm/m256/b32"            , "RVM"  , "EVEX.NDS.256.66.0F.W0 FE /r"      , "ANY AVX512F-VL"],
    ["vpaddd"           , "W:zmm {kz}, zmm, zmm/m512/b32"            , "RVM"  , "EVEX.NDS.512.66.0F.W0 FE /r"      , "ANY AVX512F"],
    ["vpaddq"           , "W:xmm {kz}, xmm, xmm/m128/b64"            , "RVM"  , "EVEX.NDS.128.66.0F.W1 D4 /r"      , "ANY AVX512F-VL"],
    ["vpaddq"           , "W:ymm {kz}, ymm, ymm/m256/b64"            , "RVM"  , "EVEX.NDS.256.66.0F.W1 D4 /r"      , "ANY AVX512F-VL"],
    ["vpaddq"           , "W:zmm {kz}, zmm, zmm/m512/b64"            , "RVM"  , "EVEX.NDS.512.66.0F.W1 D4 /r"      , "ANY AVX512F"],
    ["vpaddsb"          , "W:xmm {kz}, xmm, xmm/m128"                , "RVM"  , "EVEX.NDS.128.66.0F.WIG EC /r"     , "ANY AVX512BW-VL"],
    ["vpaddsb"          , "W:ymm {kz}, ymm, ymm/m256"                , "RVM"  , "EVEX.NDS.256.66.0F.WIG EC /r"     , "ANY AVX512BW-VL"],
    ["vpaddsb"          , "W:zmm {kz}, zmm, zmm/m512"                , "RVM"  , "EVEX.NDS.512.66.0F.WIG EC /r"     , "ANY AVX512BW"],
    ["vpaddsw"          , "W:xmm {kz}, xmm, xmm/m128"                , "RVM"  , "EVEX.NDS.128.66.0F.WIG ED /r"     , "ANY AVX512BW-VL"],
    ["vpaddsw"          , "W:ymm {kz}, ymm, ymm/m256"                , "RVM"  , "EVEX.NDS.256.66.0F.WIG ED /r"     , "ANY AVX512BW-VL"],
    ["vpaddsw"          , "W:zmm {kz}, zmm, zmm/m512"                , "RVM"  , "EVEX.NDS.512.66.0F.WIG ED /r"     , "ANY AVX512BW"],
    ["vpaddusb"         , "W:xmm {kz}, xmm, xmm/m128"                , "RVM"  , "EVEX.NDS.128.66.0F.WIG DC /r"     , "ANY AVX512BW-VL"],
    ["vpaddusb"         , "W:ymm {kz}, ymm, ymm/m256"                , "RVM"  , "EVEX.NDS.256.66.0F.WIG DC /r"     , "ANY AVX512BW-VL"],
    ["vpaddusb"         , "W:zmm {kz}, zmm, zmm/m512"                , "RVM"  , "EVEX.NDS.512.66.0F.WIG DC /r"     , "ANY AVX512BW"],
    ["vpaddusw"         , "W:xmm {kz}, xmm, xmm/m128"                , "RVM"  , "EVEX.NDS.128.66.0F.WIG DD /r"     , "ANY AVX512BW-VL"],
    ["vpaddusw"         , "W:ymm {kz}, ymm, ymm/m256"                , "RVM"  , "EVEX.NDS.256.66.0F.WIG DD /r"     , "ANY AVX512BW-VL"],
    ["vpaddusw"         , "W:zmm {kz}, zmm, zmm/m512"                , "RVM"  , "EVEX.NDS.512.66.0F.WIG DD /r"     , "ANY AVX512BW"],
    ["vpaddw"           , "W:xmm {kz}, xmm, xmm/m128"                , "RVM"  , "EVEX.NDS.128.66.0F.WIG FD /r"     , "ANY AVX512BW-VL"],
    ["vpaddw"           , "W:ymm {kz}, ymm, ymm/m256"                , "RVM"  , "EVEX.NDS.256.66.0F.WIG FD /r"     , "ANY AVX512BW-VL"],
    ["vpaddw"           , "W:zmm {kz}, zmm, zmm/m512"                , "RVM"  , "EVEX.NDS.512.66.0F.WIG FD /r"     , "ANY AVX512BW"],
    ["vpalignr"         , "W:xmm {kz}, xmm, xmm/m128, ib"            , "RVM"  , "EVEX.NDS.128.66.0F3A.WIG 0F /r ib", "ANY AVX512BW-VL"],
    ["vpalignr"         , "W:ymm {kz}, ymm, ymm/m256, ib"            , "RVM"  , "EVEX.NDS.256.66.0F3A.WIG 0F /r ib", "ANY AVX512BW-VL"],
    ["vpalignr"         , "W:zmm {kz}, zmm, zmm/m512, ib"            , "RVM"  , "EVEX.NDS.512.66.0F3A.WIG 0F /r ib", "ANY AVX512BW"],
    ["vpandd"           , "W:xmm {kz}, xmm, xmm/m128/b32"            , "RVM"  , "EVEX.NDS.128.66.0F.W0 DB /r"      , "ANY AVX512F-VL"],
    ["vpandd"           , "W:ymm {kz}, ymm, ymm/m256/b32"            , "RVM"  , "EVEX.NDS.256.66.0F.W0 DB /r"      , "ANY AVX512F-VL"],
    ["vpandd"           , "W:zmm {kz}, zmm, zmm/m512/b32"            , "RVM"  , "EVEX.NDS.512.66.0F.W0 DB /r"      , "ANY AVX512F"],
    ["vpandnd"          , "W:xmm {kz}, xmm, xmm/m128/b32"            , "RVM"  , "EVEX.NDS.128.66.0F.W0 DF /r"      , "ANY AVX512F-VL"],
    ["vpandnd"          , "W:ymm {kz}, ymm, ymm/m256/b32"            , "RVM"  , "EVEX.NDS.256.66.0F.W0 DF /r"      , "ANY AVX512F-VL"],
    ["vpandnd"          , "W:zmm {kz}, zmm, zmm/m512/b32"            , "RVM"  , "EVEX.NDS.512.66.0F.W0 DF /r"      , "ANY AVX512F"],
    ["vpandnq"          , "W:xmm {kz}, xmm, xmm/m128/b64"            , "RVM"  , "EVEX.NDS.128.66.0F.W1 DF /r"      , "ANY AVX512F-VL"],
    ["vpandnq"          , "W:ymm {kz}, ymm, ymm/m256/b64"            , "RVM"  , "EVEX.NDS.256.66.0F.W1 DF /r"      , "ANY AVX512F-VL"],
    ["vpandnq"          , "W:zmm {kz}, zmm, zmm/m512/b64"            , "RVM"  , "EVEX.NDS.512.66.0F.W1 DF /r"      , "ANY AVX512F"],
    ["vpandq"           , "W:xmm {kz}, xmm, xmm/m128/b64"            , "RVM"  , "EVEX.NDS.128.66.0F.W1 DB /r"      , "ANY AVX512F-VL"],
    ["vpandq"           , "W:ymm {kz}, ymm, ymm/m256/b64"            , "RVM"  , "EVEX.NDS.256.66.0F.W1 DB /r"      , "ANY AVX512F-VL"],
    ["vpandq"           , "W:zmm {kz}, zmm, zmm/m512/b64"            , "RVM"  , "EVEX.NDS.512.66.0F.W1 DB /r"      , "ANY AVX512F"],
    ["vpavgb"           , "W:xmm {kz}, xmm, xmm/m128"                , "RVM"  , "EVEX.NDS.128.66.0F.WIG E0 /r"     , "ANY AVX512BW-VL"],
    ["vpavgb"           , "W:ymm {kz}, ymm, ymm/m256"                , "RVM"  , "EVEX.NDS.256.66.0F.WIG E0 /r"     , "ANY AVX512BW-VL"],
    ["vpavgb"           , "W:zmm {kz}, zmm, zmm/m512"                , "RVM"  , "EVEX.NDS.512.66.0F.WIG E0 /r"     , "ANY AVX512BW"],
    ["vpavgw"           , "W:xmm {kz}, xmm, xmm/m128"                , "RVM"  , "EVEX.NDS.128.66.0F.WIG E3 /r"     , "ANY AVX512BW-VL"],
    ["vpavgw"           , "W:ymm {kz}, ymm, ymm/m256"                , "RVM"  , "EVEX.NDS.256.66.0F.WIG E3 /r"     , "ANY AVX512BW-VL"],
    ["vpavgw"           , "W:zmm {kz}, zmm, zmm/m512"                , "RVM"  , "EVEX.NDS.512.66.0F.WIG E3 /r"     , "ANY AVX512BW"],
    ["vpbroadcastb"     , "W:xmm {kz}, r32"                          , "RM"   , "EVEX.128.66.0F38.W0 7A /r"        , "ANY AVX512BW-VL"],
    ["vpbroadcastb"     , "W:xmm {kz}, xmm/m8"                       , "RM"   , "EVEX.128.66.0F38.W0 78 /r"        , "ANY AVX512BW-VL"],
    ["vpbroadcastb"     , "W:ymm {kz}, r32"                          , "RM"   , "EVEX.256.66.0F38.W0 7A /r"        , "ANY AVX512BW-VL"],
    ["vpbroadcastb"     , "W:ymm {kz}, xmm/m8"                       , "RM"   , "EVEX.256.66.0F38.W0 78 /r"        , "ANY AVX512BW-VL"],
    ["vpbroadcastb"     , "W:zmm {kz}, r32"                          , "RM"   , "EVEX.512.66.0F38.W0 7A /r"        , "ANY AVX512BW"],
    ["vpbroadcastb"     , "W:zmm {kz}, xmm/m8"                       , "RM"   , "EVEX.512.66.0F38.W0 78 /r"        , "ANY AVX512BW"],
    ["vpbroadcastd"     , "W:xmm {kz}, r32"                          , "RM"   , "EVEX.128.66.0F38.W0 7C /r"        , "ANY AVX512F-VL"],
    ["vpbroadcastd"     , "W:xmm {kz}, xmm/m32"                      , "RM"   , "EVEX.128.66.0F38.W0 58 /r"        , "ANY AVX512F-VL"],
    ["vpbroadcastd"     , "W:ymm {kz}, r32"                          , "RM"   , "EVEX.256.66.0F38.W0 7C /r"        , "ANY AVX512F-VL"],
    ["vpbroadcastd"     , "W:ymm {kz}, xmm/m32"                      , "RM"   , "EVEX.256.66.0F38.W0 58 /r"        , "ANY AVX512F-VL"],
    ["vpbroadcastd"     , "W:zmm {kz}, r32"                          , "RM"   , "EVEX.512.66.0F38.W0 7C /r"        , "ANY AVX512F"],
    ["vpbroadcastd"     , "W:zmm {kz}, xmm/m32"                      , "RM"   , "EVEX.512.66.0F38.W0 58 /r"        , "ANY AVX512F"],
    ["vpbroadcastmb2d"  , "W:xmm, k"                                 , "RM"   , "EVEX.128.F3.0F38.W0 3A /r"        , "ANY AVX512CD-VL"],
    ["vpbroadcastmb2d"  , "W:ymm, k"                                 , "RM"   , "EVEX.256.F3.0F38.W0 3A /r"        , "ANY AVX512CD-VL"],
    ["vpbroadcastmb2d"  , "W:zmm, k"                                 , "RM"   , "EVEX.512.F3.0F38.W0 3A /r"        , "ANY AVX512CD"],
    ["vpbroadcastmb2q"  , "W:xmm, k"                                 , "RM"   , "EVEX.128.F3.0F38.W1 2A /r"        , "ANY AVX512CD-VL"],
    ["vpbroadcastmb2q"  , "W:ymm, k"                                 , "RM"   , "EVEX.256.F3.0F38.W1 2A /r"        , "ANY AVX512CD-VL"],
    ["vpbroadcastmb2q"  , "W:zmm, k"                                 , "RM"   , "EVEX.512.F3.0F38.W1 2A /r"        , "ANY AVX512CD"],
    ["vpbroadcastq"     , "W:xmm {kz}, r64"                          , "RM"   , "EVEX.128.66.0F38.W1 7C /r"        , "X64 AVX512F-VL"],
    ["vpbroadcastq"     , "W:xmm {kz}, xmm/m64"                      , "RM"   , "EVEX.128.66.0F38.W1 59 /r"        , "ANY AVX512F-VL"],
    ["vpbroadcastq"     , "W:ymm {kz}, r64"                          , "RM"   , "EVEX.256.66.0F38.W1 7C /r"        , "X64 AVX512F-VL"],
    ["vpbroadcastq"     , "W:ymm {kz}, xmm/m64"                      , "RM"   , "EVEX.256.66.0F38.W1 59 /r"        , "ANY AVX512F-VL"],
    ["vpbroadcastq"     , "W:zmm {kz}, r64"                          , "RM"   , "EVEX.512.66.0F38.W1 7C /r"        , "X64 AVX512F"],
    ["vpbroadcastq"     , "W:zmm {kz}, xmm/m64"                      , "RM"   , "EVEX.512.66.0F38.W1 59 /r"        , "ANY AVX512F"],
    ["vpbroadcastw"     , "W:xmm {kz}, r32"                          , "RM"   , "EVEX.128.66.0F38.W0 7B /r"        , "ANY AVX512BW-VL"],
    ["vpbroadcastw"     , "W:xmm {kz}, xmm/m16"                      , "RM"   , "EVEX.128.66.0F38.W0 79 /r"        , "ANY AVX512BW-VL"],
    ["vpbroadcastw"     , "W:ymm {kz}, r32"                          , "RM"   , "EVEX.256.66.0F38.W0 7B /r"        , "ANY AVX512BW-VL"],
    ["vpbroadcastw"     , "W:ymm {kz}, xmm/m16"                      , "RM"   , "EVEX.256.66.0F38.W0 79 /r"        , "ANY AVX512BW-VL"],
    ["vpbroadcastw"     , "W:zmm {kz}, r32"                          , "RM"   , "EVEX.512.66.0F38.W0 7B /r"        , "ANY AVX512BW"],
    ["vpbroadcastw"     , "W:zmm {kz}, xmm/m16"                      , "RM"   , "EVEX.512.66.0F38.W0 79 /r"        , "ANY AVX512BW"],
    ["vpcmpb"           , "W:k {k}, xmm, xmm/m128, ib"               , "RVM"  , "EVEX.NDS.128.66.0F3A.W0 3F /r ib" , "ANY AVX512BW-VL"],
    ["vpcmpb"           , "W:k {k}, ymm, ymm/m256, ib"               , "RVM"  , "EVEX.NDS.256.66.0F3A.W0 3F /r ib" , "ANY AVX512BW-VL"],
    ["vpcmpb"           , "W:k {k}, zmm, zmm/m512, ib"               , "RVM"  , "EVEX.NDS.512.66.0F3A.W0 3F /r ib" , "ANY AVX512BW"],
    ["vpcmpd"           , "W:k {k}, xmm, xmm/m128/b32, ib"           , "RVM"  , "EVEX.NDS.128.66.0F3A.W0 1F /r ib" , "ANY AVX512F-VL"],
    ["vpcmpd"           , "W:k {k}, ymm, ymm/m256/b32, ib"           , "RVM"  , "EVEX.NDS.256.66.0F3A.W0 1F /r ib" , "ANY AVX512F-VL"],
    ["vpcmpd"           , "W:k {k}, zmm, zmm/m512/b32, ib"           , "RVM"  , "EVEX.NDS.512.66.0F3A.W0 1F /r ib" , "ANY AVX512F"],
    ["vpcmpeqb"         , "W:k {k}, xmm, xmm/m128"                   , "RVM"  , "EVEX.NDS.128.66.0F.WIG 74 /r"     , "ANY AVX512BW-VL"],
    ["vpcmpeqb"         , "W:k {k}, ymm, ymm/m256"                   , "RVM"  , "EVEX.NDS.256.66.0F.WIG 74 /r"     , "ANY AVX512BW-VL"],
    ["vpcmpeqb"         , "W:k {k}, zmm, zmm/m512"                   , "RVM"  , "EVEX.NDS.512.66.0F.WIG 74 /r"     , "ANY AVX512BW"],
    ["vpcmpeqd"         , "W:k {k}, xmm, xmm/m128/b32"               , "RVM"  , "EVEX.NDS.128.66.0F.W0 76 /r"      , "ANY AVX512F-VL"],
    ["vpcmpeqd"         , "W:k {k}, ymm, ymm/m256/b32"               , "RVM"  , "EVEX.NDS.256.66.0F.W0 76 /r"      , "ANY AVX512F-VL"],
    ["vpcmpeqd"         , "W:k {k}, zmm, zmm/m512/b32"               , "RVM"  , "EVEX.NDS.512.66.0F.W0 76 /r"      , "ANY AVX512F"],
    ["vpcmpeqq"         , "W:k {k}, xmm, xmm/m128/b64"               , "RVM"  , "EVEX.NDS.128.66.0F38.W1 29 /r"    , "ANY AVX512F-VL"],
    ["vpcmpeqq"         , "W:k {k}, ymm, ymm/m256/b64"               , "RVM"  , "EVEX.NDS.256.66.0F38.W1 29 /r"    , "ANY AVX512F-VL"],
    ["vpcmpeqq"         , "W:k {k}, zmm, zmm/m512/b64"               , "RVM"  , "EVEX.NDS.512.66.0F38.W1 29 /r"    , "ANY AVX512F"],
    ["vpcmpeqw"         , "W:k {k}, xmm, xmm/m128"                   , "RVM"  , "EVEX.NDS.128.66.0F.WIG 75 /r"     , "ANY AVX512BW-VL"],
    ["vpcmpeqw"         , "W:k {k}, ymm, ymm/m256"                   , "RVM"  , "EVEX.NDS.256.66.0F.WIG 75 /r"     , "ANY AVX512BW-VL"],
    ["vpcmpeqw"         , "W:k {k}, zmm, zmm/m512"                   , "RVM"  , "EVEX.NDS.512.66.0F.WIG 75 /r"     , "ANY AVX512BW"],
    ["vpcmpgtb"         , "W:k {k}, xmm, xmm/m128"                   , "RVM"  , "EVEX.NDS.128.66.0F.WIG 64 /r"     , "ANY AVX512BW-VL"],
    ["vpcmpgtb"         , "W:k {k}, ymm, ymm/m256"                   , "RVM"  , "EVEX.NDS.256.66.0F.WIG 64 /r"     , "ANY AVX512BW-VL"],
    ["vpcmpgtb"         , "W:k {k}, zmm, zmm/m512"                   , "RVM"  , "EVEX.NDS.512.66.0F.WIG 64 /r"     , "ANY AVX512BW"],
    ["vpcmpgtd"         , "W:k {k}, xmm, xmm/m128/b32"               , "RVM"  , "EVEX.NDS.128.66.0F.W0 66 /r"      , "ANY AVX512F-VL"],
    ["vpcmpgtd"         , "W:k {k}, ymm, ymm/m256/b32"               , "RVM"  , "EVEX.NDS.256.66.0F.W0 66 /r"      , "ANY AVX512F-VL"],
    ["vpcmpgtd"         , "W:k {k}, zmm, zmm/m512/b32"               , "RVM"  , "EVEX.NDS.512.66.0F.W0 66 /r"      , "ANY AVX512F"],
    ["vpcmpgtq"         , "W:k {k}, xmm, xmm/m128/b64"               , "RVM"  , "EVEX.NDS.128.66.0F38.W1 37 /r"    , "ANY AVX512F-VL"],
    ["vpcmpgtq"         , "W:k {k}, ymm, ymm/m256/b64"               , "RVM"  , "EVEX.NDS.256.66.0F38.W1 37 /r"    , "ANY AVX512F-VL"],
    ["vpcmpgtq"         , "W:k {k}, zmm, zmm/m512/b64"               , "RVM"  , "EVEX.NDS.512.66.0F38.W1 37 /r"    , "ANY AVX512F"],
    ["vpcmpgtw"         , "W:k {k}, xmm, xmm/m128"                   , "RVM"  , "EVEX.NDS.128.66.0F.WIG 65 /r"     , "ANY AVX512BW-VL"],
    ["vpcmpgtw"         , "W:k {k}, ymm, ymm/m256"                   , "RVM"  , "EVEX.NDS.256.66.0F.WIG 65 /r"     , "ANY AVX512BW-VL"],
    ["vpcmpgtw"         , "W:k {k}, zmm, zmm/m512"                   , "RVM"  , "EVEX.NDS.512.66.0F.WIG 65 /r"     , "ANY AVX512BW"],
    ["vpcmpq"           , "W:k {k}, xmm, xmm/m128/b64, ib"           , "RVM"  , "EVEX.NDS.128.66.0F3A.W1 1F /r ib" , "ANY AVX512F-VL"],
    ["vpcmpq"           , "W:k {k}, ymm, ymm/m256/b64, ib"           , "RVM"  , "EVEX.NDS.256.66.0F3A.W1 1F /r ib" , "ANY AVX512F-VL"],
    ["vpcmpq"           , "W:k {k}, zmm, zmm/m512/b64, ib"           , "RVM"  , "EVEX.NDS.512.66.0F3A.W1 1F /r ib" , "ANY AVX512F"],
    ["vpcmpub"          , "W:k {k}, xmm, xmm/m128, ib"               , "RVM"  , "EVEX.NDS.128.66.0F3A.W0 3E /r ib" , "ANY AVX512BW-VL"],
    ["vpcmpub"          , "W:k {k}, ymm, ymm/m256, ib"               , "RVM"  , "EVEX.NDS.256.66.0F3A.W0 3E /r ib" , "ANY AVX512BW-VL"],
    ["vpcmpub"          , "W:k {k}, zmm, zmm/m512, ib"               , "RVM"  , "EVEX.NDS.512.66.0F3A.W0 3E /r ib" , "ANY AVX512BW"],
    ["vpcmpud"          , "W:k {k}, xmm, xmm/m128/b32, ib"           , "RVM"  , "EVEX.NDS.128.66.0F3A.W0 1E /r ib" , "ANY AVX512F-VL"],
    ["vpcmpud"          , "W:k {k}, ymm, ymm/m256/b32, ib"           , "RVM"  , "EVEX.NDS.256.66.0F3A.W0 1E /r ib" , "ANY AVX512F-VL"],
    ["vpcmpud"          , "W:k {k}, zmm, zmm/m512/b32, ib"           , "RVM"  , "EVEX.NDS.512.66.0F3A.W0 1E /r ib" , "ANY AVX512F"],
    ["vpcmpuq"          , "W:k {k}, xmm, xmm/m128/b64, ib"           , "RVM"  , "EVEX.NDS.128.66.0F3A.W1 1E /r ib" , "ANY AVX512F-VL"],
    ["vpcmpuq"          , "W:k {k}, ymm, ymm/m256/b64, ib"           , "RVM"  , "EVEX.NDS.256.66.0F3A.W1 1E /r ib" , "ANY AVX512F-VL"],
    ["vpcmpuq"          , "W:k {k}, zmm, zmm/m512/b64, ib"           , "RVM"  , "EVEX.NDS.512.66.0F3A.W1 1E /r ib" , "ANY AVX512F"],
    ["vpcmpuw"          , "W:k {k}, xmm, xmm/m128/b64, ib"           , "RVM"  , "EVEX.NDS.128.66.0F3A.W1 3E /r ib" , "ANY AVX512BW-VL"],
    ["vpcmpuw"          , "W:k {k}, ymm, ymm/m256/b64, ib"           , "RVM"  , "EVEX.NDS.256.66.0F3A.W1 3E /r ib" , "ANY AVX512BW-VL"],
    ["vpcmpuw"          , "W:k {k}, zmm, zmm/m512/b64, ib"           , "RVM"  , "EVEX.NDS.512.66.0F3A.W1 3E /r ib" , "ANY AVX512BW"],
    ["vpcmpw"           , "W:k {k}, xmm, xmm/m128/b64, ib"           , "RVM"  , "EVEX.NDS.128.66.0F3A.W1 3F /r ib" , "ANY AVX512BW-VL"],
    ["vpcmpw"           , "W:k {k}, ymm, ymm/m256/b64, ib"           , "RVM"  , "EVEX.NDS.256.66.0F3A.W1 3F /r ib" , "ANY AVX512BW-VL"],
    ["vpcmpw"           , "W:k {k}, zmm, zmm/m512/b64, ib"           , "RVM"  , "EVEX.NDS.512.66.0F3A.W1 3F /r ib" , "ANY AVX512BW"],
    ["vpcompressd"      , "W:xmm/m128 {kz}, xmm"                     , "MR"   , "EVEX.128.66.0F38.W0 8B /r"        , "ANY AVX512F-VL"],
    ["vpcompressd"      , "W:ymm/m256 {kz}, ymm"                     , "MR"   , "EVEX.256.66.0F38.W0 8B /r"        , "ANY AVX512F-VL"],
    ["vpcompressd"      , "W:zmm/m512 {kz}, zmm"                     , "MR"   , "EVEX.512.66.0F38.W0 8B /r"        , "ANY AVX512F"],
    ["vpcompressq"      , "W:xmm/m128 {kz}, xmm"                     , "MR"   , "EVEX.128.66.0F38.W1 8B /r"        , "ANY AVX512F-VL"],
    ["vpcompressq"      , "W:ymm/m256 {kz}, ymm"                     , "MR"   , "EVEX.256.66.0F38.W1 8B /r"        , "ANY AVX512F-VL"],
    ["vpcompressq"      , "W:zmm/m512 {kz}, zmm"                     , "MR"   , "EVEX.512.66.0F38.W1 8B /r"        , "ANY AVX512F"],
    ["vpconflictd"      , "W:xmm {kz}, xmm/m128/b32"                 , "RM"   , "EVEX.128.66.0F38.W0 C4 /r"        , "ANY AVX512CD-VL"],
    ["vpconflictd"      , "W:ymm {kz}, ymm/m256/b32"                 , "RM"   , "EVEX.256.66.0F38.W0 C4 /r"        , "ANY AVX512CD-VL"],
    ["vpconflictd"      , "W:zmm {kz}, zmm/m512/b32"                 , "RM"   , "EVEX.512.66.0F38.W0 C4 /r"        , "ANY AVX512CD"],
    ["vpconflictq"      , "W:xmm {kz}, xmm/m128/b32"                 , "RM"   , "EVEX.128.66.0F38.W1 C4 /r"        , "ANY AVX512CD-VL"],
    ["vpconflictq"      , "W:ymm {kz}, ymm/m256/b32"                 , "RM"   , "EVEX.256.66.0F38.W1 C4 /r"        , "ANY AVX512CD-VL"],
    ["vpconflictq"      , "W:zmm {kz}, zmm/m512/b32"                 , "RM"   , "EVEX.512.66.0F38.W1 C4 /r"        , "ANY AVX512CD"],
    ["vpermd"           , "W:ymm {kz}, ymm, ymm/m256/b32"            , "RVM"  , "EVEX.NDS.256.66.0F38.W0 36 /r"    , "ANY AVX512F-VL"],
    ["vpermd"           , "W:zmm {kz}, zmm, zmm/m512/b32"            , "RVM"  , "EVEX.NDS.512.66.0F38.W0 36 /r"    , "ANY AVX512F-VL"],
    ["vpermi2d"         , "X:xmm {kz}, xmm, xmm/m128/b32"            , "RVM"  , "EVEX.DDS.128.66.0F38.W0 76 /r"    , "ANY AVX512F-VL"],
    ["vpermi2d"         , "X:ymm {kz}, ymm, ymm/m256/b32"            , "RVM"  , "EVEX.DDS.256.66.0F38.W0 76 /r"    , "ANY AVX512F-VL"],
    ["vpermi2d"         , "X:zmm {kz}, zmm, zmm/m512/b32"            , "RVM"  , "EVEX.DDS.512.66.0F38.W0 76 /r"    , "ANY AVX512F"],
    ["vpermi2pd"        , "W:xmm {kz}, xmm, xmm/m128/b64"            , "RVM"  , "EVEX.DDS.128.66.0F38.W1 77 /r"    , "ANY AVX512F-VL"],
    ["vpermi2pd"        , "W:ymm {kz}, ymm, ymm/m256/b64"            , "RVM"  , "EVEX.DDS.256.66.0F38.W1 77 /r"    , "ANY AVX512F-VL"],
    ["vpermi2pd"        , "W:zmm {kz}, zmm, zmm/m512/b64"            , "RVM"  , "EVEX.DDS.512.66.0F38.W1 77 /r"    , "ANY AVX512F"],
    ["vpermi2ps"        , "W:xmm {kz}, xmm, xmm/m128/b32"            , "RVM"  , "EVEX.DDS.128.66.0F38.W0 77 /r"    , "ANY AVX512F-VL"],
    ["vpermi2ps"        , "W:ymm {kz}, ymm, ymm/m256/b32"            , "RVM"  , "EVEX.DDS.256.66.0F38.W0 77 /r"    , "ANY AVX512F-VL"],
    ["vpermi2ps"        , "W:zmm {kz}, zmm, zmm/m512/b32"            , "RVM"  , "EVEX.DDS.512.66.0F38.W0 77 /r"    , "ANY AVX512F"],
    ["vpermi2q"         , "X:xmm {kz}, xmm, xmm/m128/b64"            , "RVM"  , "EVEX.DDS.128.66.0F38.W1 76 /r"    , "ANY AVX512F-VL"],
    ["vpermi2q"         , "X:ymm {kz}, ymm, ymm/m256/b64"            , "RVM"  , "EVEX.DDS.256.66.0F38.W1 76 /r"    , "ANY AVX512F-VL"],
    ["vpermi2q"         , "X:zmm {kz}, zmm, zmm/m512/b64"            , "RVM"  , "EVEX.DDS.512.66.0F38.W1 76 /r"    , "ANY AVX512F"],
    ["vpermi2w"         , "X:xmm {kz}, xmm, xmm/m128"                , "RVM"  , "EVEX.DDS.128.66.0F38.W1 75 /r"    , "ANY AVX512BW-VL"],
    ["vpermi2w"         , "X:ymm {kz}, ymm, ymm/m256"                , "RVM"  , "EVEX.DDS.256.66.0F38.W1 75 /r"    , "ANY AVX512BW-VL"],
    ["vpermi2w"         , "X:zmm {kz}, zmm, zmm/m512"                , "RVM"  , "EVEX.DDS.512.66.0F38.W1 75 /r"    , "ANY AVX512BW"],
    ["vpermilpd"        , "W:xmm {kz}, xmm, xmm/m128/b64"            , "RVM"  , "EVEX.NDS.128.66.0F38.W1 0D /r"    , "ANY AVX512F-VL"],
    ["vpermilpd"        , "W:xmm {kz}, xmm/m128/b64, ib"             , "RMI"  , "EVEX.128.66.0F3A.W1 05 /r ib"     , "ANY AVX512F-VL"],
    ["vpermilpd"        , "W:ymm {kz}, ymm, ymm/m256/b64"            , "RVM"  , "EVEX.NDS.256.66.0F38.W1 0D /r"    , "ANY AVX512F-VL"],
    ["vpermilpd"        , "W:ymm {kz}, ymm/m256/b64, ib"             , "RMI"  , "EVEX.256.66.0F3A.W1 05 /r ib"     , "ANY AVX512F-VL"],
    ["vpermilpd"        , "W:zmm {kz}, zmm, zmm/m512/b64"            , "RVM"  , "EVEX.NDS.512.66.0F38.W1 0D /r"    , "ANY AVX512F"],
    ["vpermilpd"        , "W:zmm {kz}, zmm/m512/b64, ib"             , "RMI"  , "EVEX.512.66.0F3A.W1 05 /r ib"     , "ANY AVX512F"],
    ["vpermilps"        , "W:xmm {kz}, xmm, xmm/m128/b64"            , "RVM"  , "EVEX.NDS.128.66.0F38.W0 0C /r"    , "ANY AVX512F-VL"],
    ["vpermilps"        , "W:xmm {kz}, xmm/m128/b64, ib"             , "RMI"  , "EVEX.128.66.0F3A.W0 04 /r ib"     , "ANY AVX512F-VL"],
    ["vpermilps"        , "W:ymm {kz}, ymm, ymm/m256/b64"            , "RVM"  , "EVEX.NDS.256.66.0F38.W0 0C /r"    , "ANY AVX512F-VL"],
    ["vpermilps"        , "W:ymm {kz}, ymm/m256/b64, ib"             , "RMI"  , "EVEX.256.66.0F3A.W0 04 /r ib"     , "ANY AVX512F-VL"],
    ["vpermilps"        , "W:zmm {kz}, zmm, zmm/m512/b64"            , "RVM"  , "EVEX.NDS.512.66.0F38.W0 0C /r"    , "ANY AVX512F"],
    ["vpermilps"        , "W:zmm {kz}, zmm/m512/b64, ib"             , "RMI"  , "EVEX.512.66.0F3A.W0 04 /r ib"     , "ANY AVX512F"],
    ["vpermq"           , "W:ymm {kz}, ymm, ymm/m256/b64"            , "RVM"  , "EVEX.NDS.256.66.0F38.W1 36 /r"    , "ANY AVX512F-VL"],
    ["vpermq"           , "W:ymm {kz}, ymm/m256/b64, ib"             , "RMI"  , "EVEX.256.66.0F3A.W1 00 /r ib"     , "ANY AVX512F-VL"],
    ["vpermq"           , "W:zmm {kz}, zmm, zmm/m512/b64"            , "RVM"  , "EVEX.NDS.512.66.0F38.W1 36 /r"    , "ANY AVX512F-VL"],
    ["vpermq"           , "W:zmm {kz}, zmm/m512/b64, ib"             , "RMI"  , "EVEX.512.66.0F3A.W1 00 /r ib"     , "ANY AVX512F-VL"],
    ["vpermt2d"         , "X:xmm {kz}, xmm, xmm/m128/b32"            , "RVM"  , "EVEX.DDS.128.66.0F38.W0 7E /r"    , "ANY AVX512F-VL"],
    ["vpermt2d"         , "X:ymm {kz}, ymm, ymm/m256/b32"            , "RVM"  , "EVEX.DDS.256.66.0F38.W0 7E /r"    , "ANY AVX512F-VL"],
    ["vpermt2d"         , "X:zmm {kz}, zmm, zmm/m512/b32"            , "RVM"  , "EVEX.DDS.512.66.0F38.W0 7E /r"    , "ANY AVX512F"],
    ["vpermt2pd"        , "X:xmm {kz}, xmm, xmm/m128/b64"            , "RVM"  , "EVEX.DDS.128.66.0F38.W1 7F /r"    , "ANY AVX512F-VL"],
    ["vpermt2pd"        , "X:ymm {kz}, ymm, ymm/m256/b64"            , "RVM"  , "EVEX.DDS.256.66.0F38.W1 7F /r"    , "ANY AVX512F-VL"],
    ["vpermt2pd"        , "X:zmm {kz}, zmm, zmm/m512/b64"            , "RVM"  , "EVEX.DDS.512.66.0F38.W1 7F /r"    , "ANY AVX512F"],
    ["vpermt2ps"        , "X:xmm {kz}, xmm, xmm/m128/b32"            , "RVM"  , "EVEX.DDS.128.66.0F38.W0 7F /r"    , "ANY AVX512F-VL"],
    ["vpermt2ps"        , "X:ymm {kz}, ymm, ymm/m256/b32"            , "RVM"  , "EVEX.DDS.256.66.0F38.W0 7F /r"    , "ANY AVX512F-VL"],
    ["vpermt2ps"        , "X:zmm {kz}, zmm, zmm/m512/b32"            , "RVM"  , "EVEX.DDS.512.66.0F38.W0 7F /r"    , "ANY AVX512F"],
    ["vpermt2q"         , "X:xmm {kz}, xmm, xmm/m128/b64"            , "RVM"  , "EVEX.DDS.128.66.0F38.W1 7E /r"    , "ANY AVX512F-VL"],
    ["vpermt2q"         , "X:ymm {kz}, ymm, ymm/m256/b64"            , "RVM"  , "EVEX.DDS.256.66.0F38.W1 7E /r"    , "ANY AVX512F-VL"],
    ["vpermt2q"         , "X:zmm {kz}, zmm, zmm/m512/b64"            , "RVM"  , "EVEX.DDS.512.66.0F38.W1 7E /r"    , "ANY AVX512F"],
    ["vpermt2w"         , "X:xmm {kz}, xmm, xmm/m128"                , "RVM"  , "EVEX.DDS.128.66.0F38.W1 7D /r"    , "ANY AVX512BW-VL"],
    ["vpermt2w"         , "X:ymm {kz}, ymm, ymm/m256"                , "RVM"  , "EVEX.DDS.256.66.0F38.W1 7D /r"    , "ANY AVX512BW-VL"],
    ["vpermt2w"         , "X:zmm {kz}, zmm, zmm/m512"                , "RVM"  , "EVEX.DDS.512.66.0F38.W1 7D /r"    , "ANY AVX512BW"],
    ["vpermw"           , "W:xmm {kz}, xmm, xmm/m128"                , "RVM"  , "EVEX.NDS.128.66.0F38.W1 8D /r"    , "ANY AVX512BW-VL"],
    ["vpermw"           , "W:ymm {kz}, ymm, ymm/m256"                , "RVM"  , "EVEX.NDS.256.66.0F38.W1 8D /r"    , "ANY AVX512BW-VL"],
    ["vpermw"           , "W:zmm {kz}, zmm, zmm/m512"                , "RVM"  , "EVEX.NDS.512.66.0F38.W1 8D /r"    , "ANY AVX512BW"],
    ["vpextrb"          , "W:r32/m8 , xmm, ib"                       , "MRI"  , "EVEX.128.66.0F3A.WIG 14 /r ib"    , "ANY AVX512BW"],
    ["vpextrd"          , "W:r32/m32, xmm, ib"                       , "MRI"  , "EVEX.128.66.0F3A.W0 16 /r ib"     , "ANY AVX512DQ"],
    ["vpextrq"          , "W:r64/m64, xmm, ib"                       , "MRI"  , "EVEX.128.66.0F3A.W1 16 /r ib"     , "X64 AVX512DQ"],
    ["vpextrw"          , "W:r32    , xmm, ib"                       , "RMI"  , "EVEX.128.66.0F.WIG C5 /r ib"      , "ANY AVX512BW"],
    ["vpextrw"          , "W:r32/m16, xmm, ib"                       , "MRI"  , "EVEX.128.66.0F3A.WIG 15 /r ib"    , "ANY AVX512BW"],
    ["vpgatherdd"       , "W:xmm {k}, vm32x"                         , "RM"   , "EVEX.128.66.0F38.W0 90"           , "ANY AVX512F-VL"],
    ["vpgatherdd"       , "W:ymm {k}, vm32y"                         , "RM"   , "EVEX.256.66.0F38.W0 90"           , "ANY AVX512F-VL"],
    ["vpgatherdd"       , "W:zmm {k}, vm32z"                         , "RM"   , "EVEX.512.66.0F38.W0 90"           , "ANY AVX512F"],
    ["vpgatherdpd"      , "W:xmm {k}, vm32x"                         , "RM"   , "EVEX.128.66.0F38.W1 92"           , "ANY AVX512F-VL"],
    ["vpgatherdpd"      , "W:ymm {k}, vm32y"                         , "RM"   , "EVEX.256.66.0F38.W1 92"           , "ANY AVX512F-VL"],
    ["vpgatherdpd"      , "W:zmm {k}, vm32z"                         , "RM"   , "EVEX.512.66.0F38.W1 92"           , "ANY AVX512F"],
    ["vpgatherdps"      , "W:xmm {k}, vm32x"                         , "RM"   , "EVEX.128.66.0F38.W0 92"           , "ANY AVX512F-VL"],
    ["vpgatherdps"      , "W:ymm {k}, vm32y"                         , "RM"   , "EVEX.256.66.0F38.W0 92"           , "ANY AVX512F-VL"],
    ["vpgatherdps"      , "W:zmm {k}, vm32z"                         , "RM"   , "EVEX.512.66.0F38.W0 92"           , "ANY AVX512F"],
    ["vpgatherdq"       , "W:xmm {k}, vm32x"                         , "RM"   , "EVEX.128.66.0F38.W1 90"           , "ANY AVX512F-VL"],
    ["vpgatherdq"       , "W:ymm {k}, vm32y"                         , "RM"   , "EVEX.256.66.0F38.W1 90"           , "ANY AVX512F-VL"],
    ["vpgatherdq"       , "W:zmm {k}, vm32z"                         , "RM"   , "EVEX.512.66.0F38.W1 90"           , "ANY AVX512F"],
    ["vpgatherqd"       , "W:xmm {k}, vm64x"                         , "RM"   , "EVEX.128.66.0F38.W0 91"           , "ANY AVX512F-VL"],
    ["vpgatherqd"       , "W:ymm {k}, vm64y"                         , "RM"   , "EVEX.256.66.0F38.W0 91"           , "ANY AVX512F-VL"],
    ["vpgatherqd"       , "W:zmm {k}, vm64z"                         , "RM"   , "EVEX.512.66.0F38.W0 91"           , "ANY AVX512F"],
    ["vpgatherqpd"      , "W:xmm {k}, vm32x"                         , "RM"   , "EVEX.128.66.0F38.W1 93"           , "ANY AVX512F-VL"],
    ["vpgatherqpd"      , "W:ymm {k}, vm32y"                         , "RM"   , "EVEX.256.66.0F38.W1 93"           , "ANY AVX512F-VL"],
    ["vpgatherqpd"      , "W:zmm {k}, vm32z"                         , "RM"   , "EVEX.512.66.0F38.W1 93"           , "ANY AVX512F"],
    ["vpgatherqps"      , "W:xmm {k}, vm32x"                         , "RM"   , "EVEX.128.66.0F38.W0 93"           , "ANY AVX512F-VL"],
    ["vpgatherqps"      , "W:ymm {k}, vm32y"                         , "RM"   , "EVEX.256.66.0F38.W0 93"           , "ANY AVX512F-VL"],
    ["vpgatherqps"      , "W:zmm {k}, vm32z"                         , "RM"   , "EVEX.512.66.0F38.W0 93"           , "ANY AVX512F"],
    ["vpgatherqq"       , "W:xmm {k}, vm64x"                         , "RM"   , "EVEX.128.66.0F38.W1 91"           , "ANY AVX512F-VL"],
    ["vpgatherqq"       , "W:ymm {k}, vm64y"                         , "RM"   , "EVEX.256.66.0F38.W1 91"           , "ANY AVX512F-VL"],
    ["vpgatherqq"       , "W:zmm {k}, vm64z"                         , "RM"   , "EVEX.512.66.0F38.W1 91"           , "ANY AVX512F"],
    ["vpinsrb"          , "W:xmm {kz}, xmm, r32/m8, ib"              , "RVMI" , "EVEX.NDS.128.66.0F3A.WIG 20 /r ib", "ANY AVX512BW"],
    ["vpinsrd"          , "W:xmm {kz}, xmm, r32/m32, ib"             , "RVMI" , "EVEX.NDS.128.66.0F3A.W0 22 /r ib" , "ANY AVX512DQ"],
    ["vpinsrq"          , "W:xmm {kz}, xmm, r64/m64, ib"             , "RVMI" , "EVEX.NDS.128.66.0F3A.W1 22 /r ib" , "X64 AVX512DQ"],
    ["vpinsrw"          , "W:xmm {kz}, xmm, r32/m16, ib"             , "RVMI" , "EVEX.NDS.128.66.0F.WIG C4 /r ib"  , "ANY AVX512BW"],
    ["vplzcntd"         , "W:xmm {kz}, xmm/m128/b32"                 , "RM"   , "EVEX.128.66.0F38.W0 44 /r"        , "ANY AVX512CD-VL"],
    ["vplzcntd"         , "W:ymm {kz}, ymm/m256/b32"                 , "RM"   , "EVEX.256.66.0F38.W0 44 /r"        , "ANY AVX512CD-VL"],
    ["vplzcntd"         , "W:zmm {kz}, zmm/m512/b32"                 , "RM"   , "EVEX.512.66.0F38.W0 44 /r"        , "ANY AVX512CD"],
    ["vplzcntq"         , "W:xmm {kz}, xmm/m128/b64"                 , "RM"   , "EVEX.128.66.0F38.W1 44 /r"        , "ANY AVX512CD-VL"],
    ["vplzcntq"         , "W:ymm {kz}, ymm/m256/b64"                 , "RM"   , "EVEX.256.66.0F38.W1 44 /r"        , "ANY AVX512CD-VL"],
    ["vplzcntq"         , "W:zmm {kz}, zmm/m512/b64"                 , "RM"   , "EVEX.512.66.0F38.W1 44 /r"        , "ANY AVX512CD"],
    ["vpmaddubsw"       , "W:xmm {kz}, xmm, xmm/m128"                , "RVM"  , "EVEX.NDS.128.66.0F38.WIG 04 /r"   , "ANY AVX512BW-VL"],
    ["vpmaddubsw"       , "W:ymm {kz}, ymm, ymm/m256"                , "RVM"  , "EVEX.NDS.256.66.0F38.WIG 04 /r"   , "ANY AVX512BW-VL"],
    ["vpmaddubsw"       , "W:zmm {kz}, zmm, zmm/m512"                , "RVM"  , "EVEX.NDS.512.66.0F38.WIG 04 /r"   , "ANY AVX512BW"],
    ["vpmaddwd"         , "W:xmm {kz}, xmm, xmm/m128"                , "RVM"  , "EVEX.NDS.128.66.0F.WIG F5 /r"     , "ANY AVX512BW-VL"],
    ["vpmaddwd"         , "W:ymm {kz}, ymm, ymm/m256"                , "RVM"  , "EVEX.NDS.256.66.0F.WIG F5 /r"     , "ANY AVX512BW-VL"],
    ["vpmaddwd"         , "W:zmm {kz}, zmm, zmm/m512"                , "RVM"  , "EVEX.NDS.512.66.0F.WIG F5 /r"     , "ANY AVX512BW"],
    ["vpmaxsb"          , "W:xmm {kz}, xmm, xmm/m128"                , "RVM"  , "EVEX.NDS.128.66.0F38.WIG 3C /r"   , "ANY AVX512BW-VL"],
    ["vpmaxsb"          , "W:ymm {kz}, ymm, ymm/m256"                , "RVM"  , "EVEX.NDS.256.66.0F38.WIG 3C /r"   , "ANY AVX512BW-VL"],
    ["vpmaxsb"          , "W:zmm {kz}, zmm, zmm/m512"                , "RVM"  , "EVEX.NDS.512.66.0F38.WIG 3C /r"   , "ANY AVX512BW"],
    ["vpmaxsd"          , "W:xmm {kz}, xmm, xmm/m128/b32"            , "RVM"  , "EVEX.NDS.128.66.0F38.W0 3D /r"    , "ANY AVX512F-VL"],
    ["vpmaxsd"          , "W:ymm {kz}, ymm, ymm/m256/b32"            , "RVM"  , "EVEX.NDS.256.66.0F38.W0 3D /r"    , "ANY AVX512F-VL"],
    ["vpmaxsd"          , "W:zmm {kz}, zmm, zmm/m512/b32"            , "RVM"  , "EVEX.NDS.512.66.0F38.W0 3D /r"    , "ANY AVX512F"],
    ["vpmaxsq"          , "W:xmm {kz}, xmm, xmm/m128/b64"            , "RVM"  , "EVEX.NDS.128.66.0F38.W1 3D /r"    , "ANY AVX512F-VL"],
    ["vpmaxsq"          , "W:ymm {kz}, ymm, ymm/m256/b64"            , "RVM"  , "EVEX.NDS.256.66.0F38.W1 3D /r"    , "ANY AVX512F-VL"],
    ["vpmaxsq"          , "W:zmm {kz}, zmm, zmm/m512/b64"            , "RVM"  , "EVEX.NDS.512.66.0F38.W1 3D /r"    , "ANY AVX512F"],
    ["vpmaxsw"          , "W:xmm {kz}, xmm, xmm/m128"                , "RVM"  , "EVEX.NDS.128.66.0F.WIG EE /r"     , "ANY AVX512BW-VL"],
    ["vpmaxsw"          , "W:ymm {kz}, ymm, ymm/m256"                , "RVM"  , "EVEX.NDS.256.66.0F.WIG EE /r"     , "ANY AVX512BW-VL"],
    ["vpmaxsw"          , "W:zmm {kz}, zmm, zmm/m512"                , "RVM"  , "EVEX.NDS.512.66.0F.WIG EE /r"     , "ANY AVX512BW"],
    ["vpmaxub"          , "W:xmm {kz}, xmm, xmm/m128"                , "RVM"  , "EVEX.NDS.128.66.0F.WIG DE /r"     , "ANY AVX512BW-VL"],
    ["vpmaxub"          , "W:ymm {kz}, ymm, ymm/m256"                , "RVM"  , "EVEX.NDS.256.66.0F.WIG DE /r"     , "ANY AVX512BW-VL"],
    ["vpmaxub"          , "W:zmm {kz}, zmm, zmm/m512"                , "RVM"  , "EVEX.NDS.512.66.0F.WIG DE /r"     , "ANY AVX512BW"],
    ["vpmaxud"          , "W:xmm {kz}, xmm, xmm/m128/b32"            , "RVM"  , "EVEX.NDS.128.66.0F38.W0 3F /r"    , "ANY AVX512F-VL"],
    ["vpmaxud"          , "W:ymm {kz}, ymm, ymm/m256/b32"            , "RVM"  , "EVEX.NDS.256.66.0F38.W0 3F /r"    , "ANY AVX512F-VL"],
    ["vpmaxud"          , "W:zmm {kz}, zmm, zmm/m512/b32"            , "RVM"  , "EVEX.NDS.512.66.0F38.W0 3F /r"    , "ANY AVX512F"],
    ["vpmaxuq"          , "W:xmm {kz}, xmm, xmm/m128/b64"            , "RVM"  , "EVEX.NDS.128.66.0F38.W1 3F /r"    , "ANY AVX512F-VL"],
    ["vpmaxuq"          , "W:ymm {kz}, ymm, ymm/m256/b64"            , "RVM"  , "EVEX.NDS.256.66.0F38.W1 3F /r"    , "ANY AVX512F-VL"],
    ["vpmaxuq"          , "W:zmm {kz}, zmm, zmm/m512/b64"            , "RVM"  , "EVEX.NDS.512.66.0F38.W1 3F /r"    , "ANY AVX512F"],
    ["vpmaxuw"          , "W:xmm {kz}, xmm, xmm/m128"                , "RVM"  , "EVEX.NDS.128.66.0F38.WIG 3E /r"   , "ANY AVX512BW-VL"],
    ["vpmaxuw"          , "W:ymm {kz}, ymm, ymm/m256"                , "RVM"  , "EVEX.NDS.256.66.0F38.WIG 3E /r"   , "ANY AVX512BW-VL"],
    ["vpmaxuw"          , "W:zmm {kz}, zmm, zmm/m512"                , "RVM"  , "EVEX.NDS.512.66.0F38.WIG 3E /r"   , "ANY AVX512BW"],
    ["vpminsb"          , "W:xmm {kz}, xmm, xmm/m128"                , "RVM"  , "EVEX.NDS.128.66.0F38.WIG 38 /r"   , "ANY AVX512BW-VL"],
    ["vpminsb"          , "W:ymm {kz}, ymm, ymm/m256"                , "RVM"  , "EVEX.NDS.256.66.0F38.WIG 38 /r"   , "ANY AVX512BW-VL"],
    ["vpminsb"          , "W:zmm {kz}, zmm, zmm/m512"                , "RVM"  , "EVEX.NDS.512.66.0F38.WIG 38 /r"   , "ANY AVX512BW"],
    ["vpminsd"          , "W:xmm {kz}, xmm, xmm/m128/b32"            , "RVM"  , "EVEX.NDS.128.66.0F38.W0 39 /r"    , "ANY AVX512F-VL"],
    ["vpminsd"          , "W:ymm {kz}, ymm, ymm/m256/b32"            , "RVM"  , "EVEX.NDS.256.66.0F38.W0 39 /r"    , "ANY AVX512F-VL"],
    ["vpminsd"          , "W:zmm {kz}, zmm, zmm/m512/b32"            , "RVM"  , "EVEX.NDS.512.66.0F38.W0 39 /r"    , "ANY AVX512F"],
    ["vpminsq"          , "W:xmm {kz}, xmm, xmm/m128/b64"            , "RVM"  , "EVEX.NDS.128.66.0F38.W1 39 /r"    , "ANY AVX512F-VL"],
    ["vpminsq"          , "W:ymm {kz}, ymm, ymm/m256/b64"            , "RVM"  , "EVEX.NDS.256.66.0F38.W1 39 /r"    , "ANY AVX512F-VL"],
    ["vpminsq"          , "W:zmm {kz}, zmm, zmm/m512/b64"            , "RVM"  , "EVEX.NDS.512.66.0F38.W1 39 /r"    , "ANY AVX512F"],
    ["vpminsw"          , "W:xmm {kz}, xmm, xmm/m128"                , "RVM"  , "EVEX.NDS.128.66.0F.WIG EA /r"     , "ANY AVX512BW-VL"],
    ["vpminsw"          , "W:ymm {kz}, ymm, ymm/m256"                , "RVM"  , "EVEX.NDS.256.66.0F.WIG EA /r"     , "ANY AVX512BW-VL"],
    ["vpminsw"          , "W:zmm {kz}, zmm, zmm/m512"                , "RVM"  , "EVEX.NDS.512.66.0F.WIG EA /r"     , "ANY AVX512BW"],
    ["vpminub"          , "W:xmm {kz}, xmm, xmm/m128"                , "RVM"  , "EVEX.NDS.128.66.0F DA /r"         , "ANY AVX512BW-VL"],
    ["vpminub"          , "W:ymm {kz}, ymm, ymm/m256"                , "RVM"  , "EVEX.NDS.256.66.0F DA /r"         , "ANY AVX512BW-VL"],
    ["vpminub"          , "W:zmm {kz}, zmm, zmm/m512"                , "RVM"  , "EVEX.NDS.512.66.0F DA /r"         , "ANY AVX512BW"],
    ["vpminud"          , "W:xmm {kz}, xmm, xmm/m128/b32"            , "RVM"  , "EVEX.NDS.128.66.0F38.W0 3B /r"    , "ANY AVX512F-VL"],
    ["vpminud"          , "W:ymm {kz}, ymm, ymm/m256/b32"            , "RVM"  , "EVEX.NDS.256.66.0F38.W0 3B /r"    , "ANY AVX512F-VL"],
    ["vpminud"          , "W:zmm {kz}, zmm, zmm/m512/b32"            , "RVM"  , "EVEX.NDS.512.66.0F38.W0 3B /r"    , "ANY AVX512F"],
    ["vpminuq"          , "W:xmm {kz}, xmm, xmm/m128/b64"            , "RVM"  , "EVEX.NDS.128.66.0F38.W1 3B /r"    , "ANY AVX512F-VL"],
    ["vpminuq"          , "W:ymm {kz}, ymm, ymm/m256/b64"            , "RVM"  , "EVEX.NDS.256.66.0F38.W1 3B /r"    , "ANY AVX512F-VL"],
    ["vpminuq"          , "W:zmm {kz}, zmm, zmm/m512/b64"            , "RVM"  , "EVEX.NDS.512.66.0F38.W1 3B /r"    , "ANY AVX512F"],
    ["vpminuw"          , "W:xmm {kz}, xmm, xmm/m128"                , "RVM"  , "EVEX.NDS.128.66.0F38 3A /r"       , "ANY AVX512BW-VL"],
    ["vpminuw"          , "W:ymm {kz}, ymm, ymm/m256"                , "RVM"  , "EVEX.NDS.256.66.0F38 3A /r"       , "ANY AVX512BW-VL"],
    ["vpminuw"          , "W:zmm {kz}, zmm, zmm/m512"                , "RVM"  , "EVEX.NDS.512.66.0F38 3A /r"       , "ANY AVX512BW"],
    ["vpmovb2m"         , "W:k, xmm"                                 , "RM"   , "EVEX.128.F3.0F38.W0 29 /r"        , "ANY AVX512BW-VL"],
    ["vpmovb2m"         , "W:k, ymm"                                 , "RM"   , "EVEX.256.F3.0F38.W0 29 /r"        , "ANY AVX512BW-VL"],
    ["vpmovb2m"         , "W:k, zmm"                                 , "RM"   , "EVEX.512.F3.0F38.W0 29 /r"        , "ANY AVX512BW"],
    ["vpmovd2m"         , "W:k, xmm"                                 , "RM"   , "EVEX.128.F3.0F38.W0 39 /r"        , "ANY AVX512DQ-VL"],
    ["vpmovd2m"         , "W:k, ymm"                                 , "RM"   , "EVEX.256.F3.0F38.W0 39 /r"        , "ANY AVX512DQ-VL"],
    ["vpmovd2m"         , "W:k, zmm"                                 , "RM"   , "EVEX.512.F3.0F38.W0 39 /r"        , "ANY AVX512DQ"],
    ["vpmovdb"          , "W:xmm/m32 {kz}, xmm"                      , "MR"   , "EVEX.128.F3.0F38.W0 31 /r"        , "ANY AVX512F-VL"],
    ["vpmovdb"          , "W:xmm/m64 {kz}, ymm"                      , "MR"   , "EVEX.256.F3.0F38.W0 31 /r"        , "ANY AVX512F-VL"],
    ["vpmovdb"          , "W:xmm/m128 {kz}, zmm"                     , "MR"   , "EVEX.512.F3.0F38.W0 31 /r"        , "ANY AVX512F"],
    ["vpmovdw"          , "W:xmm/m64 {kz}, xmm"                      , "MR"   , "EVEX.128.F3.0F38.W0 33 /r"        , "ANY AVX512F-VL"],
    ["vpmovdw"          , "W:xmm/m128 {kz}, ymm"                     , "MR"   , "EVEX.256.F3.0F38.W0 33 /r"        , "ANY AVX512F-VL"],
    ["vpmovdw"          , "W:ymm/m256 {kz}, zmm"                     , "MR"   , "EVEX.512.F3.0F38.W0 33 /r"        , "ANY AVX512F"],
    ["vpmovm2b"         , "W:xmm, k"                                 , "RM"   , "EVEX.128.F3.0F38.W0 28 /r"        , "ANY AVX512BW-VL"],
    ["vpmovm2b"         , "W:ymm, k"                                 , "RM"   , "EVEX.256.F3.0F38.W0 28 /r"        , "ANY AVX512BW-VL"],
    ["vpmovm2b"         , "W:zmm, k"                                 , "RM"   , "EVEX.512.F3.0F38.W0 28 /r"        , "ANY AVX512BW"],
    ["vpmovm2d"         , "W:xmm, k"                                 , "RM"   , "EVEX.128.F3.0F38.W0 38 /r"        , "ANY AVX512DQ-VL"],
    ["vpmovm2d"         , "W:ymm, k"                                 , "RM"   , "EVEX.256.F3.0F38.W0 38 /r"        , "ANY AVX512DQ-VL"],
    ["vpmovm2d"         , "W:zmm, k"                                 , "RM"   , "EVEX.512.F3.0F38.W0 38 /r"        , "ANY AVX512DQ"],
    ["vpmovm2q"         , "W:xmm, k"                                 , "RM"   , "EVEX.128.F3.0F38.W1 38 /r"        , "ANY AVX512DQ-VL"],
    ["vpmovm2q"         , "W:ymm, k"                                 , "RM"   , "EVEX.256.F3.0F38.W1 38 /r"        , "ANY AVX512DQ-VL"],
    ["vpmovm2q"         , "W:zmm, k"                                 , "RM"   , "EVEX.512.F3.0F38.W1 38 /r"        , "ANY AVX512DQ"],
    ["vpmovm2w"         , "W:xmm, k"                                 , "RM"   , "EVEX.128.F3.0F38.W1 28 /r"        , "ANY AVX512BW-VL"],
    ["vpmovm2w"         , "W:ymm, k"                                 , "RM"   , "EVEX.256.F3.0F38.W1 28 /r"        , "ANY AVX512BW-VL"],
    ["vpmovm2w"         , "W:zmm, k"                                 , "RM"   , "EVEX.512.F3.0F38.W1 28 /r"        , "ANY AVX512BW"],
    ["vpmovq2m"         , "W:k, xmm"                                 , "RM"   , "EVEX.128.F3.0F38.W1 39 /r"        , "ANY AVX512DQ-VL"],
    ["vpmovq2m"         , "W:k, ymm"                                 , "RM"   , "EVEX.256.F3.0F38.W1 39 /r"        , "ANY AVX512DQ-VL"],
    ["vpmovq2m"         , "W:k, zmm"                                 , "RM"   , "EVEX.512.F3.0F38.W1 39 /r"        , "ANY AVX512DQ"],
    ["vpmovqb"          , "W:xmm/m16 {kz}, xmm"                      , "MR"   , "EVEX.128.F3.0F38.W0 32 /r"        , "ANY AVX512F-VL"],
    ["vpmovqb"          , "W:xmm/m32 {kz}, ymm"                      , "MR"   , "EVEX.256.F3.0F38.W0 32 /r"        , "ANY AVX512F-VL"],
    ["vpmovqb"          , "W:xmm/m64 {kz}, zmm"                      , "MR"   , "EVEX.512.F3.0F38.W0 32 /r"        , "ANY AVX512F"],
    ["vpmovqd"          , "W:xmm/m64 {kz}, xmm"                      , "MR"   , "EVEX.128.F3.0F38.W0 35 /r"        , "ANY AVX512F-VL"],
    ["vpmovqd"          , "W:xmm/m128 {kz}, ymm"                     , "MR"   , "EVEX.256.F3.0F38.W0 35 /r"        , "ANY AVX512F-VL"],
    ["vpmovqd"          , "W:ymm/m256 {kz}, zmm"                     , "MR"   , "EVEX.512.F3.0F38.W0 35 /r"        , "ANY AVX512F"],
    ["vpmovqw"          , "W:xmm/m32 {kz}, xmm"                      , "MR"   , "EVEX.128.F3.0F38.W0 34 /r"        , "ANY AVX512F-VL"],
    ["vpmovqw"          , "W:xmm/m64 {kz}, ymm"                      , "MR"   , "EVEX.256.F3.0F38.W0 34 /r"        , "ANY AVX512F-VL"],
    ["vpmovqw"          , "W:xmm/m128 {kz}, zmm"                     , "MR"   , "EVEX.512.F3.0F38.W0 34 /r"        , "ANY AVX512F"],
    ["vpmovsdb"         , "W:xmm/m32 {kz}, xmm"                      , "MR"   , "EVEX.128.F3.0F38.W0 21 /r"        , "ANY AVX512F-VL"],
    ["vpmovsdb"         , "W:xmm/m64 {kz}, ymm"                      , "MR"   , "EVEX.256.F3.0F38.W0 21 /r"        , "ANY AVX512F-VL"],
    ["vpmovsdb"         , "W:xmm/m128 {kz}, zmm"                     , "MR"   , "EVEX.512.F3.0F38.W0 21 /r"        , "ANY AVX512F"],
    ["vpmovsdw"         , "W:xmm/m64 {kz}, xmm"                      , "MR"   , "EVEX.128.F3.0F38.W0 23 /r"        , "ANY AVX512F-VL"],
    ["vpmovsdw"         , "W:xmm/m128 {kz}, ymm"                     , "MR"   , "EVEX.256.F3.0F38.W0 23 /r"        , "ANY AVX512F-VL"],
    ["vpmovsdw"         , "W:ymm/m256 {kz}, zmm"                     , "MR"   , "EVEX.512.F3.0F38.W0 23 /r"        , "ANY AVX512F"],
    ["vpmovsqb"         , "W:xmm/m16 {kz}, xmm"                      , "MR"   , "EVEX.128.F3.0F38.W0 22 /r"        , "ANY AVX512F-VL"],
    ["vpmovsqb"         , "W:xmm/m32 {kz}, ymm"                      , "MR"   , "EVEX.256.F3.0F38.W0 22 /r"        , "ANY AVX512F-VL"],
    ["vpmovsqb"         , "W:xmm/m64 {kz}, zmm"                      , "MR"   , "EVEX.512.F3.0F38.W0 22 /r"        , "ANY AVX512F"],
    ["vpmovsqd"         , "W:xmm/m64 {kz}, xmm"                      , "MR"   , "EVEX.128.F3.0F38.W0 25 /r"        , "ANY AVX512F-VL"],
    ["vpmovsqd"         , "W:xmm/m128 {kz}, ymm"                     , "MR"   , "EVEX.256.F3.0F38.W0 25 /r"        , "ANY AVX512F-VL"],
    ["vpmovsqd"         , "W:ymm/m256 {kz}, zmm"                     , "MR"   , "EVEX.512.F3.0F38.W0 25 /r"        , "ANY AVX512F"],
    ["vpmovsqw"         , "W:xmm/m32 {kz}, xmm"                      , "MR"   , "EVEX.128.F3.0F38.W0 24 /r"        , "ANY AVX512F-VL"],
    ["vpmovsqw"         , "W:xmm/m64 {kz}, ymm"                      , "MR"   , "EVEX.256.F3.0F38.W0 24 /r"        , "ANY AVX512F-VL"],
    ["vpmovsqw"         , "W:xmm/m128 {kz}, zmm"                     , "MR"   , "EVEX.512.F3.0F38.W0 24 /r"        , "ANY AVX512F"],
    ["vpmovswb"         , "W:xmm/m64 {kz}, xmm"                      , "MR"   , "EVEX.128.F3.0F38.W0 20 /r"        , "ANY AVX512BW-VL"],
    ["vpmovswb"         , "W:xmm/m128 {kz}, ymm"                     , "MR"   , "EVEX.256.F3.0F38.W0 20 /r"        , "ANY AVX512BW-VL"],
    ["vpmovswb"         , "W:ymm/m256 {kz}, zmm"                     , "MR"   , "EVEX.512.F3.0F38.W0 20 /r"        , "ANY AVX512BW"],
    ["vpmovsxbd"        , "W:xmm {kz}, xmm/m32"                      , "RM"   , "EVEX.128.66.0F38.WIG 21 /r"       , "ANY AVX512F-VL"],
    ["vpmovsxbd"        , "W:ymm {kz}, xmm/m64"                      , "RM"   , "EVEX.256.66.0F38.WIG 21 /r"       , "ANY AVX512F-VL"],
    ["vpmovsxbd"        , "W:zmm {kz}, xmm/m128"                     , "RM"   , "EVEX.512.66.0F38.WIG 21 /r"       , "ANY AVX512F"],
    ["vpmovsxbq"        , "W:xmm {kz}, xmm/m16"                      , "RM"   , "EVEX.128.66.0F38.WIG 22 /r"       , "ANY AVX512F-VL"],
    ["vpmovsxbq"        , "W:ymm {kz}, xmm/m32"                      , "RM"   , "EVEX.256.66.0F38.WIG 22 /r"       , "ANY AVX512F-VL"],
    ["vpmovsxbq"        , "W:zmm {kz}, xmm/m64"                      , "RM"   , "EVEX.512.66.0F38.WIG 22 /r"       , "ANY AVX512F"],
    ["vpmovsxbw"        , "W:xmm {kz}, xmm/m64"                      , "RM"   , "EVEX.128.66.0F38.WIG 20 /r"       , "ANY AVX512BW-VL"],
    ["vpmovsxbw"        , "W:ymm {kz}, xmm/m128"                     , "RM"   , "EVEX.256.66.0F38.WIG 20 /r"       , "ANY AVX512BW-VL"],
    ["vpmovsxbw"        , "W:zmm {kz}, ymm/m256"                     , "RM"   , "EVEX.512.66.0F38.WIG 20 /r"       , "ANY AVX512BW"],
    ["vpmovsxdq"        , "W:xmm {kz}, xmm/m32"                      , "RM"   , "EVEX.128.66.0F38.W0 25 /r"        , "ANY AVX512F-VL"],
    ["vpmovsxdq"        , "W:ymm {kz}, xmm/m64"                      , "RM"   , "EVEX.256.66.0F38.W0 25 /r"        , "ANY AVX512F-VL"],
    ["vpmovsxdq"        , "W:zmm {kz}, xmm/m128"                     , "RM"   , "EVEX.512.66.0F38.W0 25 /r"        , "ANY AVX512F"],
    ["vpmovsxwd"        , "W:xmm {kz}, xmm/m64"                      , "RM"   , "EVEX.128.66.0F38.WIG 23 /r"       , "ANY AVX512F-VL"],
    ["vpmovsxwd"        , "W:ymm {kz}, xmm/m128"                     , "RM"   , "EVEX.256.66.0F38.WIG 23 /r"       , "ANY AVX512F-VL"],
    ["vpmovsxwd"        , "W:zmm {kz}, ymm/m256"                     , "RM"   , "EVEX.512.66.0F38.WIG 23 /r"       , "ANY AVX512F"],
    ["vpmovsxwq"        , "W:xmm {kz}, xmm/m32"                      , "RM"   , "EVEX.128.66.0F38.WIG 24 /r"       , "ANY AVX512F-VL"],
    ["vpmovsxwq"        , "W:ymm {kz}, xmm/m64"                      , "RM"   , "EVEX.256.66.0F38.WIG 24 /r"       , "ANY AVX512F-VL"],
    ["vpmovsxwq"        , "W:zmm {kz}, xmm/m128"                     , "RM"   , "EVEX.512.66.0F38.WIG 24 /r"       , "ANY AVX512F"],
    ["vpmovusdb"        , "W:xmm/m32 {kz}, xmm"                      , "MR"   , "EVEX.128.F3.0F38.W0 11 /r"        , "ANY AVX512F-VL"],
    ["vpmovusdb"        , "W:xmm/m64 {kz}, ymm"                      , "MR"   , "EVEX.256.F3.0F38.W0 11 /r"        , "ANY AVX512F-VL"],
    ["vpmovusdb"        , "W:xmm/m128 {kz}, zmm"                     , "MR"   , "EVEX.512.F3.0F38.W0 11 /r"        , "ANY AVX512F"],
    ["vpmovusdw"        , "W:xmm/m64 {kz}, xmm"                      , "MR"   , "EVEX.128.F3.0F38.W0 13 /r"        , "ANY AVX512F-VL"],
    ["vpmovusdw"        , "W:xmm/m128 {kz}, ymm"                     , "MR"   , "EVEX.256.F3.0F38.W0 13 /r"        , "ANY AVX512F-VL"],
    ["vpmovusdw"        , "W:ymm/m256 {kz}, zmm"                     , "MR"   , "EVEX.512.F3.0F38.W0 13 /r"        , "ANY AVX512F"],
    ["vpmovusqb"        , "W:xmm/m16 {kz}, xmm"                      , "MR"   , "EVEX.128.F3.0F38.W0 12 /r"        , "ANY AVX512F-VL"],
    ["vpmovusqb"        , "W:xmm/m32 {kz}, ymm"                      , "MR"   , "EVEX.256.F3.0F38.W0 12 /r"        , "ANY AVX512F-VL"],
    ["vpmovusqb"        , "W:xmm/m64 {kz}, zmm"                      , "MR"   , "EVEX.512.F3.0F38.W0 12 /r"        , "ANY AVX512F"],
    ["vpmovusqd"        , "W:xmm/m64 {kz}, xmm"                      , "MR"   , "EVEX.128.F3.0F38.W0 15 /r"        , "ANY AVX512F-VL"],
    ["vpmovusqd"        , "W:xmm/m128 {kz}, ymm"                     , "MR"   , "EVEX.256.F3.0F38.W0 15 /r"        , "ANY AVX512F-VL"],
    ["vpmovusqd"        , "W:ymm/m256 {kz}, zmm"                     , "MR"   , "EVEX.512.F3.0F38.W0 15 /r"        , "ANY AVX512F"],
    ["vpmovusqw"        , "W:xmm/m32 {kz}, xmm"                      , "MR"   , "EVEX.128.F3.0F38.W0 14 /r"        , "ANY AVX512F-VL"],
    ["vpmovusqw"        , "W:xmm/m64 {kz}, ymm"                      , "MR"   , "EVEX.256.F3.0F38.W0 14 /r"        , "ANY AVX512F-VL"],
    ["vpmovusqw"        , "W:xmm/m128 {kz}, zmm"                     , "MR"   , "EVEX.512.F3.0F38.W0 14 /r"        , "ANY AVX512F"],
    ["vpmovuswb"        , "W:xmm/m64 {kz}, xmm"                      , "MR"   , "EVEX.128.F3.0F38.W0 10 /r"        , "ANY AVX512BW-VL"],
    ["vpmovuswb"        , "W:xmm/m128 {kz}, ymm"                     , "MR"   , "EVEX.256.F3.0F38.W0 10 /r"        , "ANY AVX512BW-VL"],
    ["vpmovuswb"        , "W:ymm/m256 {kz}, zmm"                     , "MR"   , "EVEX.512.F3.0F38.W0 10 /r"        , "ANY AVX512BW"],
    ["vpmovw2m"         , "W:k, xmm"                                 , "RM"   , "EVEX.128.F3.0F38.W1 29 /r"        , "ANY AVX512BW-VL"],
    ["vpmovw2m"         , "W:k, ymm"                                 , "RM"   , "EVEX.256.F3.0F38.W1 29 /r"        , "ANY AVX512BW-VL"],
    ["vpmovw2m"         , "W:k, zmm"                                 , "RM"   , "EVEX.512.F3.0F38.W1 29 /r"        , "ANY AVX512BW"],
    ["vpmovwb"          , "W:xmm/m64 {kz}, xmm"                      , "MR"   , "EVEX.128.F3.0F38.W0 30 /r"        , "ANY AVX512BW-VL"],
    ["vpmovwb"          , "W:xmm/m128 {kz}, ymm"                     , "MR"   , "EVEX.256.F3.0F38.W0 30 /r"        , "ANY AVX512BW-VL"],
    ["vpmovwb"          , "W:ymm/m256 {kz}, zmm"                     , "MR"   , "EVEX.512.F3.0F38.W0 30 /r"        , "ANY AVX512BW"],
    ["vpmovzxbd"        , "W:xmm {kz}, xmm/m32"                      , "RM"   , "EVEX.128.66.0F38.WIG 31 /r"       , "ANY AVX512F-VL"],
    ["vpmovzxbd"        , "W:ymm {kz}, xmm/m64"                      , "RM"   , "EVEX.256.66.0F38.WIG 31 /r"       , "ANY AVX512F-VL"],
    ["vpmovzxbd"        , "W:zmm {kz}, xmm/m128"                     , "RM"   , "EVEX.512.66.0F38.WIG 31 /r"       , "ANY AVX512F"],
    ["vpmovzxbq"        , "W:xmm {kz}, xmm/m16"                      , "RM"   , "EVEX.128.66.0F38.WIG 32 /r"       , "ANY AVX512F-VL"],
    ["vpmovzxbq"        , "W:ymm {kz}, xmm/m32"                      , "RM"   , "EVEX.256.66.0F38.WIG 32 /r"       , "ANY AVX512F-VL"],
    ["vpmovzxbq"        , "W:zmm {kz}, xmm/m64"                      , "RM"   , "EVEX.512.66.0F38.WIG 32 /r"       , "ANY AVX512F"],
    ["vpmovzxbw"        , "W:xmm {kz}, xmm/m64"                      , "RM"   , "EVEX.128.66.0F38.WIG 30 /r"       , "ANY AVX512BW-VL"],
    ["vpmovzxbw"        , "W:ymm {kz}, xmm/m128"                     , "RM"   , "EVEX.256.66.0F38.WIG 30 /r"       , "ANY AVX512BW-VL"],
    ["vpmovzxbw"        , "W:zmm {kz}, ymm/m256"                     , "RM"   , "EVEX.512.66.0F38.WIG 30 /r"       , "ANY AVX512BW"],
    ["vpmovzxdq"        , "W:xmm {kz}, xmm/m32"                      , "RM"   , "EVEX.128.66.0F38.W0 35 /r"        , "ANY AVX512F-VL"],
    ["vpmovzxdq"        , "W:ymm {kz}, xmm/m64"                      , "RM"   , "EVEX.256.66.0F38.W0 35 /r"        , "ANY AVX512F-VL"],
    ["vpmovzxdq"        , "W:zmm {kz}, xmm/m128"                     , "RM"   , "EVEX.512.66.0F38.W0 35 /r"        , "ANY AVX512F"],
    ["vpmovzxwd"        , "W:xmm {kz}, xmm/m64"                      , "RM"   , "EVEX.128.66.0F38.WIG 33 /r"       , "ANY AVX512F-VL"],
    ["vpmovzxwd"        , "W:ymm {kz}, xmm/m128"                     , "RM"   , "EVEX.256.66.0F38.WIG 33 /r"       , "ANY AVX512F-VL"],
    ["vpmovzxwd"        , "W:zmm {kz}, ymm/m256"                     , "RM"   , "EVEX.512.66.0F38.WIG 33 /r"       , "ANY AVX512F"],
    ["vpmovzxwq"        , "W:xmm {kz}, xmm/m32"                      , "RM"   , "EVEX.128.66.0F38.WIG 34 /r"       , "ANY AVX512F-VL"],
    ["vpmovzxwq"        , "W:ymm {kz}, xmm/m64"                      , "RM"   , "EVEX.256.66.0F38.WIG 34 /r"       , "ANY AVX512F-VL"],
    ["vpmovzxwq"        , "W:zmm {kz}, xmm/m128"                     , "RM"   , "EVEX.512.66.0F38.WIG 34 /r"       , "ANY AVX512F"],
    ["vpmuldq"          , "W:xmm {kz}, xmm, xmm/m128/b64"            , "RVM"  , "EVEX.NDS.128.66.0F38.W1 28 /r"    , "ANY AVX512F-VL"],
    ["vpmuldq"          , "W:ymm {kz}, ymm, ymm/m256/b64"            , "RVM"  , "EVEX.NDS.256.66.0F38.W1 28 /r"    , "ANY AVX512F-VL"],
    ["vpmuldq"          , "W:zmm {kz}, zmm, zmm/m512/b64"            , "RVM"  , "EVEX.NDS.512.66.0F38.W1 28 /r"    , "ANY AVX512F"],
    ["vpmulhrsw"        , "W:xmm {kz}, xmm, xmm/m128"                , "RVM"  , "EVEX.NDS.128.66.0F38.WIG 0B /r"   , "ANY AVX512BW-VL"],
    ["vpmulhrsw"        , "W:ymm {kz}, ymm, ymm/m256"                , "RVM"  , "EVEX.NDS.256.66.0F38.WIG 0B /r"   , "ANY AVX512BW-VL"],
    ["vpmulhrsw"        , "W:zmm {kz}, zmm, zmm/m512"                , "RVM"  , "EVEX.NDS.512.66.0F38.WIG 0B /r"   , "ANY AVX512BW"],
    ["vpmulhuw"         , "W:xmm {kz}, xmm, xmm/m128"                , "RVM"  , "EVEX.NDS.128.66.0F.WIG E4 /r"     , "ANY AVX512BW-VL"],
    ["vpmulhuw"         , "W:ymm {kz}, ymm, ymm/m256"                , "RVM"  , "EVEX.NDS.256.66.0F.WIG E4 /r"     , "ANY AVX512BW-VL"],
    ["vpmulhuw"         , "W:zmm {kz}, zmm, zmm/m512"                , "RVM"  , "EVEX.NDS.512.66.0F.WIG E4 /r"     , "ANY AVX512BW"],
    ["vpmulhw"          , "W:xmm {kz}, xmm, xmm/m128"                , "RVM"  , "EVEX.NDS.128.66.0F.WIG E5 /r"     , "ANY AVX512BW-VL"],
    ["vpmulhw"          , "W:ymm {kz}, ymm, ymm/m256"                , "RVM"  , "EVEX.NDS.256.66.0F.WIG E5 /r"     , "ANY AVX512BW-VL"],
    ["vpmulhw"          , "W:zmm {kz}, zmm, zmm/m512"                , "RVM"  , "EVEX.NDS.512.66.0F.WIG E5 /r"     , "ANY AVX512BW"],
    ["vpmulld"          , "W:xmm {kz}, xmm, xmm/m128/b32"            , "RVM"  , "EVEX.NDS.128.66.0F38.W0 40 /r"    , "ANY AVX512F-VL"],
    ["vpmulld"          , "W:ymm {kz}, ymm, ymm/m256/b32"            , "RVM"  , "EVEX.NDS.256.66.0F38.W0 40 /r"    , "ANY AVX512F-VL"],
    ["vpmulld"          , "W:zmm {kz}, zmm, zmm/m512/b32"            , "RVM"  , "EVEX.NDS.512.66.0F38.W0 40 /r"    , "ANY AVX512F"],
    ["vpmullq"          , "W:xmm {kz}, xmm, xmm/m128/b64"            , "RVM"  , "EVEX.NDS.128.66.0F38.W1 40 /r"    , "ANY AVX512DQ-VL"],
    ["vpmullq"          , "W:ymm {kz}, ymm, ymm/m256/b64"            , "RVM"  , "EVEX.NDS.256.66.0F38.W1 40 /r"    , "ANY AVX512DQ-VL"],
    ["vpmullq"          , "W:zmm {kz}, zmm, zmm/m512/b64"            , "RVM"  , "EVEX.NDS.512.66.0F38.W1 40 /r"    , "ANY AVX512DQ"],
    ["vpmullw"          , "W:xmm {kz}, xmm, xmm/m128"                , "RVM"  , "EVEX.NDS.128.66.0F.WIG D5 /r"     , "ANY AVX512BW-VL"],
    ["vpmullw"          , "W:ymm {kz}, ymm, ymm/m256"                , "RVM"  , "EVEX.NDS.256.66.0F.WIG D5 /r"     , "ANY AVX512BW-VL"],
    ["vpmullw"          , "W:zmm {kz}, zmm, zmm/m512"                , "RVM"  , "EVEX.NDS.512.66.0F.WIG D5 /r"     , "ANY AVX512BW"],
    ["vpmuludq"         , "W:xmm {kz}, xmm, xmm/m128/b64"            , "RVM"  , "EVEX.NDS.128.66.0F.W1 F4 /r"      , "ANY AVX512F-VL"],
    ["vpmuludq"         , "W:ymm {kz}, ymm, ymm/m256/b64"            , "RVM"  , "EVEX.NDS.256.66.0F.W1 F4 /r"      , "ANY AVX512F-VL"],
    ["vpmuludq"         , "W:zmm {kz}, zmm, zmm/m512/b64"            , "RVM"  , "EVEX.NDS.512.66.0F.W1 F4 /r"      , "ANY AVX512F"],
    ["vpord"            , "W:xmm {kz}, xmm, xmm/m128/b32"            , "RVM"  , "EVEX.NDS.128.66.0F.W0 EB /r"      , "ANY AVX512F-VL"],
    ["vpord"            , "W:ymm {kz}, ymm, ymm/m256/b32"            , "RVM"  , "EVEX.NDS.256.66.0F.W0 EB /r"      , "ANY AVX512F-VL"],
    ["vpord"            , "W:zmm {kz}, zmm, zmm/m512/b32"            , "RVM"  , "EVEX.NDS.512.66.0F.W0 EB /r"      , "ANY AVX512F"],
    ["vporq"            , "W:xmm {kz}, xmm, xmm/m128/b64"            , "RVM"  , "EVEX.NDS.128.66.0F.W1 EB /r"      , "ANY AVX512F-VL"],
    ["vporq"            , "W:ymm {kz}, ymm, ymm/m256/b64"            , "RVM"  , "EVEX.NDS.256.66.0F.W1 EB /r"      , "ANY AVX512F-VL"],
    ["vporq"            , "W:zmm {kz}, zmm, zmm/m512/b64"            , "RVM"  , "EVEX.NDS.512.66.0F.W1 EB /r"      , "ANY AVX512F"],
    ["vprold"           , "W:xmm {kz}, xmm/m128/b32, ib"             , "VMI"  , "EVEX.NDD.128.66.0F.W0 72 /1 ib"   , "ANY AVX512F-VL"],
    ["vprold"           , "W:ymm {kz}, ymm/m256/b32, ib"             , "VMI"  , "EVEX.NDD.256.66.0F.W0 72 /1 ib"   , "ANY AVX512F-VL"],
    ["vprold"           , "W:zmm {kz}, zmm/m512/b32, ib"             , "VMI"  , "EVEX.NDD.512.66.0F.W0 72 /1 ib"   , "ANY AVX512F"],
    ["vprolq"           , "W:xmm {kz}, xmm/m128/b64, ib"             , "VMI"  , "EVEX.NDD.128.66.0F.W1 72 /1 ib"   , "ANY AVX512F-VL"],
    ["vprolq"           , "W:ymm {kz}, ymm/m256/b64, ib"             , "VMI"  , "EVEX.NDD.256.66.0F.W1 72 /1 ib"   , "ANY AVX512F-VL"],
    ["vprolq"           , "W:zmm {kz}, zmm/m512/b64, ib"             , "VMI"  , "EVEX.NDD.512.66.0F.W1 72 /1 ib"   , "ANY AVX512F"],
    ["vprolvd"          , "W:xmm {kz}, xmm, xmm/m128/b32"            , "RVM"  , "EVEX.NDS.128.66.0F38.W0 15 /r"    , "ANY AVX512F-VL"],
    ["vprolvd"          , "W:ymm {kz}, ymm, ymm/m256/b32"            , "RVM"  , "EVEX.NDS.256.66.0F38.W0 15 /r"    , "ANY AVX512F-VL"],
    ["vprolvd"          , "W:zmm {kz}, zmm, zmm/m512/b32"            , "RVM"  , "EVEX.NDS.512.66.0F38.W0 15 /r"    , "ANY AVX512F"],
    ["vprolvq"          , "W:xmm {kz}, xmm, xmm/m128/b64"            , "RVM"  , "EVEX.NDS.128.66.0F38.W1 15 /r"    , "ANY AVX512F-VL"],
    ["vprolvq"          , "W:ymm {kz}, ymm, ymm/m256/b64"            , "RVM"  , "EVEX.NDS.256.66.0F38.W1 15 /r"    , "ANY AVX512F-VL"],
    ["vprolvq"          , "W:zmm {kz}, zmm, zmm/m512/b64"            , "RVM"  , "EVEX.NDS.512.66.0F38.W1 15 /r"    , "ANY AVX512F"],
    ["vprord"           , "W:xmm {kz}, xmm/m128/b32, ib"             , "VMI"  , "EVEX.NDD.128.66.0F.W0 72 /0 ib"   , "ANY AVX512F-VL"],
    ["vprord"           , "W:ymm {kz}, ymm/m256/b32, ib"             , "VMI"  , "EVEX.NDD.256.66.0F.W0 72 /0 ib"   , "ANY AVX512F-VL"],
    ["vprord"           , "W:zmm {kz}, zmm/m512/b32, ib"             , "VMI"  , "EVEX.NDD.512.66.0F.W0 72 /0 ib"   , "ANY AVX512F"],
    ["vprorq"           , "W:xmm {kz}, xmm/m128/b64, ib"             , "VMI"  , "EVEX.NDD.128.66.0F.W1 72 /0 ib"   , "ANY AVX512F-VL"],
    ["vprorq"           , "W:ymm {kz}, ymm/m256/b64, ib"             , "VMI"  , "EVEX.NDD.256.66.0F.W1 72 /0 ib"   , "ANY AVX512F-VL"],
    ["vprorq"           , "W:zmm {kz}, zmm/m512/b64, ib"             , "VMI"  , "EVEX.NDD.512.66.0F.W1 72 /0 ib"   , "ANY AVX512F"],
    ["vprorvd"          , "W:xmm {kz}, xmm, xmm/m128/b32"            , "RVM"  , "EVEX.NDS.128.66.0F38.W0 14 /r"    , "ANY AVX512F-VL"],
    ["vprorvd"          , "W:ymm {kz}, ymm, ymm/m256/b32"            , "RVM"  , "EVEX.NDS.256.66.0F38.W0 14 /r"    , "ANY AVX512F-VL"],
    ["vprorvd"          , "W:zmm {kz}, zmm, zmm/m512/b32"            , "RVM"  , "EVEX.NDS.512.66.0F38.W0 14 /r"    , "ANY AVX512F"],
    ["vprorvq"          , "W:xmm {kz}, xmm, xmm/m128/b64"            , "RVM"  , "EVEX.NDS.128.66.0F38.W1 14 /r"    , "ANY AVX512F-VL"],
    ["vprorvq"          , "W:ymm {kz}, ymm, ymm/m256/b64"            , "RVM"  , "EVEX.NDS.256.66.0F38.W1 14 /r"    , "ANY AVX512F-VL"],
    ["vprorvq"          , "W:zmm {kz}, zmm, zmm/m512/b64"            , "RVM"  , "EVEX.NDS.512.66.0F38.W1 14 /r"    , "ANY AVX512F"],
    ["vpsadbw"          , "W:xmm, xmm, xmm/m128"                     , "RVM"  , "EVEX.NDS.128.66.0F.WIG F6 /r"     , "ANY AVX512BW-VL"],
    ["vpsadbw"          , "W:ymm, ymm, ymm/m256"                     , "RVM"  , "EVEX.NDS.256.66.0F.WIG F6 /r"     , "ANY AVX512BW-VL"],
    ["vpsadbw"          , "W:zmm, zmm, zmm/m512"                     , "RVM"  , "EVEX.NDS.512.66.0F.WIG F6 /r"     , "ANY AVX512BW"],
    ["vpscatterdd"      , "W:vm32x {k}, xmm"                         , "MR"   , "EVEX.128.66.0F38.W0 A0"           , "ANY AVX512F-VL"],
    ["vpscatterdd"      , "W:vm32y {k}, ymm"                         , "MR"   , "EVEX.256.66.0F38.W0 A0"           , "ANY AVX512F-VL"],
    ["vpscatterdd"      , "W:vm32z {k}, zmm"                         , "MR"   , "EVEX.512.66.0F38.W0 A0"           , "ANY AVX512F"],
    ["vpscatterdq"      , "W:vm32x {k}, xmm"                         , "MR"   , "EVEX.128.66.0F38.W1 A0"           , "ANY AVX512F-VL"],
    ["vpscatterdq"      , "W:vm32y {k}, ymm"                         , "MR"   , "EVEX.256.66.0F38.W1 A0"           , "ANY AVX512F-VL"],
    ["vpscatterdq"      , "W:vm32z {k}, zmm"                         , "MR"   , "EVEX.512.66.0F38.W1 A0"           , "ANY AVX512F"],
    ["vpscatterqd"      , "W:vm64x {k}, xmm"                         , "MR"   , "EVEX.128.66.0F38.W0 A1"           , "ANY AVX512F-VL"],
    ["vpscatterqd"      , "W:vm64y {k}, xmm"                         , "MR"   , "EVEX.256.66.0F38.W0 A1"           , "ANY AVX512F-VL"],
    ["vpscatterqd"      , "W:vm64z {k}, ymm"                         , "MR"   , "EVEX.512.66.0F38.W0 A1"           , "ANY AVX512F"],
    ["vpscatterqq"      , "W:vm64x {k}, xmm"                         , "MR"   , "EVEX.128.66.0F38.W1 A1"           , "ANY AVX512F-VL"],
    ["vpscatterqq"      , "W:vm64y {k}, ymm"                         , "MR"   , "EVEX.256.66.0F38.W1 A1"           , "ANY AVX512F-VL"],
    ["vpscatterqq"      , "W:vm64z {k}, zmm"                         , "MR"   , "EVEX.512.66.0F38.W1 A1"           , "ANY AVX512F"],
    ["vpshufb"          , "W:xmm {kz}, xmm, xmm/m128"                , "RVM"  , "EVEX.NDS.128.66.0F38.WIG 00 /r"   , "ANY AVX512BW-VL"],
    ["vpshufb"          , "W:ymm {kz}, ymm, ymm/m256"                , "RVM"  , "EVEX.NDS.256.66.0F38.WIG 00 /r"   , "ANY AVX512BW-VL"],
    ["vpshufb"          , "W:zmm {kz}, zmm, zmm/m512"                , "RVM"  , "EVEX.NDS.512.66.0F38.WIG 00 /r"   , "ANY AVX512BW"],
    ["vpshufd"          , "W:xmm {kz}, xmm, xmm/m128/b32, ib"        , "RMI"  , "EVEX.128.66.0F.W0 70 /r ib"       , "ANY AVX512F-VL"],
    ["vpshufd"          , "W:ymm {kz}, ymm, ymm/m256/b32, ib"        , "RMI"  , "EVEX.256.66.0F.W0 70 /r ib"       , "ANY AVX512F-VL"],
    ["vpshufd"          , "W:zmm {kz}, zmm, zmm/m512/b32, ib"        , "RMI"  , "EVEX.512.66.0F.W0 70 /r ib"       , "ANY AVX512F"],
    ["vpshufhw"         , "W:xmm {kz}, xmm, xmm/m128, ib"            , "RMI"  , "EVEX.128.F3.0F.WIG 70 /r ib"      , "ANY AVX512BW-VL"],
    ["vpshufhw"         , "W:ymm {kz}, ymm, ymm/m256, ib"            , "RMI"  , "EVEX.256.F3.0F.WIG 70 /r ib"      , "ANY AVX512BW-VL"],
    ["vpshufhw"         , "W:zmm {kz}, zmm, zmm/m512, ib"            , "RMI"  , "EVEX.512.F3.0F.WIG 70 /r ib"      , "ANY AVX512BW"],
    ["vpshuflw"         , "W:xmm {kz}, xmm, xmm/m128, ib"            , "RMI"  , "EVEX.128.F2.0F.WIG 70 /r ib"      , "ANY AVX512BW-VL"],
    ["vpshuflw"         , "W:ymm {kz}, ymm, ymm/m256, ib"            , "RMI"  , "EVEX.256.F2.0F.WIG 70 /r ib"      , "ANY AVX512BW-VL"],
    ["vpshuflw"         , "W:zmm {kz}, zmm, zmm/m512, ib"            , "RMI"  , "EVEX.512.F2.0F.WIG 70 /r ib"      , "ANY AVX512BW"],
    ["vpslld"           , "W:xmm {kz}, xmm, xmm/m128"                , "RVM"  , "EVEX.NDS.128.66.0F.W0 F2 /r"      , "ANY AVX512F-VL"],
    ["vpslld"           , "W:xmm {kz}, xmm/m128/b32, ib"             , "VMI"  , "EVEX.NDD.128.66.0F.W0 72 /6 ib"   , "ANY AVX512F-VL"],
    ["vpslld"           , "W:ymm {kz}, ymm, xmm/m128"                , "RVM"  , "EVEX.NDS.256.66.0F.W0 F2 /r"      , "ANY AVX512F-VL"],
    ["vpslld"           , "W:ymm {kz}, ymm/m256/b32, ib"             , "VMI"  , "EVEX.NDD.256.66.0F.W0 72 /6 ib"   , "ANY AVX512F-VL"],
    ["vpslld"           , "W:zmm {kz}, zmm, xmm/m128"                , "RVM"  , "EVEX.NDS.512.66.0F.W0 F2 /r"      , "ANY AVX512F"],
    ["vpslld"           , "W:zmm {kz}, zmm/m512/b32, ib"             , "VMI"  , "EVEX.NDD.512.66.0F.W0 72 /6 ib"   , "ANY AVX512F"],
    ["vpslldq"          , "W:xmm, xmm, xmm/m128, ib"                 , "VMI"  , "EVEX.NDD.128.66.0F 73 /7 ib"      , "ANY AVX512BW-VL"],
    ["vpslldq"          , "W:ymm, ymm, ymm/m256, ib"                 , "VMI"  , "EVEX.NDD.256.66.0F 73 /7 ib"      , "ANY AVX512BW-VL"],
    ["vpslldq"          , "W:zmm, zmm, zmm/m512, ib"                 , "VMI"  , "EVEX.NDD.512.66.0F 73 /7 ib"      , "ANY AVX512BW"],
    ["vpsllq"           , "W:xmm {kz}, xmm, xmm/m128"                , "RVM"  , "EVEX.NDS.128.66.0F.W1 F3 /r"      , "ANY AVX512F-VL"],
    ["vpsllq"           , "W:xmm {kz}, xmm/m128/b64, ib"             , "VMI"  , "EVEX.NDD.128.66.0F.W1 73 /6 ib"   , "ANY AVX512F-VL"],
    ["vpsllq"           , "W:ymm {kz}, ymm, xmm/m128"                , "RVM"  , "EVEX.NDS.256.66.0F.W1 F3 /r"      , "ANY AVX512F-VL"],
    ["vpsllq"           , "W:ymm {kz}, ymm/m256/b64, ib"             , "VMI"  , "EVEX.NDD.256.66.0F.W1 73 /6 ib"   , "ANY AVX512F-VL"],
    ["vpsllq"           , "W:zmm {kz}, zmm, xmm/m128"                , "RVM"  , "EVEX.NDS.512.66.0F.W1 F3 /r"      , "ANY AVX512F"],
    ["vpsllq"           , "W:zmm {kz}, zmm/m512/b64, ib"             , "VMI"  , "EVEX.NDD.512.66.0F.W1 73 /6 ib"   , "ANY AVX512F"],
    ["vpsllvd"          , "W:xmm {kz}, xmm, xmm/m128/b32"            , "RVM"  , "EVEX.NDS.128.66.0F38.W0 47 /r"    , "ANY AVX512F-VL"],
    ["vpsllvd"          , "W:ymm {kz}, ymm, ymm/m256/b32"            , "RVM"  , "EVEX.NDS.256.66.0F38.W0 47 /r"    , "ANY AVX512F-VL"],
    ["vpsllvd"          , "W:zmm {kz}, zmm, zmm/m512/b32"            , "RVM"  , "EVEX.NDS.512.66.0F38.W0 47 /r"    , "ANY AVX512F"],
    ["vpsllvq"          , "W:xmm {kz}, xmm, xmm/m128/b64"            , "RVM"  , "EVEX.NDS.128.66.0F38.W1 47 /r"    , "ANY AVX512F-VL"],
    ["vpsllvq"          , "W:ymm {kz}, ymm, ymm/m256/b64"            , "RVM"  , "EVEX.NDS.256.66.0F38.W1 47 /r"    , "ANY AVX512F-VL"],
    ["vpsllvq"          , "W:zmm {kz}, zmm, zmm/m512/b64"            , "RVM"  , "EVEX.NDS.512.66.0F38.W1 47 /r"    , "ANY AVX512F"],
    ["vpsllvw"          , "W:xmm {kz}, xmm, xmm/m128"                , "RVM"  , "EVEX.NDS.128.66.0F38.W1 12 /r"    , "ANY AVX512BW-VL"],
    ["vpsllvw"          , "W:ymm {kz}, ymm, ymm/m256"                , "RVM"  , "EVEX.NDS.256.66.0F38.W1 12 /r"    , "ANY AVX512BW-VL"],
    ["vpsllvw"          , "W:zmm {kz}, zmm, zmm/m512"                , "RVM"  , "EVEX.NDS.512.66.0F38.W1 12 /r"    , "ANY AVX512BW"],
    ["vpsllw"           , "W:xmm {kz}, xmm, xmm/m128"                , "RVM"  , "EVEX.NDS.128.66.0F.WIG F1 /r"     , "ANY AVX512BW-VL"],
    ["vpsllw"           , "W:xmm {kz}, xmm/m128, ib"                 , "VMI"  , "EVEX.NDD.128.66.0F.WIG 71 /6 ib"  , "ANY AVX512BW-VL"],
    ["vpsllw"           , "W:ymm {kz}, ymm, xmm/m128"                , "RVM"  , "EVEX.NDS.256.66.0F.WIG F1 /r"     , "ANY AVX512BW-VL"],
    ["vpsllw"           , "W:ymm {kz}, ymm/m256, ib"                 , "VMI"  , "EVEX.NDD.256.66.0F.WIG 71 /6 ib"  , "ANY AVX512BW-VL"],
    ["vpsllw"           , "W:zmm {kz}, zmm, xmm/m128"                , "RVM"  , "EVEX.NDS.512.66.0F.WIG F1 /r"     , "ANY AVX512BW"],
    ["vpsllw"           , "W:zmm {kz}, zmm/m512, ib"                 , "VMI"  , "EVEX.NDD.512.66.0F.WIG 71 /6 ib"  , "ANY AVX512BW"],
    ["vpsrad"           , "W:xmm {kz}, xmm, xmm/m128"                , "RVM"  , "EVEX.NDS.128.66.0F.W0 E2 /r"      , "ANY AVX512F-VL"],
    ["vpsrad"           , "W:xmm {kz}, xmm/m128/b32, ib"             , "VMI"  , "EVEX.NDD.128.66.0F.W0 72 /4 ib"   , "ANY AVX512F-VL"],
    ["vpsrad"           , "W:ymm {kz}, ymm, xmm/m128"                , "RVM"  , "EVEX.NDS.256.66.0F.W0 E2 /r"      , "ANY AVX512F-VL"],
    ["vpsrad"           , "W:ymm {kz}, ymm/m256/b32, ib"             , "VMI"  , "EVEX.NDD.256.66.0F.W0 72 /4 ib"   , "ANY AVX512F-VL"],
    ["vpsrad"           , "W:zmm {kz}, zmm, xmm/m128"                , "RVM"  , "EVEX.NDS.512.66.0F.W0 E2 /r"      , "ANY AVX512F"],
    ["vpsrad"           , "W:zmm {kz}, zmm/m512/b32, ib"             , "VMI"  , "EVEX.NDD.512.66.0F.W0 72 /4 ib"   , "ANY AVX512F"],
    ["vpsraq"           , "W:xmm {kz}, xmm, xmm/m128"                , "RVM"  , "EVEX.NDS.128.66.0F.W1 E2 /r"      , "ANY AVX512F-VL"],
    ["vpsraq"           , "W:xmm {kz}, xmm/m128/b64, ib"             , "VMI"  , "EVEX.NDD.128.66.0F.W1 72 /4 ib"   , "ANY AVX512F-VL"],
    ["vpsraq"           , "W:ymm {kz}, ymm, xmm/m128"                , "RVM"  , "EVEX.NDS.256.66.0F.W1 E2 /r"      , "ANY AVX512F-VL"],
    ["vpsraq"           , "W:ymm {kz}, ymm/m256/b64, ib"             , "VMI"  , "EVEX.NDD.256.66.0F.W1 72 /4 ib"   , "ANY AVX512F-VL"],
    ["vpsraq"           , "W:zmm {kz}, zmm, xmm/m128"                , "RVM"  , "EVEX.NDS.512.66.0F.W1 E2 /r"      , "ANY AVX512F"],
    ["vpsraq"           , "W:zmm {kz}, zmm/m512/b64, ib"             , "VMI"  , "EVEX.NDD.512.66.0F.W1 72 /4 ib"   , "ANY AVX512F"],
    ["vpsravd"          , "W:xmm {kz}, xmm, xmm/m128/b32"            , "RVM"  , "EVEX.NDS.128.66.0F38.W0 46 /r"    , "ANY AVX512F-VL"],
    ["vpsravd"          , "W:ymm {kz}, ymm, ymm/m256/b32"            , "RVM"  , "EVEX.NDS.256.66.0F38.W0 46 /r"    , "ANY AVX512F-VL"],
    ["vpsravd"          , "W:zmm {kz}, zmm, zmm/m512/b32"            , "RVM"  , "EVEX.NDS.512.66.0F38.W0 46 /r"    , "ANY AVX512F"],
    ["vpsravq"          , "W:xmm {kz}, xmm, xmm/m128/b64"            , "RVM"  , "EVEX.NDS.128.66.0F38.W1 46 /r"    , "ANY AVX512F-VL"],
    ["vpsravq"          , "W:ymm {kz}, ymm, ymm/m256/b64"            , "RVM"  , "EVEX.NDS.256.66.0F38.W1 46 /r"    , "ANY AVX512F-VL"],
    ["vpsravq"          , "W:zmm {kz}, zmm, zmm/m512/b64"            , "RVM"  , "EVEX.NDS.512.66.0F38.W1 46 /r"    , "ANY AVX512F"],
    ["vpsravw"          , "W:xmm {kz}, xmm, xmm/m128"                , "RVM"  , "EVEX.NDS.128.66.0F38.W1 11 /r"    , "ANY AVX512BW-VL"],
    ["vpsravw"          , "W:ymm {kz}, ymm, ymm/m256"                , "RVM"  , "EVEX.NDS.256.66.0F38.W1 11 /r"    , "ANY AVX512BW-VL"],
    ["vpsravw"          , "W:zmm {kz}, zmm, zmm/m512"                , "RVM"  , "EVEX.NDS.512.66.0F38.W1 11 /r"    , "ANY AVX512BW"],
    ["vpsraw"           , "W:xmm {kz}, xmm, xmm/m128"                , "RVM"  , "EVEX.NDS.128.66.0F.WIG E1 /r"     , "ANY AVX512BW-VL"],
    ["vpsraw"           , "W:xmm {kz}, xmm/m128, ib"                 , "VMI"  , "EVEX.NDD.128.66.0F.WIG 71 /4 ib"  , "ANY AVX512BW-VL"],
    ["vpsraw"           , "W:ymm {kz}, ymm, xmm/m128"                , "RVM"  , "EVEX.NDS.256.66.0F.WIG E1 /r"     , "ANY AVX512BW-VL"],
    ["vpsraw"           , "W:ymm {kz}, ymm/m256, ib"                 , "VMI"  , "EVEX.NDD.256.66.0F.WIG 71 /4 ib"  , "ANY AVX512BW-VL"],
    ["vpsraw"           , "W:zmm {kz}, zmm, xmm/m128"                , "RVM"  , "EVEX.NDS.512.66.0F.WIG E1 /r"     , "ANY AVX512BW"],
    ["vpsraw"           , "W:zmm {kz}, zmm/m512, ib"                 , "VMI"  , "EVEX.NDD.512.66.0F.WIG 71 /4 ib"  , "ANY AVX512BW"],
    ["vpsrld"           , "W:xmm {kz}, xmm, xmm/m128"                , "RVM"  , "EVEX.NDS.128.66.0F.W0 D2 /r"      , "ANY AVX512F-VL"],
    ["vpsrld"           , "W:xmm {kz}, xmm/m128/b32, ib"             , "VMI"  , "EVEX.NDD.128.66.0F.W0 72 /2 ib"   , "ANY AVX512F-VL"],
    ["vpsrld"           , "W:ymm {kz}, ymm, xmm/m128"                , "RVM"  , "EVEX.NDS.256.66.0F.W0 D2 /r"      , "ANY AVX512F-VL"],
    ["vpsrld"           , "W:ymm {kz}, ymm/m256/b32, ib"             , "VMI"  , "EVEX.NDD.256.66.0F.W0 72 /2 ib"   , "ANY AVX512F-VL"],
    ["vpsrld"           , "W:zmm {kz}, zmm, xmm/m128"                , "RVM"  , "EVEX.NDS.512.66.0F.W0 D2 /r"      , "ANY AVX512F"],
    ["vpsrld"           , "W:zmm {kz}, zmm/m512/b32, ib"             , "VMI"  , "EVEX.NDD.512.66.0F.W0 72 /2 ib"   , "ANY AVX512F"],
    ["vpsrldq"          , "W:xmm, xmm, xmm/m128, ib"                 , "VMI"  , "EVEX.NDD.128.66.0F.WIG 73 /3 ib"  , "ANY AVX512BW-VL"],
    ["vpsrldq"          , "W:ymm, ymm, ymm/m256, ib"                 , "VMI"  , "EVEX.NDD.256.66.0F.WIG 73 /3 ib"  , "ANY AVX512BW-VL"],
    ["vpsrldq"          , "W:zmm, zmm, zmm/m512, ib"                 , "VMI"  , "EVEX.NDD.512.66.0F.WIG 73 /3 ib"  , "ANY AVX512BW"],
    ["vpsrlq"           , "W:xmm {kz}, xmm, xmm/m128"                , "RVM"  , "EVEX.NDS.128.66.0F.W1 D3 /r"      , "ANY AVX512F-VL"],
    ["vpsrlq"           , "W:xmm {kz}, xmm/m128/b64, ib"             , "VMI"  , "EVEX.NDD.128.66.0F.W1 73 /2 ib"   , "ANY AVX512F-VL"],
    ["vpsrlq"           , "W:ymm {kz}, ymm, xmm/m128"                , "RVM"  , "EVEX.NDS.256.66.0F.W1 D3 /r"      , "ANY AVX512F-VL"],
    ["vpsrlq"           , "W:ymm {kz}, ymm/m256/b64, ib"             , "VMI"  , "EVEX.NDD.256.66.0F.W1 73 /2 ib"   , "ANY AVX512F-VL"],
    ["vpsrlq"           , "W:zmm {kz}, zmm, xmm/m128"                , "RVM"  , "EVEX.NDS.512.66.0F.W1 D3 /r"      , "ANY AVX512F"],
    ["vpsrlq"           , "W:zmm {kz}, zmm/m512/b64, ib"             , "VMI"  , "EVEX.NDD.512.66.0F.W1 73 /2 ib"   , "ANY AVX512F"],
    ["vpsrlvd"          , "W:xmm {kz}, xmm, xmm/m128/b32"            , "RVM"  , "EVEX.NDS.128.66.0F38.W0 45 /r"    , "ANY AVX512F-VL"],
    ["vpsrlvd"          , "W:ymm {kz}, ymm, ymm/m256/b32"            , "RVM"  , "EVEX.NDS.256.66.0F38.W0 45 /r"    , "ANY AVX512F-VL"],
    ["vpsrlvd"          , "W:zmm {kz}, zmm, zmm/m512/b32"            , "RVM"  , "EVEX.NDS.512.66.0F38.W0 45 /r"    , "ANY AVX512F"],
    ["vpsrlvq"          , "W:xmm {kz}, xmm, xmm/m128/b64"            , "RVM"  , "EVEX.NDS.128.66.0F38.W1 45 /r"    , "ANY AVX512F-VL"],
    ["vpsrlvq"          , "W:ymm {kz}, ymm, ymm/m256/b64"            , "RVM"  , "EVEX.NDS.256.66.0F38.W1 45 /r"    , "ANY AVX512F-VL"],
    ["vpsrlvq"          , "W:zmm {kz}, zmm, zmm/m512/b64"            , "RVM"  , "EVEX.NDS.512.66.0F38.W1 45 /r"    , "ANY AVX512F"],
    ["vpsrlvw"          , "W:xmm {kz}, xmm, xmm/m128"                , "RVM"  , "EVEX.NDS.128.66.0F38.W1 10 /r"    , "ANY AVX512BW-VL"],
    ["vpsrlvw"          , "W:ymm {kz}, ymm, ymm/m256"                , "RVM"  , "EVEX.NDS.256.66.0F38.W1 10 /r"    , "ANY AVX512BW-VL"],
    ["vpsrlvw"          , "W:zmm {kz}, zmm, zmm/m512"                , "RVM"  , "EVEX.NDS.512.66.0F38.W1 10 /r"    , "ANY AVX512BW"],
    ["vpsrlw"           , "W:xmm {kz}, xmm, xmm/m128"                , "RVM"  , "EVEX.NDS.128.66.0F.WIG D1 /r"     , "ANY AVX512BW-VL"],
    ["vpsrlw"           , "W:xmm {kz}, xmm/m128, ib"                 , "VMI"  , "EVEX.NDD.128.66.0F.WIG 71 /2 ib"  , "ANY AVX512BW-VL"],
    ["vpsrlw"           , "W:ymm {kz}, ymm, xmm/m128"                , "RVM"  , "EVEX.NDS.256.66.0F.WIG D1 /r"     , "ANY AVX512BW-VL"],
    ["vpsrlw"           , "W:ymm {kz}, ymm/m256, ib"                 , "VMI"  , "EVEX.NDD.256.66.0F.WIG 71 /2 ib"  , "ANY AVX512BW-VL"],
    ["vpsrlw"           , "W:zmm {kz}, zmm, xmm/m128"                , "RVM"  , "EVEX.NDS.512.66.0F.WIG D1 /r"     , "ANY AVX512BW"],
    ["vpsrlw"           , "W:zmm {kz}, zmm/m512, ib"                 , "VMI"  , "EVEX.NDD.512.66.0F.WIG 71 /2 ib"  , "ANY AVX512BW"],
    ["vpsubb"           , "W:xmm {kz}, xmm, xmm/m128"                , "RVM"  , "EVEX.NDS.128.66.0F.WIG F8 /r"     , "ANY AVX512BW-VL"],
    ["vpsubb"           , "W:ymm {kz}, ymm, ymm/m256"                , "RVM"  , "EVEX.NDS.256.66.0F.WIG F8 /r"     , "ANY AVX512BW-VL"],
    ["vpsubb"           , "W:zmm {kz}, zmm, zmm/m512"                , "RVM"  , "EVEX.NDS.512.66.0F.WIG F8 /r"     , "ANY AVX512BW"],
    ["vpsubd"           , "W:xmm {kz}, xmm, xmm/m128/b32"            , "RVM"  , "EVEX.NDS.128.66.0F.W0 FA /r"      , "ANY AVX512F-VL"],
    ["vpsubd"           , "W:ymm {kz}, ymm, ymm/m256/b32"            , "RVM"  , "EVEX.NDS.256.66.0F.W0 FA /r"      , "ANY AVX512F-VL"],
    ["vpsubd"           , "W:zmm {kz}, zmm, zmm/m512/b32"            , "RVM"  , "EVEX.NDS.512.66.0F.W0 FA /r"      , "ANY AVX512F"],
    ["vpsubq"           , "W:xmm {kz}, xmm, xmm/m128/b64"            , "RVM"  , "EVEX.NDS.128.66.0F.W1 FB /r"      , "ANY AVX512F-VL"],
    ["vpsubq"           , "W:ymm {kz}, ymm, ymm/m256/b64"            , "RVM"  , "EVEX.NDS.256.66.0F.W1 FB /r"      , "ANY AVX512F-VL"],
    ["vpsubq"           , "W:zmm {kz}, zmm, zmm/m512/b64"            , "RVM"  , "EVEX.NDS.512.66.0F.W1 FB /r"      , "ANY AVX512F"],
    ["vpsubsb"          , "W:xmm {kz}, xmm, xmm/m128"                , "RVM"  , "EVEX.NDS.128.66.0F.WIG E8 /r"     , "ANY AVX512BW-VL"],
    ["vpsubsb"          , "W:ymm {kz}, ymm, ymm/m256"                , "RVM"  , "EVEX.NDS.256.66.0F.WIG E8 /r"     , "ANY AVX512BW-VL"],
    ["vpsubsb"          , "W:zmm {kz}, zmm, zmm/m512"                , "RVM"  , "EVEX.NDS.512.66.0F.WIG E8 /r"     , "ANY AVX512BW"],
    ["vpsubsw"          , "W:xmm {kz}, xmm, xmm/m128"                , "RVM"  , "EVEX.NDS.128.66.0F.WIG E9 /r"     , "ANY AVX512BW-VL"],
    ["vpsubsw"          , "W:ymm {kz}, ymm, ymm/m256"                , "RVM"  , "EVEX.NDS.256.66.0F.WIG E9 /r"     , "ANY AVX512BW-VL"],
    ["vpsubsw"          , "W:zmm {kz}, zmm, zmm/m512"                , "RVM"  , "EVEX.NDS.512.66.0F.WIG E9 /r"     , "ANY AVX512BW"],
    ["vpsubusb"         , "W:xmm {kz}, xmm, xmm/m128"                , "RVM"  , "EVEX.NDS.128.66.0F.WIG D8 /r"     , "ANY AVX512BW-VL"],
    ["vpsubusb"         , "W:ymm {kz}, ymm, ymm/m256"                , "RVM"  , "EVEX.NDS.256.66.0F.WIG D8 /r"     , "ANY AVX512BW-VL"],
    ["vpsubusb"         , "W:zmm {kz}, zmm, zmm/m512"                , "RVM"  , "EVEX.NDS.512.66.0F.WIG D8 /r"     , "ANY AVX512BW"],
    ["vpsubusw"         , "W:xmm {kz}, xmm, xmm/m128"                , "RVM"  , "EVEX.NDS.128.66.0F.WIG D9 /r"     , "ANY AVX512BW-VL"],
    ["vpsubusw"         , "W:ymm {kz}, ymm, ymm/m256"                , "RVM"  , "EVEX.NDS.256.66.0F.WIG D9 /r"     , "ANY AVX512BW-VL"],
    ["vpsubusw"         , "W:zmm {kz}, zmm, zmm/m512"                , "RVM"  , "EVEX.NDS.512.66.0F.WIG D9 /r"     , "ANY AVX512BW"],
    ["vpsubw"           , "W:xmm {kz}, xmm, xmm/m128"                , "RVM"  , "EVEX.NDS.128.66.0F.WIG F9 /r"     , "ANY AVX512BW-VL"],
    ["vpsubw"           , "W:ymm {kz}, ymm, ymm/m256"                , "RVM"  , "EVEX.NDS.256.66.0F.WIG F9 /r"     , "ANY AVX512BW-VL"],
    ["vpsubw"           , "W:zmm {kz}, zmm, zmm/m512"                , "RVM"  , "EVEX.NDS.512.66.0F.WIG F9 /r"     , "ANY AVX512BW"],
    ["vpternlogd"       , "X:xmm {kz}, xmm/m128/b32, ib"             , "RVMI" , "EVEX.DDS.128.66.0F3A.W0 25 /r ib" , "ANY AVX512F-VL"],
    ["vpternlogd"       , "X:ymm {kz}, ymm/m256/b32, ib"             , "RVMI" , "EVEX.DDS.256.66.0F3A.W0 25 /r ib" , "ANY AVX512F-VL"],
    ["vpternlogd"       , "X:zmm {kz}, zmm/m512/b32, ib"             , "RVMI" , "EVEX.DDS.512.66.0F3A.W0 25 /r ib" , "ANY AVX512F"],
    ["vpternlogq"       , "X:xmm {kz}, xmm/m128/b64, ib"             , "RVMI" , "EVEX.DDS.128.66.0F3A.W1 25 /r ib" , "ANY AVX512F-VL"],
    ["vpternlogq"       , "X:ymm {kz}, ymm/m256/b64, ib"             , "RVMI" , "EVEX.DDS.256.66.0F3A.W1 25 /r ib" , "ANY AVX512F-VL"],
    ["vpternlogq"       , "X:zmm {kz}, zmm/m512/b64, ib"             , "RVMI" , "EVEX.DDS.512.66.0F3A.W1 25 /r ib" , "ANY AVX512F"],
    ["vptestmb"         , "W:k {k}, xmm, xmm/m128"                   , "RVM"  , "EVEX.NDS.128.66.0F38.W0 26 /r"    , "ANY AVX512BW-VL"],
    ["vptestmb"         , "W:k {k}, ymm, ymm/m256"                   , "RVM"  , "EVEX.NDS.256.66.0F38.W0 26 /r"    , "ANY AVX512BW-VL"],
    ["vptestmb"         , "W:k {k}, zmm, zmm/m512"                   , "RVM"  , "EVEX.NDS.512.66.0F38.W0 26 /r"    , "ANY AVX512BW"],
    ["vptestmd"         , "W:k {k}, xmm, xmm/m128/b32"               , "RVM"  , "EVEX.NDS.128.66.0F38.W0 27 /r"    , "ANY AVX512F-VL"],
    ["vptestmd"         , "W:k {k}, ymm, ymm/m256/b32"               , "RVM"  , "EVEX.NDS.256.66.0F38.W0 27 /r"    , "ANY AVX512F-VL"],
    ["vptestmd"         , "W:k {k}, zmm, zmm/m512/b32"               , "RVM"  , "EVEX.NDS.512.66.0F38.W0 27 /r"    , "ANY AVX512F"],
    ["vptestmq"         , "W:k {k}, xmm, xmm/m128/b64"               , "RVM"  , "EVEX.NDS.128.66.0F38.W1 27 /r"    , "ANY AVX512F-VL"],
    ["vptestmq"         , "W:k {k}, ymm, ymm/m256/b64"               , "RVM"  , "EVEX.NDS.256.66.0F38.W1 27 /r"    , "ANY AVX512F-VL"],
    ["vptestmq"         , "W:k {k}, zmm, zmm/m512/b64"               , "RVM"  , "EVEX.NDS.512.66.0F38.W1 27 /r"    , "ANY AVX512F"],
    ["vptestmw"         , "W:k {k}, xmm, xmm/m128"                   , "RVM"  , "EVEX.NDS.128.66.0F38.W1 26 /r"    , "ANY AVX512BW-VL"],
    ["vptestmw"         , "W:k {k}, ymm, ymm/m256"                   , "RVM"  , "EVEX.NDS.256.66.0F38.W1 26 /r"    , "ANY AVX512BW-VL"],
    ["vptestmw"         , "W:k {k}, zmm, zmm/m512"                   , "RVM"  , "EVEX.NDS.512.66.0F38.W1 26 /r"    , "ANY AVX512BW"],
    ["vptestnmb"        , "W:k {k}, xmm, xmm/m128"                   , "RVM"  , "EVEX.NDS.128.F3.0F38.W0 26 /r"    , "ANY AVX512BW-VL"],
    ["vptestnmb"        , "W:k {k}, ymm, ymm/m256"                   , "RVM"  , "EVEX.NDS.256.F3.0F38.W0 26 /r"    , "ANY AVX512BW-VL"],
    ["vptestnmb"        , "W:k {k}, zmm, zmm/m512"                   , "RVM"  , "EVEX.NDS.512.F3.0F38.W0 26 /r"    , "ANY AVX512BW"],
    ["vptestnmd"        , "W:k {k}, xmm, xmm/m128/b32"               , "RVM"  , "EVEX.NDS.128.F3.0F38.W0 27 /r"    , "ANY AVX512F-VL"],
    ["vptestnmd"        , "W:k {k}, ymm, ymm/m256/b32"               , "RVM"  , "EVEX.NDS.256.F3.0F38.W0 27 /r"    , "ANY AVX512F-VL"],
    ["vptestnmd"        , "W:k {k}, zmm, zmm/m512/b32"               , "RVM"  , "EVEX.NDS.512.F3.0F38.W0 27 /r"    , "ANY AVX512F"],
    ["vptestnmq"        , "W:k {k}, xmm, xmm/m128/b64"               , "RVM"  , "EVEX.NDS.128.F3.0F38.W1 27 /r"    , "ANY AVX512F-VL"],
    ["vptestnmq"        , "W:k {k}, ymm, ymm/m256/b64"               , "RVM"  , "EVEX.NDS.256.F3.0F38.W1 27 /r"    , "ANY AVX512F-VL"],
    ["vptestnmq"        , "W:k {k}, zmm, zmm/m512/b64"               , "RVM"  , "EVEX.NDS.512.F3.0F38.W1 27 /r"    , "ANY AVX512F"],
    ["vptestnmw"        , "W:k {k}, xmm, xmm/m128"                   , "RVM"  , "EVEX.NDS.128.F3.0F38.W1 26 /r"    , "ANY AVX512BW-VL"],
    ["vptestnmw"        , "W:k {k}, ymm, ymm/m256"                   , "RVM"  , "EVEX.NDS.256.F3.0F38.W1 26 /r"    , "ANY AVX512BW-VL"],
    ["vptestnmw"        , "W:k {k}, zmm, zmm/m512"                   , "RVM"  , "EVEX.NDS.512.F3.0F38.W1 26 /r"    , "ANY AVX512BW"],
    ["vpunpckhbw"       , "W:xmm {kz}, xmm, xmm/m128"                , "RVM"  , "EVEX.NDS.128.66.0F.WIG 68 /r"     , "ANY AVX512BW-VL"],
    ["vpunpckhbw"       , "W:ymm {kz}, ymm, ymm/m256"                , "RVM"  , "EVEX.NDS.256.66.0F.WIG 68 /r"     , "ANY AVX512BW-VL"],
    ["vpunpckhbw"       , "W:zmm {kz}, zmm, zmm/m512"                , "RVM"  , "EVEX.NDS.512.66.0F 68 /r"         , "ANY AVX512BW"],
    ["vpunpckhdq"       , "W:xmm {kz}, xmm, xmm/m128/b32"            , "RVM"  , "EVEX.NDS.128.66.0F.W0 6A /r"      , "ANY AVX512F-VL"],
    ["vpunpckhdq"       , "W:ymm {kz}, ymm, ymm/m256/b32"            , "RVM"  , "EVEX.NDS.256.66.0F.W0 6A /r"      , "ANY AVX512F-VL"],
    ["vpunpckhdq"       , "W:zmm {kz}, zmm, zmm/m512/b32"            , "RVM"  , "EVEX.NDS.512.66.0F.W0 6A /r"      , "ANY AVX512F"],
    ["vpunpckhqdq"      , "W:xmm {kz}, xmm, xmm/m128/b64"            , "RVM"  , "EVEX.NDS.128.66.0F.W1 6D /r"      , "ANY AVX512F-VL"],
    ["vpunpckhqdq"      , "W:ymm {kz}, ymm, ymm/m256/b64"            , "RVM"  , "EVEX.NDS.256.66.0F.W1 6D /r"      , "ANY AVX512F-VL"],
    ["vpunpckhqdq"      , "W:zmm {kz}, zmm, zmm/m512/b64"            , "RVM"  , "EVEX.NDS.512.66.0F.W1 6D /r"      , "ANY AVX512F"],
    ["vpunpckhwd"       , "W:xmm {kz}, xmm, xmm/m128"                , "RVM"  , "EVEX.NDS.128.66.0F.WIG 69 /r"     , "ANY AVX512BW-VL"],
    ["vpunpckhwd"       , "W:ymm {kz}, ymm, ymm/m256"                , "RVM"  , "EVEX.NDS.256.66.0F.WIG 69 /r"     , "ANY AVX512BW-VL"],
    ["vpunpckhwd"       , "W:zmm {kz}, zmm, zmm/m512"                , "RVM"  , "EVEX.NDS.512.66.0F 69 /r"         , "ANY AVX512BW"],
    ["vpunpcklbw"       , "W:xmm {kz}, xmm, xmm/m128"                , "RVM"  , "EVEX.NDS.128.66.0F.WIG 60 /r"     , "ANY AVX512BW-VL"],
    ["vpunpcklbw"       , "W:ymm {kz}, ymm, ymm/m256"                , "RVM"  , "EVEX.NDS.256.66.0F.WIG 60 /r"     , "ANY AVX512BW-VL"],
    ["vpunpcklbw"       , "W:zmm {kz}, zmm, zmm/m512"                , "RVM"  , "EVEX.NDS.512.66.0F 60 /r"         , "ANY AVX512BW"],
    ["vpunpckldq"       , "W:xmm {kz}, xmm, xmm/m128/b32"            , "RVM"  , "EVEX.NDS.128.66.0F.W0 62 /r"      , "ANY AVX512F-VL"],
    ["vpunpckldq"       , "W:ymm {kz}, ymm, ymm/m256/b32"            , "RVM"  , "EVEX.NDS.256.66.0F.W0 62 /r"      , "ANY AVX512F-VL"],
    ["vpunpckldq"       , "W:zmm {kz}, zmm, zmm/m512/b32"            , "RVM"  , "EVEX.NDS.512.66.0F.W0 62 /r"      , "ANY AVX512F"],
    ["vpunpcklqdq"      , "W:xmm {kz}, xmm, xmm/m128/b64"            , "RVM"  , "EVEX.NDS.128.66.0F.W1 6C /r"      , "ANY AVX512F-VL"],
    ["vpunpcklqdq"      , "W:ymm {kz}, ymm, ymm/m256/b64"            , "RVM"  , "EVEX.NDS.256.66.0F.W1 6C /r"      , "ANY AVX512F-VL"],
    ["vpunpcklqdq"      , "W:zmm {kz}, zmm, zmm/m512/b64"            , "RVM"  , "EVEX.NDS.512.66.0F.W1 6C /r"      , "ANY AVX512F"],
    ["vpunpcklwd"       , "W:xmm {kz}, xmm, xmm/m128"                , "RVM"  , "EVEX.NDS.128.66.0F.WIG 61 /r"     , "ANY AVX512BW-VL"],
    ["vpunpcklwd"       , "W:ymm {kz}, ymm, ymm/m256"                , "RVM"  , "EVEX.NDS.256.66.0F.WIG 61 /r"     , "ANY AVX512BW-VL"],
    ["vpunpcklwd"       , "W:zmm {kz}, zmm, zmm/m512"                , "RVM"  , "EVEX.NDS.512.66.0F 61 /r"         , "ANY AVX512BW"],
    ["vpxord"           , "W:xmm {kz}, xmm, xmm/m128/b32"            , "RVM"  , "EVEX.NDS.128.66.0F.W0 EF /r"      , "ANY AVX512F-VL"],
    ["vpxord"           , "W:ymm {kz}, ymm, ymm/m256/b32"            , "RVM"  , "EVEX.NDS.256.66.0F.W0 EF /r"      , "ANY AVX512F-VL"],
    ["vpxord"           , "W:zmm {kz}, zmm, zmm/m512/b32"            , "RVM"  , "EVEX.NDS.512.66.0F.W0 EF /r"      , "ANY AVX512F"],
    ["vpxorq"           , "W:xmm {kz}, xmm, xmm/m128/b64"            , "RVM"  , "EVEX.NDS.128.66.0F.W1 EF /r"      , "ANY AVX512F-VL"],
    ["vpxorq"           , "W:ymm {kz}, ymm, ymm/m256/b64"            , "RVM"  , "EVEX.NDS.256.66.0F.W1 EF /r"      , "ANY AVX512F-VL"],
    ["vpxorq"           , "W:zmm {kz}, zmm, zmm/m512/b64"            , "RVM"  , "EVEX.NDS.512.66.0F.W1 EF /r"      , "ANY AVX512F"],
    ["vrangepd"         , "W:xmm {kz}, xmm, xmm/m128/b64, ib"        , "RVMI" , "EVEX.NDS.128.66.0F3A.W1 50 /r ib" , "ANY AVX512DQ-VL"],
    ["vrangepd"         , "W:ymm {kz}, ymm, ymm/m256/b64, ib"        , "RVMI" , "EVEX.NDS.256.66.0F3A.W1 50 /r ib" , "ANY AVX512DQ-VL"],
    ["vrangepd"         , "W:zmm {kz}, zmm, zmm/m512/b64, ib {sae}"  , "RVMI" , "EVEX.NDS.512.66.0F3A.W1 50 /r ib" , "ANY AVX512DQ"],
    ["vrangeps"         , "W:xmm {kz}, xmm, xmm/m128/b32, ib"        , "RVMI" , "EVEX.NDS.128.66.0F3A.W0 50 /r ib" , "ANY AVX512DQ-VL"],
    ["vrangeps"         , "W:ymm {kz}, ymm, ymm/m256/b32, ib"        , "RVMI" , "EVEX.NDS.256.66.0F3A.W0 50 /r ib" , "ANY AVX512DQ-VL"],
    ["vrangeps"         , "W:zmm {kz}, zmm, zmm/m512/b32, ib {sae}"  , "RVMI" , "EVEX.NDS.512.66.0F3A.W0 50 /r ib" , "ANY AVX512DQ"],
    ["vrangesd"         , "W:xmm {kz}, xmm, xmm/m64, ib {sae}"       , "RVMI" , "EVEX.NDS.LIG.66.0F3A.W1 51 /r ib" , "ANY AVX512DQ"],
    ["vrangess"         , "W:xmm {kz}, xmm, xmm/m32, ib {sae}"       , "RVMI" , "EVEX.NDS.LIG.66.0F3A.W0 51 /r ib" , "ANY AVX512DQ"],
    ["vrcp14pd"         , "W:xmm {kz}, xmm/m128/b64"                 , "RM"   , "EVEX.128.66.0F38.W1 4C /r"        , "ANY AVX512F-VL"],
    ["vrcp14pd"         , "W:ymm {kz}, ymm/m256/b64"                 , "RM"   , "EVEX.256.66.0F38.W1 4C /r"        , "ANY AVX512F-VL"],
    ["vrcp14pd"         , "W:zmm {kz}, zmm/m512/b64"                 , "RM"   , "EVEX.512.66.0F38.W1 4C /r"        , "ANY AVX512F"],
    ["vrcp14ps"         , "W:xmm {kz}, xmm/m128/b32"                 , "RM"   , "EVEX.128.66.0F38.W0 4C /r"        , "ANY AVX512F-VL"],
    ["vrcp14ps"         , "W:ymm {kz}, ymm/m256/b32"                 , "RM"   , "EVEX.256.66.0F38.W0 4C /r"        , "ANY AVX512F-VL"],
    ["vrcp14ps"         , "W:zmm {kz}, zmm/m512/b32"                 , "RM"   , "EVEX.512.66.0F38.W0 4C /r"        , "ANY AVX512F"],
    ["vrcp14sd"         , "W:xmm {kz}, xmm, xmm/m64"                 , "RVM"  , "EVEX.NDS.LIG.66.0F38.W1 4D /r"    , "ANY AVX512F"],
    ["vrcp14ss"         , "W:xmm {kz}, xmm, xmm/m32"                 , "RVM"  , "EVEX.NDS.LIG.66.0F38.W0 4D /r"    , "ANY AVX512F"],
    ["vrcp28pd"         , "W:zmm {kz}, zmm/m512/b64 {sae}"           , "RM"   , "EVEX.512.66.0F38.W1 CA /r"        , "ANY AVX512ER"],
    ["vrcp28ps"         , "W:zmm {kz}, zmm/m512/b32 {sae}"           , "RM"   , "EVEX.512.66.0F38.W0 CA /r"        , "ANY AVX512ER"],
    ["vrcp28sd"         , "W:xmm {kz}, xmm, xmm/m64 {sae}"           , "RVM"  , "EVEX.NDS.LIG.66.0F38.W1 CB /r"    , "ANY AVX512ER"],
    ["vrcp28ss"         , "W:xmm {kz}, xmm, xmm/m32 {sae}"           , "RVM"  , "EVEX.NDS.LIG.66.0F38.W0 CB /r"    , "ANY AVX512ER"],
    ["vreducepd"        , "W:xmm {kz}, xmm/m128/b64, ib"             , "RMI"  , "EVEX.128.66.0F3A.W1 56 /r ib"     , "ANY AVX512DQ-VL"],
    ["vreducepd"        , "W:ymm {kz}, ymm/m256/b64, ib"             , "RMI"  , "EVEX.256.66.0F3A.W1 56 /r ib"     , "ANY AVX512DQ-VL"],
    ["vreducepd"        , "W:zmm {kz}, zmm/m512/b64, ib"             , "RMI"  , "EVEX.512.66.0F3A.W1 56 /r ib"     , "ANY AVX512DQ"],
    ["vreduceps"        , "W:xmm {kz}, xmm/m128/b32, ib"             , "RMI"  , "EVEX.128.66.0F3A.W0 56 /r ib"     , "ANY AVX512DQ-VL"],
    ["vreduceps"        , "W:ymm {kz}, ymm/m256/b32, ib"             , "RMI"  , "EVEX.256.66.0F3A.W0 56 /r ib"     , "ANY AVX512DQ-VL"],
    ["vreduceps"        , "W:zmm {kz}, zmm/m512/b32, ib"             , "RMI"  , "EVEX.512.66.0F3A.W0 56 /r ib"     , "ANY AVX512DQ"],
    ["vreducesd"        , "W:xmm {kz}, xmm, xmm/m64, ib"             , "RVMI" , "EVEX.NDS.LIG.66.0F3A.W1 57 /r ib" , "ANY AVX512DQ"],
    ["vreducess"        , "W:xmm {kz}, xmm, xmm/m32, ib"             , "RVMI" , "EVEX.NDS.LIG.66.0F3A.W0 57 /r ib" , "ANY AVX512DQ"],
    ["vrndscalepd"      , "W:xmm {kz}, xmm/m128/b64, ib"             , "RMI"  , "EVEX.128.66.0F3A.W1 09 /r ib"     , "ANY AVX512F-VL"],
    ["vrndscalepd"      , "W:ymm {kz}, ymm/m256/b64, ib"             , "RMI"  , "EVEX.256.66.0F3A.W1 09 /r ib"     , "ANY AVX512F-VL"],
    ["vrndscalepd"      , "W:zmm {kz}, zmm/m512/b64, ib {sae}"       , "RMI"  , "EVEX.512.66.0F3A.W1 09 /r ib"     , "ANY AVX512F"],
    ["vrndscaleps"      , "W:xmm {kz}, xmm/m128/b32, ib"             , "RMI"  , "EVEX.128.66.0F3A.W0 08 /r ib"     , "ANY AVX512F-VL"],
    ["vrndscaleps"      , "W:ymm {kz}, ymm/m256/b32, ib"             , "RMI"  , "EVEX.256.66.0F3A.W0 08 /r ib"     , "ANY AVX512F-VL"],
    ["vrndscaleps"      , "W:zmm {kz}, zmm/m512/b32, ib {sae}"       , "RMI"  , "EVEX.512.66.0F3A.W0 08 /r ib"     , "ANY AVX512F"],
    ["vrndscalesd"      , "W:xmm {kz}, xmm, xmm/m64, ib {sae}"       , "RVMI" , "EVEX.NDS.LIG.66.0F3A.W1 0B /r ib" , "ANY AVX512F"],
    ["vrndscaless"      , "W:xmm {kz}, xmm, xmm/m32, ib {sae}"       , "RVMI" , "EVEX.NDS.LIG.66.0F3A.W0 0A /r ib" , "ANY AVX512F"],
    ["vrsqrt14pd"       , "W:xmm {kz}, xmm/m128/b64"                 , "RM"   , "EVEX.128.66.0F38.W1 4E /r"        , "ANY AVX512F-VL"],
    ["vrsqrt14pd"       , "W:ymm {kz}, ymm/m256/b64"                 , "RM"   , "EVEX.256.66.0F38.W1 4E /r"        , "ANY AVX512F-VL"],
    ["vrsqrt14pd"       , "W:zmm {kz}, zmm/m512/b64"                 , "RM"   , "EVEX.512.66.0F38.W1 4E /r"        , "ANY AVX512F"],
    ["vrsqrt14ps"       , "W:xmm {kz}, xmm/m128/b32"                 , "RM"   , "EVEX.128.66.0F38.W0 4E /r"        , "ANY AVX512F-VL"],
    ["vrsqrt14ps"       , "W:ymm {kz}, ymm/m256/b32"                 , "RM"   , "EVEX.256.66.0F38.W0 4E /r"        , "ANY AVX512F-VL"],
    ["vrsqrt14ps"       , "W:zmm {kz}, zmm/m512/b32"                 , "RM"   , "EVEX.512.66.0F38.W0 4E /r"        , "ANY AVX512F"],
    ["vrsqrt14sd"       , "W:xmm {kz}, xmm, xmm/m64"                 , "RVM"  , "EVEX.NDS.LIG.66.0F38.W1 4F /r"    , "ANY AVX512F"],
    ["vrsqrt14ss"       , "W:xmm {kz}, xmm, xmm/m32"                 , "RVM"  , "EVEX.NDS.LIG.66.0F38.W0 4F /r"    , "ANY AVX512F"],
    ["vrsqrt28pd"       , "W:zmm {kz}, zmm/m512/b64 {sae}"           , "RM"   , "EVEX.512.66.0F38.W1 CC /r"        , "ANY AVX512ER"],
    ["vrsqrt28ps"       , "W:zmm {kz}, zmm/m512/b32 {sae}"           , "RM"   , "EVEX.512.66.0F38.W0 CC /r"        , "ANY AVX512ER"],
    ["vrsqrt28sd"       , "W:xmm {kz}, xmm, xmm/m64 {sae}"           , "RVM"  , "EVEX.NDS.LIG.66.0F38.W1 CD /r"    , "ANY AVX512ER"],
    ["vrsqrt28ss"       , "W:xmm {kz}, xmm, xmm/m32 {sae}"           , "RVM"  , "EVEX.NDS.LIG.66.0F38.W0 CD /r"    , "ANY AVX512ER"],
    ["vscalefpd"        , "W:xmm {kz}, xmm, xmm/m128/b64"            , "RVM"  , "EVEX.NDS.128.66.0F38.W1 2C /r"    , "ANY AVX512F-VL"],
    ["vscalefpd"        , "W:ymm {kz}, ymm, ymm/m256/b64"            , "RVM"  , "EVEX.NDS.256.66.0F38.W1 2C /r"    , "ANY AVX512F-VL"],
    ["vscalefpd"        , "W:zmm {kz}, zmm, zmm/m512/b64 {er}"       , "RVM"  , "EVEX.NDS.512.66.0F38.W1 2C /r"    , "ANY AVX512F"],
    ["vscalefps"        , "W:xmm {kz}, xmm, xmm/m128/b32"            , "RVM"  , "EVEX.NDS.128.66.0F38.W0 2C /r"    , "ANY AVX512F-VL"],
    ["vscalefps"        , "W:ymm {kz}, ymm, ymm/m256/b32"            , "RVM"  , "EVEX.NDS.256.66.0F38.W0 2C /r"    , "ANY AVX512F-VL"],
    ["vscalefps"        , "W:zmm {kz}, zmm, zmm/m512/b32 {er}"       , "RVM"  , "EVEX.NDS.512.66.0F38.W0 2C /r"    , "ANY AVX512F"],
    ["vscalefsd"        , "W:xmm {kz}, xmm, xmm/m64 {er}"            , "RVM"  , "EVEX.NDS.LIG.66.0F38.W1 2D /r"    , "ANY AVX512F"],
    ["vscalefss"        , "W:xmm {kz}, xmm, xmm/m32 {er}"            , "RVM"  , "EVEX.NDS.LIG.66.0F38.W0 2D /r"    , "ANY AVX512F"],
    ["vscatterdpd"      , "W:vm32x {k}, xmm"                         , "MR"   , "EVEX.128.66.0F38.W1 A2"           , "ANY AVX512F-VL"],
    ["vscatterdpd"      , "W:vm32x {k}, ymm"                         , "MR"   , "EVEX.256.66.0F38.W1 A2"           , "ANY AVX512F-VL"],
    ["vscatterdpd"      , "W:vm32y {k}, zmm"                         , "MR"   , "EVEX.512.66.0F38.W1 A2"           , "ANY AVX512F"],
    ["vscatterdps"      , "W:vm32x {k}, xmm"                         , "MR"   , "EVEX.128.66.0F38.W0 A2"           , "ANY AVX512F-VL"],
    ["vscatterdps"      , "W:vm32y {k}, ymm"                         , "MR"   , "EVEX.256.66.0F38.W0 A2"           , "ANY AVX512F-VL"],
    ["vscatterdps"      , "W:vm32z {k}, zmm"                         , "MR"   , "EVEX.512.66.0F38.W0 A2"           , "ANY AVX512F"],
    ["vscatterpf0dpd"   , "R:vm32y {k}"                              , "M"    , "EVEX.512.66.0F38.W1 C6 /5"        , "ANY AVX512PF"],
    ["vscatterpf0dps"   , "R:vm32z {k}"                              , "M"    , "EVEX.512.66.0F38.W0 C6 /5"        , "ANY AVX512PF"],
    ["vscatterpf0qpd"   , "R:vm64z {k}"                              , "M"    , "EVEX.512.66.0F38.W1 C7 /5"        , "ANY AVX512PF"],
    ["vscatterpf0qps"   , "R:vm64z {k}"                              , "M"    , "EVEX.512.66.0F38.W0 C7 /5"        , "ANY AVX512PF"],
    ["vscatterpf1dpd"   , "R:vm32y {k}"                              , "M"    , "EVEX.512.66.0F38.W1 C6 /6"        , "ANY AVX512PF"],
    ["vscatterpf1dps"   , "R:vm32z {k}"                              , "M"    , "EVEX.512.66.0F38.W0 C6 /6"        , "ANY AVX512PF"],
    ["vscatterpf1qpd"   , "R:vm64z {k}"                              , "M"    , "EVEX.512.66.0F38.W1 C7 /6"        , "ANY AVX512PF"],
    ["vscatterpf1qps"   , "R:vm64z {k}"                              , "M"    , "EVEX.512.66.0F38.W0 C7 /6"        , "ANY AVX512PF"],
    ["vscatterqpd"      , "W:vm64x {k}, xmm"                         , "MR"   , "EVEX.128.66.0F38.W1 A3"           , "ANY AVX512F-VL"],
    ["vscatterqpd"      , "W:vm64y {k}, ymm"                         , "MR"   , "EVEX.256.66.0F38.W1 A3"           , "ANY AVX512F-VL"],
    ["vscatterqpd"      , "W:vm64z {k}, zmm"                         , "MR"   , "EVEX.512.66.0F38.W1 A3"           , "ANY AVX512F"],
    ["vscatterqps"      , "W:vm64x {k}, xmm"                         , "MR"   , "EVEX.128.66.0F38.W0 A3"           , "ANY AVX512F-VL"],
    ["vscatterqps"      , "W:vm64y {k}, xmm"                         , "MR"   , "EVEX.256.66.0F38.W0 A3"           , "ANY AVX512F-VL"],
    ["vscatterqps"      , "W:vm64z {k}, ymm"                         , "MR"   , "EVEX.512.66.0F38.W0 A3"           , "ANY AVX512F"],
    ["vshuff32x4"       , "W:ymm {kz}, ymm, ymm/m256/b32, ib"        , "RVMI" , "EVEX.NDS.256.66.0F3A.W0 23 /r ib" , "ANY AVX512F-VL"],
    ["vshuff32x4"       , "W:zmm {kz}, zmm, zmm/m512/b32, ib"        , "RVMI" , "EVEX.NDS.512.66.0F3A.W0 23 /r ib" , "ANY AVX512F"],
    ["vshuff64x2"       , "W:ymm {kz}, ymm, ymm/m256/b64, ib"        , "RVMI" , "EVEX.NDS.256.66.0F3A.W1 23 /r ib" , "ANY AVX512F-VL"],
    ["vshuff64x2"       , "W:zmm {kz}, zmm, zmm/m512/b64, ib"        , "RVMI" , "EVEX.NDS.512.66.0F3A.W1 23 /r ib" , "ANY AVX512F"],
    ["vshufi32x4"       , "W:ymm {kz}, ymm, ymm/m256/b32, ib"        , "RVMI" , "EVEX.NDS.256.66.0F3A.W0 43 /r ib" , "ANY AVX512F-VL"],
    ["vshufi32x4"       , "W:zmm {kz}, zmm, zmm/m512/b32, ib"        , "RVMI" , "EVEX.NDS.512.66.0F3A.W0 43 /r ib" , "ANY AVX512F"],
    ["vshufi64x2"       , "W:ymm {kz}, ymm, ymm/m256/b64, ib"        , "RVMI" , "EVEX.NDS.256.66.0F3A.W1 43 /r ib" , "ANY AVX512F-VL"],
    ["vshufi64x2"       , "W:zmm {kz}, zmm, zmm/m512/b64, ib"        , "RVMI" , "EVEX.NDS.512.66.0F3A.W1 43 /r ib" , "ANY AVX512F"],
    ["vshufpd"          , "W:xmm {kz}, xmm, xmm/m128/b32, ib"        , "RVMI" , "EVEX.NDS.128.66.0F.W1 C6 /r ib"   , "ANY AVX512F-VL"],
    ["vshufpd"          , "W:ymm {kz}, ymm, ymm/m256/b32, ib"        , "RVMI" , "EVEX.NDS.256.66.0F.W1 C6 /r ib"   , "ANY AVX512F-VL"],
    ["vshufpd"          , "W:zmm {kz}, zmm, zmm/m512/b32, ib"        , "RVMI" , "EVEX.NDS.512.66.0F.W1 C6 /r ib"   , "ANY AVX512F"],
    ["vshufps"          , "W:xmm {kz}, xmm, xmm/m128/b64, ib"        , "RVMI" , "EVEX.NDS.128.0F.W0 C6 /r ib"      , "ANY AVX512F-VL"],
    ["vshufps"          , "W:ymm {kz}, ymm, ymm/m256/b64, ib"        , "RVMI" , "EVEX.NDS.256.0F.W0 C6 /r ib"      , "ANY AVX512F-VL"],
    ["vshufps"          , "W:zmm {kz}, zmm, zmm/m512/b64, ib"        , "RVMI" , "EVEX.NDS.512.0F.W0 C6 /r ib"      , "ANY AVX512F"],
    ["vsqrtpd"          , "W:xmm {kz}, xmm/m128/b64"                 , "RM"   , "EVEX.128.66.0F.W1 51 /r"          , "ANY AVX512F-VL"],
    ["vsqrtpd"          , "W:ymm {kz}, ymm/m256/b64"                 , "RM"   , "EVEX.256.66.0F.W1 51 /r"          , "ANY AVX512F-VL"],
    ["vsqrtpd"          , "W:zmm {kz}, zmm/m512/b64 {er}"            , "RM"   , "EVEX.512.66.0F.W1 51 /r"          , "ANY AVX512F"],
    ["vsqrtps"          , "W:xmm {kz}, xmm/m128/b32"                 , "RM"   , "EVEX.128.0F.W0 51 /r"             , "ANY AVX512F-VL"],
    ["vsqrtps"          , "W:ymm {kz}, ymm/m256/b32"                 , "RM"   , "EVEX.256.0F.W0 51 /r"             , "ANY AVX512F-VL"],
    ["vsqrtps"          , "W:zmm {kz}, zmm/m512/b32 {er}"            , "RM"   , "EVEX.512.0F.W0 51 /r"             , "ANY AVX512F"],
    ["vsqrtsd"          , "W:xmm {kz}, xmm, xmm/m64 {er}"            , "RVM"  , "EVEX.NDS.LIG.F2.0F.W1 51 /r"      , "ANY AVX512F"],
    ["vsqrtss"          , "W:xmm {kz}, xmm, xmm/m32 {er}"            , "RVM"  , "EVEX.NDS.LIG.F3.0F.W0 51 /r"      , "ANY AVX512F"],
    ["vsubpd"           , "W:xmm {kz}, xmm, xmm/m128/b64"            , "RVM"  , "EVEX.NDS.128.66.0F.W1 5C /r"      , "ANY AVX512F-VL"],
    ["vsubpd"           , "W:ymm {kz}, ymm, ymm/m256/b64"            , "RVM"  , "EVEX.NDS.256.66.0F.W1 5C /r"      , "ANY AVX512F-VL"],
    ["vsubpd"           , "W:zmm {kz}, zmm, zmm/m512/b64 {er}"       , "RVM"  , "EVEX.NDS.512.66.0F.W1 5C /r"      , "ANY AVX512F"],
    ["vsubps"           , "W:xmm {kz}, xmm, xmm/m128/b32"            , "RVM"  , "EVEX.NDS.128.0F.W0 5C /r"         , "ANY AVX512F-VL"],
    ["vsubps"           , "W:ymm {kz}, ymm, ymm/m256/b32"            , "RVM"  , "EVEX.NDS.256.0F.W0 5C /r"         , "ANY AVX512F-VL"],
    ["vsubps"           , "W:zmm {kz}, zmm, zmm/m512/b32 {er}"       , "RVM"  , "EVEX.NDS.512.0F.W0 5C /r"         , "ANY AVX512F"],
    ["vsubsd"           , "W:xmm {kz}, xmm, xmm/m64 {er}"            , "RVM"  , "EVEX.NDS.LIG.F2.0F.W1 5C /r"      , "ANY AVX512F"],
    ["vsubss"           , "W:xmm {kz}, xmm, xmm/m32 {er}"            , "RVM"  , "EVEX.NDS.LIG.F3.0F.W0 5C /r"      , "ANY AVX512F"],
    ["vucomisd"         , "R:xmm, xmm/m64 {sae}"                     , "RM"   , "EVEX.LIG.66.0F.W1 2E /r"          , "ANY AVX512F OF=0 SF=0 ZF=W AF=0 PF=W CF=W"],
    ["vucomiss"         , "R:xmm, xmm/m32 {sae}"                     , "RM"   , "EVEX.LIG.0F.W0 2E /r"             , "ANY AVX512F OF=0 SF=0 ZF=W AF=0 PF=W CF=W"],
    ["vunpckhpd"        , "W:xmm {kz}, xmm, xmm/m128/b64"            , "RVM"  , "EVEX.NDS.128.66.0F.W1 15 /r"      , "ANY AVX512F-VL"],
    ["vunpckhpd"        , "W:ymm {kz}, ymm, ymm/m256/b64"            , "RVM"  , "EVEX.NDS.256.66.0F.W1 15 /r"      , "ANY AVX512F-VL"],
    ["vunpckhpd"        , "W:zmm {kz}, zmm, zmm/m512/b64"            , "RVM"  , "EVEX.NDS.512.66.0F.W1 15 /r"      , "ANY AVX512F"],
    ["vunpckhps"        , "W:xmm {kz}, xmm, xmm/m128/b32"            , "RVM"  , "EVEX.NDS.128.0F.W0 15 /r"         , "ANY AVX512F-VL"],
    ["vunpckhps"        , "W:ymm {kz}, ymm, ymm/m256/b32"            , "RVM"  , "EVEX.NDS.256.0F.W0 15 /r"         , "ANY AVX512F-VL"],
    ["vunpckhps"        , "W:zmm {kz}, zmm, zmm/m512/b32"            , "RVM"  , "EVEX.NDS.512.0F.W0 15 /r"         , "ANY AVX512F"],
    ["vunpcklpd"        , "W:xmm {kz}, xmm, xmm/m128/b64"            , "RVM"  , "EVEX.NDS.128.66.0F.W1 14 /r"      , "ANY AVX512F-VL"],
    ["vunpcklpd"        , "W:ymm {kz}, ymm, ymm/m256/b64"            , "RVM"  , "EVEX.NDS.256.66.0F.W1 14 /r"      , "ANY AVX512F-VL"],
    ["vunpcklpd"        , "W:zmm {kz}, zmm, zmm/m512/b64"            , "RVM"  , "EVEX.NDS.512.66.0F.W1 14 /r"      , "ANY AVX512F"],
    ["vunpcklps"        , "W:xmm {kz}, xmm, xmm/m128/b32"            , "RVM"  , "EVEX.NDS.128.0F.W0 14 /r"         , "ANY AVX512F-VL"],
    ["vunpcklps"        , "W:ymm {kz}, ymm, ymm/m256/b32"            , "RVM"  , "EVEX.NDS.256.0F.W0 14 /r"         , "ANY AVX512F-VL"],
    ["vunpcklps"        , "W:zmm {kz}, zmm, zmm/m512/b32"            , "RVM"  , "EVEX.NDS.512.0F.W0 14 /r"         , "ANY AVX512F"],
    ["vxorpd"           , "W:xmm {kz}, xmm, xmm/m128/b64"            , "RVM"  , "EVEX.NDS.128.66.0F.W1 57 /r"      , "ANY AVX512DQ-VL"],
    ["vxorpd"           , "W:ymm {kz}, ymm, ymm/m256/b64"            , "RVM"  , "EVEX.NDS.256.66.0F.W1 57 /r"      , "ANY AVX512DQ-VL"],
    ["vxorpd"           , "W:zmm {kz}, zmm, zmm/m512/b64"            , "RVM"  , "EVEX.NDS.512.66.0F.W1 57 /r"      , "ANY AVX512DQ"],
    ["vxorps"           , "W:xmm {kz}, xmm, xmm/m128/b32"            , "RVM"  , "EVEX.NDS.128.0F.W0 57 /r"         , "ANY AVX512DQ-VL"],
    ["vxorps"           , "W:ymm {kz}, ymm, ymm/m256/b32"            , "RVM"  , "EVEX.NDS.256.0F.W0 57 /r"         , "ANY AVX512DQ-VL"],
    ["vxorps"           , "W:zmm {kz}, zmm, zmm/m512/b32"            , "RVM"  , "EVEX.NDS.512.0F.W0 57 /r"         , "ANY AVX512DQ"]
  ]
};

}).apply(this, typeof module === "object" && module && module.exports
  ? [module, "exports"] : [this.asmdb || (this.asmdb = {}), "x86data"]);
